#include "framecpp/config.h"

extern "C" {

#include <strings.h>
#include <zlib.h>

} // extern "C"

#include <algorithm>
#include <iterator>
#include <memory>
#include <sstream>
#include <stdexcept>
#include <string>

#include "general/util.hh"
#include "general/autoarray.hh"
#include "general/types.hh"

#include "framecpp/Common/Compression.hh"

using namespace FrameCPP;

#define LM_INFO 0

#if LM_DEBUG
#define AT( ) std::cerr << "INFO: " << __FILE__ << " " << __LINE__ << std::endl;
#else
#define AT( )
#endif

namespace
{
  typedef General::hash_map< INT_4U, INT_4U > data_type_size_mapping;

  using namespace FrameCPP::Compression;

  static INT_4S ZERO_COMPRESS_SHORT_BSIZE = 12;

  data_type_size_mapping& size_map_init( )
  {
    static data_type_size_mapping m;

#define MAP( A, B) \
    m[ A ] = sizeof( B )

    MAP( FR_VECT_C, CHAR );
    MAP( FR_VECT_2S, INT_2S );
    MAP( FR_VECT_8R, REAL_8 );
    MAP( FR_VECT_4R, REAL_4 );
    MAP( FR_VECT_4S, INT_4S );
    MAP( FR_VECT_8S, INT_8S );
    MAP( FR_VECT_8C, COMPLEX_8 );
    MAP( FR_VECT_16C, COMPLEX_16 );
    // MAP( FR_VECT_STRING, 
    MAP( FR_VECT_2U, INT_2U );
    MAP( FR_VECT_4U, INT_4U );
    MAP( FR_VECT_8U, INT_8U );
    MAP( FR_VECT_1U, CHAR_U );

#undef MAP
  
    return m;
  }

  static const data_type_size_mapping&	size_map = size_map_init( );

  template < typename _Map >
  inline typename _Map::key_type
  reverse_lookup( typename _Map::mapped_type Key,
		  const _Map& Map )
  {
    for ( typename _Map::const_iterator
	    cur = Map.begin( ),
	    last = Map.end( );
	  cur != last;
	  ++cur )
    {
      if ( cur->second == Key )
      {
	return cur->first;
      }
    }
    
    std::ostringstream	msg;

    msg << "Cannot reverse map: " << Key;
    throw std::range_error( msg.str( ) );
  }

  template < typename T >
  inline INT_2U
  bits( T Value )
  {
    //-------------------------------------------------------------------
    // Take care of the special case where there are not bits in Value
    //   This is to be considered to have 1 bit.
    //-------------------------------------------------------------------
    if ( Value == 0 )
    {
      return 1;
    }

    register INT_2U b = sizeof( T ) * 8;	/* 8 bits per byte */
    register T test = 1 << (b - 1);		/* Shift # bits - 1 */

    //-------------------------------------------------------------------
    // Here is another special case where the Value has the most
    //   significant bit for the type set.
    //-------------------------------------------------------------------
    if ( Value & test )
    {
      return b;
    }
    ++b;
    //-------------------------------------------------------------------
    // Compute the number of bits by testing till the high bit is
    //   found
    //-------------------------------------------------------------------
    while( ! ( Value & test ) )
    {
      --b;
      test >>= 1;
    }
    return b;
  }

  INT_2U
  data_size( INT_4U Type )
  {
    data_type_size_mapping::const_iterator cur = size_map.find( Type );

    if ( cur == size_map.end( ) )
    {
      std::ostringstream	msg;

      msg << "unknown size for data type: " << Type;

      throw std::range_error( msg.str( ) );
    }
    return cur->second;
  }

  template < class T >
  inline std::string dump( const void* Input, INT_8U NData )
  {
    const T* input( reinterpret_cast< const T* >( Input ) );
    std::ostringstream	oss;

    std::copy( input, &( input[ NData ] ),
	       std::ostream_iterator< const T >( oss, " ") );
	  
    return oss.str( );
  }

  template < >
  inline std::string dump< CHAR_U >( const void* Input, INT_8U NData )
  {
    const CHAR_U* input( reinterpret_cast< const CHAR_U* >( Input ) );
    std::ostringstream	oss;

    std::copy( input, &( input[ NData ] ),
	       std::ostream_iterator< const int >( oss, " ") );
	  
    return oss.str( );
  }

  void FrVectZExpandShort( INT_2S *out, INT_2U *data, INT_8U nData );
  int  FrVectZCompShort ( INT_2U *out,
			  INT_8U *compL, INT_2S *data, INT_8U nData,
			  INT_8U bSize);
  void FrVectZExpandInt( INT_4S *out, INT_4U *data, INT_8U nData );
  int  FrVectZCompInt ( INT_4U *out,
			INT_8U *compL, INT_4S *data, INT_8U nData,
			INT_8U bSize);

  void swap_data( INT_8U NData,
		  FrameCPP::Compression::data_types_type Type,
		  CHAR_U* Buffer );

  void
  DiffCompress( CHAR_U* Data, INT_4U NData, INT_2U Type )
  {
    using namespace ::FrameCPP::Compression;

    switch( Type )
    {
    case FR_VECT_1U:
      Differential::Encode< CHAR_U >( Data, NData );
      break;
    case FR_VECT_C:
      Differential::Encode< CHAR >( Data, NData );
      break;
      
    case FR_VECT_2S:
      Differential::Encode< INT_2S >( Data, NData );
      break;
      
    case FR_VECT_2U:
      Differential::Encode< INT_2U >( Data, NData );
      break;
      
    case FR_VECT_4S:
      Differential::Encode< INT_4S >( Data, NData );
      break;
      
    case FR_VECT_4U:
      Differential::Encode< INT_4U >( Data, NData );
      break;
      
    case FR_VECT_8S:
      Differential::Encode< INT_8S >( Data, NData );
      break;
      
    case FR_VECT_8U:
      Differential::Encode< INT_8U >( Data, NData );
      break;
      
    case FR_VECT_4R:
      Differential::Encode< INT_4S >( Data, NData );
      break;
      
    case FR_VECT_8R:
      Differential::Encode< INT_8S >( Data, NData );
      break;
      
    case FR_VECT_8C:
      Differential::Encode< INT_4S >( Data, NData * 2);
      break;
      
    case FR_VECT_16C:
      Differential::Encode< INT_8S >( Data, NData * 2);
      break;
    }
  }

  void
  DiffUncompress( CHAR_U* Data, INT_4U NData, data_types_type Type )
  {
    using namespace ::FrameCPP::Compression;

    if ( NData > 1 )
    {
      switch( Type )
      {
      case FR_VECT_C:
	Differential::Decode<CHAR>( Data, NData );
	break;
      case FR_VECT_1U:
	Differential::Decode<CHAR_U>( Data, NData );
	break;
      case FR_VECT_2S:
	Differential::Decode<INT_2S>( Data, NData );
	break;
      case FR_VECT_2U:
	Differential::Decode<INT_2U>( Data, NData );
	break;
      case FR_VECT_4S:
	Differential::Decode<INT_4S>( Data, NData );
	break;
      case FR_VECT_4U:
	Differential::Decode<INT_4U>( Data, NData );
	break;
      case FR_VECT_8S:
	Differential::Decode<INT_8S>( Data, NData );
	break;
      case FR_VECT_8U:
	Differential::Decode<INT_8U>( Data, NData );
	break;
      case FR_VECT_4R:
	Differential::Decode<INT_4S>( Data, NData );
	break;
      case FR_VECT_8R:
	Differential::Decode<INT_8S>( Data, NData );
	break;
      case FR_VECT_8C:
	Differential::Decode<INT_4S>( Data, NData * 2 );
	break;
      case FR_VECT_16C:
	Differential::Decode<INT_8S>( Data, NData * 2 );
	break;
      default:
	{
	  std::ostringstream	msg;
	  
	  msg << "unable to perform differntial decoding on data type: "
	      << Type;
	  throw std::range_error( msg.str( ) );
	}
	break;
      }
    }
  }

  INT_4U
  DiffGZip( CHAR_U*& Data,
	    INT_8U& DataLength,
	    INT_8U& DataByteLength,
	    INT_4U DataType,
	    INT_4U Level,
	    const compress_type_mapping& CompressionMapping )
  {
    switch( DataType )
    {
    case FR_VECT_C:
    case FR_VECT_1U:
    case FR_VECT_2S:
    case FR_VECT_2U:
    case FR_VECT_4S:
    case FR_VECT_4U:
      DiffCompress( Data, DataLength, DataType );
      GZip::Compress( Data, DataByteLength, Level );
      return (INT_4U)reverse_lookup( MODE_DIFF_GZIP, CompressionMapping );
      break;
    default:
      GZip::Compress( Data, DataByteLength, Level );
      return (INT_4U)reverse_lookup( MODE_GZIP, CompressionMapping );
      break;
    }
  }

} // namespace anonymous

void FrameCPP::
ZeroSuppressCompressShort( CHAR_U*& Data, INT_8U& NBytes, INT_4U NData )
{
  INT_8U bufferSize = (INT_8U)( NBytes * 1.01 + 13 );
  General::AutoArray<INT_2U> buffer( new INT_2U[ bufferSize ] );
    
  int res = FrVectZCompShort( buffer.get( ), &bufferSize,
			      reinterpret_cast< INT_2S* >( Data ),
			      NData, ZERO_COMPRESS_SHORT_BSIZE );
  if ( res != 0 )
  {
    throw std::logic_error( "logical error: zero suppression failed" );
  }

  if ( Data )
  {
    delete[] Data;
  }
  Data = new CHAR_U[ bufferSize ];
  memcpy( Data, buffer.get( ), bufferSize );
  NBytes = bufferSize;
}

void FrameCPP::
ZeroSuppressUncompressShort( CHAR_U*& Data, INT_8U& NBytes, INT_4U NData )
{
  INT_8U	length( sizeof( INT_2S ) * NData );
  General::AutoArray<CHAR_U> buffer( new CHAR_U[ length ] );

  FrVectZExpandShort( reinterpret_cast< INT_2S* >( buffer.get( ) ),
		      reinterpret_cast< INT_2U* >( Data ),
		      NData );
    
  // Assign the data
  if ( Data )
  {
    delete[] Data;
  }
  Data = buffer.release( );
  NBytes = length;
}

void FrameCPP::
ZeroSuppressCompressInt( CHAR_U*& Data, INT_8U& NBytes, INT_4U NData )
{
  INT_8U bufferSize = (INT_8U)( NBytes * 1.01 + 13 );
  General::AutoArray<INT_4U> buffer( new INT_4U[ bufferSize ] );
    
  int res = FrVectZCompInt( buffer.get( ), &bufferSize,
			    reinterpret_cast< INT_4S* >( Data ),
			    NData, 8 );
  if ( res != 0 )
  {
    throw std::logic_error( "logical error: zero suppression failed" );
  }
    
  if ( Data )
  {
    delete[] Data;
  }
  Data = new CHAR_U[ bufferSize ];
  memcpy( Data, buffer.get( ), bufferSize );
  NBytes = bufferSize;
}

void FrameCPP::
ZeroSuppressUncompressInt( CHAR_U*& Data, INT_8U& NBytes, INT_4U NData )
{
  // This may throw std::bad_alloc
  INT_8U	length( NData * sizeof( INT_4U ) );
  General::AutoArray<CHAR_U> buffer( new CHAR_U[ length ] );

  FrVectZExpandInt( reinterpret_cast< INT_4S* >( buffer.get( ) ),
		    reinterpret_cast< INT_4U* >( Data ),
		    NData );

  // Assign the data
  if ( Data )
  {
    delete[] Data;
  }
  Data = buffer.release( );
  NBytes = length;
}

namespace FrameCPP
{
  namespace Compression
  {
    namespace Differential
    {
      template< class T >
      void
      Encode( CHAR_U* Data, INT_4U NData )
      {
	if ( NData > 1 )
	{
	  T*	cur( reinterpret_cast< T* >( Data ) );
	  T*	pre( reinterpret_cast< T* >( Data ) );

	  //---------------------------------------------------------------
	  // Go backwards to eliminate the necisity of temporaries
	  //---------------------------------------------------------------
	  cur = &( cur[ NData - 1 ] );
	  pre = &( pre[ NData - 2 ] );
	  while( cur != (T*)Data )
	  {
	    *cur -= *pre;
	    --cur;
	    --pre;
	  }
	}
      }

      template< class T >
      void
      Decode( CHAR_U* Data, INT_4U NData )
      {
	if ( NData > 1 )
	{
	  T*	cur( reinterpret_cast< T* >( Data ) );
	  T*	pre( reinterpret_cast< T* >( Data ) );

	  //---------------------------------------------------------------
	  // Go forward for the decoding process
	  //---------------------------------------------------------------
	  cur = &( cur[ 1 ] );
	  pre = &( pre[ 0 ] );

	  while( --NData > 0 )
	  {
	    *cur += *pre;
	    ++cur;
	    ++pre;
	  }
	}
      }
    }

    namespace GZip
    {
      //-----------------------------------------------------------------
      //
      // The data must be in an uncompressed state before this is called.
      //
      //!param: size_t level - Compression level.
      //
      void Compress( CHAR_U*& Data, INT_8U& NBytes, size_t Level )
      {
	if ( Level == 0 )
	{
	  return;
	}
	if ( Level > 9 )
	{
	  Level = 9;
	}
    
	uLongf bufferSize = uLongf( NBytes * 1.001 + 13 );
    
	CHAR_U* buffer = new CHAR_U[ bufferSize ];
#if HAVE_LIBZ_COMPRESS2
	::compress2( buffer, &bufferSize, Data, NBytes, Level );
#else
	::compress( buffer, &bufferSize, Data, NBytes );
#endif

	delete[] Data;
	Data = new CHAR_U[ bufferSize ];
	memcpy( Data, buffer, bufferSize );
	NBytes = bufferSize;
	delete[] buffer;
	buffer = (unsigned char*)NULL;
      } // func - Compress

      void
      Expand( CHAR_U*& Data,
	      INT_8U& DataLength,
	      const INT_8U ExpandedLength )
      {
	//---------------------------------------------------------------
	// This may throw std::bad_alloc
	//---------------------------------------------------------------
	General::AutoArray< CHAR_U >
	  buffer( new CHAR_U[ ExpandedLength ] );
	uLongf uncompressedLength( ExpandedLength );

	int error = uncompress( buffer.get( ),
				&uncompressedLength,
				Data,
				DataLength );

	//---------------------------------------------------------------
	// Check the return status from the uncompress
	//---------------------------------------------------------------
	if ( error != Z_OK )
	{
	  switch( error )
	  {
	  case Z_MEM_ERROR:
	    throw std::bad_alloc();
	    break;
	
	  case Z_BUF_ERROR:
	  case Z_DATA_ERROR:
	    throw std::runtime_error( "Compressed data is corrupt." );
	    break;
                    
	  default:
	    throw std::runtime_error( "An unknown error occurred during "
				      "uncompression." );
	  }
	}

	//---------------------------------------------------------------
	// Make sure we uncompressed the correct number of bytes.
	//---------------------------------------------------------------
	if ( uncompressedLength != ExpandedLength )
	{
	  throw std::range_error( "Compressed data is corrupt" );
	}

	//---------------------------------------------------------------
	// Successfully expanded the buffer, so return the results
	//---------------------------------------------------------------
	//-------------------------------------------------------------------
	// Release previous memory used
	//-------------------------------------------------------------------
	if ( Data )
	{
	  // Release the memory first
	  delete[] Data;
	}
	//-------------------------------------------------------------------
	// Store the new data
	//-------------------------------------------------------------------
	Data = buffer.release( );
	DataLength = ExpandedLength;
      } // func - Exapnd

    } // namespace - GZip

    void
    Compress( INT_4U& CompressionMode,
	      INT_4U Level,
	      const compress_type_mapping& CompressionMapping,
	      INT_4U DataType,
	      const data_type_mapping& DataTypeMapping,
	      CHAR_U*& Data,
	      INT_8U& DataLength,
	      INT_8U& DataByteLength )
    {
      INT_4U compression_mode_return = CompressionMode;
      {
	//---------------------------------------------------------------
	// Translate the compression mode
	//---------------------------------------------------------------
	compress_type_mapping::const_iterator c
	  = CompressionMapping.find( CompressionMode );

	if ( c == CompressionMapping.end( ) )
	{
	  std::ostringstream	msg;

	  msg << "unknown compression mode: " << CompressionMode;
	  throw std::range_error( msg.str( ) );
	}
	CompressionMode = c->second;
      }
      {
	//---------------------------------------------------------------
	// Translate the data type
	//---------------------------------------------------------------
	data_type_mapping::const_iterator d
	  = DataTypeMapping.find( DataType );

	if ( d == DataTypeMapping.end( ) )
	{
	  std::ostringstream	msg;

	  msg << "unknown compression mode: " << DataType;
	  throw std::range_error( msg.str( ) );
	}
	DataType = d->second;
      }
      //---------------------------------------------------------------------
      // Establish the compression mode
      //---------------------------------------------------------------------

      AT( );
      switch( CompressionMode )
      {
      case MODE_GZIP:
	AT( );
	GZip::Compress( Data, DataByteLength, Level );
	break;
      case MODE_DIFF:
	AT( );
	DiffCompress( Data, DataLength, DataType );
	break;
      case MODE_DIFF_GZIP:
	AT( );
	compression_mode_return
	  = DiffGZip( Data, DataLength, DataByteLength, DataType,
		      Level, CompressionMapping );
	break;
      case MODE_ZERO_SUPPRESS_SHORT:
	AT( );
	switch( DataType )
	{
	case FR_VECT_2S:
	case FR_VECT_2U:
	  DiffCompress( Data, DataLength, DataType );
	  try
	  {
	    ZeroSuppressCompressShort( Data, DataByteLength, DataLength );
	  }
	  catch( const FrUncompressable& e )
	  {
	    // Abandon compression and leave the buffer as it was
	    DiffUncompress( Data, DataLength, (data_types_type)DataType );
	    goto raw_return;
	  }
	  break;
	default:
	  goto raw_return;
	  break;
	}
	break;
      case MODE_ZERO_SUPPRESS_SHORT_GZIP_OTHER:
	AT( );
	switch( DataType )
	{
	case FR_VECT_2S:
	case FR_VECT_2U:
	  DiffCompress( Data, DataLength, DataType );
	  try
	  {
	    ZeroSuppressCompressShort( Data, DataByteLength, DataLength );
	  }
	  catch( const FrUncompressable& e )
	  {
	    // Abandon compression and leave the buffer as it was
	    DiffUncompress( Data, DataLength, (data_types_type)DataType );
	    goto raw_return;
	  }
	  break;
	default:
	  GZip::Compress( Data, DataByteLength, Level );
	  compression_mode_return
	    = (INT_4U)reverse_lookup( MODE_GZIP,
				      CompressionMapping );
	  break;
	}
	break;
      case MODE_ZERO_SUPPRESS_INT_FLOAT:
      case MODE_ZERO_SUPPRESS_OTHERWISE_GZIP:
	AT( );
	switch( DataType )
	{
	case FR_VECT_2S:
	case FR_VECT_2U:
	  DiffCompress( Data, DataLength, DataType );
	  try
	  {
	    ZeroSuppressCompressShort( Data, DataByteLength, DataLength );
	  }
	  catch( const FrUncompressable& e )
	  {
	    // Abandon compression and leave the buffer as it was
	    DiffUncompress( Data, DataLength, (data_types_type)DataType );
	    goto raw_return;
	  }
	  compression_mode_return
	    = (INT_4U)reverse_lookup( MODE_ZERO_SUPPRESS_SHORT,
				      CompressionMapping );
	  break;
	case FR_VECT_4R:
	case FR_VECT_4S:
	case FR_VECT_4U:
	  DiffCompress( Data, DataLength, DataType );
	  try
	  {
	    ZeroSuppressCompressInt( Data, DataByteLength, DataLength );
	    if ( MODE_ZERO_SUPPRESS_OTHERWISE_GZIP
		 == CompressionMode )
	    {
	      compression_mode_return
		= (INT_4U)reverse_lookup( MODE_ZERO_SUPPRESS_INT_FLOAT,
					  CompressionMapping );
	    }
	  }
	  catch( const FrUncompressable& e )
	  {
	    // Abandon compression and leave the buffer as it was
	    DiffUncompress( Data, DataLength, (data_types_type)DataType );
	    goto raw_return;
	  }
	  break;
	default:
	  if ( MODE_ZERO_SUPPRESS_OTHERWISE_GZIP
	       == CompressionMode )
	  {
	    GZip::Compress( Data, DataByteLength, Level );
	    compression_mode_return
	      = (INT_4U)reverse_lookup( MODE_GZIP,
					CompressionMapping );
	  }
	  else
	  {
	    goto raw_return;
	  }
	  break;
	}
	break;
      default:
	AT( );
	break;
      }
      CompressionMode = compression_mode_return;
      return;

    raw_return:
      CompressionMode
	= (INT_4U)reverse_lookup( MODE_RAW,
				  CompressionMapping );
      return;
    }

    void
    Expand( bool NativeOrder,
	    INT_4U CompressionMode,
	    const compress_type_mapping& CompressionMapping,
	    INT_4U DataType,
	    const data_type_mapping& DataTypeMapping,
	    CHAR_U*& Data,
	    INT_8U& DataLength,
	    INT_8U& DataByteLength
	    )
    {
      {
	//---------------------------------------------------------------
	// Translate the compression mode
	//---------------------------------------------------------------
	compress_type_mapping::const_iterator c
	  = CompressionMapping.find( CompressionMode );

	if ( c == CompressionMapping.end( ) )
	{
	  std::ostringstream	msg;

	  msg << "unknown compression mode: " << CompressionMode;
	  throw std::range_error( msg.str( ) );
	}
	CompressionMode = c->second;
      }
      {
	//---------------------------------------------------------------
	// Translate the data type
	//---------------------------------------------------------------
	data_type_mapping::const_iterator d
	  = DataTypeMapping.find( DataType );

	if ( d == DataTypeMapping.end( ) )
	{
	  std::ostringstream	msg;

	  msg << "unknown compression mode: " << DataType;
	  throw std::range_error( msg.str( ) );
	}
	DataType = d->second;
      }
      AT( );
      // If it's not compressed, then it might need to be byte swapped   
      if( CompressionMode == MODE_RAW )
      {
	AT( );
	swap_data( DataLength, (data_types_type)DataType, Data );
	AT( );
	return;
      }

      AT( );
      // See if the data should be z-uncompressed (zero suppress method)
      // for short data only
      //
      if( ( CompressionMode == MODE_ZERO_SUPPRESS_SHORT_GZIP_OTHER ||
	    CompressionMode == MODE_ZERO_SUPPRESS_SHORT ) &&
	  ( DataType == FR_VECT_2S || DataType == FR_VECT_2U ) )
      {
	AT( );
	if ( ! NativeOrder )
	{
	  AT( );
	  swap_data( DataByteLength/sizeof( INT_2S ),
		     (data_types_type)DataType, Data );
	  NativeOrder = true;
	}

	ZeroSuppressUncompressShort( Data, DataByteLength, DataLength );
      }
      else if ( ( CompressionMode == MODE_ZERO_SUPPRESS_INT_FLOAT ) &&
		( DataType == FR_VECT_4S
		  || DataType == FR_VECT_4U
		  || DataType == FR_VECT_4R ) )
      {
	AT( );
	if ( ! NativeOrder )
	{
	  AT( );
	  swap_data( DataByteLength/sizeof( INT_4S ),
		     (data_types_type)DataType, Data );
	  NativeOrder = true;
	}
	ZeroSuppressUncompressInt( Data, DataByteLength, DataLength );
      }
      // Determine if we need to gunzip the data.
      if ( ( CompressionMode == MODE_GZIP ) ||
	   ( CompressionMode == MODE_DIFF_GZIP ) || 
	   ( CompressionMode == MODE_ZERO_SUPPRESS_SHORT_GZIP_OTHER &&
	     ( DataType != FR_VECT_2S
	       || DataType != FR_VECT_2U ) ) )
      {
	//---------------------------------------------------------------
	// Gunzip the buffer
	//---------------------------------------------------------------
	AT( );
	const INT_8U	length = DataLength * data_size( DataType );
	GZip::Expand( Data,		/* compressed data		*/
		      DataByteLength,	/* size of compressed data	*/
		      length		/* expected data length		*/
		      );
      }
    
      if ( ! NativeOrder )
      {
	AT( );
	swap_data( DataLength, (data_types_type)DataType, Data );
	NativeOrder = true;
      }

      // Integrate the data if necessary.
      if ( ( CompressionMode == MODE_DIFF ) ||
	   ( CompressionMode == MODE_DIFF_GZIP ) ||
	   ( CompressionMode == MODE_ZERO_SUPPRESS_SHORT_GZIP_OTHER ) ||
	   ( CompressionMode == MODE_ZERO_SUPPRESS_SHORT &&
	     ( DataType == FR_VECT_2S || DataType == FR_VECT_2U ) ) ||
	   ( CompressionMode == MODE_ZERO_SUPPRESS_INT_FLOAT &&
	     ( DataType == FR_VECT_4S ||
	       DataType == FR_VECT_4U ||
	       DataType == FR_VECT_4R ) ) )
      {
	DiffUncompress( Data, DataLength, (data_types_type)DataType );
      }
    } // func - Expand

  } // namespace - Compression
} // namespace - FrameCPP 

namespace
{
  int FrVectZCompShort( INT_2U *out, 
			INT_8U *compL,
			INT_2S *data,
			INT_8U nData,
			INT_8U bSize)
  {
    INT_2U nBits, uData, limit;
    INT_2S pOut;
    INT_2S max;
    INT_8U iIn, iOut, i, maxOut;

    static const INT_2S wMax[] = {
      0, 0,
      1,      3,      7,     0xf,
      0x1f,   0x3f,   0x7f,  0xff,
      0x1ff,  0x3ff,  0x7ff, 0xfff,
      0x1fff, 0x3fff, 0x7fff
    };
    
    maxOut = *compL/2;
    /*------------------------- store the bloc size -----*/
    out[0] = bSize;
    iOut = 0;
    pOut = 16;
    /*--------------------------------- store the data --*/
    iIn = 0;
    while( iIn < nData )
    {
      /*------------- tune the size of the last bloc -----*/
      if(iIn + bSize > nData)
      {
	bSize = nData - iIn;
      }

      /*--------- get the maximum amplitude --------------*/
      
      max = 0;
      for( i=0; i<bSize; i++ )
      {
	if(data[iIn+i] == MIN_INT_2S )
	{
	  throw FrZCompRangeError( );
	}
	if(data[iIn+i] > 0)
	{
	  max = max |  data[iIn + i];
	}
	else
	{
	  max = max | -data[iIn + i];
	}
      }

      /*----- determine the number of bits needed -( (2*(max-1)  < 2**nBits)) */

#if 0
      nBits = bits( max );
#else /* 1 */
      if(max > 127) 
      {
	if(max > 2047) 
	{
	  if(max > 8191)
	  {
	    if (max > 16383)
	    {
	      nBits = 16;
	    }
	    else
	    {
	      nBits = 15;
	    }
	  }
	  else
	  {
	    if(max > 4095)
	    {
	      nBits = 14;
	    }
	    else
	    {
	      nBits = 13;
	    }
	  }
	}
	else
	{
	  if(max > 511)
	  {
	    if(max > 1023)
	    {
	      nBits = 12;
	    }
	    else
	    {
	      nBits = 11;
	    }
	  }
	  else
	  {
	    if(max > 255)
	    {
	      nBits = 10;
	    }
	    else
	    {
	      nBits = 9;
	    }
	  }
	}
      }
      else
      {
	if(max > 7) 
	{
	  if(max > 31)
	  {
	    if(max > 63)
	    {
	      nBits = 8;
	    }
	    else
	    {
	      nBits = 7;
	    }
	  }
	  else
	  {
	    if(max > 15)
	    {
	      nBits = 6;
	    }
	    else
	    {
	      nBits = 5;
	    }
	  }
	}
	else
	{
	  if(max > 1)
	  {
	    if(max > 3)
	    {
	      nBits = 4;
	    }
	    else
	    {
	      nBits = 3;
	    }
	  }
	  else
	  {
	    if(max > 0)
	    {
	      nBits = 2;
	    }
	    else
	    {
	      nBits = 1;
	    }
	  }
	}
      }
#endif /* 1 */
      
      /*---- encode the data size - we store nBits - 1 in 4 bits---*/

      out[iOut] = out[iOut] | ((nBits-1) << pOut);
      if(pOut > 12)
      {
	iOut ++;
	if (iOut >= maxOut)
	{
#if LM_INFO
	  std::cerr << "INFO: uncompressable: "
		    << " iOut: " << iOut
		    << " maxOut: " << maxOut
		    << std::endl;
#endif
	  throw FrUncompressable( "FrVectZCompShort: iOut >= maxOut" );
	}
	pOut = pOut - 16;
	out[iOut] = (nBits-1) >> - pOut;
      }
      pOut  = pOut +4;
      
      /*----------------------- encode the data itself ------*/
      if(nBits > 1)
      {
	limit = 16 - nBits;
	for(i = 0; i<bSize; i++)
	{
	  uData = (INT_2U) (data[iIn+i] + wMax[nBits]);
	  
	  out[iOut] = out[iOut] | (uData << pOut);
	  if(pOut > limit)
	  {
	    iOut ++;
	    if (iOut >= maxOut)
	    {
#if LM_INFO
	      std::cerr << "INFO: uncompressable: "
			<< " iOut: " << iOut
			<< " maxOut: " << maxOut
			<< std::endl;
#endif
	      throw FrUncompressable( "FrVectZCompShort: iOut >= maxOut" );
	    }
	    pOut = pOut - 16;
	    out[iOut] = uData >> -pOut;}
	  pOut  = pOut + nBits;
	}
      } 
      /*----------------------------- increase pointer -------*/
      
      iIn = iIn + bSize;
    }

    *compL = 2*(iOut+1);

    return(0);
  }

  int
  FrVectZCompInt( INT_4U*	out, 
		  INT_8U*	compL,
		  INT_4S*	data,
		  INT_8U	nData,
		  INT_8U	bSize )
  {
    INT_4U nBits, uData, limit;
    INT_4U pOut;
    INT_4S max;
    INT_8U  iIn, iOut, i, maxOut;
    static const INT_4S wMax[] = {
      0        ,
      0        ,1         ,3         ,7,
      0xf      ,0x1f      ,0x3f      ,0x7f,
      0xff     ,0x1ff     ,0x3ff     ,0x7ff,
      0xfff    ,0x1fff    ,0x3fff    ,0x7fff,
      0xffff   ,0x1ffff   ,0x3ffff   ,0x7ffff,
      0xfffff  ,0x1fffff  ,0x3fffff  ,0x7fffff,
      0xffffff ,0x1ffffff ,0x3ffffff ,0x7ffffff,
      0xfffffff,0x1fffffff,0x3fffffff,0x7fffffff
    };

    maxOut = *compL/4;
    /*----------------------------store the bloc size--*/
    out[0] = bSize;
    iOut = 0;
    pOut = 16;
    /*------------------------------- store the data --*/
    iIn = 0;
    while(iIn<nData)
    {
      /*------------ tune the size of the last bloc -----*/

      if(iIn + bSize > nData)
      {
	bSize = nData - iIn;
      }

      /*-------- get the maximum amplitude --------------*/
      max = 0;
     
      for( i=0; i < bSize; i++)
      {
	if( data[iIn+i] == MIN_INT_4S )
	{
#if LM_INFO
	  std::cerr << "INFO: iIn+i: " << ( iIn+i )
		    << " " << iIn << " " << i
		    << " data: " << data[iIn+i]
		    << std::endl;
#endif
	  throw FrameCPP::FrZCompIRangeError( );
	}
	if(data[iIn+i] > 0)
	{
	  max = max |  data[iIn + i];
	}
	else
	{
	  max = max | -data[iIn + i];
	}
      }

      /*----- determine the number of bits needed - (2*(max-1)  < 2**nBits)   */
    
#if 0
      nBits = bits( max );
#else /* 1 */
      if(max > 0x7fff)
      {
	if(max > 0x7fffff) 
	{
	  if(max > 0x7ffffff) 
	  {
	    if(max > 0x1fffffff)
	    {
	      if(max > 0x3fffffff)
	      {
		nBits = 32;
	      }
	      else
	      {
		nBits = 31;
	      }
	    }
	    else
	    {
	      if(max > 0xfffffff)
	      {
		nBits = 30;
	      }
	      else
	      {
		nBits = 29;
	      }
	    }
	  }
	  else
	  {
	    if(max > 0x1ffffff)
	    {
	      if(max > 0x3ffffff)
	      {
		nBits = 28;
	      }
	      else
	      {
		nBits = 27;
	      }
	    }
	    else
	    {
	      if(max > 0xffffff)
	      {
		nBits = 26;
	      }
	      else
	      {
		nBits = 25;
	      }
	    }
	  }
	}
	else
	{
	  if(max > 0x7ffff) 
	  {
	    if(max > 0x1fffff)
	    {
	      if(max > 0x3fffff)
	      {
		nBits = 24;
	      }
	      else
	      {
		nBits = 23;
	      }
	    }
	    else
	    {
	      if(max > 0xfffff)
	      {
		nBits = 22;
	      }
	      else
	      {
		nBits = 21;
	      }
	    }
	  }
	  else
	  {
	    if(max > 0x1ffff)
	    {
	      if(max > 0x3ffff)
	      {
		nBits = 20;
	      }
	      else
	      {
		nBits = 19;
	      }
	    }
	    else
	    {
	      if(max > 0xffff)
	      {
		nBits = 18;
	      }
	      else
	      {
		nBits = 17;
	      }
	    }
	  }
	}
      }
      else
      {
	if(max > 0x7f) 
	{
	  if(max > 0x7ff) 
	  {
	    if(max > 0x1fff)
	    {
	      if(max > 0x3fff)
	      {
		nBits = 16;
	      }
	      else
	      {
		nBits = 15;
	      }
	    }
	    else
	    {
	      if(max > 0xfff)
	      {
		nBits = 14;
	      }
	      else
	      {
		nBits = 13;
	      }
	    }
	  }
	  else
	  {
	    if(max > 0x1ff)
	    {
	      if(max > 0x3ff)
	      {
		nBits = 12;
	      }
	      else
	      {
		nBits = 11;
	      }
	    }
	    else
	    {
	      if(max > 0xff)
	      {
		nBits = 10;
	      }
	      else
	      {
		nBits = 9;
	      }
	    }
	  }
	}
	else
	{
	  if(max > 0x7) 
	  {
	    if(max > 0x1f)
	    {
	      if(max > 0x3f)
	      {
		nBits = 8;
	      }
	      else
	      {
		nBits = 7;
	      }
	    }
	    else
	    {
	      if(max > 0xf)
	      {
		nBits = 6;
	      }
	      else
	      {
		nBits = 5;
	      }
	    }
	  }
	  else
	  {
	    if(max > 1)
	    {
	      if(max > 3)
	      {
		nBits = 4;
	      }
	      else
	      {
		nBits = 3;
	      }
	    }
	    else
	    {
	      if(max > 0)
	      {
		nBits = 2;
	      }
	      else
	      {
		nBits = 1;
	      }
	    }
	  }
	}
      }
#endif /* 1 */
      
      /*---- encode the data size ---- we store nBits - 1 in 4 bits---*/

      if(pOut != 32)
      {
	out[iOut] = out[iOut] | ((nBits-1) << pOut);
      }
      if(pOut > 27)
      {
	iOut ++;
	if(iOut >= maxOut)
	{
#if LM_INFO
	  std::cerr << "INFO: uncompressable: "
		    << " iOut: " << iOut
		    << " maxOut: " << maxOut
		    << std::endl;
#endif
	  throw FrUncompressable( "FrVectZCompInt: iOut >= maxOut" );
	}
	pOut = pOut - 32;
	out[iOut] = (nBits-1) >> - pOut;}
      pOut  = pOut + 5;
      /*--------- encode the data itself ------*/
      if(nBits > 1)
      {
	limit = 32 - nBits;
	for(i = 0; i<bSize; i++)
	{
	  uData = (unsigned int) (data[iIn+i] + wMax[nBits]);
	  
	  if( pOut != 32) out[iOut] = out[iOut] | (uData << pOut);
	  if( pOut > limit)
	  {
	    iOut ++;
	    if(iOut >= maxOut)
   	    {
#if LM_INFO
	      std::cerr << "INFO: uncompressable: "
			<< " iOut: " << iOut
			<< " maxOut: " << maxOut
			<< std::endl;
#endif
	      throw FrUncompressable( "FrVectZCompInt: iOut >= maxOut" );
	    }
	    pOut = pOut - 32;
	    out[iOut] = uData >> -pOut;
	  }
	  pOut  = pOut + nBits;
	}
      } 
      /*------------------------- increase pointer -------*/
  
      iIn = iIn + bSize;
    }

    *compL = 4*(iOut+1);

    return(0);
  }

  void
  FrVectZExpandShort( INT_2S*	out,
		      INT_2U*	data, 
		      INT_8U	nData)
  {
    static const INT_2S wMax[] = {
      0,      0,
      1,      3,      7,     0xf,
      0x1f,   0x3f,   0x7f,  0xff,
      0x1ff,  0x3ff,  0x7ff, 0xfff,
      0x1fff, 0x3fff, 0x7fff
    };
    static const INT_2U mask[] = {
      0,
      1,      3,      7,      0xf,
      0x1f,   0x3f,   0x7f,   0xff,
      0x1ff,  0x3ff,  0x7ff,  0xfff,
      0x1fff, 0x3fff, 0x7fff, 0xffff
    };

    union {
      INT_4U	int_4u;
      INT_2U	int_2u[2];
    } iBuf;

    INT_2U nBits, pIn, uData, *buf;
    INT_8U i, iIn, iOut, bSize;

    /*------------------ retrieve the bloc size -----*/
    bSize = data[0];
    iIn = 1;
    pIn = 0;
    /*------------- retrieve the data  --------------*/

    buf = (INT_2U *) &(iBuf.int_2u[0]);
    iOut = 0;
    do
    {
      /*-------- extract nBits ---(we check if data is in 1 or 2 words) ------*/
        
      if(pIn <= 12)
      {
	uData = data[iIn] >> pIn;
	pIn = pIn + 4;
      }
      else
      {
	buf[0] = data[iIn];
	iIn++;
#if WORDS_BIGENDIAN
	buf[1] = buf[0];
	buf[0] = data[iIn];
#else /* WORDS_BIGENDIAN */
	buf[1] = data[iIn];
#endif /* WORDS_BIGENDIAN */
	uData = iBuf.int_4u >> pIn;
	pIn = pIn - 12;
      }
      
      nBits = 1 + (mask[4] & uData);
      if(nBits == 1)
	nBits = 0;

      /*----------------------------extract data ----------*/
    
      for(i=0; i<bSize; i++)
      {
	if(iOut >= nData)
	{
	  break;
	}

	if(pIn + nBits <= 16)
	{
	  uData = data[iIn] >> pIn;
	  pIn = pIn + nBits;
	}
	else
	{
	  buf[0] = data[iIn];
	  iIn++;
#if WORDS_BIGENDIAN
	  buf[1] = buf[0];
	  buf[0] = data[iIn];
#else /* WORDS_BIGENDIAN */
	  buf[1] = data[iIn];
#endif /* WORDS_BIGENDIAN */
	  uData = iBuf.int_4u >> pIn;
	  pIn = pIn + nBits - 16;
	}

	out[iOut] = (mask[nBits] & uData) - wMax[nBits];
	iOut++;
      }
    }
    while(iOut<nData);
  }

  void
  FrVectZExpandInt( INT_4S*	out, 
		    INT_4U*	data, 
		    INT_8U	nData )
  {
    INT_4U nBits, pIn, uData;
    static const INT_4U wMax[] = {
      0,
      0        ,1         ,3         ,7,
      0xf      ,0x1f      ,0x3f      ,0x7f,
      0xff     ,0x1ff     ,0x3ff     ,0x7ff,
      0xfff    ,0x1fff    ,0x3fff    ,0x7fff,
      0xffff   ,0x1ffff   ,0x3ffff   ,0x7ffff,
      0xfffff  ,0x1fffff  ,0x3fffff  ,0x7fffff,
      0xffffff ,0x1ffffff ,0x3ffffff ,0x7ffffff,
      0xfffffff,0x1fffffff,0x3fffffff,0x7fffffff
    };
    static const INT_4U mask[] = {
      0,
      1         ,3         ,7,         0xf,
      0x1f      ,0x3f      ,0x7f,      0xff,
      0x1ff     ,0x3ff     ,0x7ff,     0xfff,
      0x1fff    ,0x3fff    ,0x7fff,    0xffff,
      0x1ffff   ,0x3ffff   ,0x7ffff,   0xfffff,
      0x1fffff  ,0x3fffff  ,0x7fffff,  0xffffff,
      0x1ffffff ,0x3ffffff ,0x7ffffff, 0xfffffff,
      0x1fffffff,0x3fffffff,0x7fffffff,0xffffffff
    };

    INT_4S  i, iIn, bSize;
    INT_4U iOut;

    /*---------------- retrieve the bloc size -----*/
    bSize = data[0] & 0xffff;

    iIn = 0;
    pIn = 16;

    /*------------------ retrieve the data---------*/

    iOut = 0;
    do
    {
      /*-------- extract nBits -(we check if data is in 1 or 2 words) ------*/
        
      if(pIn <= 27)
      {
	uData = data[iIn] >> pIn;
	pIn = pIn + 5;
      }
      else
      {
	uData = (data[iIn] >> pIn) & mask[32-pIn];
	iIn++;
	uData += data[iIn] << (32-pIn);
	pIn = pIn - 27;
      }

      nBits = 1 + (0x1f & uData);
      if(nBits == 1)
      {
	nBits = 0;
      }

      /*----------------------extract data ----------*/
      for(i=0; i<bSize; i++)
      {
	if(iOut >= nData)
	{
	  break;
	}
	
	if(pIn + nBits <= 32)
	{
	  uData = data[iIn] >> pIn;
	  pIn = pIn + nBits;
	}
	else
	{
	  uData = (data[iIn] >> pIn) & mask[32-pIn];
	  iIn++;
	  uData += data[iIn] << (32-pIn);
	  pIn = pIn + nBits - 32;
	}

	out[iOut] = (mask[nBits] & uData) - wMax[nBits];
	iOut++;
      }
    }
    while(iOut<nData);
  }

  void
  swap_data( INT_8U NData,
	     FrameCPP::Compression::data_types_type Type,
	     CHAR_U* Buffer )
  {
    using namespace FrameCPP::Compression;

    switch( Type )
    {
    case FR_VECT_2S:
    case FR_VECT_2U:
      reverse<2>( Buffer, NData );
      break;
    case FR_VECT_4R:
    case FR_VECT_4S:
    case FR_VECT_4U:
      reverse<4>( Buffer, NData );
      break;
    case FR_VECT_8R:
    case FR_VECT_8S:
    case FR_VECT_8U:
      reverse<8>( Buffer, NData );
      break;
    case FR_VECT_8C:
      reverse<4>( Buffer, NData * 2);
      break;
    case FR_VECT_16C:
      reverse<8>( Buffer, NData * 2);
      break;
    case FR_VECT_STRING:
      {
	CHAR_U*	pos( Buffer );
	INT_2U*	len( (INT_2U*)(Buffer ) );
	for ( INT_4U x = 0; x < NData; x++ )
	{
	  reverse< 2 >( pos, 1 );
	  pos += *len + 2;
	  len = (INT_2U*)pos;
	}
      }
      break;
    default:
      break;
    }
  }
} // namespace anonymous

#define INSTANTIATE( LM_TYPE ) \
  template void FrameCPP::Compression::Differential::Encode< LM_TYPE >( CHAR_U* Data, INT_4U NData ); \
  template void FrameCPP::Compression::Differential::Decode< LM_TYPE >( CHAR_U* Data, INT_4U NData )


INSTANTIATE( CHAR_U );
INSTANTIATE( CHAR );
INSTANTIATE( INT_2S );
INSTANTIATE( INT_2U );
INSTANTIATE( INT_4S );
INSTANTIATE( INT_4U );
INSTANTIATE( INT_8S );
INSTANTIATE( INT_8U );
