#include "framecpp/config.h"

#include <iomanip>
#include <list>
#include <map>
#include <sstream>

#include "general/autoarray.hh"
#include "general/fstream.hh"
#include "general/util.hh"

#include "framecpp/FrameCPP.hh"

#include "framecpp/Common/MD5Sum.hh"

#if FRAME_SPEC_CURRENT == 6
#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/FrEndOfFile.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/Util.hh"
#else /* FRAME_SPEC_CURRENT */
#error Unsupported frame spec: FRAME_SPEC_CURRENT
#endif /* FRAME_SPEC_CURRENT */

#include "framecpp/FrameH.hh"
#include "framecpp/FrVect.hh"
#include "framecpp/CheckSum.hh"
#include "framecpp/DetectorMap.hh"
#include "framecpp/IFrameStream.hh"
#include "framecpp/Verify.hh"

#define	DEBUG_MD5SUM		0
#define	DEBUG_FILE_CHECKSUM	0

using namespace FrameCPP::Version;
using FrameCPP::CheckSum;
using FrameCPP::CheckSumCRC;
using FrameCPP::Common::MD5Sum;
using FrameCPP::Version::FrEndOfFile;
using FrameCPP::Version::IFrameStream;


namespace
{
  const INT_8U x123456789abcdef( 0x123456789abcdefLL );
  const bool DEFAULT_CHECK_MD5SUM( false );

  template< class T >
  inline void
  check_magic_number( const unsigned char* const Source, const T Ref )
  {
    T buf;

    //-------------------------------------------------------------------
    // Load the buffer
    //-------------------------------------------------------------------
    memcpy(&buf, Source, sizeof( buf ) );
    //-------------------------------------------------------------------
    // Check the bytes
    //-------------------------------------------------------------------
    if ( Ref != buf )
    {
      //-----------------------------------------------------------------
      // Reverse the bytes and check again
      //-----------------------------------------------------------------
      reverse< sizeof( buf ) >( &buf, 1 );
      if ( Ref != buf )
      {
	//---------------------------------------------------------------
	// Pattern does not match either BIGENDIAN or LITTLEENDIAN so
	//   it must not be a frame file.
	//---------------------------------------------------------------
	throw std::runtime_error( "non-frame frame file" );
      }
    }
  }
} // namespace - anonymous

const char* FrameCPP::Verify::m_error_strings[] = 
  {
    "no failure",
    "checksum error",
    "unreadable file error",
    "unsupported frame version error",
    "frame does not conform to the frame specification",
    "the type of checksum being requested was unknown.",
    "no checksum was supplied at the end of the file",
    "one of the internal frame structures does not pass its own consitancy check",
    "the file is missing the FrEndOfFile structure required by the frame specification.",
    "One or more channels have a non-zero dataValid field.",
    "One or more pieces of the filename meta data is invalid."
  };

const char* FrameCPP::Verify::m_error_type_strings[] =
  {
    "NO_ERROR",
    "CHECKSUM_ERROR",
    "FILE_OPEN_ERROR",
    "UNSUPPORTED_FRAME_SPEC",
    "FRAME_SPEC_CONFORMANCE",
    "UNSUPPORTED_CHECKSUM_TYPE",
    "NO_CHECKSUM",
    "INVALID_FRAME_STRUCTURE",
    "FILE_TRUNCATION",
    "DATA_INVALID",
    "METADATA_INVALID",
  };

FrameCPP::Verify::
Verify( )
  : m_buffer_size( INT_8U( 0 ) ),
    m_check_data_valid( false ),
    m_check_file_checksum( true ),
    m_check_frame_checksum( false ),
    m_check_md5sum( DEFAULT_CHECK_MD5SUM ),
    m_must_have_eof_checksum( true ),
    m_strict( true ),
    m_use_memory_mapped_io( true ),
    m_validate_metadata( true )
{
}

std::string FrameCPP::Verify::
StrError( FrameCPP::Verify::error_type ErrNum )
{
  if ( ( ErrNum >= NO_ERROR ) && ( ErrNum <= MAX_ERROR ) )
  {
    return m_error_strings[ ErrNum ];
  }
  std::ostringstream	msg;

  msg << "FrameCPP::Verify: Unknown error code: " << ErrNum;

  return msg.str( );
}

std::string FrameCPP::Verify::
StrErrorType( FrameCPP::Verify::error_type ErrNum )
{
  if ( ( ErrNum >= NO_ERROR ) && ( ErrNum <= MAX_ERROR ) )
  {
    return m_error_type_strings[ ErrNum ];
  }
  std::ostringstream	msg;

  msg << "FrameCPP::Verify: Unknown error_type code: " << ErrNum;

  return msg.str( );
}

FrameCPP::Verify::error_type FrameCPP::Verify::
operator()( const std::string& Filename )
{
  //---------------------------------------------------------------------
  // Local declarations
  //---------------------------------------------------------------------
  General::AutoArray<char>		io_buffer;

  //---------------------------------------------------------------------
  // Do initialization
  //---------------------------------------------------------------------
  m_error_info.str( "" );

  //---------------------------------------------------------------------
  // Open the frame file
  //---------------------------------------------------------------------
  General::ifstream ifile( Filename.c_str( ) );

  if( ! ifile.good( ) )
  {
    // An error has happened in opening the stream.
    return FILE_OPEN_ERROR;
  }

  //---------------------------------------------------------------------
  // Establish the buffer size
  //---------------------------------------------------------------------

  if ( m_buffer_size > 0 )
  {
    io_buffer.reset( new char[m_buffer_size] );
    ifile.setbuf( io_buffer.get( ), m_buffer_size );
  }
  ifile.rdbuf( )->UseMemoryMappedIO( m_use_memory_mapped_io );

  std::auto_ptr< FrameCPP::FrameFilename >	ff;
  if ( ValidateMetadata( ) )
  {
    ff.reset( new FrameCPP::FrameFilename( Filename.c_str( ) ) );
  }
  return this->operator()( ifile, ff );
}

FrameCPP::Verify::error_type FrameCPP::Verify::
operator()( std::istream& Stream,
	    std::auto_ptr< FrameCPP::FrameFilename > FrameFilename )
{
  //---------------------------------------------------------------------
  // Do initialization
  //---------------------------------------------------------------------
  error_type	retval = NO_ERROR;

  m_error_info.str( "" );

  //---------------------------------------------------------------------
  // Local parameter declarations
  //---------------------------------------------------------------------
  std::map< INT_2U, std::string >	class_names_by_id;
  std::map< std::string, INT_2U >	class_names_by_name;
  std::list< INT_2U >			class_id_index;
  std::list< std::string >		class_name_index;
  std::map< std::string, INT_2U >	class_names_to_class_id;
  bool 					seen_fr_end_of_file = false;

  class_names_to_class_id[ "FrSH" ] = FrBase::CLASS_FR_SH;

  //---------------------------------------------------------------------
  // Information needed to construct a dictionary
  //---------------------------------------------------------------------

  std::auto_ptr< Dictionary::definition_type >	def;
  Dictionary	dictionary( GetClassIO( ), CLASS_LAST_ENTRY );
  dictionary.TrackObjects( false );

  //---------------------------------------------------------------------
  // Create Input Frame Stream
  //---------------------------------------------------------------------

  try
  {
    IFrameStream	frame_stream( Stream, IFrameStream::MODE_SEQUENTIAL );
    
    //-------------------------------------------------------------------
    // Ensure that the steam is good.
    //-------------------------------------------------------------------
    if( ! frame_stream.good( ) )
    {
      // An error has happened in opening the stream.
      return FILE_OPEN_ERROR;
    }

    //-------------------------------------------------------------------
    // If strict, then ensure no part of the frame spec has been
    //   violated.
    //-------------------------------------------------------------------
    if ( m_strict )
    {
      try
      {
	check_magic_number( frame_stream.Get8ByteEndianness( ),
			    x123456789abcdef );
      }
      catch( ... )
      {
	return FRAME_SPEC_CONFORMANCE;
      }
    }

    //-------------------------------------------------------------------
    // Establish that CRC checksumming is requested
    //-------------------------------------------------------------------
    MD5Sum	md5;

    if ( m_check_md5sum )
    {
      // Enable checking of MD5Sum
      frame_stream.SetMD5Sum( md5 );
    }
    else
    {
      // Disable checking of MD5Sum
      frame_stream.SetMD5Sum( );
    }

    if ( m_check_file_checksum )
    {
      frame_stream.SetChecksumMethodFile( CheckSum::CRC );
    }
      
    bool 	calculating_frame_checksum = false ;
    FrBase	base( FrBase::CLASS_NULL );

    //-------------------------------------------------------------------
    // Read through all objects
    //-------------------------------------------------------------------
    while( frame_stream.good( ) )
    {
#if DEBUG_MD5SUM
      MD5Sum	md5_here( md5 );
      std::cerr << "DEBUG: pos: " << frame_stream.tellg( )
		<< " md5sum: " << md5_here
		<< std::endl
		<< std::flush;
#endif /* DEBUG_MD5SUM */
#if DEBUG_FILE_CHECKSUM
      CheckSum*	file_crc = frame_stream.GetCheckSumFile( ).Clone( );
      std::cerr << "DEBUG: pos: " << frame_stream.tellg( )
		<< " file crc: " << file_crc->value( )
		<< std::endl
		<< std::flush;
#endif /* DEBUG_FILE_CHECKSUM */
      if ( ( m_check_frame_checksum == true ) &&
	   ( calculating_frame_checksum == false ) )
      {
	// Clear the buffer and get read for the time when the checksum will
	// be calculated
	frame_stream.SetChecksumMethodFrame( CheckSum::CRC );
      }
      frame_stream.SetReadBoundry( frame_stream.tellg( ) +
				   base.Bytes( frame_stream.GetVersion( ) ) );
      base.read( frame_stream );
      
      if ( frame_stream.good( ) )
      {
	//---------------------------------------------------------------
	// Add any completed definitions to Dictionary
	//---------------------------------------------------------------
	if ( ( base.GetClassId( ) != FrBase::CLASS_FR_SE ) &&
	     ( def.get( ) ) )
	{
	  // Save the current definition
	  dictionary.AddDefinition( def.release( ) );
	}
	//-------------------------------------------------------------
	// Try to read in the structure using real read routines.
	//-------------------------------------------------------------
	try
	{
	  frame_stream.SetReadBoundry( frame_stream.tellg( ) +
				       base.Skip( frame_stream ) );
	  if ( ( frame_stream.GetVersion( ) < 6 ) &&
	       ( base.GetClassId( ) != FrBase::CLASS_NULL ) &&
	       ( base.GetClassId( ) != FrBase::CLASS_FR_SE ) &&
	       ( base.GetClassId( ) != FrBase::CLASS_FR_SH ) )
	  {
	    //:TODO: Need to set up calculation of CRC/MD5Sum for versions
	    //:TODO:   prior to Version 6 frame spec
	    INT_8U				here = frame_stream.tellg( );
	    INT_8U				len = base.Skip( frame_stream );
	    General::AutoArray< CHAR_U >	buffer( new CHAR_U[ len ] );;
	    bool				calc_checksum = true;

	    switch( dictionary.
		    MapStreamClassToLibraryClass( base.GetClassId( ) ) )
	    {
	    case CLASS_FR_END_OF_FILE:
	      calc_checksum = false;
	      break;
	    default:
	      calc_checksum = true;
	      break;
	    }

	    frame_stream.read( buffer.get( ), len, calc_checksum );
	    switch( dictionary.
		    MapStreamClassToLibraryClass( base.GetClassId( ) ) )
	    {
	    case CLASS_FR_END_OF_FILE:
	      {
		if ( frame_stream.IsCalculatingMD5Sum( ) )
		{
		  md5.Update( buffer.get( ), len );
		}
		if ( frame_stream.IsCalculatingChecksumFile() )
		{
		  INT_4U	offset = INT_4U( 0 );

		  offset += sizeof( INT_4U );	// NFrames
		  offset += sizeof( INT_4U );	// NBytes
		  offset += sizeof( INT_4U );	// mChkType
		  for ( unsigned int x = 0;
			x < sizeof( INT_4U );
			++x, ++offset )
		  {
		    buffer[ offset ] = '\0';
		  }
		  frame_stream.GetCheckSumFile( ).calc( buffer.get( ), len );
		}
#if DEBUG_MD5SUM
		MD5Sum	md5_here( md5 );
		std::cerr << "DEBUG: pos: " << frame_stream.tellg( )
			  << " md5sum: " << md5_here
			  << std::endl
			  << std::flush;
#endif /* DEBUG_MD5SUM */
#if DEBUG_FILE_CHECKSUM
		CheckSum*	file_crc =
		  frame_stream.GetCheckSumFile( ).Clone( );
		std::cerr << "DEBUG: pos: " << frame_stream.tellg( )
			  << " file crc: " << file_crc->value( )
			  << std::endl
			  << std::flush;
#endif /* DEBUG_FILE_CHECKSUM */
	      }
	      break;
	    default:
	      break;
	    }
		
	    Stream.seekg( here, std::ios::beg );
	  }
	  std::auto_ptr< FrBase >	obj( dictionary.Read( frame_stream,
							      base ) );
	  switch ( dictionary.
		   MapStreamClassToLibraryClass( base.GetClassId( ) ) )
	  {
	  case CLASS_FRAME_H:
	    // Start of frame.
	    calculating_frame_checksum = true;
	    if ( ValidateMetadata( ) )
	    {
	      const FrameH*
		frameh( dynamic_cast< const FrameH* >( obj.get( ) ) );
	      
	      if ( FrameFilename.get( ) != (FrameCPP::FrameFilename*)NULL )
	      {
		// Verify the start time
		GPSTime	ff_start( FrameFilename->G( ), 0 );
		GPSTime ff_end( ff_start + FrameFilename->T( ) );

		if ( ( ff_start > frameh->GetGTime( ) ) ||
		     ( ff_end < GPSTime( frameh->GetGTime( ) +
					 frameh->GetDt( ) ) ) )
		{
		  m_error_info << "  METADATA_VALID ERROR: Either the start time or the duration is invalid"
			       << std::endl;
		  retval = METADATA_INVALID;
		}
		//-------------------------------------------------------
		// Setup for Detector validation
		//-------------------------------------------------------
		m_detector_validation.resize( 0 );
		m_detector_validation_excess.resize( 0 );
		m_channel_validation.resize( 0 );
		m_channel_validation_excess.resize( 0 );
		std::string s( FrameFilename->S( ) );

		for ( std::string::const_iterator
			cur = s.begin( ),
			last  = s.end( );
		      cur != last;
		      ++cur )
		{
		  if ( *cur != 'Z' )
		  {
		    m_detector_validation[ *cur ] = 0;
		    m_channel_validation[ *cur ] = 0;
		  }
		}
	      } // if - FrameFilename
	      if ( frameh->GetGTime( ).GetLeapSeconds( ) !=
		   frameh->GetULeapS( ) )
	      {
		m_error_info << "  METADATA_VALID ERROR:"
			     << " Value of leap seconds is invalid for the frame's start time"
			     << " [ ULeapS:"
			     << " read: " << frameh->GetULeapS( )
			     << " expected: " << frameh->GetGTime( ).GetLeapSeconds( )
			     << " GPSTime: " << frameh->GetGTime( )
			     << "]"
			     << std::endl;
		retval = METADATA_INVALID;
	      }
	    } // if ( ValidateMetadata( ) )
	    break;
	  case CLASS_FR_ADC_DATA:
	    if ( m_check_data_valid )
	    {
	      const FrAdcData*
		adc( dynamic_cast< const FrAdcData* >( obj.get( ) ) );
	      if ( adc )
	      {
		if ( adc->GetDataValid( ) != 0 )
		{
		  std::ios_base::fmtflags saved = m_error_info.flags( );
		  
		  m_error_info << "  DATA_VALID ERROR: value: 0x"
			       << std::hex
			       << std::setw(4)
			       << std::setfill('0')
			       << adc->GetDataValid( )
			       << " channel: " << adc->GetName( )
			       << std::endl;
		  m_error_info.flags( saved );
		}
		retval = DATA_INVALID;
	      }
	    }
	    if ( ( ValidateMetadata( ) ) &&
		 ( FrameFilename.get( ) != (FrameCPP::FrameFilename*)NULL ) )

	    {
	      const FrAdcData*
		adc( dynamic_cast< const FrAdcData* >( obj.get( ) ) );
	      if ( adc )
	      {
		const std::string& name = adc->GetName( );
		if ( ( name.length( ) >= 3 ) &&
		     ( name[ 2 ] == ':' ) )
		{
		  set_channel_site( name[0] );
		}
	      }
	    }
	    break;
	  case CLASS_FR_DETECTOR:
	    if ( ( ValidateMetadata( ) ) &&
		 ( FrameFilename.get( ) != (FrameCPP::FrameFilename*)NULL ) )

	    {
	      const FrDetector*
		detector( dynamic_cast< const FrDetector* >( obj.get( ) ) );
	      if ( detector )
	      {
		const char*	d = detector->GetPrefix( );
		if ( ( *d == ' ' ) || ( *d == '\0' ) )
		{
		  d = DETECTOR_MAP::DetectorToPrefix( detector->GetName( ) );
		}
		detector_validation_type::iterator
		  pos = m_detector_validation.find( *d );
		if ( pos == m_detector_validation.end( ) )
		{
		  // Don't store this if the prefix is unknown
		  if ( !( *d == ' ' ) &&
		       !( *d == '\0' ) &&
		       !( *d == 'Z' ) )
		  {
		    m_detector_validation_excess[ *d ]++;
		  }
		}
		else
		{
		  (*pos).second++;
		}
	      }
	    }
	    break;
	  case CLASS_FR_END_OF_FILE:
	    {
	      const FrEndOfFile*
		eof( dynamic_cast< const FrEndOfFile* >( obj.get( ) ) );
		
	      if ( eof )
	      {
		seen_fr_end_of_file = true;
		
		if ( m_check_file_checksum )
		{
		  if ( eof->GetChkType( ) == FrameCPP::CheckSum::NONE )
		  {
		    if ( m_must_have_eof_checksum )
		    {
		      return NO_CHECKSUM;
		    }
		  }
		  else if ( eof->GetChkType( ) ==
			    (INT_4U)frame_stream.GetCheckSumFile( ).GetType( ) )
		  {
		    if ( eof->GetChkSum( ) !=
			 frame_stream.GetCheckSumFile( ).value( ) )
		    {
		      m_error_info << " CHECKSUM_ERROR: FR_END_OF_FILE: "
				   << " read: "
				   <<  eof->GetChkSum( )
				   << std::hex << std::showbase
				   << "(" << eof->GetChkSum( ) << ")"
				   << std::dec << std::noshowbase
				   << " computed: "
				   << frame_stream.GetCheckSumFile( ).value( )
				   << std::hex << std::showbase << "("
				   << frame_stream.GetCheckSumFile( ).value( )
				   << ")" << std::dec << std::noshowbase
				   << std::endl;
		      return CHECKSUM_ERROR;
		    }
		  }
		  else
		  {
		    return UNSUPPORTED_CHECKSUM_TYPE;
		    break;
		  }
		} // if m_check_file_checksum
	      } // eof
	    }
	    break;
	  case CLASS_FR_END_OF_FRAME:
	    {
	      if ( ( ValidateMetadata( ) ) &&
		   ( FrameFilename.get( ) != (FrameCPP::FrameFilename*)NULL ) )
	      {
		for ( detector_validation_type::const_iterator
			cur = m_detector_validation.begin( ),
			last = m_detector_validation.end( );
		      cur != last;
		      ++cur )
		{
		  if ( (*cur).second == 0 )
		  {
		    m_error_info << " METADATA_VALID ERROR:"
				 << " No detector information for site: "
				 << (*cur).first
				 << std::endl;
		    retval = METADATA_INVALID;
		  }
		}
		for ( detector_validation_type::const_iterator
			cur = m_detector_validation_excess.begin( ),
			last = m_detector_validation_excess.end( );
		      cur != last;
		      ++cur )
		{
		  m_error_info << " METADATA_VALID ERROR:"
			       << " Detector information missing in site component: "
			       << (*cur).first
			       << std::endl;
		  retval = METADATA_INVALID;
		}
		for ( detector_validation_type::const_iterator
			cur = m_channel_validation.begin( ),
			last = m_channel_validation.end( );
		      cur != last;
		      ++cur )
		{
		  if ( (*cur).second == 0 )
		  {
		    m_error_info << " METADATA_VALID ERROR:"
				 << " No channels refer to site: "
				 << (*cur).first
				 << std::endl;
		    retval = METADATA_INVALID;
		  }
		}
		for ( detector_validation_type::const_iterator
			cur = m_channel_validation_excess.begin( ),
			last = m_channel_validation_excess.end( );
		      cur != last;
		      ++cur )
		{
		  m_error_info << " METADATA_VALID ERROR:"
			       << " Channel references unlisted site: "
			       << (*cur).first
			       << std::endl;
		  retval = METADATA_INVALID;
		}
	      }
	      if ( m_check_frame_checksum )
	      {
		const FrEndOfFrame*
		  eof( dynamic_cast< const FrEndOfFrame* >( obj.get( ) ) );
		
		calculating_frame_checksum = false;
		if ( eof )
		{
		  if ( eof->GetChkType( ) == FrameCPP::CheckSum::NONE )
		  {
		    // Do nothing
		  }
		  else if ( eof->GetChkType( ) ==
			    (INT_4U)frame_stream.GetCheckSumFrame( ).GetType( ) )
		  {
		    if ( eof->GetChkSum( ) !=
			 frame_stream.GetCheckSumFrame( ).value( ) )
		    {
		      m_error_info << " CHECKSUM_ERROR: FR_END_OF_FRAME: expected: "
				   << frame_stream.GetCheckSumFrame( ).value( )
				   << " received: " <<  eof->GetChkSum( )
				   << std::endl;
		      return CHECKSUM_ERROR;
		    }
		  }
		  else
		  {
		    return UNSUPPORTED_CHECKSUM_TYPE;
		    break;
		  }
		}
	      }
	    }
	    break;
	  case FrBase::CLASS_FR_SE:
	    {
	      if ( def.get( ) )
	      {
		const FrSE* se( dynamic_cast< const FrSE* >( obj.get( ) ) );
		
		if ( se )
		{
		  def->elements.push_back( Dictionary::
					   element_type( se->GetName( ),
							 se->GetClass( ),
							 se->GetComment( ) ) );
		}
	      }
	    }
	    break;
	  case FrBase::CLASS_FR_SH:
	    {
	      // Start new definition
	      if ( def.get( ) )
	      {
		// Save the current definition
		dictionary.AddDefinition( def.release( ) );
	      }
	      // Create an FrSH
	      const FrSH* sh( dynamic_cast< const FrSH* >( obj.get( ) ) );
		
	      if ( sh )
	      {
		def.reset( new Dictionary::definition_type( sh->GetName( ),
							    sh->GetClass( ),
							    sh->GetComment( ) ) );
		class_names_by_id[ sh->GetClass( ) ] = sh->GetName( );
		class_id_index.push_back( sh->GetClass( ) );
		class_names_by_name[ sh->GetName( ) ] = sh->GetClass( );
		class_name_index.push_back( sh->GetName( ) );
	      }
	    }
	    break;
	  case CLASS_FR_PROC_DATA:
	    if ( ( ValidateMetadata( ) ) &&
		 ( FrameFilename.get( ) != (FrameCPP::FrameFilename*)NULL ) )

	    {
	      const FrProcData*
		channel( dynamic_cast< const FrProcData* >( obj.get( ) ) );
	      if ( channel )
	      {
		const std::string& name = channel->GetName( );
		if ( ( name.length( ) >= 3 ) &&
		     ( name[ 2 ] == ':' ) )
		{
		  set_channel_site( name[0] );
		}
	      }
	    }
	    break;
	  case CLASS_FR_SIM_DATA:
	    if ( ( ValidateMetadata( ) ) &&
		 ( FrameFilename.get( ) != (FrameCPP::FrameFilename*)NULL ) )

	    {
	      const FrSimData*
		channel( dynamic_cast< const FrSimData* >( obj.get( ) ) );
	      if ( channel )
	      {
		const std::string& name = channel->GetName( );
		if ( ( name.length( ) >= 3 ) &&
		     ( name[ 2 ] == ':' ) )
		{
		  set_channel_site( name[0] );
		}
	      }
	    }
	    break;
	  case CLASS_FR_VECT:
	    {
	      FrVect* v( dynamic_cast< FrVect* >( obj.get( ) ) );
	      v->GetData( );
	    }
	    break;
	  }
	  
	}
	catch( const std::exception& e )
	{
	  m_error_info << "  INVALID_FRAME_STRUCTURE: " << e.what( );
	  return INVALID_FRAME_STRUCTURE;
	}
	catch( ... )
	{
	  return INVALID_FRAME_STRUCTURE;
	}
      }
    } // while freme_stream.good
#if DEBUG_MD5SUM
    std::cerr << "DEBUG: MD5Sum: " << md5 << std::flush;
    if ( FrameFilename.get( ) != (FrameCPP::FrameFilename*)NULL )
    {
      std::cerr << "  "
		<< FrameFilename->S( ) << "-"
		<< FrameFilename->D( ) << "-"
		<< FrameFilename->G( ) << "-"
		<< FrameFilename->T( )
		<< FrameFilename->Ext( );
    }
    std::cerr << std::endl;
#endif /* DEBUG_MD5SUM */
  }
  catch( const std::exception& e )
  {
    m_error_info << "  FILE_OPEN_ERROR: " << e.what( );
    return FILE_OPEN_ERROR;
  }
  catch( ... )
  {
    return FILE_OPEN_ERROR;
  }
  //---------------------------------------------------------------------
  // Ensure that the end of file structure was seen
  //---------------------------------------------------------------------
  if ( seen_fr_end_of_file == false )
  {
    return FILE_TRUNCATION;
  }
  return retval;
}
