#ifndef FRAME_CPP_INTERFACE__VERIFY_HH
#define FRAME_CPP_INTERFACE__VERIFY_HH
#include "framecpp/config.h"

#include <memory>
#include <sstream>
#include <string>

#include "general/types.hh"
#include "general/hash_map.hh"
#include "framecpp/FrameFilename.hh"

namespace FrameCPP
{
  class Verify
  {
  public:
    enum error_type {
      NO_ERROR = 0,
      CHECKSUM_ERROR,
      FILE_OPEN_ERROR,
      UNSUPPORTED_FRAME_SPEC,
      FRAME_SPEC_CONFORMANCE,
      UNSUPPORTED_CHECKSUM_TYPE,
      NO_CHECKSUM,
      INVALID_FRAME_STRUCTURE,
      FILE_TRUNCATION,
      DATA_INVALID,
      METADATA_INVALID,
      MAX_ERROR = METADATA_INVALID
    };

    //: Default constructor
    Verify( );

    //: Query the size of the I/O buffer
    INT_8U BufferSize( ) const;

    //: Change the size of the I/O buffer
    void   BufferSize( INT_8U Bytes );

    //: Query if the data valid field of the FrAdcData must be zero
    bool CheckDataValid( ) const;

    //: Change requirement if the data valid field of the FrAdcData must be zero
    void CheckDataValid( bool Check );

    //: Query if checksums will be calculated for a frame file
    bool CheckFileChecksum( ) const;

    //: Change requirement for calculating the checksum of a frame file
    void CheckFileChecksum( bool Check );

    //: Query if checksums will be calculated for the frames of a frame file
    bool CheckFrameChecksum( ) const;

    //: Change requirement for calculating the checksum of frames of a frame file
    void CheckFrameChecksum( bool Check );

    //: Query if md5sum will be calculated for the frame file
    bool CheckMD5Sum( ) const;

    //: Change requirement for calculating the md5sum value for the frame file
    void CheckMD5Sum( bool Chekc );

    //: Retrieve additional infomation about an error.
    std::string ErrorInfo( ) const;

    //: Query if the frame file must have the EOF checksum structure.
    bool MustHaveEOFChecksum( ) const;

    //: Change requirement of the frame file having the EOF checksum structure
    void MustHaveEOFChecksum( bool Confomance );

    //: Query strict conformance requirement
    bool Strict( ) const;

    //: Change strict conformance requirement
    void Strict( bool Strictness );

    //: Return string describing error code
    static std::string StrError( error_type ErrorNum );

    //: Retrieve the string representation of the error_type
    static std::string StrErrorType( const error_type Error );

    //: Query if memory mapped I/O will be used
    bool UseMemoryMappedIO( ) const;

    //: Specify if to use memory mapped I/O.
    void UseMemoryMappedIO( bool UseMemoryMappedIO );

    //: Query if metadata should be validated
    bool ValidateMetadata( ) const;

    //: Specify if to validate metadata
    void  ValidateMetadata( bool ValidateMetadata );

    //: Verify a frame file
    error_type operator()( const std::string& Filename );

    //: Verify a frame stream
    error_type operator()( std::istream& Stream,
			   std::auto_ptr< FrameCPP::FrameFilename > FrameFilename = std::auto_ptr< FrameCPP::FrameFilename >() );

  private:
    typedef General::hash_map< char, int >	detector_validation_type;

    static const char* m_error_strings[];
    static const char* m_error_type_strings[];

    detector_validation_type	m_detector_validation;
    detector_validation_type	m_detector_validation_excess;

    detector_validation_type	m_channel_validation;
    detector_validation_type	m_channel_validation_excess;

    INT_8U		m_buffer_size;
    bool		m_check_data_valid;
    bool		m_check_file_checksum;
    bool		m_check_frame_checksum;
    bool		m_check_md5sum;
    std::ostringstream	m_error_info;
    bool		m_must_have_eof_checksum;
    bool		m_strict;
    bool		m_use_memory_mapped_io;
    bool		m_validate_metadata;

    void set_channel_site( char Site );
  };

  //!retrun: INT_8U - return size in bytes of the I/O buffer.
  inline INT_8U Verify::
  BufferSize( ) const
  {
    return m_buffer_size;
  }

  //!param: INT_8U - establish the size in bytes of the I/O buffer. A
  //+	value of 0 sets the buffer size to the system specified default.
  inline void Verify::
  BufferSize( INT_8U Bytes )
  {
    m_buffer_size = Bytes;
  }

  //!return: bool - state of checking that the data valid field is zero
  inline bool Verify::
  CheckDataValid( ) const
  {
    return m_check_data_valid;
  }

  //!param: bool Check - true if the data valid field must be zero, false
  //+	otherwise.
  inline void Verify::
  CheckDataValid( bool Check )
  {
    m_check_data_valid = Check;
  }

  //!return: bool - true if checksums for frame files need to be
  //+	performed, false otherwise.
  inline bool Verify::
  CheckFileChecksum( ) const
  {
    return m_check_file_checksum;
  }

  //!param: bool Check - true if checksums for frame files need to be
  //+	performed, false otherwise.
  inline void Verify::
  CheckFileChecksum( bool Check )
  {
    m_check_file_checksum = Check;
  }

  //!return: bool - true if checksums for individual frames need to be
  //+	performed, false otherwise.
  inline bool Verify::
  CheckFrameChecksum( ) const
  {
    return m_check_frame_checksum;
  }

  //!param: bool Check - true if checksums for individual frames need to be
  //+	performed, false otherwise.
  inline void Verify::
  CheckFrameChecksum( bool Check )
  {
    m_check_frame_checksum = Check;
  }

  //!return: bool - true if md5sum for the frame file will be
  //+	performed, false otherwise.
  inline bool Verify::
  CheckMD5Sum( ) const
  {
    return m_check_md5sum;
  }

  //!param: bool Check - true if md5sum for the frame file should be
  //+	performed, false otherwise.
  inline void Verify::
  CheckMD5Sum( bool Check )
  {
    m_check_md5sum = Check;
  }

  //!return: const std::string&  - Additional error information
  inline std::string Verify::
  ErrorInfo( ) const
  {
    return m_error_info.str( );
  }

  //!return: bool - true if the frame file must have the EOF checksum
  //+	 structure, false otherwise
  inline bool Verify::
  MustHaveEOFChecksum( ) const
  {
    return m_must_have_eof_checksum;
  }

  //!param: bool Conformance - true if the frame file must have the EOF
  //+	  checksum structure, false otherwise
  inline void Verify::
  MustHaveEOFChecksum( bool Confomance )
  {
    m_must_have_eof_checksum = Confomance;
  }

  //!return: bool - state of strictness. True for strict conformance to
  //+	 the frame spec; false otherwise.
  inline bool Verify::
  Strict( ) const
  {
    return m_strict;
  }

  //!param: bool Strictness - true if the frame file must conform to all
  //+	details of the frame spec; false otherwise.
  inline void Verify::
  Strict( bool Strictness )
  {
    m_strict = Strictness;

  }
  //!return: bool - true if memory mapped I/O will be used
  inline bool Verify::
  UseMemoryMappedIO( ) const
  {
    return m_use_memory_mapped_io;
  }

  //!param: bool UseMemoryMappedIO - true if memory mapped I/O should be
  //+	 used; false otherwise.
  inline void Verify::
  UseMemoryMappedIO( bool UseMemoryMappedIO )
  {
    m_use_memory_mapped_io = UseMemoryMappedIO;
  }

  //!return: bool - true if metadata should be validated
  inline bool Verify::
  ValidateMetadata( ) const
  {
    return m_validate_metadata;
  }

  //!param: bool ValidateMetadata - true if frame metadata to be verified against
  //+	 against the filename
  inline void Verify::
  ValidateMetadata( bool ValidateMetadata )
  {
    m_validate_metadata = ValidateMetadata;
  }

  inline void Verify::
  set_channel_site( char Site )
  {
    detector_validation_type::iterator
      pos = m_channel_validation.find( Site );
    if ( pos == m_channel_validation.end( ) )
    {
      m_channel_validation_excess[ Site ]++;
    }
    else
    {
      (*pos).second++;
    }
  }
}

#endif /* FRAME_CPP_INTERFACE__VERIFY_HH */
