#ifndef FrameCppAdcDataHH
#define FrameCppAdcDataHH


#include "framecpp/config.h"
#include "framecpp/Version3_4_5/base.hh"
#include "framecpp/Version3_4_5/sh.hh"
#include "framecpp/Version3_4_5/vect.hh"
#include "framecpp/Version3_4_5/time.hh"

#include "framecpp/FrAdcDataBase.hh"
#include "framecpp/SearchContainer.hh"


namespace FrameCPP {

namespace Version_6
{
  class FrAdcData;
}

namespace Version_3_4_5 {


class FrameReaderBase;

//-----------------------------------------------------------------------------
//
//: ADC Data
//
// This class represents ADC data.  It is modeled after the FrAdcData
// structure in the Frame Specification, LIGO-T970130-B-E.
//
// <P>The AdcData class encapsulates the following attributes for an ADC
// channel:
//
// <table border=2 cellpadding=4>
//   <tr><td> <B>Attribute</B></td><td><B>Description</B>                        </td></tr>
//   <tr><td> name         </td><td>Channel name                                 </td></tr>
//   <tr><td> comment      </td><td>Comment                                      </td></tr>
//   <tr><td> channelGroup </td><td>ADC group number                             </td></tr>
//   <tr><td> channelNumber</td><td>Channel number                               </td></tr>
//   <tr><td> nBits        </td><td>Number of bits in A/D output                    </td></tr>
//   <tr><td> bias         </td><td>DC bias on channel (Volts @ ADC_counts = 0)     </td></tr>
//   <tr><td> slope        </td><td>ADC calibration: input units/count              </td></tr>
//   <tr><td> units        </td><td>Adc calibrartion: input units for slope         </td></tr>
//   <tr><td> sampleRate   </td><td>Data acquisistion rate, samples / sec           </td></tr>
//   <tr><td> timeOffset   </td><td>For triggered data lasting less than one frame, the
//                          integer seconds & nanoseconds start time relative to
//                          frame start.                                    </td></tr>
//   <tr><td> fShift       </td><td>Frequency shift if signal has been heterodyned before
//                          ADC:<p>
//                          fShift = (f_heterodyne - fNyquist@sampleRate)   </td></tr>
//   <tr><td> phase        </td></td> Phase of heterodyning signal at start of data set   </td></tr>
//   <tr><td> dataValid    </td><td>Data valid flag: dataValid = 0 means data is valid,
//                          non-zero implies data is suspect/not valid.     </td></tr>
// </table>
//
// <p>This data is available through accessor methods.  The data can't be
// modified, however.  The only exceptions to this are the comment field (which
// can be appended to), and the dataValid flag.
//
// <p>The AdcData object is also a container for two types of data:
//
// <table border=2 cellpadding=4>
//   <tr><td> <B>Data</B></td><td><B>Description</B>                             </td></tr>
//   <tr><td> data</td><td>A container of Frame Vect objects which contain the sampled
//                 data.                                                 </td></tr>
//   <tr><td> aux </td><td>A container of Vect objects holding user-provided
//                 information. Its use is generic.                      </td></tr>
// </table>
//
// <p>The data is accessed through the container objects inside AdcData.  These
// are accessed via the <code>refData</code> and <code>refAux</code> methods.
//
//!todo: What about adding mutators?
//
class AdcData : public Base, public FrAdcDataBase
{
public:

    /* Container Macros */
    //!ignore_begin:

    DECLARE_DEFAULT_SIMPLE_CONTAINER( Data, data, FrameCPP::Version_3_4_5::Vect );
    DECLARE_SIMPLE_CONTAINER( Aux, aux, Vect );
    //!ignore_end:

    /* Constructors/Destructor */
    //-----------------------------------------------------------------------------
    //
    //: Default Constructor.
    //
    // This creates an AdcData object with the following properties:
    // <ul type=disc>
    //     <li> name = "none" </li>
    //     <li> comment = "" </li>
    //     <li> groupNumber = 0 </li>
    //     <li> channelNumber = 0 </li>
    //     <li> nBits = 8 </li>
    //     <li> bias = 0 </li>
    //     <li> slope = 1.0 </li>
    //     <li> units = "" (dimensionless) </li>
    //     <li> sampleRate = 0.0 </li>
    //     <li> timeOffset = 0 seconds, 0 nanoseconds </li>
    //     <li> fShift = 0.0 </li>
    //     <li> Phase = 0.0 </li>
    //     <li> dataValid = 0 </li>
    //     <li> No other data. </li>
    // </ul>
    //
    //!exc: None.   
    AdcData();

    //-----------------------------------------------------------------------------
    //
    //: Copy Constructor.
    //
    // A deep copy is performed for owned objects.
    //
    //!param: const AdcData& adcData - Object to be copied.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    AdcData( const AdcData& adcData );

    //-----------------------------------------------------------------------------
    //
    //: Constructor.
    //
    // This creates an AdcData object with the specified properties.
    // Additionally, the comment is set to "" (no comment).
    //
    //!param: const std::string& name - Chanel name
    //!param: INT_4U channelGroup - Channel groups
    //!param: INT_4U channelNumber - Channel number
    //!param: INT_4U nBits - Number of bits in A/D output
    //!param: const REAL_8& sampleRate - Data acquisition rate, samples/sec.
    //!param: REAL_4 bias - DC Bias on channel.  Default: 0.0.
    //!param: REAL_4 slope - Input units/count.  Default: 1.0.
    //!param: const std::string& units - Units for slope.  Default: "counts".
    //!param: const REAL_8& fShift - Frequency shift if signal has been
    //+     heterodyned before ADC.  Default: 0.0.
    //!param: const REAL_4& phase - Heterodyning signal phase. Default: 0.0.
    //!param: const Time& timeOffset - Time offset relative to frame start.
    //+     Default: Time( 0, 0 ) (no offset).
    //!param: bool dataValid - Data valid flag.  Default: 0 (data is valid)
    //
    //!exc: None.
    //
    AdcData( const std::string& name, INT_4U group, INT_4U channel, INT_4U nBits,
             const REAL_8& sampleRate, REAL_4 bias = 0.0, REAL_4 slope = 1.0,
             const std::string& units = "counts", const REAL_8& fShift = 0.0,
             const Time& timeOffset = Time( 0, 0 ), INT_2U dataValid = 0,
	     const REAL_4& phase = 0.0 );


    /* Operator Overloads */
    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    // A deep copy is performed for owned objects.
    //
    //!param: const AdcData& adcData - Obejct to be assigned.
    //
    //!return: const AdcData& -- This object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const AdcData& operator=( const AdcData& adcData );

    //-----------------------------------------------------------------------------
    //
    //: Equal comparison.
    //
    // Determines whether two AdcData objects are equal. For this comparison,
    // all of the data is compared.
    //
    //!param: const AdcData& adcData - The object to compare with.
    //
    //!return: bool -- true if the objects are equal.
    //
    //!kw: const
    //
    //!exc: None.
    bool operator==( const AdcData& adcData ) const;
    bool operator!=( const AdcData& adcData ) const;

    //
    //: += Assignment Operator.
    //
    //!param: const AdcData& adcData -
    //
    //!return: const AdcData&
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: frame_mismatch - ADC mismatch
    //
    const AdcData& operator+=( const AdcData& adcData );
    
    /* Required Methods */
    virtual INT_2U getClassId() const;

    //-----------------------------------------------------------------------------
    //
    //: Structure Header
    //
    // Constructs the Structure Header for this class. 
    //
    //!return: SH -- ADC data structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH();

    //-----------------------------------------------------------------------------
    //
    //: Structure Header
    //
    // Constructs the data format version 4 Structure Header for this class. 
    //
    //!return: SH -- Structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH4();

    //-----------------------------------------------------------------------------
    //
    //: Structure Header
    //
    // Constructs the data format version 3 Structure Header for this class. 
    //
    //!return: SH -- Structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH3();

    /* General Accessors */

    /* gcc 2.95.1 can't handle a typedef with this if the throw spec is
       present (e.g.
       typedef SearchContainer< AdcData, &AdcData::getName > dc; ).
       
       const std::string& getName() const throw(); */
    const std::string& getName() const;
    const std::string& getComment() const;
    INT_4U getChannelGroup() const;
    INT_4U getChannelNumber() const;
    INT_4U getNBits() const;
    REAL_4 getBias() const;
    REAL_4 getSlope() const;
    const std::string& getUnits() const;
    REAL_8 getSampleRate() const;
    const Time& getTimeOffset() const;
    REAL_8 getFShift() const;
    REAL_4 getPhase() const;
    INT_2U getDataValid() const;

    //
    //: Get data valid flag (known as OverRange before)
    //
    // Gets the data valid flag.
    //
    // <p>dataValid = 0 -> ADC data valid.
    // <br>dataValid != 0 -> ADC data suspect/not valid.
    //
    //!return: INT_2U -- Data valid flag.
    //
    //!kw: inline const
    //
    //!exc: None.
    INT_2U getOverRange() const;
    INT_2U getInstance() const;

    /* Data Accessors */
    FrameCPP::Version_3_4_5::AdcData::DataContainer& refData();
    const AdcData::DataContainer& refData() const;

    /* Aux Accessors */
    AdcData::AuxContainer& refAux();
    const AdcData::AuxContainer& refAux() const;

    /* Mutators */
    void appendComment( const std::string& comment );
    void setDataValid( INT_2U flag );
    void setBias( REAL_4 bias );
    void setSlope( REAL_4 slope );
    void setTimeOffset( const Time& time );

private:
  friend class Dictionary;
  friend class FrameReadPlan;
  friend class FrameCPP::Version_6::FrAdcData;

    //-----------------------------------------------------------------------------
    //
    //: Reads an AdcData object from an FrameReaderBase object.
    //
    // This method
    // creates a dynamically allocated object.  This method is called indirectly
    // by the FrameReader class.  This is a private method, access is granted
    // through the ObjectProxy< AdcData > class which is a friend.
    //
    // <p>*NOTE* This method allocates memory and returns a pointer to it.  The
    // user is responsible for freeing this memory.
    //
    //!param: FrameReaderBase& in - Input data stream object.
    //
    //!return: AdcData* -- Newly allocated.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failed.
    //
    static AdcData* read( FrameReaderBase& in );
    //: Version 4 frame reading method.
    static AdcData* read4( FrameReaderBase& in );
    //: Version 3 frame reading method.
    static AdcData* read3( FrameReaderBase& in );

    //-----------------------------------------------------------------------------
    //
    //: Write an AdcData object
    //
    // Writes an AdcData object to an Output object.
    //
    //!param: Output& out out - Output data stream object.
    //
    //!exc: write_failure - Write failure occured.
    //
    virtual void write( Output& out ) const;

private:
    //: Channel name.
    std::string mName;
    //: Comment
    std::string mComment;
    //: Channel group
    INT_4U mChannelGroup;
    //: Channel number
    INT_4U mChannelNumber;
    //: Number of bits in A/D output
    INT_4U mNBits;
    //: DC Bias on channel
    REAL_4 mBias;
    //: Slope: Input units/count
    REAL_4 mSlope;
    //: Units for slope
    std::string mUnits;
    //: Acquisition rate, samples/sec
    REAL_8 mSampleRate;
    //: Time offset from frame start
    Time mTimeOffset;
    //: Frequency shift (if heterodyned before ADC)
    REAL_8 mFShift;
    //: Phase of heterodyning signal at start of dataset
    REAL_4 mPhase;
    //: Data valid flag.
    INT_2U mDataValid;
    //: Adc data.
    FrameCPP::Version_3_4_5::Container< FrameCPP::Version_3_4_5::Vect > mData;
    //: User-provided data
    FrameCPP::Version_3_4_5::Container< FrameCPP::Version_3_4_5::Vect > mAux;
    //: This object's instance number (when read from file)
    INT_2U mInstance;
};    



/******************/
/* Inline Methods */
/******************/


//-----------------------------------------------------------------------------
//
//: Not-equal comparison.
//
// Determines whether two adc objects are not equal.  All contained data is
// compared.
//
//!param: const AdcData& adcData - The object to compare with.
//
//!return: bool -- True if the objects are not equal.
//
//!kw: inline const
//
//!exc: None.
//   
inline bool AdcData::operator!=( const AdcData& adcData ) const
{
    return !( (*this) == adcData );
}


//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Comment
//
// Gets the comment.
//
//!return: const std::string& -- Comment.
//
//!exc: None.
//      
inline const std::string& AdcData::getComment() const
{
    return mComment;
}


//-----------------------------------------------------------------------------
//
//: Group number
//
// Gets channel group number.
//
//!return: INT_4U -- Group number.
//
//!kw: inline const
//
//!exc: None.
//      
inline INT_4U AdcData::getChannelGroup() const
{
    return mChannelGroup;
}


//-----------------------------------------------------------------------------
//
//: Channel number
//
// Gets the channel number.
//
//!return: INT_4U -- Channel number.
//
//!kw: inline const
//
//!exc: None.
//      
inline INT_4U AdcData::getChannelNumber() const
{
    return mChannelNumber;
}


//-----------------------------------------------------------------------------
//
//: Number of bits
//
// Gets the number of bits in A/D output.
//
//!return: INT_4U -- Number of bits.
//
//!kw: inline const
//
//!exc: None.
//      
inline INT_4U AdcData::getNBits() const
{
    return mNBits;
}


//-----------------------------------------------------------------------------
//
//: Bias
//
// Gets the bias.  This is the DC bias on channel (Volts @ ADC_counts = 0).
//
//!return: REAL_4 -- Number of bits.
//
//!kw: inline const
//
//!exc: None.
//      
inline REAL_4 AdcData::getBias() const
{
    return mBias;
}


//-----------------------------------------------------------------------------
//
//: Slope
//
// Gets the slope.  ADC calibration: input units/count.
//
//!return: REAL_4 -- Slope.
//
//!kw: inline const
//
//!exc: None.
//      
inline REAL_4 AdcData::getSlope() const
{
    return mSlope;
}


//-----------------------------------------------------------------------------
//
//: Slope Units
//
// Gets the units for the slope.
//
//!return: const std::string& -- Slope units.
//
//!kw: inline const
//
//!exc: None.
//      
inline const std::string& AdcData::getUnits() const
{
    return mUnits;
}


//-----------------------------------------------------------------------------
//
//: Sample Rate
//
// Returns the data aquisition rate, samples/sec.
//
//!return: REAL_8 -- Sampling rate.
//
//!kw: inline const
//
//!exc: None.
//      
inline REAL_8 AdcData::getSampleRate() const
{
    return mSampleRate;
}


//-----------------------------------------------------------------------------
//
//: Time Offset
//
// Returns the time offset. For triggered data lasting less than one frame,
// this is the integer seconds and nanoseconds start time relative to the
// frame start.
//
//!return: const Time& -- Time offset.
//
//!kw: inline const
//
//!exc: None.
//      
inline const Time& AdcData::getTimeOffset() const
{
    return mTimeOffset;
}


//-----------------------------------------------------------------------------
//
//: Phase of heterodyning signal at start of dataset
//
// In the heterodyning process the real time series is multiplied by
// cos[2*PI*fShift(t-t0)+phase] to get the real part and by
// -sin[2*PI*fShift(t-t0)+phase] (note the minus sign) to get the 
// imaginary part of the resulting complex time series.
// The time origin t0 is the beginning of the frame.
//
//!return: REAL_4 -- Phase.
//
//!kw: inline const
//
//!exc: None.
//      
inline REAL_4 AdcData::getPhase() const
{
    return mPhase;
}

//-----------------------------------------------------------------------------
//
//: Frequency shift
//
// This is the frequency shift if the signal has been heterodyned before the
// ADC:
//
// fShift = (f_heterodyne - fNyquist@sampleRate).
//
//!return: REAL_8 -- Frequency shift.
//
//!kw: inline const
//
//!exc: None.
//      
inline REAL_8 AdcData::getFShift() const
{
    return mFShift;
}


//-----------------------------------------------------------------------------
//
//: Data valid flag
//
// Gets the data valid flag.
//
// <p>dataValid = 0 -> ADC data valid.
// <br>dataValid != 0 -> ADC data suspect/not valid.
//
//!return: INT_2U -- Data valid flag.
//
//!kw: inline const
//
//!exc: None.
//      
inline INT_2U AdcData::getDataValid() const
{
    return mDataValid;
}

   
inline INT_2U AdcData::getOverRange() const
{
   return getDataValid();
}
   

//-----------------------------------------------------------------------------
//
//: Instance number.
//
// Gets ADC structure instance (valid only if frame read from disk)
//
//!return: INT_2U -- Instance number.
//
//!kw: inline const
//
//!exc: None.
//      
inline INT_2U AdcData::getInstance() const
{
    return mInstance;
}


//-----------------------------------------------------------------------------
//
//: Auxiliary Info container
//
// Get the container which contains user-provided information.
//
//!return: SearchContainer< Vect, &Vect::getName >& -- Aux data.
//
//!kw: inline
//
//!exc: None.
//      
inline AdcData::AuxContainer& AdcData::refAux()
{
    return mAux;
}


//-----------------------------------------------------------------------------
//
//: Auxiliary Info container
//
// Get the container which contains user-provided information.
//
//!return: const SearchContainer< Vect, &Vect::getName >& -- Aux data.
//
//!kw: inline const
//
//!exc: None.
//      
inline const
AdcData::AuxContainer& AdcData::refAux() const
{
    return mAux;
}


//-----------------------------------------------------------------------------
//
//: Data Container
//
// Get the container which contains the sampled data.
//
//!return: SearchContainer< Vect, &Vect::getName >& -- Adc data.
//
//!kw: inline
//
//!exc: None.
//      
inline AdcData::DataContainer& AdcData::refData()
{
    return mData;
}

//-----------------------------------------------------------------------------
//
//: Data Container
//
// Get the container which contains the sampled data.
//
//!return: const SearchContainer< Vect, &Vect::getName >& -- Adc data.
//
//!kw: inline const
//
//!exc: None.
//      
inline const
AdcData::DataContainer& AdcData::refData() const
{
    return mData;
}



//-----------------------------------------------------------------------------
// Mutators
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Appends a comment.
//
// This method adds a comment followed by a newline to the
// end of the current list of comments.
//
//!param: const std::string& comment - The comment to append.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
//!kw: inline
//
inline void AdcData::appendComment( const std::string& comment )
{
    if ( mComment.size() > 0 ) {
	 mComment += "\n" + comment;
    } else
	 mComment = comment;
}


//-----------------------------------------------------------------------------
//
//: Set the data valid flag.
//
//!param: INT_2U flag - Set if data is suspect.
//
//!kw: inline
//
//!exc: None.
//      
inline void AdcData::setDataValid( INT_2U flag )
{
  mDataValid = flag;
}


//-----------------------------------------------------------------------------
//
//: Set the bias
//
//!param:  bias - Signal conversion offset.
//
//!kw: inline
//
//!exc: None.
//      
inline void AdcData::setBias( REAL_4 bias )
{
  mBias = bias;
}


//-----------------------------------------------------------------------------
//
//: Set the slope
//
//!param:  slope - Signal conversion linear coefficient.
//
//!kw: inline
//
//!exc: None.
//      
inline void AdcData::setSlope( REAL_4 slope )
{
  mSlope = slope;
}

 
//-----------------------------------------------------------------------------
//
//: Set time offset
//
//!param:  time - Time offset.
//
//!kw: inline
//
//!exc: None.
//      
inline void AdcData::setTimeOffset( const Time& time )
{
  mTimeOffset = time;
}

} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // FrameCppAdcDataHH
