#ifndef FrameCppContainerBaseHH
#define FrameCppContainerBaseHH


#include "framecpp/config.h"
#include "errors.hh"

namespace FrameCPP {

namespace Version_3_4_5 {

class Base;
class Output;


//-----------------------------------------------------------------------------
//
//: The base class for all containers.
//
// <p>Most of the LIGO Frame components are containers which contain other
// frame components.  For example, the RawData object is a container for 
// AdcData, SerData, log messages, and other information.  In a frame file,
// these containers are implemented as linked-lists.  However, in the FrameCPP
// library these are implemented by a container class, "Container".  This class
// standardizes the interface and maximizes code re-use.
//
// <p>ContainerBase is the base class for these container objects.  The
// container classes themselves are templates, so this class allows the user to
// perform queries on the containers which do not require knowledge of the
// template type (the template type specifies what Frame class is stored in
// the container).
//
// <p>The methods defined are (they are all pure virtual):
// <ul type=disc>
//     <li> write - Enable the container to be written to an output object.
//          This is used by the Frame Object classes when they are being
//          written.  It is not meant to be called by the user. </li>
//     <li> operator[] - Return a pointer to a contained object.  The parameter
//          to this method is a zero-offset index into the container.  The
//          object is returned as a pointer to Base (the base class for all
//          frame classes) by index. </li>
//     <li> size - Returns the number of elements in the container. </li>
// </ul>
//
// <p> The destructor is also declared virtual so that the container can be
// destructed with only a pointer to the ContainerBase parent.
//
//!todo: Implement iterators which dereference to Base*'s.  This will allow the
//!todo: STL iterator model to be used with this class.
//
class ContainerBase
{
public:

    //: Destructor
    virtual ~ContainerBase();

    /* Operator Overloads  */
    //: Container indexing.
    //!exc: range_error
    virtual Base* operator[]( unsigned int index ) = 0;
    //: Container indexing.
    //!exc: range_error   
    virtual const Base* operator[]( unsigned int index ) const = 0;

    /* Accessors */
    //: Get container size.
    //!exc: None.
    virtual unsigned int size() const = 0;
};

   
} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // FrameCppContainerBaseHH
