#include "framecpp/config.h"

#include <math.h>
#include "daqframe.hh"
#include <dimension.hh>

using namespace FrameCPP::Version_3_4_5;


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
//!exc: None.
//   
DaqFrame::DaqFrame()
  : Frame(), Input(), catHint(0)
{
}


void DaqFrame::
TriggerADC( FrAdcDataBase* ADC, bool Activate )
{
  AdcData* adc( dynamic_cast< AdcData* >( ADC ) );
  if ( adc )
  {
    daqTriggerADC( adc, Activate, false /* duplicates */ );
  }
}

//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
// This performs a deep copy of owned elements.
//
//!param: const Frame& frame - Object to be copied.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
DaqFrame::DaqFrame( const Frame& frame )
  : Frame( frame ), Input()
{
}


//-----------------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Destructor
//
//!exc: None.
//
DaqFrame::~DaqFrame()
{
}


//-----------------------------------------------------------------------------
//
//: Concatenate active ADCs in this frame with another's from `in'
//
// For use with the DAQ system. Cats active ADC channels with data from the `in' stream.
// Input stream is checked to be a frame file with the same configuration as
// this one. This is done by checking the run number, which is incremented
// when the DAQ system is reconfigured.
//
// :TODO: should check if input file size changed and throw "cannot_update" exception if it did
//
//!param: std::istream* in - frame file, random access
//
//!exc: read_failure - Read failed.
//!exc: cannot_update - Frame configuration changed.
//!exc: cannot_cat - Data is not contiguous in time.
//
void DaqFrame::daqCat( std::istream* in )
{
  mStream = in;

  // check the run number, it must be the same.
  // call to daqFrameRun() triggers file header loading in FrameCPP::Version_3_4_5::Input
  if ( getRun() != daqFrameRun() )
    {
      throw cannot_update ( "Run number changed" );
    }

  // check continuity of data in time
  //
  if ( Time( INT_4S( mDt ) + mGTime.getSec( ), mGTime.getNSec( ) ) !=
       Time(daqFrameGTimeS( ), daqFrameGTimeN( ) ) )
    {
      throw cannot_cat ( "Data is not contiguous" );
    }

  REAL_8 deltaT = daqFrameDt();
  mDt += deltaT;


  // cat active ADCs
  //
  for ( unsigned int i=0 ; i < adcUpdateVect.size(); i++ )
    {
      INT_4U adc_offset = adcUpdateVect[i].adc_offset;
      adcUpdateVect[i].adc->setDataValid( adcUpdateVect[i].adc->getDataValid() |  daqAdcDataValid(adc_offset)); // OR data valid flag

      // Set data byteswap flag to false to avoid swapping bytes in getData()
      // on data that gets written over by next read().
      //
      adcUpdateVect[i].vect->setDataNeedsByteSwap(false);
      seekVectorData ( adcUpdateVect[i].offset );
      // :FIXME: this fails for STRING data type

      INT_4U srate = ((INT_4U)deltaT) * (INT_4U)(adcUpdateVect[i].adc->getSampleRate()); // sampling rate of this ADC structure
      INT_4U ndata = adcUpdateVect[i].vect->getNData(); // number of elements in the vector currently
      INT_4U tsize = Vect::getTypeSize( adcUpdateVect[i].vect->getType() ); // size of one data sample

      // :TODO: need to extend adcUpdateVect[i].vect->mData size to fit new data in.
      // This is done relying on this->catHint variable to avoid excessive memory copying.
      if ( ndata/srate >=  adcUpdateVect[i].catVectDataSize) {
	int ninc = catHint > 1? catHint: 1; // how many elements expected to be concatenated will determine new data space size
	// Grow `adcUpdateVect[i].vect->mData' by ninc*tsize*srate bytes
	unsigned char* d = new unsigned char[ (ndata + ninc * srate) * tsize ]; // allocate new expanded data space
	memcpy( d, adcUpdateVect[i].vect->getData(), ndata * tsize ); // store all of the data in the new data space
	adcUpdateVect[i].vect->setData( d, false ); // old data is release and new data is set 
	adcUpdateVect[i].catVectDataSize += ninc; // rememeber how much preallocated space is left
      }
      mStream->read( (char*) adcUpdateVect[i].vect->getData() + ndata * tsize, srate * tsize ); // read next ADC data into the vector's data space
      adcUpdateVect[i].vect->setNData( ndata + srate ); // reflect expanded data space of the vector
      adcUpdateVect[i].vect->setNBytes( adcUpdateVect[i].vect->getNBytes() + srate * tsize ); // reflect expanded data space of the vector

      std::vector<FrameCPP::Version_3_4_5::Dimension> dims;
      dims.push_back( FrameCPP::Version_3_4_5::Dimension( ndata + srate,
		  adcUpdateVect[i].vect->getDimension(0).getDx(),
		  adcUpdateVect[i].vect->getDimension(0).getUnitX(),
		  adcUpdateVect[i].vect->getDimension(0).getStartX() ));

      adcUpdateVect[i].vect->setDimension( dims );

      // Set byteswap flag to whatever reference frame header dictates
      //
      adcUpdateVect[i].vect->setDataNeedsByteSwap( mFileHeader->byteSwapNeeded() );
    }
}


//-----------------------------------------------------------------------------
//
//: Update this frame with another from `in'
//
// For use with the DAQ system. Updates active ADC channels and all the rest
// of required data attributes of this frame with data from the `in' stream.
// Input stream is checked to be a frame file with the same configuration as
// this one. This is done by checking the run number, which is incremented
// when the DAQ system is reconfigured.
//
// :TODO: should check if input file size changed and throw "cannot_update" exception if it did
//
//!param: istream* in - frame file, random access
//
//!exc: read_failure - Read failed.
//!exc: cannot_update - Frame configuration changed.
//
void DaqFrame::daqUpdate( std::istream* in )
{
  mStream = in;

  // check the run number, it must be the same.
  // call to daqFrameRun() triggers file header loading in FrameCPP::Version_3_4_5::Input
  if ( getRun() != daqFrameRun() )
    {
      throw cannot_update ( "Run number changed" );
    }

  // update frame header attributes
  mFrame = daqFrameNumber();
  mGTime.setSec(daqFrameGTimeS());
  mGTime.setNSec(daqFrameGTimeN());
  mULeapS = daqFrameULeapS();
  //  mDt = daqFrameDt();

  // update active ADCs -- all the attributes that can change and the data
  //
  for ( unsigned int i=0 ; i < adcUpdateVect.size(); i++ )
    {
      INT_4U adc_offset = adcUpdateVect[i].adc_offset;
      adcUpdateVect[i].adc->setBias(daqAdcBias(adc_offset));
      adcUpdateVect[i].adc->setSlope(daqAdcSlope(adc_offset));
      adcUpdateVect[i].adc->setDataValid(daqAdcDataValid(adc_offset));

      // Set data byteswap flag to false to avoid swapping bytes in getData()
      // on data that gets written over by next read().
      //
      adcUpdateVect[i].vect->setDataNeedsByteSwap(false);
      //      mStream->seekg( adcUpdateVect[i].offset );
      seekVectorData ( adcUpdateVect[i].offset );
// :FIXME: this fails for STRING data type
      mStream->read( (char*) adcUpdateVect[i].vect->getData(),
		     adcUpdateVect[i].vect->getNData()
		     * Vect::getTypeSize( adcUpdateVect[i].vect->getType() ));

      // Set byteswap flag to whatever refernce frame header dictates
      //
      adcUpdateVect[i].vect->setDataNeedsByteSwap( mFileHeader->byteSwapNeeded() );
    }
}

// :TODO: make this code rely on the fact that the data vector goes right after
// the ADC structure. Make it work with the TOC 
// 
//!exc: cannot_update
//   
void DaqFrame::setDaqMaps( std::map<INT_4U, adc_map_info_struct>& omi, std::multimap<INT_2U, INT_2U>& imi )
{
  RawData* raw( getRawData() );
  if ( raw == NULL )
    {
      throw cannot_update( "No Raw Data structure" );
    }

  // Copy the maps over
  //
  this->instanceOffsetMap = omi;
  this->classInstanceMultimap = imi;

  // Replace vector (class<<16|instance) "offset" with real offset
  //
  {
    std::pair <IMI, IMI> g
      = classInstanceMultimap.equal_range( ID_ADCDATA );
    for ( IMI p = g.first; p != g.second; ++p )
      {
	INT_4U adc_instance = p->second;
	std::map<INT_4U, adc_map_info_struct>::iterator q = instanceOffsetMap.find (p->first << 16 | adc_instance );
	adc_map_info_struct* adc_info = &( q->second );
	OMI r = instanceOffsetMap.find ( adc_info->vect_offset );
	if ( r == instanceOffsetMap.end() )
	  {
	    throw cannot_update( "Logic error: vector instance referenced from an ADC was not found" );
	  }
	adc_info->vect_offset = r->second.offset;
      }
  }

  // Get the frame header offset
  //

  std::pair <IMI, IMI> g
    = classInstanceMultimap.equal_range( ID_FRAME );
  if ( g.first != g.second )
    {
      OMI q = instanceOffsetMap.find (g.first->first << 16 | g.first->second);
      frame_offset = q->second.offset;
    }
  else 
    {
      throw cannot_update( "Frame header not found" );
    }  
}


//-----------------------------------------------------------------------------
//
//: Start updating all ADCs
//
// Activates all ADC present in the frame for update in daqUpdate()
//
//!exc: std::bad_alloc - Memory allocation failed
//!exc: not_found_error
//   
void DaqFrame::daqActivateAllADCs()
{
  adcUpdateVect.clear();
  if ( getRawData() == NULL )
	return;
  RawData::AdcDataContainer& adc( getRawData()->refAdc() );

  for ( unsigned int i = 0;
	i < adc.size();
	i++ )
    {
      // allow duplicates, since the vector was cleared
      //
      daqTriggerADC ( adc[i], true, true );
    }
}

//-----------------------------------------------------------------------------
//
//: Stop updating all ADCs
//
// Deactivate all ADC present in the frame.
// No ADCs wil be updated by daqUpdate()
//
//!exc: None.
//      
void DaqFrame::daqResetADCs()
{
  adcUpdateVect.clear();
  if ( getRawData() == NULL )
	return;
  RawData::AdcDataContainer& adc( getRawData()->refAdc() );

  for ( unsigned int i = 0;
	i < adc.size();
	i++ )
    {
      // allow duplicates, since the vector was cleared
      //
      daqTriggerADC ( adc[i], false, true );
    }
}

//-----------------------------------------------------------------------------
//: (De)activate an ADC
//
// Single ADC channel (de)activated referenced by its name
//
//!param: char* adcName - ADC channel name
//!param: bool activate - true to activate, false to deactivate
//
//!exc: not_found_error - ADC channel not found
//!exc: std::bad_alloc - Memory allocation failed
//   
void DaqFrame::daqTriggerADC( const char* adcName, bool activate )
{
  if ( getRawData() == NULL )
    {
      throw not_found_error("no RawData in the frame");
    }
  RawData::AdcDataContainer& adc( getRawData()->refAdc() );
  std::pair< RawData::adcData_hash_iterator,
    RawData::adcData_hash_iterator > p( adc.hashFind( adcName ) );

  if ( p.first == p.second )
    {
      throw not_found_error("no ADC channels");
    }

  for ( RawData::adcData_hash_iterator hiter = p.first;
	hiter != p.second; ++hiter )
    {
      daqTriggerADC( hiter->second, activate, false );
    }
}

//-----------------------------------------------------------------------------
//: (De)activate an ADC
//
// Single ADC channel (de)activated referenced by its index
//
//!param: unsigned int idx - ADC channel index
//!param: bool activate - true to activate, false to deactivate
//
//!exc: not_found_error - ADC channel not found
//!exc: std::bad_alloc - Memory allocation failed
//
void DaqFrame::daqTriggerADC( unsigned int idx, bool activate )
{
  if ( getRawData() == NULL )
    {
      throw not_found_error( "no RawData in the frame" );
    }
  RawData::AdcDataContainer& adc( getRawData()->refAdc() );
  if ( idx >= adc.size() )
    {
      throw not_found_error( "index out of range" );
    }
  daqTriggerADC ( adc[idx], activate, false );
}


//-----------------------------------------------------------------------------
//: (De)activate a range of ADCs
//
// Range if ADC channels (de)activated
//
//!param: unsigned int idx - first ADC index
//!param: unsigned int last - last ADC index
//!param: bool activate - true to activate, false to deactivate
//
//!exc: not_found_error - ADC channel not found
//!exc: std::bad_alloc - Memory allocation failed
//
void DaqFrame::daqTriggerADC( unsigned int idx, unsigned int last, 
			      bool activate )
{
  if ( getRawData() == NULL )
    {
      throw not_found_error("no RawData in the frame");
    }
  RawData::AdcDataContainer& adc( getRawData()->refAdc() );
  if ( idx >= adc.size() || last >= adc.size() )
    {
      throw not_found_error( "index out of range" );
    }
  for ( unsigned int i = idx ; i <= last; i++ )
    {
      daqTriggerADC ( adc[i], activate, false );
    }
}


//   
//!exc: not_found_error
//!exc: std::bad_alloc - Memory allocation failed
//   
void DaqFrame::daqTriggerADC ( AdcData* adc,
			       bool activate, bool duplicates )
{

  // Get ADC instance number and find the corresponding map info data
  //
  INT_2U adc_instance = adc->getInstance();
   std::map<INT_4U, adc_map_info_struct>::iterator q
     = instanceOffsetMap.find (ID_ADCDATA << 16 | adc_instance);
  adc_map_info_struct* adc_info = &(q->second);

  // Do not create duplicate update vector elements.
  // If active adc is activated again or inactive is deactivated again,
  // we are done here.
  //
  if ( !duplicates && adc_info->active == activate )
    {
      return;
    }

  adc_info->active = activate;

  if ( activate ) 
    {
      // Add new update vector element
      //

      Vect* vect( adc->refData()[0] );
      
      if ( vect->getCompress() != Vect::NONE )
	{
	  throw not_found_error( "Compressed data vectors are not supported" );
	}
      adc_update_struct aus;
      // This should find vector next to the ADC structure
      aus.offset = adc_info->vect_offset;
      aus.vect=vect;
      aus.adc_offset = adc_info->offset;
      aus.adc=adc;
      aus.catVectDataSize = 1; // One data vector preallocated initally, i.e. just the original data, nothing additional
      adcUpdateVect.push_back( aus );
    }
  else
    {
      // This part is quite costly if vector size is big
      //
      for ( unsigned int i = 0; i < adcUpdateVect.size(); i++ )
	{
	  if ( adc_info->vect_offset == adcUpdateVect[i].offset )
	    {
	      adcUpdateVect.erase( adcUpdateVect.begin() + i );
	      break;
	    }
	}
    }
}

