#ifndef DaqFrameCppFrameHH
#define DaqFrameCppFrameHH

#include "framecpp/FrameCPP.hh"
#include "framecpp/config.h"

#include "frame.hh"
#include "input.hh"

namespace FrameCPP
{
  namespace Version_3_4_5
  {
    class DaqFrame;

    typedef struct adc_map_info_struct {
      INT_4U offset;
      INT_4U vect_offset;
      bool active; // whether this adc is actively update
    } adc_map_info_struct;


//-----------------------------------------------------------------------------
//
//: Data Acquisition Frame
//
// DAQ frame is a Frame with extensions to update from another frame file.
// This is done for interfacing with LIGO DAQ (Data Acquisition) to ease
// repeated frame reading. Only the first frame is read and parsed in full.
// This first reference frame is DaqFrame object.
// Subsequent frames are identical to the reference frame in their structure.
// DaqUpdate() method is used to refresh reference frame with the data
// from the new frame. Currently frame header attributes and ADC channels
// are refreshed. Only active ADC channels are refreshed.
// Activation is done with trigger methods present in DaqFrame class.
//
// :FIXME: :TODO: Make daqUpdate() work with version 3 frames
class DaqFrame : public Frame, private Input
{
public:

  /* Constructors */
  //-----------------------------------------------------------------------------
  //
  //: Default Constructor.
  //
  //!exc: None.
  //
  DaqFrame();

  //-----------------------------------------------------------------------------
  //: (De)activate an ADC
  //
  // Single ADC channel (de)activated referenced by its name
  //
  //!param: char* adcName - ADC channel name
  //!param: bool activate - true to activate, false to deactivate
  //!param: bool duplicates - allow duplicate update vector elements
  //
  //!exc: not_found_error - ADC channel not found
  //!exc: std::bad_alloc - Memory allocation failed
  //
  void TriggerADC( FrAdcDataBase*, bool activate );

  //-----------------------------------------------------------------------------
  //
  //: Copy Constructor.
  //
  // This performs a deep copy of owned elements.
  //
  //!param: const Frame& frame - Object to be copied.
  //
  //!exc: std::bad_alloc - Memory allocation failed.
  //
   DaqFrame( const Frame& frame );

  //: Destructor
  virtual ~DaqFrame();

  /* Mutators */
  //!exc: cannot_update - Frame configuration changed.   
  void setDaqMaps( std::map<INT_4U, adc_map_info_struct>& omi,
	           std::multimap<INT_2U, INT_2U>& imi );
  inline void setFileHeader( FileHeader* fh)
  {
    delete mFileHeader;
    mFileHeader = 0;
   
    mFileHeader = new FileHeader( *fh );
  }


  //-----------------------------------------------------------------------------
  //
  //: Update this frame with another from `in'
  //
  // For use with the DAQ system. Updates active ADC channels and all the rest
  // of required data attributes of this frame with data from the `in' stream.
  // Input stream is checked to be a frame file with the same configuration as
  // this one. This is done by checking the run number, which is incremented
  // when the DAQ system is reconfigured.
  //
  // :TODO: should check if input file size changed and throw "cannot_update" exception if it did
  //
  //!param: std::istream* in - frame file, random access
  //
  //!exc: read_failure - Read failed.
  //!exc: cannot_update - Frame configuration changed.
  //
  void daqUpdate( std::istream* in );


  //-----------------------------------------------------------------------------
  //
  //: Concatenate active ADCs in this frame with another's from `in'
  //
  // For use with the DAQ system. Cats active ADC channels with data from the `in' stream.
  // Input stream is checked to be a frame file with the same configuration as
  // this one. This is done by checking the run number, which is incremented
  // when the DAQ system is reconfigured.
  //
  // :TODO: should check if input file size changed and throw "cannot_update" exception if it did
  //
  //!param: std::istream* in - frame file, random access
  //
  //!exc: read_failure - Read failed.
  //!exc: cannot_update - Frame configuration changed.
  //!exc: cannot_cat - Data is not contiguous in time.
  //
  void daqCat( std::istream* in );

  //-----------------------------------------------------------------------------
  //
  //: Start updating all ADCs
  //
  // Activates all ADC present in the frame for update in daqUpdate()
  //
  //!exc: std::bad_alloc - Memory allocation failed
  //!exc: not_found_error - ADC channel not found.
  //
   void daqActivateAllADCs();

  
  //-----------------------------------------------------------------------------
  //
  //: Stop updating all ADCs
  //
  // Deactivate all ADC present in the frame.
  // No ADCs wil be updated by daqUpdate()
  //
  //!exc: None.
  //
  void daqResetADCs();

  //-----------------------------------------------------------------------------
  //: (De)activate an ADC
  //
  // Single ADC channel (de)activated referenced by its name
  //
  //!param: char* adcName - ADC channel name
  //!param: bool activate - true to activate, false to deactivate
  //!param: bool duplicates - allow duplicate update vector elements
  //
  //!exc: not_found_error - ADC channel not found
  //!exc: std::bad_alloc - Memory allocation failed
  //
  void daqTriggerADC( AdcData*, bool activate, bool duplicates = false );

  //-----------------------------------------------------------------------------
  //: (De)activate an ADC
  //
  // Single ADC channel (de)activated referenced by its name
  //
  //!param: const char* adcName - ADC channel name
  //!param: bool activate - true to activate, false to deactivate
  //
  //!exc: not_found_error - ADC channel not found
  //!exc: std::bad_alloc - Memory allocation failed
  void daqTriggerADC( const char* adcName, bool activate );

  //-----------------------------------------------------------------------------
  //: (De)activate an ADC
  //
  // Single ADC channel (de)activated referenced by its index
  //
  //!param: unsigned int idx - ADC channel index
  //!param: bool activate - true to activate, false to deactivate
  //
  //!exc: not_found_error - ADC channel not found
  //!exc: std::bad_alloc - Memory allocation failed
  //
  void daqTriggerADC( unsigned int idx, bool activate );

  //-----------------------------------------------------------------------------
  //: (De)activate a range of ADCs
  //
  // Range if ADC channels (de)activated
  //
  //!param: unsigned int idx - first ADC index
  //!param: unsigned int last - last ADC index
  //!param: bool activate - true to activate, false to deactivate
  //
  //!exc: not_found_error - ADC channel not found
  //!exc: std::bad_alloc - Memory allocation failed
  //
  void daqTriggerADC( unsigned int idx, unsigned int last, bool activate );

  //: Print current ADC channel update vector
  void printActiveChannels ( std::ostream& ost )
    {
      for (unsigned int i = 0; i < adcUpdateVect.size(); i++ )
	{
	  ost << adcUpdateVect[i].offset << '\t'
	      << adcUpdateVect[i].vect->getNData() << '\t'
	      << adcUpdateVect[i].vect->getType() << '\t'
	      << adcUpdateVect[i].vect->getData() << std::endl;
	}
    }

  /* Map of active ADCs */
  typedef struct adc_update_struct {
    INT_4U offset;
    Vect* vect;
    INT_4U catVectDataSize; // shows how many vectors were preallocated for concatenation (in units of (INT_4U)(adcUpdateVect[i].adc->getSampleRate()) * Vect::getTypeSize( adcUpdateVect[i].vect->getType() ) )
    INT_4U adc_offset;
    AdcData* adc;
  } adc_update_struct;

  //: Get ADC update vector.
  //!exc: None.
  inline const std::vector<adc_update_struct>& refAdcUpdateVect() const { return adcUpdateVect; }

  //: Set frame concat hint
  void setCatHint( INT_4U hint ) { catHint = hint; }

  //: Get frame concat hint
  inline INT_4U getCatHint() { return catHint; }

private:
  std::map<INT_4U, adc_map_info_struct> instanceOffsetMap;
  typedef std::map<INT_4U, adc_map_info_struct>::const_iterator OMI;
  std::multimap<INT_2U, INT_2U> classInstanceMultimap;
  typedef std::multimap<INT_2U, INT_2U>::const_iterator IMI;
  //: Offset to the start of frame.
  INT_4U frame_offset;
  //: Adc update vector.
  std::vector<adc_update_struct> adcUpdateVect;
  //: Hint on how many frames will be concatenated.
  INT_4U catHint;

  /* Frame header accessors */

  /*
    Both nameLength() and name() can be used in all
    structures that had `name' as the third element/
    The list of such strucutres include:
    FrSH, FrSE, Frame, FrADCData, FrDetector, FrHistory,
    FrRawData, FrProcData, FrSerData, FrStatData, FrVect,
    FrEvent, FrSummary.
  */

  //: Get length of name at `offset'.
  inline int daqNameLength (int offset) {
    mStream->seekg (offset);
    INT_2U nameLength;
    //    mStream->read (&nameLength, 2);
    (*this) >> nameLength;
    return nameLength;
  }

#ifdef not_def
  inline char *daqName (int offset) {
    return mStream->str () + offset + 2;
  }


  inline char *daqFrameName () {
    return daqName (frame_offset);
  }
#endif

  //: Get frame run number.
  inline INT_4S daqFrameRun () {
    mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2);
    INT_4S run; 
    (*this) >> run;
    return run;
  }

  //: Get frame number.
  inline INT_4U daqFrameNumber () {
    mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 4);
    INT_4U frame; 
    (*this) >> frame;
    return frame;
  }

  /*
    The following frame header accessors musbe called after the frame file
   header was loaded by one of the above calls
  */

  //: Get frame data quality word.
  inline INT_4U daqFrameDataQuality () {
    if (mFileHeader->getDataFormatVersion() < 4)
      return 0; // there is no data quality word in version 3

    mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 8);
    INT_4U dq;
    (*this) >> dq;
    return dq;
  }

  //: Get frame start time seconds.
  inline INT_4U daqFrameGTimeS () {
    if (mFileHeader->getDataFormatVersion() < 4)
      mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 8);
    else
      mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 12);
    INT_4U gps;
    (*this) >> gps;
    return gps;
  }

  //: Get frame start time nanoseconds.
  inline INT_4U daqFrameGTimeN () {
    if (mFileHeader->getDataFormatVersion() < 4)
      mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 12);
    else
      mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 16);
    INT_4U gps; 
    (*this) >> gps;
    return gps;
  }

  //: Get frame leap seconds.
  inline INT_2U daqFrameULeapS () {
    if (mFileHeader->getDataFormatVersion() < 4)
      mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 16);
    else
      mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 20);
    INT_2U s; 
    (*this) >> s;
    return s;
  }

  //: Get frame local time.
  inline INT_4S daqFrameLocalTime () {
    if (mFileHeader->getDataFormatVersion() < 4)
      mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 18);
    else
      mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 22);
    INT_4S s;
    (*this) >> s;
    return s;
  }

  //: Get fame delta time.
  inline REAL_8 daqFrameDt () {
    if (mFileHeader->getDataFormatVersion() < 4)
      mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 22);
    else
      mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 26);
    REAL_8 s; 
    (*this) >> s;
    return s;
  }

  /* ADC structure accessors */
#ifdef not_def
  inline char *adcName (int offset) {
    return name (offset);
  }

  inline char *adcComment (int offset) {
    return name (offset + daqNameLength (offset) + 2);
  }
#endif

  /* skip over the `name' and the `comment' strings */
#define ADC_NAME_COMMENT_OFFSET_DECL(offset) \
  int NL = offset + daqNameLength (offset) + 2
#define ADC_NAME_COMMENT_OFFSET (NL + daqNameLength (NL) + 2)

  //: Get adc crate number.
  inline INT_4U daqAdcCrate (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    mStream->seekg (ADC_NAME_COMMENT_OFFSET);
    INT_4U crate;
    (*this) >> crate;
    return crate;
  }

  //: Get adc channel number.
  inline INT_4U daqAdcChannel (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    mStream->seekg (ADC_NAME_COMMENT_OFFSET + 4);
    INT_4U cnum;
    (*this) >> cnum;
    return cnum;
  }

  //: Get adc number of bits.
  inline INT_4U daqAdcNBits (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    mStream->seekg (ADC_NAME_COMMENT_OFFSET + 8);
    INT_4U nbits;
    (*this) >> nbits;
    return nbits;
  }

  //: Get adc bias.
  inline REAL_4 daqAdcBias (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    mStream->seekg (ADC_NAME_COMMENT_OFFSET + 12);
    REAL_4 bias;
    (*this) >> bias;
    return bias;
  }

  //: get adc slope.
  inline REAL_4 daqAdcSlope (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    mStream->seekg (ADC_NAME_COMMENT_OFFSET + 16);
    REAL_4 slope;
    (*this) >> slope;
    return slope;
  }

#ifdef not_def
  inline char *daqAdcUnits (int offset) {
    ADC_NAME_COMMENT_OFFSET_DECL(offset);
    return name (ADC_NAME_COMMENT_OFFSET + 20);
  }
#endif


#undef ADC_NAME_COMMENT_OFFSET_DECL
#undef ADC_NAME_COMMENT_OFFSET

#define ADC_OFFSET_HEADER_END_DECL(offset) \
  INT_4U headerLength; \
  mStream->seekg (offset - 8); /* move to `header length' */ \
/* mStream->read (&headerLength, 4);  get the length */ \
  (*this) >> headerLength;
#define ADC_OFFSET_HEADER_END (offset - 8 + headerLength)

  //: Get adc sampling rate.
  inline REAL_8 daqAdcSampleRate (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    mStream->seekg (ADC_OFFSET_HEADER_END - 12 - 26);
    REAL_8 a;
    (*this) >> a;
    return a;
  }

  //: Get adc time offset seconds.
  inline INT_4U daqAdcTimeOffsetS (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    mStream->seekg (ADC_OFFSET_HEADER_END - 12 - 18);
    INT_4U a;
    (*this) >> a;
    return a;
  }

  //: Get adc time offset nanoseconds.
  inline INT_4U daqAdcTimeOffsetN (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    mStream->seekg (ADC_OFFSET_HEADER_END - 12 - 14);
    INT_4U a;
    (*this) >> a;
    return a;
  }

  //: Get adc frequency shift.
  inline REAL_8 daqAdcFShift (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    mStream->seekg (ADC_OFFSET_HEADER_END - 12 - 10);
    REAL_8 fs;
    (*this) >> fs;
    return fs;
  }

  //: Get adc data valid word.
  inline INT_2U daqAdcDataValid (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    mStream->seekg (ADC_OFFSET_HEADER_END - 12 - 2);
    INT_2U ovr;
    (*this) >> ovr;
    return ovr;
  }

  //: Given the offset find the ADC data vector ptr.
  inline INT_4U adcDataVectorPtr (int offset) {
    ADC_OFFSET_HEADER_END_DECL(offset);
    //    std::cout << "adc structure length is " << std::hex << headerLength << dec << std::endl
    mStream->seekg (ADC_OFFSET_HEADER_END - 12); // adjust to data ve
    INT_2U classId, instanceId;
    //    mStream->read (&classId, 2); // read class id
    //    mStream->read (&instanceId, 2); // read instance id
    (*this) >> classId >> instanceId;
    return classId << 16 | instanceId;
  }

#undef ADC_OFFSET_HEADER_END_DECL
#undef ADC_OFFSET_HEADER_END

  /* Vector structure accessors */
  //:  Move to the beginning of vector data.
  //
  // Given the offset to the start of the vector structure,
  // move the input strstream position to the beginning of the data.
  inline void seekVectorData (int offset) {
    mStream->seekg (offset + daqNameLength (offset) + 2 + 12);
  }

#ifdef not_def
  inline char *vectorData (int offset) {
    return mStream->str () + offset + daqNameLength (offset) + 2 + 12;
  }
#endif

};

} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // DaqFrameCppFrameHH
