#ifndef FrameCppDaqReaderHH
#define FrameCppDaqReaderHH


#include "framereaderbase.hh"
#include "time.hh"
#include "location.hh"
#include "daqframe.hh"

namespace FrameCPP {

namespace Version_3_4_5 {

class Time;
 

//-----------------------------------------------------------------------------
//
//: Data Acquisition Reader
//
// Read the DaqFrame. readFrame() reads the frame fully
// and creates the structure map in it.
// If the frame file happen to have more than one frame in it,
// then the second and all subsequent frames have the same structure,
// which is different from the first frame's structure.
// 
//
class DaqReader : private FrameReaderBase
{
public:

  enum ErrorCode
  {
    BADSTREAM = FrameReaderBase::BADSTREAM,
    SKIPFAILED = FrameReaderBase::SKIPFAILED,
    ENDOFFILE = FrameReaderBase::ENDOFFILE
  };
    
  //: Constructor
  //
  //!exc: read_failure
  //
  DaqReader( std::istream& in, bool autoReset = false );

  // Destructor
  //
  //!exc: None.
  //
  virtual ~DaqReader();

  FrameReaderBase::getError;
  FrameReaderBase::errorBegin;
  FrameReaderBase::errorEnd;
  FrameReaderBase::errorQuantity;
  FrameReaderBase::errorFirst;
  FrameReaderBase::errorLast;
  FrameReaderBase::clearErrors;
  FrameReaderBase::setErrorWatch;

  FrameReaderBase::getNBytes;

  //-----------------------------------------------------------------------------
  //
  //: Read a reference DaqFrame and build an update map along the way.
  //
  //!return: Frame*
  //
  //!exc: bad_alloc - Memory allocation failed.
  //!exc: read_failure - Read failure.
  //
  DaqFrame* readFrame();

  /* Accessors */
  FrameReaderBase::getFrames;
  FrameReaderBase::isAutoReset;
  FrameReaderBase::setAutoReset;
  FrameReaderBase::reset;

  //: Overrides object header reader in the superclass.
  //   
  //!exc: read_failure
  //
   FrameCPP::Version_3_4_5::ObjectHeader readNextObjectHeader();
  
  /* Maps (from frame builder's imageframewriter.hh) */

  //
  // Mapping works like this: (classId << 16 | instanceId) -> offset.
  //<br>
  // classId identifies a type of structure,
  // instanceId is ++ for every newly output structure of class classId.
  // `offset' is a number of bytes to the start of the structure.
  //<br>
  // (classId, instanceID) pairs are found in the frame image
  // and are specified in the documentation as PTR_STRUCT.
  // The pair serves as a pointer.
  //<br>
  // Map allows a quick lookup of the offset to the start of a structure
  // knowing the instance and class IDs.

  std::map<INT_4U, adc_map_info_struct> instanceOffsetMap;
  typedef std::map<INT_4U, adc_map_info_struct>::const_iterator OMI;

  // Multimap is constructed to allow a lookup of
  // all the instances of a particular class.
  //<br>
  // The mapping works like this: classId -> instanceId.
  //<br>
  // Now, if you need to find all the instance of, say, ADC class,
  // you go and iterate over IMI for ID_ADCDATA,
  // then for every instance Id found yo can get the offset
  // from the instanceOffsetMap, using OMI.
  
  std::multimap<INT_2U, INT_2U> classInstanceMultimap;
  typedef std::multimap<INT_2U, INT_2U>::const_iterator IMI;

//!ignore_begin:
#define ADC_OFFSET_HEADER_END_DECL(offset) \
  INT_4U headerLength; \
  mStream->seekg (offset - 8); /* move to `header length' */ \
/*  mStream->read (&headerLength, 4); get the length */ \
  (*this) >> headerLength;
#define ADC_OFFSET_HEADER_END (offset - 8 + headerLength)
//!ignore_end:

  //: Given the offset find the ADC data vector ptr.
  inline INT_4U adcDataVectorPtr (int offset);
};


//: Constructor
//   
//!exc: read_failure   
//   
inline DaqReader::DaqReader( std::istream& in, bool autoReset )
   : FrameReaderBase( in, autoReset )
{
}



//: Given the offset find the ADC data vector ptr.
inline INT_4U DaqReader::adcDataVectorPtr (int offset) {
  ADC_OFFSET_HEADER_END_DECL(offset);
  //    cout << "adc structure length is " << hex << headerLength << dec << endl
  mStream->seekg (ADC_OFFSET_HEADER_END - 12); // adjust to data ve
  INT_2U classId, instanceId;
  // mStream->read (&classId, 2); // read class id
  // mStream->read (&instanceId, 2); // read instance id
  (*this) >> classId >> instanceId;
  return ID_VECT << 16 | instanceId;
}

#undef ADC_OFFSET_HEADER_END_DECL
#undef ADC_OFFSET_HEADER_END


} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppDaqReaderHH
