#ifndef FrameCppDetectorHH
#define FrameCppDetectorHH

#include "framecpp/config.h"
#include "framecpp/Version3_4_5/base.hh"
#include "framecpp/Version3_4_5/statdata.hh"
#include "framecpp/Version3_4_5/table.hh"

#include "framecpp/SearchContainer.hh"

namespace FrameCPP {

namespace Version_6
{
  class FrDetector;
}

namespace Version_3_4_5 {

class FrameReaderBase;


//-----------------------------------------------------------------------------
//
//: This class represents a detector.
//
// It is modeled after the FrDetector structure in the Frame Specification,
// LIGO-T970130-B-E.
//
class Detector : public Base
{
public:
    //!ignore_begin:
    /* Container Macros */
    DECLARE_CONTAINER( More, more, Vect, getName );
    DECLARE_CONTAINER( MoreTable, moreTable, Table, getName );
    DECLARE_CONTAINER( StatData, statData, StatData, getName );
    //!ignore_end:

    /* Constructors */
    //-----------------------------------------------------------------------------
    //
    //: Default Constructor.
    //
    // This creates an Detector object with the following properties:
    // <ul>
    //     <li> name = "none" </li>
    //     <li> location is 0 degrees longitude, 0 degrees latitude </li>
    //     <li> elevation = 0 meters </li>
    //     <li> armXazimuth = 0 </li>
    //     <li> armYazimuth = 90. </li>
    // </ul>
    //
    //!exc: None.
    //
    Detector();

    //-----------------------------------------------------------------------------
    //
    // Copy Constructor.  This performs a deep copy of the object.
    //
    //!param: const Detector& detector - Detector to be copied.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
   Detector( const Detector& detector );

    //-----------------------------------------------------------------------------
    //
    //: Constructor.
    //
    // This creates a Detector object with the specified attributes.
    //
    //!param: const std::string& name - Instrument name
    //!param: REAL_8 longitude - Instrument Longitude
    //!param: REAL_8 latitude - Instrument Latitude
    //!param: REAL_4 elevation - Vertex elevation
    //!param: REAL_4 armXazimuth - Orientation of X arm, degrees CCW from East.
    //!param: REAL_4 armYazimuth - Orientation of Y arm, degrees CCW from East.
    //!param: REAL_4 armXaltitude - Orientation of X arm, degrees above horizon.
    //!param: REAL_4 armYaltutude - Orientation of Y arm, degrees above horizon.
    //!param: REAL_4 armXmidpoint - Midpoint of X cavity.
    //!param: REAL_4 armYmidpoint - Midpoint of Y cavity.
    //!param: INT_4S localTime - Local seasonal time minus UTC (seconds)
    //!param: INT_4U dataQuality - Data quality word.
    //!param: std::string& qaBitList - Data quality bit definition list.
    //
    //!exc: None.
    //
    Detector( const std::string& name,
	      REAL_8 longitude, REAL_8 latitude, REAL_4 elevation,
	      REAL_4 armXazimuth, REAL_4 armYazimuth,
	      REAL_4 armXaltitude, REAL_4 armYaltitude,
	      REAL_4 armXmidpoint, REAL_4 armYmidpoint,
	      INT_4S localTime, INT_4U dataQuality, const std::string& qaBitList );
    
    /* Destructor */
    //!exc: None.
    //
   virtual ~Detector();
    
    /* Operator Overloads */
    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    //!param: const Detector& detector - Detector to be assigned.
    //
    //!return: const Detector& -- This object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const Detector& operator=( const Detector& detector );

    //-----------------------------------------------------------------------------
    //
    //: += Operator.
    //
    //!param: const Detector& detector - Detector to be assigned.
    //
    //!return: const Detector& -- This object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const Detector& operator+=( const Detector& detector );

    //-----------------------------------------------------------------------------
    //
    //: Equal comparison.
    //
    // Determines whether two Detector objects are equal. For this comparison,
    // all of the data is compared.
    //
    //!param: const Detector& detector - The object to compare with.
    //
    //!return: bool -- true if the objects are equal.
    //
    //!kw: const
    //
    //!exc: None.   
    bool operator==( const Detector& detector ) const;
    bool operator!=( const Detector& detector ) const;
    
    /* Required Methods */
    //!exc: None.
    virtual INT_2U getClassId() const;

    //-----------------------------------------------------------------------------
    //
    //: Constructs the Structure Header for this class.  
    //
    //!return: SH -- Detector structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
   static SH getSH();

    //-----------------------------------------------------------------------------
    //
    //: Constructs frame data format version 4 Structure Header for this class.  
    //
    //!return: SH -- Structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH4();

    //-----------------------------------------------------------------------------
    //
    //: Constructs frame data format version 3 Structure Header for this class.  
    //
    //!return: SH -- Structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
   static SH getSH3();

    /* Accessors */
    const std::string& getName() const;
   
    //!exc: None.
    REAL_8 getLongitude() const;
   
    //!exc: None.
    REAL_8 getLatitude() const;
   
    //!exc: None.   
    REAL_4 getElevation() const;
   
    //!exc: None.   
    REAL_4 getArmXAzimuth() const;
   
    //!exc: None.   
    REAL_4 getArmYAzimuth() const;
   
    //!exc: None.   
    REAL_4 getArmXAltitude() const;
   
    //!exc: None.   
    REAL_4 getArmYAltitude() const;
   
    //!exc: None.   
    REAL_4 getArmXMidpoint() const;
   
    //!exc: None.   
    REAL_4 getArmYMidpoint() const;
   
    //!exc: None.   
    INT_4S getLocalTime() const;

    //!exc: None.   
    INT_4U getDataQuality() const;
   
    const std::string& getQaBitList() const;

    //!exc: None.   
    MoreContainer& refMore();
   
    //!exc: None.   
    const MoreContainer& refMore() const;
   
    //!exc: None.   
    MoreTableContainer& refMoreTable();
   
    //!exc: None.   
    const MoreTableContainer& refMoreTable() const;

    //!exc: None.   
    StatDataContainer& refStatData();

    //!exc: None.   
    const StatDataContainer& refStatData() const;

    //-----------------------------------------------------------------------------
    //
    //: Find StatData corresponding to a particular time.
    //
    //!param: INT_4U time - Time for which to look.
    //
    //!return: StatData* -- Found data.
    //
    //!exc: not_found_error - Data not found.
    //
   StatData* findStatData( INT_4U time );

    //-----------------------------------------------------------------------------
    //
    //: Find StatData corresponding to a particular time.
    //
    //!param: INT_4U time - Time for which to look.
    //
    //!return: const StatData* -- Found data.
    //
    //!exc: not_found_error - Data not found.
    //
    const StatData* findStatData( INT_4U time ) const;

private:
  friend class Dictionary;
  friend class FrameCPP::Version_6::FrDetector;

    //-----------------------------------------------------------------------------
    //
    //: Reads an Detector object from an FrameReaderBase object.
    //
    // This method creates a dynamically allocated object.
    //
    // <p>*NOTE* This method allocates memory and returns a pointer to it.  The
    // user is responsible for freeing this memory.
    //
    //!param: FrameReaderBase& in - Input data stream object.
    //
    //!return: Detector* -- Newly allocated.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure.
    //
    static Detector* read( FrameReaderBase& in );
   
    //: Version 4 frame input method.
    //   
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure.
    //   
    static Detector* read4( FrameReaderBase& in );

    //: Version 3 frame input method.
    //   
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure.
    //      
    static Detector* read3( FrameReaderBase& in );

    //-----------------------------------------------------------------------------
    //
    //: Writes an Detector object to an Output object.
    //
    //!param: Output& out - Output data stream object.
    //
    //!exc: write_failure - Write failure occured.
    //
    virtual void write( Output& out ) const;
    
private:
    //: Instrument name.
    std::string mName;
    //: Instrument Longitude
    REAL_8 mLongitude;
    //: Instrument Latitude
    REAL_8 mLatitude;
    //: Vertex elevation.
    REAL_4 mElevation;
    //:  Orientation of X arm, degrees CCW from East.
    REAL_4 mArmXAzimuth;
    //:  Orientation of Y arm, degrees CCW from East.
    REAL_4 mArmYAzimuth;
    //: Pitch angle of X arm.
    REAL_4 mArmXAltitude;
    //: Pitch angle of Y arm.
    REAL_4 mArmYAltitude;
    //: Midpoint of X cavity
    REAL_4 mArmXMidpoint;
    //: Midpoint of Y cavity
    REAL_4 mArmYMidpoint;
    //: Local timezone offset from UTC (local seazonal time minus UTC), seconds
    INT_4S mLocalTime;
    //: Top level quality of data
    INT_4U mDataQuality;
    //: Data quality bit definitions
    std::string mQaBitList;
    //: User data.
    MoreContainer mMore;
    //: User table data.
    MoreTableContainer mMoreTable;
    //: Static data.
    StatDataContainer mStatData;
};


/******************/
/* Inline Methods */
/******************/


//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Not-equal comparison.
//
//!param: const Detector& detector - The object to compare with.
//
//!return: bool -- True if the objects are not equal.
//
//!kw: inline const
//
//!exc: None.
//      
inline bool Detector::operator!=( const Detector& detector ) const
{
    return !( *this == detector );
}



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------
   

//-----------------------------------------------------------------------------
//
//: Gets the instrument's longitude.
//
//!return: REAL_8 -- Detector vertex longitude.
//
//!kw: inline const
//
//!exc: None.
//      
inline REAL_8 Detector::getLongitude() const
{
    return mLongitude;
}


//-----------------------------------------------------------------------------
//
//: Gets the instrument's latitude.
//
//!return: REAL_8 -- Detector vertex latitude.
//
//!kw: inline const
//
//!exc: None.
//      
inline REAL_8 Detector::getLatitude() const
{
    return mLatitude;
}


//-----------------------------------------------------------------------------
//
//: Gets the vertex elevation in meters.
//
// This is relative to the WGS84 ellipsoid.
//
//!return: REAL_4 -- Elevation.
//
//!kw: inline const
//
//!exc: None.
//      
inline REAL_4 Detector::getElevation() const
{
    return mElevation;
}

//-----------------------------------------------------------------------------
//
//: Gets local time offset.
//
// Returns a difference between local seazonal time
// and UTC.
//
//!return: INT_4S -- local time (seconds)
//
//!kw: inline const
//
//!exc: None.
//      
inline INT_4S Detector::getLocalTime() const
{
    return mLocalTime;
}

//-----------------------------------------------------------------------------
//
//: Gets data quality word.
//
// A logical 32-bit word to denote top level quality of data.
//
//!return: INT_4U -- data quality word.
//
//!kw: inline const
//
//!exc: None.
//      
inline INT_4U Detector::getDataQuality() const
{
    return mDataQuality;
}

//-----------------------------------------------------------------------------
//
//: Gets the data quality word bit definition list.
//
//!return: const std::string& -- bit description.
//
//!kw: inline const
//
inline const std::string& Detector::getQaBitList() const
{
    return mQaBitList;
}


//-----------------------------------------------------------------------------
//
//: Gets the orientation of the X arm.
//
// This is in degrees CCW from east.
//
//!return: REAL_4 -- Arm X orientation.
//
//!kw: inline const
//
//!exc: None.
//      
inline REAL_4 Detector::getArmXAzimuth() const
{
    return mArmXAzimuth;
}


//-----------------------------------------------------------------------------
//
//: Gets the orientation of the Y arm.
//
// This is in degrees CCW from east.
//
//!return: REAL_4 -- Arm Y orientation.
//
//!kw: inline const
//
//!exc: None.
//      
inline REAL_4 Detector::getArmYAzimuth() const
{
    return mArmYAzimuth;
}


//-----------------------------------------------------------------------------
//
//: Gets the altitude angle of the X arm.
//
// This is in degrees above horizon.
//
//!return: REAL_4 -- Arm X altitude angle.
//
//!kw: inline const
//
//!exc: None.
//      
inline REAL_4 Detector::getArmXAltitude() const
{
    return mArmXAltitude;
}


//-----------------------------------------------------------------------------
//
//: Gets the altitude angle of the Y arm.
//
// This is in degrees above horizon.
//
//!return: REAL_4 -- Arm Y altitude angle.
//
//!kw: inline const
//
//!exc: None.
//      
inline REAL_4 Detector::getArmYAltitude() const
{
    return mArmYAltitude;
}


//-----------------------------------------------------------------------------
//
//: Gets the midpoint of the X cavity.
//
// Distance between the detector vertex and the middle of the X cavity
// with respect to the detector vertex (meters) (should be zero fo bars).
//
//!return: REAL_4 -- distance in meters.
//
//!kw: inline const
//
//!exc: None.
//      
inline REAL_4 Detector::getArmXMidpoint() const
{
    return mArmXMidpoint;
}


//-----------------------------------------------------------------------------
//
//: Gets the midpoint of the Y cavity.
//
// Distance between the detector vertex and the middle of the Y cavity
// with respect to the detector vertex (meters) (should be zero fo bars).
//
//!return: REAL_4 -- distance in meters.
//
//!kw: inline const
//
//!exc: None.
//      
inline REAL_4 Detector::getArmYMidpoint() const
{
    return mArmYMidpoint;
}


//-----------------------------------------------------------------------------
//
//: Gets the container for user-defined data.
//
//!return: MoreContainer& -- User data.
//
//!kw: inline
//
//!exc: None.
//      
inline Detector::MoreContainer&
Detector::refMore()
{
    return mMore;
}


//-----------------------------------------------------------------------------
//
//: Gets the container for user-defined data.
//
//!return: const MoreContainer& -- User data.
//
//!kw: inline const
//
//!exc: None.
//      
inline
const Detector::MoreContainer&
Detector::refMore() const
{
    return mMore;
}


//-----------------------------------------------------------------------------
//
//: Gets the container for user-defined table data.
//
//!return: MoreTableContainer&  -- Table data.
//
//!kw: inline
//
//!exc: None.
//      
inline Detector::MoreTableContainer&
Detector::refMoreTable()
{
    return mMoreTable;
}


//-----------------------------------------------------------------------------
//
//: Gets the container for user-defined table data.
//
//!return: const MoreTableContainer& -- Table data.
//
//!kw: inline const
//
//!exc: None.
//      
inline
const Detector::MoreTableContainer&
Detector::refMoreTable() const
{
    return mMoreTable;
}


//-----------------------------------------------------------------------------
//
//: Gets the container for StatData.
//
//!return: StatDataContainer& -- Static data.
//
//!kw: inline
//
//!exc: None.
//      
inline
Detector::StatDataContainer&
Detector::refStatData()
{
    return mStatData;
}


//-----------------------------------------------------------------------------
//
//: Gets the container for StatData.
//
//!return: const StatDataContainer& -- Static data.
//
//!kw: inline const
//
//!exc: None.
//      
inline const
Detector::StatDataContainer&
Detector::refStatData() const
{
    return mStatData;
}

 
} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // FrameCppDetectorHH
