#include "framecpp/config.h"

#include "dictionary.hh"
#include "adcdata.hh"
#include "detector.hh"
#include "endoffile.hh"
#include "endofframe.hh"
#include "history.hh"
#include "msg.hh"
#include "procdata.hh"
#include "rawdata.hh"
#include "serdata.hh"
#include "simdata.hh"
#include "simevent.hh"
#include "statdata.hh"
#include "summary.hh"
#include "event.hh"
#include "vect.hh"
#include "frame.hh"
#include "toc.hh"

using namespace FrameCPP::Version_3_4_5;


namespace FrameCPP
{
  namespace Version_3_4_5 {
    const int Library::FRAMECPP_VERSION( 5 );
    const int Library::FRAMECPP_REVISION( 0 );

    /* Registry of supported dictionaries */
    Library library;
  } // namespace - Version_3_4_5

}// namespace - FrameCPP


//-----------------------------------------------------------------------------
// Initialization
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
//: Constructor.
//
// This method initializes the dictionary by registering all of the
// frame class objects (those which are written as structures to a frame file).
//
//!exc: std::bad_alloc - Memory allocation failed in initialization.
//!exc: read_failure - Read failed.   
//
Dictionary::Dictionary(int version)
{
  if ( version == 5 )
    {
      registerClass< AdcData >();
      registerClass< Detector >();
      registerClass< EndOfFile >();
      registerClass< EndOfFrame >();
      registerClass< Frame >();
      registerVer3Class< History >();
      registerClass< Msg >();
      registerClass< ProcData >();
      registerClass< RawData >();
      registerClass< SerData >();
      registerClass< SimData >();
      registerClass< StatData >();
      registerClass< Summary >();
      registerClass< Event >();
      registerClass< Vect >();
      registerClass< Table >();
      registerClass< SimEvent >();
      registerClass< TOC >();      
    }
  else if ( version == 3 )
    {
      registerVer3Class< AdcData >();
      registerVer3Class< Detector >();
      registerVer3Class< EndOfFile >();
      registerVer3Class< EndOfFrame >();
      registerVer3Class< Frame >();
      registerVer3Class< History >();
      registerVer3Class< Msg >();
      registerVer3Class< ProcData >();
      registerVer3Class< RawData >();
      registerVer3Class< SerData >();
      registerVer3Class< SimData >();
      registerVer3Class< StatData >();
      registerVer3Class< Summary >();
      registerVer3Class< Event >();
      registerVer3Class< Vect >();
    }
  else if ( version == 4 )
    {
      registerVer4Class< AdcData >();
      registerVer4Class< Detector >();
      registerVer4Class< EndOfFile >();
      registerVer4Class< EndOfFrame >();
      registerVer4Class< Frame >();
      registerVer3Class< History >();
      registerVer3Class< Msg >();
      registerVer4Class< ProcData >();
      registerClass< RawData >();
      registerClass< SerData >();
      registerVer4Class< SimData >();
      registerClass< StatData >();
      registerVer4Class< Summary >();
      registerVer4Class< Event >();
      registerVer4Class< Vect >();
      registerVer4Class< Table >();
      registerClass< SimEvent >();
      registerVer4Class< TOC >();
    }
  else
    {
      throw read_failure( "version is not supported", Input::BADSTREAM );
    }
  mVersion = version;
}


//-----------------------------------------------------------------------------
//
//: Registers a structure header and read method with the Dictionary.
//
//!param: const SH& sh - Structure header.
//!param: ReadPointer read - Read method.
//
//!exc: std::bad_alloc - Memory allocation failed.
//   
void Dictionary::registerClass( const SH& sh, ReadPointer read )
{
    mInfo.insert( std::make_pair( sh.getClass(), ClassInfo( sh, read ) ) );
}



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Find an element by class ID.
//
//!param: INT_2U classId - The ID of the desired element.
//
//!return: iterator - The iterator for the desired element.  This is end()
//+     if the ID didn't exist.
//
//!exc: None.
//   
Dictionary::iterator Dictionary::find( INT_2U classId )
{
    return mInfo.find( classId );
}

//-----------------------------------------------------------------------------
//
//: Find an element by class ID.
//
//!param: INT_2U classId - The ID of the desired element.
//
//!return: const_iterator - The iterator for the desired element.  This is 
//+     end() if the ID didn't exist.
//
//!exc: None.
//      
Dictionary::const_iterator Dictionary::find( INT_2U classId ) const
{
    return mInfo.find( classId );
}


//-----------------------------------------------------------------------------
//
//: Find an element by class name.
//
// If more than one element exists with the same name this will only return an
// iterator to the first element.
//
//!param: const std::string& name - The name of the desired element.
//
//!return: iterator - The iterator for the desired element.  This is end()
//+     if the name didn't exist.
//
//!todo: Rewrite the insert methods so that they will fail if the class name
//!todo: already exists.
//
//!exc: None.
//      
Dictionary::iterator Dictionary::find( const std::string& name )
{
    iterator i = mInfo.begin();
    while( ( i->second.getSH().getName() != name ) && ( i != mInfo.end() ) )
    {
        i++;
    }

    return i;
}


//-----------------------------------------------------------------------------
//
//: Find an element by class name.
//
// If more than one element exists with the same name this will only return an
// iterator to the first element.
//
//!param: const std::string& name - The name of the desired element.
//
//!return: const_iterator - The iterator for the desired element.  This is
//+     end() if the name didn't exist.
//
//!todo: Rewrite the insert methods so that they will fail if the class name
//!todo: already exists.
//
//!exc: None.
//      
Dictionary::const_iterator Dictionary::find( const std::string& name ) const
{
    const_iterator i = mInfo.begin();
    while( ( i->second.getSH().getName() != name ) && ( i != mInfo.end() ) )
    {
        i++;
    }

    return i;
}



//-----------------------------------------------------------------------------
// Mutators
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Erase the element with the given class name.
//
// If the name doesn't exist then nothing is erased.  If multiple classes
// exist with the same name then only the first is removed.
//
//!param: const std::string& name - The name of the element to remove.
//
//!todo: Rewrite the insert methods so that they will fail if the class name
//!todo: already exists.
//
//!exc: None.
//      
void Dictionary::erase( const std::string& name )
{
    iterator i = mInfo.begin();
    while( ( i->second.getSH().getName() != name ) && ( i != mInfo.end() ) )
    {
        ++i;
    }

    if ( i != end() )
    {
        mInfo.erase( i );
    }
}


//-----------------------------------------------------------------------------
//
//: Call the read function for the class with the given ID.
//
//!param: Input& in - The Input object to read from.
//!param: INT_2U classId - The ID of the class to read.
//
//!return: Base* - The pointer to the instantiated object.  The memory for this
//+     object is allocated by this method, the user is responsible for
//+     destructing this object.  If the the class ID could not be found
//+     this will be NULL.
//
//!todo: Should the Input& parameter be change to FrameReaderBase&?
//
//!exc: not_found_error - Data is not found.
//!exc: read_failure - Read failed.
//   
Base* Dictionary::readElement( Input& in, INT_2U classId ) const
{
    const_iterator i = find( classId );

    if ( i == end() )
    {
        return 0;
    }

    return (*(i->second.getReadFunction()))( in );
}


//-----------------------------------------------------------------------------
//
//: Call the read function for the class with the given name.
//
//!param: Input& in - The Input object to read from.
//!param: const std::string& name - The name of the class to read.
//
//!return: Base* - The pointer to the instantiated object.  The memory for this
//+     object is allocated by this method, the user is responsible for
//+     destructing this object.  If the the class ID could not be found
//+     this will be NULL.
//
//!todo: Should the Input& parameter be change to FrameReaderBase&?
//
//!exc: not_found_error - Data is not found.
//!exc: read_failure - Read failed.
//      
Base* Dictionary::readElement( Input& in, const std::string& name ) const
{
    const_iterator i = find( name );
    if ( i == end() )
    {
        throw( not_found_error( "The given class name was not found." ) );
    }

    return (*(i->second.getReadFunction()))( in );
}

