#ifndef FrameCppDictionaryHH
#define FrameCppDictionaryHH


// System Includes
#include <map>

// Local Includes
#include "sh.hh"
#include "classinfo.hh"
#include "errors.hh"

namespace FrameCPP {

namespace Version_3_4_5 {

  class Library;

//-----------------------------------------------------------------------------
//
//: Frame Object Dictionary
//
// Dictonary stores the dictionary elements for the recognized frame classes.
// In addition, the Dictionary also stores a pointer to the read methods for
// these classes.
//
// <p>This static class loads itself with this information when it is
// constructed.
//
// <p>The class is used by the Output object to write structure headers &
// elements.  It is used by the Input object to identify structures on the
// input stream and to locate the proper read method with which to load them.
//
class Dictionary
{
public:

    /* Typedef's */
    typedef std::map< INT_2U, ClassInfo >::iterator iterator;
    typedef std::map< INT_2U, ClassInfo >::const_iterator const_iterator;
    typedef ClassInfo::ReadPointer ReadPointer;
    
    /* Constructors */
    //-----------------------------------------------------------------------------
    //
    //: Constructor.
    //
    // This method initializes the dictionary by registering all of the
    // frame class objects (those which are written as structures to a frame file).
    //
    //!exc: std::bad_alloc - Memory allocation failed in initialization.
    //!exc: read_failure - Read failed.
    //
    Dictionary(int version);

    /* Class Registration */

    //-----------------------------------------------------------------------------
    //
    //: Registers a structure header and read method with the Dictionary.
    //
    //!param: const SH& sh - Structure header.
    //!param: ReadPointer read - Read method.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //   
    void registerClass( const SH& sh, ReadPointer read );
    template< class T > void registerClass();

    //
    //: Registers a frame spec vesion 4 class with the dictionary.
    //
    // This method inserts information about a frame class into the Dictionary.
    // The frame class must define the following two methods:
    //
    // 1. static SH getSH() - A method returning the Structure Header for the
    //    class.
    // 2. static T* read( FrameReaderBase& ) - The read method for the class.
    //
    //!exc: std::bad_alloc - Memory allocation failed during the insertion.
    //
    //!todo: This should return something indicating if the information was
    //!todo: inserted (if the class was already registered, it is not inserted).
    //
    template< class T > inline void registerVer4Class()
    { registerClass( T::getSH4(), (ReadPointer)&T::read4 ); }

    //: Registers a frame spec vesion 3 class with the dictionary.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    template< class T > inline void registerVer3Class()
    { registerClass( T::getSH3(), (ReadPointer)&T::read3 ); }

    /* Accessors */
    //-----------------------------------------------------------------------------
    //
    //: Find an element by class ID.
    //
    //!param: INT_2U classId - The ID of the desired element.
    //
    //!return: iterator -- The iterator for the desired element.  This is end()
    //+     if the ID didn't exist.
    //
    //!exc: None.
    //
    iterator find( INT_2U classId );

    //-----------------------------------------------------------------------------
    //
    //: Find an element by class ID.
    //
    //!param: INT_2U classId - The ID of the desired element.
    //
    //!return: const_iterator -- The iterator for the desired element.  This is 
    //+     end() if the ID didn't exist.
    //
    //!exc: None.
    //
   const_iterator find( INT_2U classId ) const;

    //-----------------------------------------------------------------------------
    //
    //: Find an element by class name.
    //
    // If more than one element exists with the same name this will only return an
    // iterator to the first element.
    //
    //!param: const std::string& name - The name of the desired element.
    //
    //!return: iterator -- The iterator for the desired element.  This is end()
    //+     if the name didn't exist.
    //
    //!todo: Rewrite the insert methods so that they will fail if the class name
    //!todo: already exists.
    //
    //!exc: None.
    //   
    iterator find( const std::string& name );

    //-----------------------------------------------------------------------------
    //
    //: Find an element by class name.
    //
    // If more than one element exists with the same name this will only return an
    // iterator to the first element.
    //
    //!param: const std::string& name - The name of the desired element.
    //
    //!return: const_iterator -- The iterator for the desired element.  This is
    //+     end() if the name didn't exist.
    //
    //!todo: Rewrite the insert methods so that they will fail if the class name
    //!todo: already exists.
    //
    //!exc: None.
    //   
    const_iterator find( const std::string& name ) const;

    //!exc: None.
    Dictionary::iterator begin();
   
    //!exc: None.   
    Dictionary::const_iterator begin() const;
   
    //!exc: None.   
    Dictionary::iterator end();
   
    //!exc: None.   
    Dictionary::const_iterator end() const;
   
    //!exc: None.   
    unsigned int size() const;
  
    //
    //: Return frame data format version
    //
    //!exc: None.
    //   
    int getVersion() const { return mVersion; };

    /* Mutators */
    //!exc: std::bad_alloc - Memory allocation failed.
    bool insert( const SH& sh, ReadPointer read ); 

    //!exc: None.   
    void erase( iterator start, iterator finish );
   
    //!exc: None.   
    void erase( iterator pos );
   
    //!exc: None.   
    void erase( INT_2U classId );

    //-----------------------------------------------------------------------------
    //
    //: Erase the element with the given class name.
    //
    // If the name doesn't exist then nothing is erased.  If multiple classes
    // exist with the same name then only the first is removed.
    //
    //!param: const std::string& name - The name of the element to remove.
    //
    //!todo: Rewrite the insert methods so that they will fail if the class name
    //!todo: already exists.
    //
    //!exc: None.
    //
    void erase( const std::string& name );
   
    /* Read methods */
    //-----------------------------------------------------------------------------
    //
    //: Call the read function for the class with the given ID.
    //
    //!param: Input& in - The Input object to read from.
    //!param: INT_2U classId - The ID of the class to read.
    //
    //!return: Base* -- The pointer to the instantiated object.  The memory for this
    //+     object is allocated by this method, the user is responsible for
    //+     destructing this object.  If the the class ID could not be found
    //+     this will be NULL.
    //
    //!todo: Should the Input& parameter be change to FrameReaderBase&?
    //
    //!exc: not_found_error - Data is not found.
    //!exc: read_failure - Read failed.
    //
    Base* readElement( Input& in, INT_2U classId ) const;

    //-----------------------------------------------------------------------------
    //
    //: Call the read function for the class with the given name.
    //
    //!param: Input& in - The Input object to read from.
    //!param: const std::string& name - The name of the class to read.
    //
    //!return: Base* -- The pointer to the instantiated object.  The memory for this
    //+     object is allocated by this method, the user is responsible for
    //+     destructing this object.  If the the class ID could not be found
    //+     this will be NULL.
    //
    //!todo: Should the Input& parameter be change to FrameReaderBase&?
    //
    //!exc: not_found_error - Data is not found.
    //!exc: read_failure - Read failed.
    //
    Base* readElement( Input& in, const std::string& name ) const;

private:
    
  std::map< INT_2U, ClassInfo > mInfo;
  //: Frame data format version described by this dictionary
  int mVersion;
};


//-----------------------------------------------------------------------------
//
//: Registers a frame class with the dictionary.
//
// This method inserts information about a frame class into the Dictionary.
// The frame class must define the following two methods:
//
// 1. static SH getSH() - A method returning the Structure Header for the class.
// 2. static T* read( FrameReaderBase& ) - The read method for the class.
//
//!exc: std::bad_alloc - Memory allocation failed during the insertion.
//
//!todo: This should return something indicating if the information was
//!todo: inserted (if the class was already registered, it is not inserted).
//
template< class T >
void Dictionary::registerClass()
{
    registerClass( T::getSH(), (ReadPointer)&T::read );
}



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Returns an iterator at the beginning of the dictionary.
//
//!return: iterator -- An iterator pointing to the beginning of the dictionary.
//
//!exc: None.
//   
inline Dictionary::iterator Dictionary::begin()
{
    return mInfo.begin();
}


//-----------------------------------------------------------------------------
//
//: Returns a constant iterator at the beginning of the dictionary.
//
//!return: const_iterator -- A constant iterator pointing to the beginning of
//+     the dictionary.
//
//!exc: None.
//   
inline Dictionary::const_iterator Dictionary::begin() const
{
    return mInfo.begin();
}


//-----------------------------------------------------------------------------
//
//: Returns an iterator at the end of the dictionary.
//
//!return: iterator -- An iterator pointing to the end of the dictionary.
//
//!exc: None.
//   
inline Dictionary::iterator Dictionary::end()
{
    return mInfo.end();
}


//-----------------------------------------------------------------------------
//
//: Returns a constant iterator at the end of the dictionary.
//
//!return: const_iterator -- A constant iterator pointing to the end of the
//!return:     dictionary.
//
//!exc: None.
//   
inline Dictionary::const_iterator Dictionary::end() const
{
    return mInfo.end();
}


//-----------------------------------------------------------------------------
//
//: The number of elements in the dictionary.
//
//!return: unsigned int -- The number of elements in the dictionary.
//
//!exc: None.
//      
inline unsigned int Dictionary::size() const
{
    return mInfo.size();
}



//-----------------------------------------------------------------------------
// Mutators
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Inserts a Structure Header with its read pointer into the dictionary.
//
// If a structure header with the same class ID already exists in the
// dictionary then nothing will be added and the method will return false.
//
//!param: const SH& sh - the structure header to add.
//!param: ReadPointer rp - A pointer to the read function which whould be used
//+     to read this class.
//
//!return: bool -- True if the data was inserted, false if a structure header
//+     with the same Class ID already exists in the dictionary (the
//+     data will not be inserted).
//
//!exc: std::bad_alloc - Memory allocation failed during insertion.
//
//!todo: Rewrite the insert methods so that they will fail if the class name
//!todo: already exists.
//
inline bool Dictionary::insert( const SH& sh, ReadPointer rp )
{
    return mInfo.insert(
        std::make_pair( sh.getClass(), ClassInfo( sh, rp ) ) ).second;
}


//-----------------------------------------------------------------------------
//
//: Erase elements within a range.
//
// Erases all of the dictionary elements between two iterators:
// [start,finish).  If finish is < start then the result is undefined.
//
//!param: iterator start - The first element to delete.
//!param: iterator finish - The final element, it is not deleted.
//
//!todo: Check for finish < start (or find out what STL does).
//   
//!exc: None.
//      
inline void Dictionary::erase( iterator start, iterator finish )
{
    mInfo.erase( start, finish );
}


//-----------------------------------------------------------------------------
//
//: Erase an element.
//
//!param: iterator pos - The element to delete.  If this is out-of-range then
//+     the behavior is undefined.
//
//!todo: Check for out-of-range (or find out what STL does).
//
//!exc: None.
//      
inline void Dictionary::erase( iterator pos )
{
    mInfo.erase( pos );
}


//-----------------------------------------------------------------------------
//
//: Erase an element with the specified ID.
//
//!param: INT_2U classId - The ID of the element to erase.  If the ID doesn't
//!param:    exist then nothing is erased.
//
//!exc: None.
//      
inline void Dictionary::erase( INT_2U classId )
{
    mInfo.erase( classId );
}
 

//-----------------------------------------------------------------------------
//
//: Contains Dictionary for every supported Frame Data Specification version
//
// <p>This static class loads itself with this information when it is
// constructed.
//
// <p>The class is used by the Input object to find dictionary for certain
// frame data format version.
//
class Library {

public:
  Library(): dictionary3(3), dictionary4(4), dictionary5(5) {}

  //: Current FrameCPP data format version
  static const int FRAMECPP_VERSION;
  
  //: Current FrameCPP software minor version
  static const int FRAMECPP_REVISION;
   
  //: Get a dictionary for the version
  //!exc: read_failure - Read failed.
  Dictionary* getVersionDictionary( int version = FRAMECPP_VERSION )
   {
      if ( version == FRAMECPP_VERSION )
	{
	  return &dictionary5;
	}
      else if ( version == 4 )
	{
	  return &dictionary4;
	}
      else if ( version == 3 )
	{
	  return &dictionary3;
	}

      throw read_failure( "version not supported", 0);
    }
  //: Get a dictinary for the latest frame format version
  inline Dictionary* getCurrentVersionDictionary() { return &dictionary5; }

private:

  /* Specific dictionaries for data format versions */
  //: Version 3 frame dictionary.
  Dictionary dictionary3;
  //: Version 4 frame dictionary.
  Dictionary dictionary4;
  //: Version 5 frame dictionary.
  Dictionary dictionary5;
};

extern Library library;

} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppDictionaryHH
