#ifndef FrameCppDimensionHH
#define FrameCppDimensionHH


// System Includes
#include <memory>
#include <string>

#include "framecpp/config.h"

// Local Includes
#include "types.hh"

namespace FrameCPP {

namespace Version_3_4_5 {

//-----------------------------------------------------------------------------
//
//: Dimension information for a Vect.
//
// This is a simple class storing dimension information for a Vect object.
// It stores the following:
// <p><ol>
//     <li> The dimension length. </li>
//     <li> Scale factor </li>
//     <li> The units (unit per step size along the coordintate). </li>
// </ol>
//
class Dimension
{
public:

    /* Constructors/Destructor */
    //!exc: None.
    Dimension();
    
    //!exc: std::bad_alloc - Memory allocation failed.
    Dimension( const Dimension& dim );
   
    //!exc: std::bad_alloc - Memory allocation failed.   
    explicit Dimension( INT_4U nx, REAL_8 dx = 1.0, const std::string& unitX = "",
                        REAL_8 startX = 0.0 );
    
    /* Operator Overloads */
    //-----------------------------------------------------------------------------
    //
    //: Assignment operator.
    //
    //!param: const Dimension& dim - The object to assign from.
    //
    //!return: const Dimension& -- This object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const Dimension& operator=( const Dimension& dim );

    //-----------------------------------------------------------------------------
    //
    //: Equal comparison.
    //
    //!param: const Dimension& dim - The object to compare with.
    //
    //!return: bool -- true if the objects are equal, false otherwise.
    //
    //!exc: None.
    //
    bool operator==( const Dimension& dim ) const;

    //-----------------------------------------------------------------------------
    //
    //: Not equal comparison.
    //
    //!param: const Dimension& dim - The object to compare with.
    //
    //!return: bool -- false if the objects are equal, true otherwise.
    //
    //!exc: None.
    //
    bool operator!=( const Dimension& dim ) const;
    
    /* Accessors */
    //!exc: None.
    INT_4U getNx() const;
   
    //!exc: None.   
    REAL_8 getDx() const;
   
    //!exc: None.   
    REAL_8 getStartX() const;
   
    //!exc: None.   
    const std::string& getUnitX() const;
    
private:
    //: Dimension length.
    INT_4U mNx;
    //: Scale factor.
    REAL_8 mDx;
    //: Scale factor in ASCII.
    std::string mUnitX;
    //: Origin.
    REAL_8 mStartX;
};




/******************/
/* Inline Methods */
/******************/


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates a dimension object with the following values:
//
//   Dimension Length (Nx) - 1
//   Scale Factor (Dx) - 1.0
//   Units (UnitX) - ""
//
//!exc: None.   
//   
inline Dimension::Dimension()
        : mNx( 1 ), mDx( 1.0 ),  mUnitX( "" ), mStartX( .0 )
{
}


//-----------------------------------------------------------------------------
//
//: Copy constructor.
//
//!param: const Dimension& dim - The Dimension to copy from.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
inline Dimension::Dimension( const Dimension& dim )
   : mNx( dim.mNx ), mDx( dim.mDx ), mUnitX( dim.mUnitX ), 
     mStartX( dim.mStartX )
{
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
// This creates a Dimension object with the specified attributes.
//
//!param: INT_4U nx - The dimension length.
//!param: REAL_8 dx - The Scale Factor. Default: 1.0
//!param: const std::string& unitX - The units (unit per step size). Default: ""
//
//!exc: std::bad_alloc - Memory allocation failed.
//   
inline Dimension::Dimension(INT_4U nx, REAL_8 dx,
			    const std::string& unitX, REAL_8 startx )
  : mNx( nx ), mDx( dx ), mUnitX( unitX ), mStartX( startx )
{
}



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Get the dimension length.
//
//!return: INT_4U -- The dimension length.
//
//!exc: None.   
//      
inline INT_4U Dimension::getNx() const
{
    return mNx;
}


//-----------------------------------------------------------------------------
//
//: Get the scale factor.
//
//!return: REAL_8 -- The scale factor.
//
//!exc: None.   
//      
inline REAL_8 Dimension::getDx() const
{
    return mDx;
}

//-----------------------------------------------------------------------------
//
//: Get data set origin.
//
//!return: REAL_8 -- Data set origin.
//
//!exc: None.   
//      
inline REAL_8 Dimension::getStartX() const
{
    return mStartX;
}


//-----------------------------------------------------------------------------
//
//: Get the units.
//
//!return: const std::string& -- The units.
//
//!exc: None.   
//      
inline const std::string& Dimension::getUnitX() const
{
    return mUnitX;
}

 
} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // FrameCppDimensionHH
