#include "framecpp/config.h"

#include "endoffile.hh"
#include "input.hh"
#include "output.hh"


using FrameCPP::Version_3_4_5::EndOfFile;


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates an EndOfFile object with the following properties:
// <ul>
//     <li> nFrames = 0 </li>
//     <li> nBytes = 0 </li>
//     <li> chkType = 0 </li>
//     <li> chkSum = 0 </li>
//     <li> seekTOC = 0 </li>
// </ul>
//
//!exc: None.
//     
EndOfFile::EndOfFile()
        : Base(), mNFrames( 0 ), mNBytes( 0 ),
	  mChkType( 0 ), mChkSum( 0 ), mSeekTOC( 0 )
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
// This performs a deep copy of the object.
//
//!param: const EndOfFile& eof - Object to be copied.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
EndOfFile::EndOfFile( const EndOfFile& eof )
        : Base( eof ), mNFrames( eof.mNFrames ),
          mNBytes( eof.mNBytes ), mChkType( eof.mChkType ),
          mChkSum( eof.mChkSum ), mSeekTOC( eof.mSeekTOC )
{
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: INT_4U frames - The number of frame in the file.
//!param: INT_4U bytes - The number of bytes in this file.
//!param: INT_4U chktype - Checksum scheme: 0 = none, 1 = CRC.
//!param: INT_4U chksum - File checksum.
//!param: INT_4U seekTOC - Bytes to back up to the beginning of the TOC.
//
//!exc: None.      
EndOfFile::EndOfFile(
    INT_4U frames, INT_4U bytes, INT_4U chktype, INT_4U chksum, INT_4U seekTOC ) 
        : mNFrames( frames ), mNBytes( bytes ), mChkType( chktype ),
          mChkSum( chksum ), mSeekTOC( seekTOC )
{
}
   
   
//-----------------------------------------------------------------------------
//
//: Destructor.
//
// This method currently does nothing.
//
//!exc: None.
//   
EndOfFile::~EndOfFile()
{
}
   

#ifdef not_def
//-----------------------------------------------------------------------------
//
//: Input Constructor.
//
// This private constructor initializes the object from
// an Input object.  This constructor assumes that the length, class, and
// instance have already been read.
//
//!param: Input & in -
//
//!exc: std::bad_alloc
//!exc: read_failure
//
EndOfFile::EndOfFile( Input& in )
        : Base(), mNFrames( 0 ), mNBytes( 0 ), mChkType( 0 ), mChkSum( 0 )
{
    in >> mNFrames >> mNBytes >> mChkType >> mChkSum;
}
#endif


//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
//!param: const EndOfFile& eof - Object to assign.
//
//!return: const EndOfFile& - This object.
//
//!exc: None.
//
const EndOfFile& EndOfFile::operator=( const EndOfFile& eof )
{
    if ( this != &eof )
    {
        mNFrames = eof.mNFrames;
        mNBytes = eof.mNBytes;
        mChkType = eof.mChkType;
        mChkSum = eof.mChkSum;
        mSeekTOC = eof.mSeekTOC;
    }

    return *this;
}



//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Constructs the Structure Header for this class.
//
//!return: SH - Structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH EndOfFile::getSH()
{
  SH sh( "FrEndOfFile", ID_ENDOFFILE, "End of File" );
  sh.appendSE( SE( "nFrames", "INT_4U", "" ) );
  sh.appendSE( SE( "nBytes", "INT_4U", "" ) );
  sh.appendSE( SE( "chkType", "INT_4U", "" ) );
  sh.appendSE( SE( "chkSum", "INT_4U", "" ) );
  sh.appendSE( SE( "seekTOC", "INT_4U", "" ) );
  return sh;
}


//-----------------------------------------------------------------------------
//
//: Constructs frame data format version 4 Structure Header for this class.  
//
//!return: SH - Structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH EndOfFile::getSH4()
{
  SH sh( "FrEndOfFile", ID_ENDOFFILE, "End of File" );
  sh.appendSE( SE( "nFrames", "INT_4U", "" ) );
  sh.appendSE( SE( "nBytes", "INT_4U", "" ) );
  sh.appendSE( SE( "chkFlag", "INT_4U", "" ) );
  sh.appendSE( SE( "chkSum", "INT_4U", "" ) );
  sh.appendSE( SE( "seekTOC", "INT_4U", "" ) );
  return sh;
}


//-----------------------------------------------------------------------------
//
//: Constructs frame data format version 3 Structure Header for this class.  
//
//!return: SH - Structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH EndOfFile::getSH3()
{
    SH sh( "FrEndOfFile", ID_ENDOFFILE, "End of File" );
    sh.appendSE( SE( "nFrames", "INT_4U", "" ) );
    sh.appendSE( SE( "nBytes", "INT_4U", "" ) );
    sh.appendSE( SE( "chkFlag", "INT_4U", "" ) );
    sh.appendSE( SE( "chkSum", "INT_4U", "" ) );

    return sh;
}


//-----------------------------------------------------------------------------
//
//: Reads an EndOfFile object from an Input object.
//
// This method creates a dynamically allocated object.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: Input& in - Input data stream object.
//
//!return: EndOfFile* - Newly created.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//
EndOfFile* EndOfFile::read4( Input& in )
{
  EndOfFile& eof = *new EndOfFile();

  in >> eof.mNFrames >> eof.mNBytes >> eof.mChkType >> eof.mChkSum >> eof.mSeekTOC;
  return &eof;
}

//: Version 3 frame input method.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//   
EndOfFile* EndOfFile::read3( Input& in )
{
  EndOfFile& eof = *new EndOfFile();
  in >> eof.mNFrames >> eof.mNBytes >> eof.mChkType >> eof.mChkSum;
  return &eof;
}


//-----------------------------------------------------------------------------
//
//: Writes an EndOfFile object to an Output object.
//
//!param: Output& out - Output data stream object.
//
//!exc: write_failure - Write failed.
//
void EndOfFile::write( Output& out ) const
{
    out.startRecord( ID_ENDOFFILE,
		     24
		     + 4, // for seekTOC integer
		     this );
    out << mNFrames << mNBytes << mChkType << mChkSum << mSeekTOC;
    out.endRecord( this );
}

   
//-----------------------------------------------------------------------------
//
//: Gets the class ID.
//
//!return: INT_2U -- Class ID.
//
//!exc: None.
//      
INT_2U EndOfFile::getClassId() const
{
    return ID_ENDOFFILE;
}
   
