#ifndef FrameCppEndOfFileHH
#define FrameCppEndOfFileHH


#include "base.hh"
#include "sh.hh"


namespace FrameCPP {

namespace Version_6 {
  class FrEndOfFile;
} // namespace - Version_6

namespace Version_3_4_5 {

//-----------------------------------------------------------------------------
//
//: This class represents the end of a file.
//
// It is modeled after the FrEndOfFile structure in the Frame Specification,
// LIGO-T970130-B-E.
//
class EndOfFile : Base
{
public:

    /* Constructors */
    //-----------------------------------------------------------------------------
    //
    //: Default Constructor.
    //
    // This creates an EndOfFile object with the following properties:
    // <ul>
    //     <li> nFrames = 0 </li>
    //     <li> nBytes = 0 </li>
    //     <li> chkType = 0 </li>
    //     <li> chkSum = 0 </li>
    //     <li> seekTOC = 0 </li>
    // </ul>
    //
    //!exc: None.   
    //      
    EndOfFile();

    //-----------------------------------------------------------------------------
    //
    //: Copy Constructor.
    //
    // This performs a deep copy of the object.
    //
    //!param: const EndOfFile& eof - Object to be copied.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    EndOfFile( const EndOfFile& eof );


    //-----------------------------------------------------------------------------
    //
    //: Constructor.
    //
    //!param: INT_4U frames - The number of frame in the file.
    //!param: INT_4U bytes - The number of bytes in this file.
    //!param: INT_4U chktype - Checksum scheme: 0 = none, 1 = CRC.
    //!param: INT_4U chksum - File checksum.
    //!param: INT_4U seekTOC - Bytes to back up to the beginning of the TOC.
    //
    //!exc: None.   
    //         
    EndOfFile( INT_4U frames, INT_4U bytes, INT_4U chktype, INT_4U chksum, INT_4U seekTOC );
    
    /* Destructor */
    //!exc: None.   
    //         
    virtual ~EndOfFile();

    /* Operator Overloads */
    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    //!param: const EndOfFile& eof - Object to assign.
    //
    //!return: const EndOfFile& -- This object.
    //
    //!exc: None.
    //
    const EndOfFile& operator=( const EndOfFile& eof );

    /* Required Methods */
    //!exc: None.   
    //         
    virtual INT_2U getClassId() const;

    //-----------------------------------------------------------------------------
    //
    //: Constructs the Structure Header for this class.
    //
    //!return: SH -- Structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH();

    //: Constructs frame data format version 4 Structure Header for this class.  
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //   
    static SH getSH4();
   
    //: Constructs frame data format version 3 Structure Header for this class.  
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //   
    static SH getSH3();

    /* Accessors */
    //!exc: None.      
    INT_4U getNFrames() const;
   
    //!exc: None.      
    INT_4U getNBytes() const;
   
    //!exc: None.         
    INT_4U getChkType() const;
   
    //!exc: None.         
    INT_4U getChkSum() const;
   
    //!exc: None.         
    INT_4U getSeekTOC() const;

private:
    friend class Dictionary;
    friend class FrameCPP::Version_6::FrEndOfFile;

    //-----------------------------------------------------------------------------
    //
    //: Reads an EndOfFile object from an Input object.
    //
    // This method creates a dynamically allocated object.
    //
    // <p>*NOTE* This method allocates memory and returns a pointer to it.  The
    // user is responsible for freeing this memory.
    //
    //!param: Input& in - Input data stream object.
    //
    //!return: EndOfFile* -- Newly created.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //
   static EndOfFile* read( Input& in ){ return read4( in ); }

    //: Version 4 frame input method.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //   
    static EndOfFile* read4( Input& in );

    //: Version 3 frame input method.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //   
    static EndOfFile* read3( Input& in );

    //-----------------------------------------------------------------------------
    //
    //: Writes an EndOfFile object to an Output object.
    //
    //!param: Output& out - Output data stream object.
    //
    //!exc: write_failure - Write failed.
    //
    virtual void write( Output& out ) const;
    
private:

    //: The number of frame in the file.
    INT_4U mNFrames;
    //: The number of bytes in this file.
    INT_4U mNBytes;
    //: Checksum scheme: 0 = none, 1 = CRC.
    INT_4U mChkType;
    //: File checksum.
    INT_4U mChkSum;
    //: Bytes to back up to the beginning of the TOC.
    INT_4U mSeekTOC;
};
    


/******************/
/* Inline Methods */
/******************/

//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Gets the number of frames in this file.
//
//!return: INT_4U -- Number of frames.
//
//!exc: None.
//      
inline INT_4U EndOfFile::getNFrames() const
{
    return mNFrames;
}


//-----------------------------------------------------------------------------
//
//: Gets the total number of bytes in this file.
//
// This is zero if it is not computed.
//
//!return: INT_4U -- Number of bytes.
//
//!exc: None.
//      
inline INT_4U EndOfFile::getNBytes() const
{
    return mNBytes;
}


//-----------------------------------------------------------------------------
//
//: Returns the checksum type.
//
// This is one is CRC checksum was calculated, zero if not.
// All other values are unused as of version 5 of frame data format.
//
//!return: INT_4U -- Checksum type.
//
//!exc: None.
//      
inline INT_4U EndOfFile::getChkType() const
{
    return mChkType;
}


//-----------------------------------------------------------------------------
//
//: Gets the checksum value.
//
//!return: INT_4U -- Checksum.
//
//!exc: None.
//      
inline INT_4U EndOfFile::getChkSum() const
{
    return mChkSum;
}


//-----------------------------------------------------------------------------
//
//: Gets the number of bytes to back up to the beginning of the TOC
//
//!return: INT_4U -- TOC offset.
//
//!exc: None.
//      
inline INT_4U EndOfFile::getSeekTOC() const
{
    return mSeekTOC;
}

 
} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppEndOfFileHH
