#ifndef FrameCppEndOfFrameHH
#define FrameCppEndOfFrameHH


#include "base.hh"
#include "sh.hh"


namespace FrameCPP {

namespace Version_6
{
  class FrEndOfFrame;
}
namespace Version_3_4_5 {

class FrameReaderBase;


//-----------------------------------------------------------------------------
//
//: This class represents the end of a frame.
//
// It is modeled after the FrEndOfFrame structure in the Frame Specification,
// LIGO-T970130-B-E.
//
class EndOfFrame : public Base
{
public:

    /* Constructors */
    //-----------------------------------------------------------------------------
    //
    //: Default Constructor.
    //
    // This creates an EndOfFrame object with the following properties:
    // <ul>
    //     <li> run = 0 </li>
    //     <li> frame = 0 </li>
    //     <li> chkType = 0 </li>
    //     <li> chkSum = 0 </li>
    // </ul>
    //
    //!exc: None.
    //
    EndOfFrame();

    //-----------------------------------------------------------------------------
    //
    //: Copy Constructor.
    //
    // This performs a deep copy of the object.
    //
    //!param: const EndOfFrame& eofr - The object to be copied.
    //
    //!exc: None.
    //   
    EndOfFrame( const EndOfFrame& eof );

    //-----------------------------------------------------------------------------
    //
    //: Constructor.
    //
    //!param: INT_4S run - Frame run number.
    //!param: INT_4U frame - Frame number
    //!param: INT_4U chktype - Frame checksum scheme.
    //!param: INT_4U chksum - File checksum.
    //
    //!exc: None.
    //   
    EndOfFrame( INT_4S run, INT_4U frame, INT_4U chktype, INT_4U chksum );
    
    /* Destructor */
    //!exc: None.
    //   
    virtual ~EndOfFrame();

    /* Operator Overloads */

    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    //!param: const EndOfFrame& eofr - Object to be assigned.
    //
    //!return: const EndOfFrame& -- This object.
    //
    //!exc: None.
    //
    const EndOfFrame& operator=( const EndOfFrame& eof );

    /* Required Methods */
    //!exc: None.
    virtual INT_2U getClassId() const;

    //-----------------------------------------------------------------------------
    //
    //: Constructs the Structure Header for this class.  
    //
    //!return: SH -- End of frame structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH();

    //-----------------------------------------------------------------------------
    //
    //: Constructs frame data format version 4 Structure Header for this class.  
    //
    //!return: SH -- End of frame structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH4();


    //-----------------------------------------------------------------------------
    //
    //: Constructs frame data format version 3 Structure Header for this class.  
    //
    //!return: SH -- End of frame structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH3();

    /* Accessors */
    //!exc: None.
    INT_4S getRun() const;
   
    //!exc: None.   
    INT_4U getFrame() const;
   
    //!exc: None.   
    INT_4U getChkType() const;
   
    //!exc: None.   
    INT_4U getChkSum() const;

private:
  friend class Dictionary;
  friend class FrameCPP::Version_6::FrEndOfFrame;

    /* This method will be accessed by the Dictionary class but is
       not available for public use. */

    //-----------------------------------------------------------------------------
    //
    //: Reads an EndOfFrame object from an FrameReaderBase object.
    //
    // This method creates a dynamically allocated object.
    //
    // <p>*NOTE* This method allocates memory and returns a pointer to it.  The
    // user is responsible for freeing this memory.
    //
    //!param: FrameReaderBase& in - Input data stream object.
    //
    //!return: EndOfFrame* -- Newly created.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //
    static EndOfFrame* read( FrameReaderBase& in );
   
    //: Version 4 Frame read method.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //   
    static EndOfFrame* read4( FrameReaderBase& in );
   
    //: Version 3 Frame read method.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //   
    static EndOfFrame* read3( FrameReaderBase& in );

private:

    //-----------------------------------------------------------------------------
    //
    //: Writes an EndOfFrame object to an Output object.
    //
    //!param: Output& out out - Output data stream object.
    //
    //!exc: write_failure - Write failure occured.
    //
    virtual void write( Output& out ) const;

    //: Frame run number.
    INT_4S mRun;
    //: Frame number.
    INT_4U mFrame;
    //: Checksum schemes: 0 = none, 1 = CRC.
    INT_4U mChkType;
    //: File checksum.
    INT_4U mChkSum;
};
    


/******************/
/* Inline Methods */
/******************/

//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Get the run number.
//
//!return: INT_4S -- Run number.
//
//!exc: None.
//      
inline INT_4S EndOfFrame::getRun() const
{
    return mRun;
}


//-----------------------------------------------------------------------------
//
//: Get the frame number.
//
//!return: INT_4U -- Frame number.
//
//!exc: None.
//      
inline INT_4U EndOfFrame::getFrame() const
{
    return mFrame; 
}


//-----------------------------------------------------------------------------
//
//: Returns the checksum type.
//
// This is one is CRC checksum was calculated, zero if not.
// All other values are unused as of version 5 of frame data format.
//
//!return: INT_4U -- Checksum type.
//
//!exc: None.
//      
inline INT_4U EndOfFrame::getChkType() const
{
    return mChkType;
}


//-----------------------------------------------------------------------------
//
//: Gets frame checksum value.
//
//!return: INT_4U -- Frame checksum.
//
//!exc: None.
//      
inline INT_4U EndOfFrame::getChkSum() const
{
    return mChkSum;
}

 
} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppEndOfFrameHH
