#include "framecpp/config.h"

#include <stdexcept>

#include "event.hh"
#include "framereader.hh"
#include "output.hh"


using FrameCPP::Version_3_4_5::Event;
using FrameCPP::Version_3_4_5::Vect;
using FrameCPP::Version_3_4_5::Container;
using FrameCPP::Version_3_4_5::Base;

using std::runtime_error;



//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates an Event object with the following properties:
// <ul type=disc>
//     <li> name = "none" </li>
//     <li> comment = "" </li>
//     <li> inputs = "" </li>
//     <li> GTime = 0 seconds, 0 nanoseconds </li>
//     <li> timeBefore = 0 </li>
//     <li> timeAfter = 0 </li>
//     <li> eventStatus = 0 </li>
//     <li> amplitude = 0 </li>
//     <li> probability = -1 </li>
//     <li> statistics = "" </li>
//     <li> parameters = empty list </li>
//     <li> No other data. </li>
// </ul>
//
//!exc: None.
//      
Event::Event()
        : Base(),
          mName( "none" ), mComment( "" ), mInputs( "" ),
          mGTime( 0, 0 ),
	  mTimeBefore( 0. ), mTimeAfter( 0. ),
	  mEventStatus( 0 ), mAmplitude( 0. ),
	  mProbability( -1.0 ),
          mStatistics( "" ), mParameters(), mData()
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
// This performs a deep copy of the object.
//
//!param: const Event& eventData - Event data to be copied.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
Event::Event( const Event& eventData )
        : Base( eventData ),
          mName( eventData.mName ),
          mComment( eventData.mComment ), mInputs( eventData.mInputs ),
          mGTime( eventData.mGTime ),
	  mTimeBefore( eventData.mTimeBefore ), mTimeAfter( eventData.mTimeAfter ),
	  mEventStatus( eventData.mEventStatus ), mAmplitude( eventData.mAmplitude ),
	  mProbability( eventData.mProbability ),
          mStatistics( eventData.mStatistics ),
          mParameters( eventData.mParameters ),
	  mData( eventData.mData )
{
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
// Creates a Event object with the specified attributes but no data.
//
//!param: const std::string& name - Event name.
//!param: const std::string& comment - Event description.
//!param: const std::string& inputs - Input channels and filter parameters to
//+     event process.
//!param: const Time& time - Event occurrence time.
//!param: REAL_4 timeBefore -  Signal duration before event occurrence.
//!param: REAL_4 timeAfter - Signar dutaion after event occurrence.
//!param: INT_4U eventStatus - Defined by event.
//!param: REAL_4 amplitude - Continuous output amplitude returned by event.
//!param: REAL_4 prob - Likelihood estimate of event.
//!param: const std::string& statistics - Statistical description.
//!param: const std::vector<std::pair<std::string, REAL_4> >& parameters - Additional event parameters.
//
//!exc: None.
//      
Event::Event(
    const std::string& name, const std::string& comment, const std::string& inputs,
    const Time& time,
    REAL_4 timeBefore, REAL_4 timeAfter,
    INT_4U eventStatus, REAL_4 amplitude,
    REAL_4 prob,
    const std::string& statistics,
    const std::vector<std::pair<std::string, REAL_4> >& parameters)
        : Base(), 
          mName( name ), mComment( comment ), mInputs( inputs ),
          mGTime( time ),
	  mTimeBefore( timeBefore ), mTimeAfter( timeAfter ),
	  mEventStatus( eventStatus ), mAmplitude( amplitude ),
	  mProbability( prob ),
	  mStatistics( statistics ), mParameters(parameters), mData()
{
}


//-----------------------------------------------------------------------------
//
//: Event name.
//
// Gets the event name.
//
//!return: const std::string& -- The event name.
//
const std::string& Event::getName() const
{
    return mName;
}



//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
// A deep copy is performed for owned data.
//
//!param: const Event& eventData - Data to be copied.
//
//!return: const Event& - This event.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const Event& Event::operator=( const Event& eventData )
{
    if ( this != &eventData )
    {
        mName = eventData.mName;
        mComment = eventData.mComment;
        mInputs = eventData.mInputs;
        mGTime = eventData.mGTime;
	mTimeBefore = eventData.mTimeBefore;
	mTimeAfter = eventData.mTimeAfter;
	mEventStatus = eventData.mEventStatus;
	mAmplitude = eventData.mAmplitude;
        mProbability = eventData.mProbability;
        mStatistics = eventData.mStatistics;
        mParameters = eventData.mParameters;
        mData = eventData.mData;
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: += Operator.
//
//!param: const Event& event - Data to be added.
//
//!return: const Event& -- This object.
//
const Event& Event::operator+=( const Event& event )
{
   throw runtime_error( "Unimplemented Event::operator+=() is called." );
  
   return *this;
}


//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Structure Header.
//
// Constructs the Structure Header for this class.
//
//!return: SH - Structure header for Event Data.
//
//!exc: std::bad_alloc - Memory allocation failed
//
FrameCPP::Version_3_4_5::SH Event::getSH()
{
  SH sh( "FrEvent", ID_EVENT, "Event" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "comment", "STRING", "" ) );
  sh.appendSE( SE( "inputs", "STRING", "" ) );
  sh.appendSE( SE( "GTimeS", "INT_4U", "" ) );
  sh.appendSE( SE( "GTimeN", "INT_4U", "" ) );
  sh.appendSE( SE( "timeBefore", "REAL_4", "" ) );
  sh.appendSE( SE( "timeAfter", "REAL_4", "" ) );
  sh.appendSE( SE( "eventStatus", "INT_4U", "" ) );
  sh.appendSE( SE( "amplitude", "REAL_4", "" ) );
  sh.appendSE( SE( "probability", "REAL_4", "" ) );
  sh.appendSE( SE( "statistics", "STRING", "" ) );
  sh.appendSE( SE( "nParam", "INT_2U", "" ) );
  sh.appendSE( SE( "parameters", "REAL_4 *", "" ) );
  sh.appendSE( SE( "parameterNames", "STRING *", "" ) );
  sh.appendSE( SE( "data", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "table", "PTR_STRUCT(FrTable *)", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrEvent *)", "" ) );
  return sh;
}

//-----------------------------------------------------------------------------
//
//: Structure Header.
//
// Constructs frame data format vesion 4 Structure Header for this class.
//
//!return: SH - Structure header for Trigger Data.
//
//!exc: std::bad_alloc - Memory allocation failed
//
FrameCPP::Version_3_4_5::SH Event::getSH4() 
{
  SH sh( "FrTrigData", ID_EVENT, "Trigger Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "comment", "STRING", "" ) );
  sh.appendSE( SE( "inputs", "STRING", "" ) );
  sh.appendSE( SE( "GTimeS", "INT_4U", "" ) );
  sh.appendSE( SE( "GTimeN", "INT_4U", "" ) );
  sh.appendSE( SE( "timeBefore", "REAL_4", "" ) );
  sh.appendSE( SE( "timeAfter", "REAL_4", "" ) );
  sh.appendSE( SE( "triggerStatus", "INT_4U", "" ) );
  sh.appendSE( SE( "amplitude", "REAL_4", "" ) );
  sh.appendSE( SE( "probability", "REAL_4", "" ) );
  sh.appendSE( SE( "statistics", "STRING", "" ) );
  sh.appendSE( SE( "data", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "table", "PTR_STRUCT(FrTable *)", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrTrigData *)", "" ) );
  return sh;
}

//-----------------------------------------------------------------------------
//
//: Structure Header
//
// Constructs frame data format vesion 3 Structure Header for this class.
//
//!return: SH - Structure header for Trigger Data.
//
//!exc: std::bad_alloc - Memory allocation failed
//
FrameCPP::Version_3_4_5::SH Event::getSH3()
{
    SH sh( "FrTrigData", ID_EVENT, "Trigger Data" );
    sh.appendSE( SE( "name", "STRING", "" ) );
    sh.appendSE( SE( "comment", "STRING", "" ) );
    sh.appendSE( SE( "inputs", "STRING", "" ) );
    sh.appendSE( SE( "GTimeS", "INT_4U", "" ) );
    sh.appendSE( SE( "GTimeN", "INT_4U", "" ) );
    sh.appendSE( SE( "bvalue", "INT_4U", "" ) );
    sh.appendSE( SE( "rvalue", "REAL_4", "" ) );
    sh.appendSE( SE( "probability", "REAL_4", "" ) );
    sh.appendSE( SE( "statistics", "STRING", "" ) );
    sh.appendSE( SE( "data", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "next", "PTR_STRUCT(FrTrigData *)", "" ) );

    return sh;
}


//-----------------------------------------------------------------------------
//
//: Reads an Event object from an Input object.
//
// This method creates a dynamically allocated object.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: Input& in - Frame reader.
//
//!return: Event* - Newly allocated and read data.
//
//!exc: std::bad_alloc - Memory allocation failed
//!exc: read_failure - Read failed
//
Event* Event::read( FrameReaderBase& in )
{
  Event& trig = *new Event();
  INT_2U nParam;

  in >> trig.mName >> trig.mComment >> trig.mInputs >> trig.mGTime
     >> trig.mTimeBefore >> trig.mTimeAfter
     >> trig.mEventStatus >> trig.mAmplitude
     >> trig.mProbability >> trig.mStatistics 
     >> nParam;

  for ( unsigned int i = 0; i < nParam; i++ ) {
    std::string name;
    in >> name;
    trig.mParameters.push_back(std::pair<std::string, REAL_4>( name, 0. ));
  }

  for ( unsigned int i = 0; i < nParam; i++ ) {
    in >> trig.mParameters[i].second;
  }

  in >> trig.mData >> trig.mTable;
  
  in.readNextNodeReference();
    
  return &trig;
}

//: Read FrameCPP version 4 Trigger Data structure.
//
//!exc: std::bad_alloc - Memory allocation failed
//!exc: read_failure - Read failed
//   
Event* Event::read4( FrameReaderBase& in )
{
  Event& trig = *new Event();

  in >> trig.mName >> trig.mComment >> trig.mInputs >> trig.mGTime
     >> trig.mTimeBefore >> trig.mTimeAfter
     >> trig.mEventStatus >> trig.mAmplitude
     >> trig.mProbability >> trig.mStatistics >> trig.mData >> trig.mTable;

  
  in.readNextNodeReference();
    
  return &trig;
}

//: Read FrameCPP version 3 Trigger Data structure.
//
//!exc: std::bad_alloc - Memory allocation failed
//!exc: read_failure - Read failed
//   
Event* Event::read3( FrameReaderBase& in )
{
    Event& trig = *new Event();
    INT_4U bvalue;
    REAL_4 rvalue;

    in >> trig.mName >> trig.mComment >> trig.mInputs >> trig.mGTime >> bvalue >> rvalue
       >> trig.mProbability >> trig.mStatistics >> trig.mData;

    // Is this appropriate ???
    trig.mAmplitude = rvalue;

    in.readNextNodeReference();
    return &trig;
}


//-----------------------------------------------------------------------------
//
//: Writes an Event object to an Output object.
//
//!param: Output& out out - Output data object.
//
//!exc: write_failure - Write failed
//
void Event::write( Output& out ) const
{
  unsigned int paramStringDataSize = 0;
  
  for ( unsigned int i = 0; i < mParameters.size(); i++ ) 
    paramStringDataSize += 3 + mParameters[i].first.size();
  
  out.startRecord( ID_EVENT,
		   62
		   + sizeof(REAL_4) * mParameters.size()
		   + paramStringDataSize
		   + mName.size() + mComment.size()
		   + mInputs.size() + mStatistics.size(), this );
  
  out << mName << mComment << mInputs << mGTime 
      << mTimeBefore << mTimeAfter << mEventStatus << mAmplitude
      << mProbability << mStatistics << (INT_2U)mParameters.size();
  
  for ( unsigned int i = 0; i < mParameters.size(); i++ )
    out << mParameters[i].first;

  for ( unsigned int i = 0; i < mParameters.size(); i++ )
    out << mParameters[i].second;
  
  out	<< mData << mTable;
  
  out.endRecord( this );
}


//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two Event objects are equal. For this comparison,
// all of the data is compared.
//
//!param: const Event& eventData - The object to compare with.
//
//!return: bool - True if the objects are equal.
//
//!kw: const
//
//!exc: None.
//   
bool Event::operator==( const Event& eventData ) const
{
    return ( mName == eventData.mName &&
             mComment == eventData.mComment &&
             mInputs == eventData.mInputs &&
             mGTime == eventData.mGTime &&
	     mTimeBefore == eventData.mTimeBefore &&
	     mTimeAfter == eventData.mTimeAfter &&
	     mEventStatus == eventData.mEventStatus &&
	     mAmplitude == eventData.mAmplitude &&
             mProbability == eventData.mProbability &&
             mStatistics == eventData.mStatistics &&
             mParameters == eventData.mParameters &&
             mData == eventData.mData &&
	     mTable == eventData.mTable );
}

   
//-----------------------------------------------------------------------------
//
//: Class ID
//
// Gets the class ID.
//
//!return: INT_2U -- Event Data class ID.
//
//!exc: None.
//   
INT_2U Event::getClassId() const
{
    return ID_EVENT;
}


   
