#ifndef FrameCppFileHeaderHH
#define FrameCppFileHeaderHH


#include "types.hh"
#include "errors.hh"
#include "dictionary.hh"

#include "framecpp/util.hh"

namespace FrameCPP {

namespace Version_6
{
  class IFrameStream;
}

namespace Version_3_4_5 {

class Input;
class Output;


//-----------------------------------------------------------------------------
//
//: The File header.
//
// This class reresents the 40-byte file header for a frame file.  It contains
// information about how data is formatted within the file, as well as
// information about the machine used to write the file.
//
// <p>The FileHeader class also provides the ability to read from a frame file.
// Data conversions (such as byte-ordering) are taken care of by this class.
//
class FileHeader
{
public:

    /* Constructors */

    //-----------------------------------------------------------------------------
    //
    //: Copy constructor.
    //
    //!param: const FileHeader& header - Object to be copied.
    //
    //!except: std::bad_alloc - Memory allocation failed.
    //
    FileHeader( const FileHeader& header );

    //-----------------------------------------------------------------------------
    //
    //: Input constructor.
    //
    // This constructs a FileHeader by reading it from an Input object.
    //
    //!param: Input& in - Input data stream object.
    //
    //!exc: std::bad_alloc - Memory allocation failed
    //!exc: read_failure - Read failed.
    //
    FileHeader( Input& in );
    
    /* Destructor */
    //-----------------------------------------------------------------------------
    //
    //: Destructor.
    //
    //!exc: None.
    virtual ~FileHeader();

    /* Operator Overloads */

    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    //!param: const FileHeader& header - Object to be assigned.
    //
    //!return: const FileHeader& -- This object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const FileHeader& operator=( const FileHeader& header );

    /* I/O */

    //-----------------------------------------------------------------------------
    //
    //: Writes specified frame data format version file header to output.
    //
    //!param: out (Output&) - output frame file
    //!param: out (version) - data format version
    //
    //!except: write_failure - Data write failed.
    //
    static void writeLocal( Output& out, int version );
   
    //: Read new file header from the input file
    //!exc: std::bad_alloc - Memory allocation failed
    //!exc: read_failure - Read failed.
    //   
    static FileHeader* read( Input& in );

    /* Read methods */
    //: Read an integer, doing byte swapping if necessary
    //!exc: read_failure - Read failed.   
    bool readINT2U( Input& in, INT_2U* v, unsigned int n = 1 );

    //: Read an integer, doing byte swapping if necessary
    //!exc: read_failure - Read failed.   
    bool readINT2S( Input& in, INT_2S* v, unsigned int n = 1 );

    //: Read an integer, doing byte swapping if necessary
    //!exc: read_failure - Read failed.   
    bool readINT4U( Input& in, INT_4U* v, unsigned int n = 1 );

    //: Read an integer, doing byte swapping if necessary
    //!exc: read_failure - Read failed.   
    bool readINT4S( Input& in, INT_4S* v, unsigned int n = 1 );

    //: Read an integer, doing byte swapping if necessary
    //!exc: read_failure - Read failed.   
    bool readINT8U( Input& in, INT_8U* v, unsigned int n = 1 );

    //: Read an integer, doing byte swapping if necessary
    //!exc: read_failure - Read failed.   
    bool readINT8S( Input& in, INT_8S* v, unsigned int n = 1 );

    //: Read a real, doing byte swapping if necessary
    //!exc: read_failure - Read failed.   
    bool readREAL4( Input& in, REAL_4* v, unsigned int n = 1 );

    //: Read a real, doing byte swapping if necessary
    //!exc: read_failure - Read failed.   
    bool readREAL8( Input& in, REAL_8* v, unsigned int n = 1 );

    /* Accessors  */
    //: Get "IGWD" identifier from the frame file header
    //!exc: None.
    const char* getIdentifier() const;
   
    //: Get frame file data format version
    //!exc: None.   
    INT_2U getDataFormatVersion() const;
   
    //: Get library minor version field
    //!exc: None.   
    INT_2U getLibMinorVersion() const;
   
    //: Get two byte integer size
    //!exc: None.   
    INT_2U getInt2Size() const;
   
    //: Get four byte integer size
    //!exc: None.   
    INT_2U getInt4Size() const;
   
    //: Get eight byte integer size
    //!exc: None.   
    INT_2U getInt8Size() const;
   
    //: Get four byte real size
    //!exc: None.   
    INT_2U getReal4Size() const;

    //: Get eight byte real size
    //!exc: None.   
    INT_2U getReal8Size() const;
   
    //: Get sample two byte integer
    //!exc: None.   
    const unsigned char* getInt2Sample() const;
   
    //: Get sample four byte integer 
    //!exc: None.   
    const unsigned char* getInt4Sample() const;
   
    //: Get sample eight byte integer
    //!exc: None.   
    const unsigned char* getInt8Sample() const;
   
    //: Get sample four byte real
    //!exc: None.   
    const unsigned char* getReal4Sample() const;
   
    //: Get sample eight byte real
    //!exc: None.   
    const unsigned char* getReal8Sample() const;
   
    //: get ASCII string
    //!exc: None.   
    const char* getAscii() const;

    //: See whether file input does byte swapping
    //!exc: None.   
    bool byteSwapNeeded() const {
#ifdef WORDS_BIGENDIAN    
      return mInt2UReader == &FileHeader::readInt2U_L;
#else
      return mInt2UReader == &FileHeader::readInt2U_B;
#endif
    }

    //: Get a dictionary for this file header
    //!exc: read_failure - Read failed.   
    inline Dictionary* getDictionary()
      {
        if ( mData == 0 )
	  {
	    throw read_failure( "version not supported", 0);
	  }
      
        return library.getVersionDictionary( mData[ 5 ] );
      }

private:
    friend class FrameCPP::Version_6::IFrameStream;

    FileHeader( const char* Header );

    //!exc: None.   
    void update();
   
    //!exc: None.   
    void checkInt2();
   
    //!exc: None.   
    void checkInt4();

    //!exc: None.
    void checkInt8();
    
    //!exc: read_failure - Read failed.      
    void readInt2U_B( Input& in, INT_2U* v, unsigned int n );
   
    //!exc: read_failure - Read failed.   
    void readInt2U_L( Input& in, INT_2U* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readInt2S_B( Input& in, INT_2S* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readInt2S_L( Input& in, INT_2S* v, unsigned int n );

    //!exc: read_failure - Read failed.   
    void readInt4U_4B( Input& in, INT_4U* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readInt4U_4L( Input& in, INT_4U* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readInt4U_2B( Input& in, INT_4U* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readInt4U_2L( Input& in, INT_4U* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readInt4S_4B( Input& in, INT_4S* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readInt4S_4L( Input& in, INT_4S* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readInt4S_2B( Input& in, INT_4S* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readInt4S_2L( Input& in, INT_4S* v, unsigned int n );

    //!exc: read_failure - Read failed.   
    void readInt8U_8B( Input& in, INT_8U* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readInt8U_8L( Input& in, INT_8U* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readInt8U_4B( Input& in, INT_8U* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readInt8U_4L( Input& in, INT_8U* v, unsigned int n );
   
    //!exc: read_failure - Read failed.      
    void readInt8S_8B( Input& in, INT_8S* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readInt8S_8L( Input& in, INT_8S* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readInt8S_4B( Input& in, INT_8S* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readInt8S_4L( Input& in, INT_8S* v, unsigned int n );

    //!exc: read_failure - Read failed.       
    void readReal4_B( Input& in, REAL_4* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readReal4_L( Input& in, REAL_4* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readReal8_B( Input& in, REAL_8* v, unsigned int n );

    //!exc: read_failure - Read failed.      
    void readReal8_L( Input& in, REAL_8* v, unsigned int n );

    /* Pointers to reader methods */
    void (FileHeader::*mInt2UReader)( Input&, INT_2U*, unsigned int );
    void (FileHeader::*mInt2SReader)( Input&, INT_2S*, unsigned int );
    void (FileHeader::*mInt4UReader)( Input&, INT_4U*, unsigned int );
    void (FileHeader::*mInt4SReader)( Input&, INT_4S*, unsigned int );
    void (FileHeader::*mInt8UReader)( Input&, INT_8U*, unsigned int );
    void (FileHeader::*mInt8SReader)( Input&, INT_8S*, unsigned int );
    void (FileHeader::*mReal4Reader)( Input&, REAL_4*, unsigned int );
    void (FileHeader::*mReal8Reader)( Input&, REAL_8*, unsigned int );

    //: Frame file header data.
    unsigned char* mData;
};




//-----------------------------------------------------------------------------
//!exc: read_failure - Read failed.      
//   
inline
bool FileHeader::readINT2U( Input& in, INT_2U* v, unsigned int n )
{
    if ( mInt2UReader != 0 )
    {
        (this->*mInt2UReader)( in, v, n );
        return true;
    }

    return false;
}

    
//-----------------------------------------------------------------------------
//!exc: read_failure - Read failed.      
//      
inline
bool FileHeader::readINT2S( Input& in, INT_2S* v, unsigned int n )
{
    if ( mInt2SReader != 0 )
    {
        (this->*mInt2SReader)( in, v, n );
        return true;
    }
    
    return false;
}

    
//-----------------------------------------------------------------------------
//!exc: read_failure - Read failed.      
//      
inline
bool FileHeader::readINT4U( Input& in, INT_4U* v, unsigned int n )
{
    if ( mInt4UReader != 0 )
    {
        (this->*mInt4UReader)( in, v, n );
        return true;
    }
    
    return false;
}

    
//-----------------------------------------------------------------------------
//!exc: read_failure - Read failed.      
//      
inline
bool FileHeader::readINT4S( Input& in, INT_4S* v, unsigned int n )
{
    if ( mInt4SReader != 0 )
    {
        (this->*mInt4SReader)( in, v, n );
        return true;
    }
    
    return false;
}

    
//-----------------------------------------------------------------------------
//!exc: read_failure - Read failed.      
//      
inline
bool FileHeader::readINT8U( Input& in, INT_8U* v, unsigned int n )
{
    if ( mInt8UReader != 0 )
    {
        (this->*mInt8UReader)( in, v, n );
        return true;
    }
    
    return false;
}

    
//-----------------------------------------------------------------------------
//!exc: read_failure - Read failed.      
//      
inline
bool FileHeader::readINT8S( Input& in, INT_8S* v, unsigned int n )
{
    if ( mInt8SReader != 0 )
    {
        (this->*mInt8SReader)( in, v, n );
        return true;
    }
    
    return false;
}

    
//-----------------------------------------------------------------------------
//!exc: read_failure - Read failed.      
//      
inline
bool FileHeader::readREAL4( Input& in, REAL_4* v, unsigned int n )
{
    if ( mReal4Reader != 0 )
    {
        (this->*mReal4Reader)( in, v, n );
        return true;
    }
    
    return false;
}

    
//-----------------------------------------------------------------------------
//!exc: read_failure - Read failed.      
//      
inline
bool FileHeader::readREAL8( Input& in, REAL_8* v, unsigned int n )
{
    if ( mReal8Reader != 0 )
    {
        (this->*mReal8Reader)( in, v, n );
        return true;
    }
    
    return false;
}


//-----------------------------------------------------------------------------
//!exc: None
//      
inline const char* FileHeader::getIdentifier() const
{
    return (char*)mData;
}


//-----------------------------------------------------------------------------
//!exc: None
//         
inline INT_2U FileHeader::getDataFormatVersion() const
{
    return mData[ 5 ];
}


//-----------------------------------------------------------------------------
//!exc: None
//         
inline INT_2U FileHeader::getLibMinorVersion() const
{
    return mData[ 6 ];
}


//-----------------------------------------------------------------------------
//!exc: None
//         
inline INT_2U FileHeader::getInt2Size() const
{
    return mData[ 7 ];
}


//-----------------------------------------------------------------------------
//!exc: None
//         
inline INT_2U FileHeader::getInt4Size() const
{
    return mData[ 8 ];
}


//-----------------------------------------------------------------------------
//!exc: None
//         
inline INT_2U FileHeader::getInt8Size() const
{
    return mData[ 9 ];
}


//-----------------------------------------------------------------------------
//!exc: None
//         
inline INT_2U FileHeader::getReal4Size() const
{
    return mData[ 10 ];
}


//-----------------------------------------------------------------------------
//!exc: None
//         
inline INT_2U FileHeader::getReal8Size() const
{
    return mData[ 11 ];
}


//-----------------------------------------------------------------------------
//!exc: None
//         
inline const unsigned char* FileHeader::getInt2Sample() const
{
    return ( mData + 12 );
}


//-----------------------------------------------------------------------------
//!exc: None
//         
inline const unsigned char* FileHeader::getInt4Sample() const
{
    return ( mData + 14 );
}


//-----------------------------------------------------------------------------
//!exc: None
//         
inline const unsigned char* FileHeader::getInt8Sample() const
{
    return ( mData + 18 );
}


//-----------------------------------------------------------------------------
//!exc: None
//         
inline const unsigned char* FileHeader::getReal4Sample() const
{
    return ( mData + 26 );
}


//-----------------------------------------------------------------------------
//!exc: None
//         
inline const unsigned char* FileHeader::getReal8Sample() const
{
    return ( mData + 30 );
}


//-----------------------------------------------------------------------------
//!exc: None
//         
inline const char* FileHeader::getAscii() const
{
    return (char*)( mData + 38 );
}
    
 
} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppFileHeaderHH
