#if HAVE_CONFIG_H
#include "framecpp/config.h"
#endif /* HAVE_CONFIG_H */


#include <sys/types.h>

#include <cmath>

#include "frame.hh"
#include "framereaderbase.hh"
#include "output.hh"

using namespace FrameCPP::Version_3_4_5;
using std::floor;


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates a Frame with the following characteristics:
// <ul>
//     <li>name = "none"</li>
//     <li>run = 0</li>
//     <li>frame = 0</li>
//     <li>dataQuality = 0</li>
//     <li>GTime = 0 seconds, 0 nanoseconds</li>
//     <li>ULeapS = 0</li>
//     <li>dt = 0</li>
//     <li>No other data.</li>
// </ul>
//
//!exc: None.
//         
Frame::Frame() 
        : Base(), mName( "none" ), mRun( 0 ), mFrame( 0 ), mDataQuality( 0 ), mGTime( 0, 0 ),
          mULeapS( 0 ), mDt( 0 ), mType(), mUser(),
          mDetectorSim(), mDetectorProc(), mHistory(), mRawData( 0 ),
          mProcData(), mSimData(), mEvent(), mSummary(),
          mAux(), mTable(),
	  mOwnsRawData( true )
{
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
// Constructs a frame with the specified parameters but no data.
//
//!param: const std::string& name - Project or experiment name.
//!param: INT_4S run - Run number.
//!param: INT_4U frame - Frame number.
//!param: INT_4U dataQuality - data quality word.
//!param: const Time& time - GPS start time.
//!param: INT_2U uLeapS - Leap seconds in the epoch in which the frame was
//+     written.
//!param: const REAL_8& dt - Frame length in seconds.
//
Frame::Frame( const std::string& name, INT_4S run, INT_4U frame, const Time& time,
              INT_2U uLeapS, const REAL_8& dt, INT_4U dqual )
        : mName( name ), mRun( run ), mFrame( frame ), mDataQuality( dqual ), mGTime( time ),
          mULeapS( uLeapS ), mDt( dt ), mType(),
          mUser(), mDetectorSim(), mDetectorProc(), mHistory(),
          mRawData( 0 ), mProcData(), mSimData(), mEvent(),
          mSummary(), mAux(), mTable(),
	  mOwnsRawData( true )
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
// This performs a deep copy of owned elements.
//
//!param: const Frame& frame - A frame to be copied.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
Frame::Frame( const Frame& frame ) 
        : Base( frame ), mName( frame.mName ), mRun( frame.mRun ),
          mFrame( frame.mFrame ), mDataQuality( frame.mDataQuality ),
	  mGTime( frame.mGTime ),
          mULeapS( frame.mULeapS ),
          mDt( frame.mDt ), mType( frame.mType ), mUser( frame.mUser ),
          mDetectorSim( frame.mDetectorSim ),
	  mDetectorProc( frame.mDetectorProc ),
	  mHistory( frame.mHistory ),
          mRawData( 0 ), mProcData( frame.mProcData ),
          mSimData( frame.mSimData ),
          mEvent( frame.mEvent ), mSummary( frame.mSummary ),
          mAux( frame.mAux ),
	  mTable( frame.mTable ), 
	  mOwnsRawData( frame.mOwnsRawData )
{
    try
    {
      if ( frame.mRawData != NULL)
	{
	  if ( mOwnsRawData )
	    {
	      mRawData = new RawData( *frame.mRawData );
	    }
	  else
	    {
	      mRawData = frame.mRawData;
	    }
	}
    }
    catch( ... )
    {
        if ( mOwnsRawData )
        {
           delete mRawData;
           mRawData = 0;
        }

        throw;
    }
}

//-----------------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Destructor.
//
// This deletes the rawdata.
//
//!exc: None.
//         
Frame::~Frame()
{
    if ( mOwnsRawData )
    {
       delete mRawData;
       mRawData = 0;
    }
}


//-----------------------------------------------------------------------------
//
//: Get Name
//
// Gets the name.  This is the name of the project or other experiment
// description.
//
//!return: const std::string& -- Frame name.
//
const std::string& Frame::getName() const
{
    return mName;
}



//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
//!param: const Frame& frame - Object to be assigned.
//
//!return: const Frame& - This frame.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const Frame& Frame::operator=( const Frame& frame )
{
    if ( this != &frame )
    {
        mName = frame.mName;
        mRun = frame.mRun;
        mFrame = frame.mFrame;
        mDataQuality = frame.mDataQuality;
        mGTime = frame.mGTime;
        mULeapS = frame.mULeapS;
        mDt = frame.mDt;
        mType = frame.mType;
        mUser = frame.mUser;
        mDetectorSim = frame.mDetectorSim;
        mDetectorProc = frame.mDetectorProc;
        mHistory = frame.mHistory;
        mProcData = frame.mProcData;
        mSimData = frame.mSimData;
        mEvent= frame.mEvent;
        mSummary = frame.mSummary;
        mAux = frame.mAux;
        mTable = frame.mTable;
        mOwnsRawData = frame.mOwnsRawData;
        
        try
        {
            if ( mOwnsRawData )
            {
                mRawData = new RawData( *frame.mRawData );
            }
            else
            {
                mRawData = frame.mRawData;
            }
        }
        catch( ... )
        {
            if ( mOwnsRawData )
            {
               delete mRawData;
               mRawData = 0;
            }

            throw;
        }
    }
    
    return *this;
}

//-----------------------------------------------------------------------------
//
//: += Assignment Operator.
//
//!param: const Frame& frame -
//
//!return: const Frame&
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: frame_mismatch - Couldn't add `frame' (time or ADC mismatch)
//
const Frame& Frame::operator+=( const Frame& frame )
{
  // Check the `frame' is next to `*this' in time
  //
  double flr = floor( getDt() );
  Time nft( (INT_4U) flr, (INT_4U) ( getDt() - flr ) );
  double frame_flr = floor( frame.getDt() );
  Time frame_nft( (INT_4U) frame_flr, (INT_4U) ( frame.getDt() - frame_flr ) );
  if ( ( getGTime() + nft ) != frame_nft )
  {
    throw frame_mismatch( "Out of sequence in time." );
  }

  // Frame name should match
  //
  if ( getName() != frame.getName() )
  {
    throw frame_mismatch( "Name different" );
  }

  // Run should be the same for both frame
  //
  if ( getRun() != frame.getRun() )
  {
    throw frame_mismatch( "Run different" );
  }

  // Frame counter should be serial
  //
  if ( ( getFrame() + 1 ) != frame.getFrame() )
  {
    throw frame_mismatch( "Count out of sequence" );
  }
  
#ifdef not_def
  FIXME: clarify Detector structure usage

  // Detectors should be the same
  //
  // fixme: check for NULLs before doing this
  if ( *getDetectSim() != *frame.getDetectSim() )
  {
    throw frame_mismatch( "" );
  }

  // the same for getDetectProc()
#endif

#ifdef not_def
  *refType() += *frame.refType();
  *refUser() += *frame.refUser();  
  *refHistory() += *frame.refHistory();

  // Fixme: check for RawData NULLs
#endif

  *getRawData() += *frame.getRawData();

#ifdef not_def
  *refProcData() += *frame.refProcData();
  *refSimData() += *frame.refSimData();
  *refEvent() += *frame.refEvent();
  *refSummary() += *frame.refSummary();
  *refAux() += *frame.refAux();
  *refTable() += *frame.refTable();
#endif

  mDt += frame.getDt();
  
  return *this;
}

//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
//: Get Structure Header
//
// Constructs the Structure Header for this class.
// Frame Data format version 3.
//
//!return: SH - Frame structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
SH Frame::getSH3() 
{
    SH sh( "FrameH", ID_FRAME, "Frame Header" );
    sh.appendSE( SE( "name", "STRING", "" ) );
    sh.appendSE( SE( "run", "INT_4U", "" ) );
    sh.appendSE( SE( "frame", "INT_4U", "" ) );
    sh.appendSE( SE( "GTimeS", "INT_4U", "" ) );
    sh.appendSE( SE( "GTimeN", "INT_4U", "" ) );
    sh.appendSE( SE( "ULeapS", "INT_2U", "" ) );
    sh.appendSE( SE( "localTime", "INT_4S", "" ) );
    sh.appendSE( SE( "dt", "REAL_8", "" ) );
    sh.appendSE( SE( "type", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "user", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "detectSim", "PTR_STRUCT(FrDetector *)", "" ) );
    sh.appendSE( SE( "detectProc", "PTR_STRUCT(FrDetector *)", "" ) );
    sh.appendSE( SE( "history", "PTR_STRUCT(FrHistory *)", "" ) );
    sh.appendSE( SE( "rawData", "PTR_STRUCT(FrRawData *)", "" ) );
    sh.appendSE( SE( "procData", "PTR_STRUCT(FrProcData *)", "" ) );
    sh.appendSE( SE( "strain", "PTR_STRUCT(FrProcData *)", "" ) );
    sh.appendSE( SE( "simData", "PTR_STRUCT(FrSimData *)", "" ) );
    sh.appendSE( SE( "trigData", "PTR_STRUCT(FrTrigData *)", "" ) );
    sh.appendSE( SE( "summaryData", "PTR_STRUCT(FrSummary *)", "" ) );
    sh.appendSE( SE( "auxData", "PTR_STRUCT(FrVect *)", "" ) );
    return sh;
}

//-----------------------------------------------------------------------------
//
//: Get Structure Header
//
// Constructs the Structure Header for this class.
// Frame Data format version 4.
//
//!return: SH - Frame structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
SH Frame::getSH4()
{
    SH sh( "FrameH", ID_FRAME, "Frame Header" );
    sh.appendSE( SE( "name", "STRING", "" ) );
    sh.appendSE( SE( "run", "INT_4S", "" ) );
    sh.appendSE( SE( "frame", "INT_4U", "" ) );
    sh.appendSE( SE( "dataQuality", "INT_4U", "" ) );
    sh.appendSE( SE( "GTimeS", "INT_4U", "" ) );
    sh.appendSE( SE( "GTimeN", "INT_4U", "" ) );
    sh.appendSE( SE( "ULeapS", "INT_2U", "" ) );
    sh.appendSE( SE( "localTime", "INT_4S", "" ) );
    sh.appendSE( SE( "dt", "REAL_8", "" ) );
    sh.appendSE( SE( "type", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "user", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "detectSim", "PTR_STRUCT(FrDetector *)", "" ) );
    sh.appendSE( SE( "detectProc", "PTR_STRUCT(FrDetector *)", "" ) );
    sh.appendSE( SE( "history", "PTR_STRUCT(FrHistory *)", "" ) );
    sh.appendSE( SE( "rawData", "PTR_STRUCT(FrRawData *)", "" ) );
    sh.appendSE( SE( "procData", "PTR_STRUCT(FrProcData *)", "" ) );
    sh.appendSE( SE( "strain", "PTR_STRUCT(FrProcData *)", "" ) );
    sh.appendSE( SE( "simData", "PTR_STRUCT(FrSimData *)", "" ) );
    sh.appendSE( SE( "trigData", "PTR_STRUCT(FrTrigData *)", "" ) );
    sh.appendSE( SE( "simEvent", "PTR_STRUCT(FrSimEvent *)", "" ) );
    sh.appendSE( SE( "summaryData", "PTR_STRUCT(FrSummary *)", "" ) );
    sh.appendSE( SE( "auxData", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "auxTable", "PTR_STRUCT(FrTable *)", "" ) );
    return sh;
}

//-----------------------------------------------------------------------------
//
//: Get Structure Header
//
// Constructs the Structure Header for this class.  
//
//!return: SH - Frame structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
SH Frame::getSH()
{
    SH sh( "FrameH", ID_FRAME, "Frame Header" );
    sh.appendSE( SE( "name", "STRING", "" ) );
    sh.appendSE( SE( "run", "INT_4S", "" ) );
    sh.appendSE( SE( "frame", "INT_4U", "" ) );
    sh.appendSE( SE( "dataQuality", "INT_4U", "" ) );
    sh.appendSE( SE( "GTimeS", "INT_4U", "" ) );
    sh.appendSE( SE( "GTimeN", "INT_4U", "" ) );
    sh.appendSE( SE( "ULeapS", "INT_2U", "" ) );
    sh.appendSE( SE( "dt", "REAL_8", "" ) );
    sh.appendSE( SE( "type", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "user", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "detectSim", "PTR_STRUCT(FrDetector *)", "" ) );
    sh.appendSE( SE( "detectProc", "PTR_STRUCT(FrDetector *)", "" ) );
    sh.appendSE( SE( "history", "PTR_STRUCT(FrHistory *)", "" ) );
    sh.appendSE( SE( "rawData", "PTR_STRUCT(FrRawData *)", "" ) );
    sh.appendSE( SE( "procData", "PTR_STRUCT(FrProcData *)", "" ) );
    sh.appendSE( SE( "simData", "PTR_STRUCT(FrSimData *)", "" ) );
    sh.appendSE( SE( "trigData", "PTR_STRUCT(FrEvent *)", "" ) );
    sh.appendSE( SE( "simEvent", "PTR_STRUCT(FrSimEvent *)", "" ) );
    sh.appendSE( SE( "summaryData", "PTR_STRUCT(FrSummary *)", "" ) );
    sh.appendSE( SE( "auxData", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "auxTable", "PTR_STRUCT(FrTable *)", "" ) );
    return sh;
}

//-----------------------------------------------------------------------------
//
//: Read a frame
//
// Reads a Frame object from an Input object.  This method creates a
// dynamically allocated Frame object.  This frame is initialized from the
// given Input object.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: Input& in - Input data stream object.
//
//!return: Frame* - Newly created.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failed.
//
Frame* Frame::read( FrameReaderBase& in )
{
  Frame& frame = *new Frame();
  try
  {
    void*	rawData = &(frame.mRawData);

    in >> frame.mName >> frame.mRun >> frame.mFrame
       >> frame.mDataQuality 
       >> frame.mGTime >> frame.mULeapS 
       >> frame.mDt
       >> frame.mType
       >> frame.mUser
       >> frame.mDetectorSim
       >> frame.mDetectorProc
       >> frame.mHistory
       >> (Base**)rawData >> frame.mProcData >> frame.mSimData
       >> frame.mEvent >> frame.mSimEvent
       >> frame.mSummary >> frame.mAux >> frame.mTable;

    return &frame;
  }
  catch( ... )
  {
    delete &frame;
    throw;
  }
}

//: Version 4 frame read.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failed.
//   
Frame* Frame::read4( FrameReaderBase& in )
{
  INT_4S localTime, strain;
  Frame& frame = *new Frame();
  try
  {
    void*	rawData = &(frame.mRawData);

    in >> frame.mName >> frame.mRun >> frame.mFrame
       >> frame.mDataQuality 
       >> frame.mGTime >> frame.mULeapS >> localTime
       >> frame.mDt
       >> frame.mType
       >> frame.mUser
       >> frame.mDetectorSim
       >> frame.mDetectorProc
       >> frame.mHistory
       >> (Base**)rawData
       >> frame.mProcData
       >> strain
       >> frame.mSimData
       >> frame.mEvent >> frame.mSimEvent
       >> frame.mSummary >> frame.mAux >> frame.mTable;
    return &frame;
  }
  catch( ... )
  {
    delete &frame;
    throw;
  }
}


//: Version 3 frame read.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failed.
//   
Frame* Frame::read3( FrameReaderBase& in )
{
  INT_4S localTime, strain;
  INT_4U run;
  Frame& frame = *new Frame();
  try
  {
    void*	rawData = &(frame.mRawData);

    in >> frame.mName
       >> run
       >> frame.mFrame
       >> frame.mGTime
       >> frame.mULeapS
       >> localTime
       >> frame.mDt
       >> frame.mType
       >> frame.mUser
       >> frame.mDetectorSim
       >> frame.mDetectorProc
       >> frame.mHistory
       >> (Base**)rawData
       >> frame.mProcData
       >> strain
       >> frame.mSimData
       >> frame.mEvent
       >> frame.mSummary
       >> frame.mAux;

    frame.mRun = (INT_4S) run;
    return &frame;
  }
  catch( ... )
  {
    delete &frame;
    throw;
  }
}



//-----------------------------------------------------------------------------
//
//: Write a frame.
//
// Writes a frame to an Output object.
//
//!param: Output& out out - Output data stream object.
//
//!exc: write_failure - Write failed.
//
void Frame::write( Output& out ) const 
{
    out.startRecord(
		    ID_FRAME, 89 + mName.size() 
		    - 4 // for the removed mStrain
		    - 4 // for the removed mLocalTime
		    + 4 //  Table reference
		    + 4 // data quality word
		    + 4 // SimEvent reference
		    , this );
    out << mName << mRun << mFrame << mDataQuality << mGTime << mULeapS 
        << mDt << mType << mUser << mDetectorSim << mDetectorProc << mHistory
        << mRawData << mProcData << mSimData << mEvent << mSimEvent
        << mSummary << mAux << mTable
      ;
    out.endRecord( this );
}


//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
//!param: const Frame& frame - Frame which which to compare.
//
//!return: bool - True if frames are equal.
//
//!exc: None.
//   
bool Frame::operator==( const Frame& frame ) const 
{
    if ( this == &frame )
    {
        return true;
    }

    if ( !( mRawData == frame.mRawData ||
            ( ( mRawData != 0 && frame.mRawData != 0 ) &&
              ( *mRawData == *frame.mRawData ) ) ) )
    {
        return false;
    }
    
    return ( ( mName == frame.mName ) && ( mRun == frame.mRun ) &&
             ( mFrame == frame.mFrame ) &&
	     ( mDataQuality == frame.mDataQuality ) &&
	     ( mGTime == frame.mGTime ) &&
             ( mULeapS == frame.mULeapS ) && ( mDt == frame.mDt ) &&
             ( mType == frame.mType ) && ( mUser == frame.mUser ) &&
             ( mDetectorSim == frame.mDetectorSim ) &&
             ( mDetectorProc == frame.mDetectorProc ) &&
             ( mHistory == frame.mHistory ) &&
             ( mProcData == frame.mProcData ) &&
             ( mSimData == frame.mSimData ) &&
             ( mEvent == frame.mEvent ) &&
             ( mSummary == frame.mSummary ) &&
             ( mAux == frame.mAux )
	     && ( mTable == frame.mTable )
	     );
}

    
//-----------------------------------------------------------------------------
//
//: Set raw data
//
//!param: const RawData& r - RawData to set.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
void Frame::setRawData( const RawData& r ) 
{
    if ( mOwnsRawData )
    {
        delete mRawData;
        mRawData = 0;
    }
    
    mRawData = new RawData( r );
    mOwnsRawData = true;
}


//-----------------------------------------------------------------------------
//
//: Set raw data
//
//!param: const RawData* r - RawData to set.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
void Frame::setRawData( const RawData* r ) 
{
    if ( mOwnsRawData )
    {
        delete mRawData;
        mRawData = 0;
    }
    
    mRawData = new RawData( *r );
    mOwnsRawData = true;
}


//-----------------------------------------------------------------------------
//
//: Set raw data
//
//!param: RawData* r - RawData to ste.
//!param: bool allocate - Will allocate a copy of `r'.
//!param: bool owns -  Frame will own its RawData.
//
//!exc: std::bad_alloc
//
void Frame::setRawData( RawData* r, bool allocate, bool owns )
{
    if ( mOwnsRawData )
    {
        delete mRawData;
        mRawData = 0;
    }
    
    if ( allocate )
        mRawData = new RawData( *r );
    else
        mRawData = r;

    mOwnsRawData = owns;
}


//-----------------------------------------------------------------------------
//
//: Get Class ID
//
// Gets the class ID.
//
//!return: INT_2U -- Frame class ID.
//
//!exc: None.
//   
INT_2U Frame::getClassId() const
{
    return ID_FRAME;
}
    
