#include "framecpp/config.h"

#include "types.hh"
#include "framewritertoc.hh"
#include "frame.hh"
#include "sh.hh"
#include "adcdata.hh"
#include "serdata.hh"
#include "table.hh"
#include "msg.hh"
#include "rawdata.hh"
#include "detector.hh"

using namespace FrameCPP::Version_3_4_5;

//: Constructs an object and writes the fileheader.
//!exc: write_failure    
FrameWriterTOC::FrameWriterTOC( std::ostream& out )
  : TOC(), Output( out ), tocSize( 0 ), raw_seen( false )
{}


//: Destructs the object.
//!exc: write_failure       
FrameWriterTOC::~FrameWriterTOC()
{
  close();
}


//-----------------------------------------------------------------------------
//
//: Begins a record.
//
// This writes out the length, class and instance id of the
// object and also initializes the ObjectInfo structure.
//
// Overloaded from Output class. TOC hosekeeping code.
//
//!exc: write_failure       
//
void FrameWriterTOC::startRecord( INT_2U classId, INT_4U length, const FrameCPP::Version_3_4_5::Base* b )
{
  const std::string lbl = "Logic Error. TOC type conversion failed.";

  // Start output record first
  // The very next thing the caller function does is to start writing its elements
  //
  Output::startRecord( classId, length, b );

  switch (classId) {
  case ID_TOC:
    tocSize = length; // save this number for use at the "end of file" struct later
    break;
  case ID_SH:
    {
      const SH* sh = dynamic_cast< const SH* >( b );
      if ( 0 == sh ) {
	throw write_failure( lbl );
      }
      SHid.push_back( sh->getClass() );
      SHname.push_back( sh->getName() );
    }
    break;
  case ID_FRAME:
    {
      const Frame* fp = dynamic_cast< const Frame* >( b );
      if ( 0 == fp ) {
	throw write_failure( lbl );
      }
      if ( frame.empty() ) {
	// First frame header
	ULeapS = fp->getULeapS();
      }
      dataQuality.push_back( fp->getDataQuality() );
      GTime.push_back( Time( fp->getGTime().getSec(), fp->getGTime().getNSec() ));
      dt.push_back( fp->getDt() );
      runs.push_back( fp->getRun() );
      frame.push_back( fp->getFrame() );
      position.push_back( ((unsigned long)mOut.tellp()) - 8 /* 8 is for the length,class,instance vars */ );

      // Push empty elements into the position std::vectors
      positionADC.push_back( 0 );
      positionSer.push_back( 0 );
      positionTable.push_back( 0 );
      positionMsg.push_back( 0 );
    }
    raw_seen = false;
    break;
  case ID_DETECTOR:
    {
      const Detector* det = dynamic_cast< const Detector* >( b );
      if ( 0 == det ) {
	throw write_failure( lbl );
      }
      // Capture only the first occurrence of a detector
      if ( detectorNamePositionMap.find( det->getName() ) == detectorNamePositionMap.end() )
	detectorNamePositionMap[ det->getName() ] = ((unsigned long)mOut.tellp()) - 8;
    }
    break;
  case ID_MSG:
    {
      const Msg* ser = dynamic_cast< const Msg* >( b );
      if ( 0 == ser ) {
	throw write_failure( lbl );
      }
      if ( 0 == positionMsg[ positionMsg.size() - 1 ] )
	{
	  positionMsg[ positionMsg.size() - 1 ] = ((unsigned long)mOut.tellp()) - 8;
	}
    }
    break;
  case ID_ADCDATA:
    {
      const AdcData* adc = dynamic_cast< const AdcData* >( b );
      if ( 0 == adc ) {
	throw write_failure( lbl );
      }
      if ( 0 == positionADC[ positionADC.size() - 1 ] )
	{
	  positionADC[ positionADC.size() - 1 ] = ((unsigned long)mOut.tellp()) - 8;
	}

      
      ANPINC q = adcNamePositionMap.find ( adc->getName() );
      if ( q == adcNamePositionMap.end() ) {

	// Insert new std::vector into the map
	//
	std::vector< INT_4U > iv;
	iv.push_back( adc->getChannelNumber() );
	iv.push_back( adc->getChannelGroup()  );

	std::vector<INT_8U> v;

	// Grow position std::vector inserting zeroed gaps for those frames that
	// do not have named adc channel
	v.resize( frame.size() );
	v[ v.size() - 1 ] = ((unsigned long)mOut.tellp()) - 8;
	std::pair< std::vector<INT_8U>, std::vector<INT_4U> > p( v, iv );
	adcNamePositionMap.insert(
	     std::pair< std::string, std::pair< std::vector<INT_8U>, std::vector<INT_4U> > >( adc->getName(), p ));
      } else {
	// Change position std::vector in place
	//

	// Grow position std::vector inserting zeroed gaps for those frames that
	// do not have named adc channel
	q->second.first.resize( frame.size() );
	q->second.first[ q->second.first.size() - 1 ] = ((unsigned long)mOut.tellp()) - 8;
      }
    }
    break;
  case ID_RAWDATA:
    {
      raw_seen = true;
    }
    break;
  case ID_TABLE:
    {
      if ( raw_seen ) // First table after Raw data 
	{
	  const Table* table = dynamic_cast< const Table* >( b );
	  if ( 0 == table ) {
	    throw write_failure( lbl );
	  }
	  positionTable[ positionTable.size() - 1 ] =  ((unsigned long)mOut.tellp()) - 8;
	  raw_seen = false;
	}
    }
    break;
  case ID_STATDATA:
    {
      const StatData* stat = dynamic_cast< const StatData* >( b );
      if ( 0 == stat ) {
	throw write_failure( lbl );
      }

      std::vector< std::pair< std::vector< INT_4U >, INT_8U > > v;
      std::vector< INT_4U > vv;
      vv.push_back( stat->getTimeStart() );
      vv.push_back( stat->getTimeEnd() );
      vv.push_back( stat->getVersion() );

      SNPINC q = statNamePositionMap.find ( stat->getName() );
      if ( q == statNamePositionMap.end() ) {
	// Insert new std::pair into the map. This is a new static data structure.
	//
	v.push_back( std::pair< std::vector< INT_4U >, INT_8U>( vv, ((unsigned long)mOut.tellp()) - 8 ));
	const Detector* dp = stat->getDetector();
	std::pair< std::string, std::vector< std::pair< std::vector< INT_4U >, INT_8U > > > p( dp?dp->getName():"", v );
	statNamePositionMap.insert(
	   std::pair< std::string, std::pair< std::string, std::vector< std::pair< std::vector< INT_4U >, INT_8U > > > >( stat->getName(), p ));
      } else {
	// Another static data was found with the same name.
	// This means the current static is the next version of it.
	// Append another elemnt to the std::vector of vesions in the map.
	//
	q->second.second.push_back( std::pair< std::vector< INT_4U >, INT_8U>( vv, ((unsigned long)mOut.tellp()) - 8 ));
      }
    }
    break;
  case ID_PROCDATA:
    {
      const ProcData* proc = dynamic_cast< const ProcData* >( b );
      if ( 0 == proc ) {
	throw write_failure( lbl );
      }
      
      PNPINC q = procNamePositionMap.find ( proc->getName() );
      if ( q == procNamePositionMap.end() ) {

	// Insert new std::vector into the map
	//

	std::vector<INT_8U> v;

	// Grow position std::vector inserting zeroed gaps for those frames that
	// do not have named proc channel
	v.resize( frame.size() );
	v[ v.size() - 1 ] = ((unsigned long)mOut.tellp()) - 8;
	procNamePositionMap.insert( std::pair< std::string, std::vector<INT_8U> >( proc->getName(), v ));
      } else {
	// Change position std::vector in place
	//

	// Grow position std::vector inserting zeroed gaps for those frames that
	// do not have named proc channel
	q->second.resize( frame.size() );
	q->second[ q->second.size() - 1 ] = ((unsigned long)mOut.tellp()) - 8;
      }

    }
    break;
  case ID_SIMDATA:
    {
      const SimData* sim = dynamic_cast< const SimData* >( b );
      if ( 0 == sim ) {
	throw write_failure( lbl );
      }
      
      SIMNPINC q = simNamePositionMap.find ( sim->getName() );
      if ( q == simNamePositionMap.end() ) {

	// Insert new std::vector into the map
	//

	std::vector<INT_8U> v;

	// Grow position std::vector inserting zeroed gaps for those frames that
	// do not have named sim channel
	v.resize( frame.size() );
	v[ v.size() - 1 ] = ((unsigned long)mOut.tellp()) - 8;
	simNamePositionMap.insert( std::pair< std::string, std::vector<INT_8U> >( sim->getName(), v ));
      } else {
	// Change position std::vector in place
	//

	// Grow position std::vector inserting zeroed gaps for those frames that
	// do not have named sim channel
	q->second.resize( frame.size() );
	q->second[ q->second.size() - 1 ] = ((unsigned long)mOut.tellp()) - 8;
      }

    }
    break;
  case ID_SERDATA:
    {
      const SerData* ser = dynamic_cast< const SerData* >( b );
      if ( 0 == ser ) {
	throw write_failure( lbl );
      }
      if ( 0 == positionSer[ positionSer.size() - 1 ] )
	{
	  positionSer[ positionSer.size() - 1 ] = ((unsigned long)mOut.tellp()) - 8;
	}
      
      PNPINC q = serNamePositionMap.find ( ser->getName() );
      if ( q == serNamePositionMap.end() ) {

	// Insert new std::vector into the map
	//

	std::vector<INT_8U> v;

	// Grow position std::vector inserting zeroed gaps for those frames that
	// do not have named ser channel
	v.resize( frame.size() );
	v[ v.size() - 1 ] = ((unsigned long)mOut.tellp()) - 8;
	serNamePositionMap.insert( std::pair< std::string, std::vector<INT_8U> >( ser->getName(), v ));
      } else {
	// Change position std::vector in place
	//

	// Grow position std::vector inserting zeroed gaps for those frames that
	// do not have named ser channel
	q->second.resize( frame.size() );
	q->second[ q->second.size() - 1 ] = ((unsigned long)mOut.tellp()) - 8;
      }
    }
    break;
  case ID_EVENT:
    {
      const Event* event = dynamic_cast< const Event* >( b );
      if ( 0 == event ) {
	throw write_failure( lbl );
      }
      eventNamePositionMap[ event->getName() ].push_back( TOC::eventAttr( event->getGTime(), event->getAmplitude(), ((INT_8U)mOut.tellp()) - 8 ));
    }
    break;
  case ID_SIMEVENT:
    {
      const SimEvent* event = dynamic_cast< const SimEvent* >( b );
      if ( 0 == event ) {
	throw write_failure( lbl );
      }
      simEventNamePositionMap[ event->getName() ].push_back( TOC::eventAttr( event->getGTime(), event->getAmplitude(), ((INT_8U)mOut.tellp()) - 8 ));
    }
    break;
  case ID_SUMMARY:
    {
      const Summary* sum = dynamic_cast< const Summary* >( b );
      if ( 0 == sum ) {
	throw write_failure( lbl );
      }

      PNPINC q = sumNamePositionMap.find ( sum->getName() );
      if ( q == sumNamePositionMap.end() ) {

	// Insert new std::vector into the map
	//

	std::vector<INT_8U> v;

	// Grow position std::vector inserting zeroed gaps for those frames that
	// do not have named sum channel
	v.resize( frame.size() );
	v[ v.size() - 1 ] = ((unsigned long)mOut.tellp()) - 8;
	sumNamePositionMap.insert( std::pair< std::string, std::vector<INT_8U> >( sum->getName(), v ));
      } else {
	// Change position std::vector in place
	//

	// Grow position std::vector inserting zeroed gaps for those frames that
	// do not have named sum channel
	q->second.resize( frame.size() );
	q->second[ q->second.size() - 1 ] = ((unsigned long)mOut.tellp()) - 8;
      }
    }
    break;
  default:
    break;
  }
}

//: Write the TOC and close the file
void FrameWriterTOC::close( unsigned long dummy )
{
  if ( mClosed )
    {
      return;
    }

  // Write the TOC out to a file
  TOC::write ( *this );

  // Write the end of file structure with a pointer to the TOC
  Output::close( tocSize );
}
