#ifndef FrameCppHistoryHH
#define FrameCppHistoryHH


#include "framecpp/config.h"
#include "base.hh"
#include "sh.hh"


namespace FrameCPP {

namespace Version_6
{
  class FrHistory;
}

namespace Version_3_4_5 {

class FrameReaderBase;


//-----------------------------------------------------------------------------
//
//: This class represents history data.
//
// It is modeled after the FrHistory structure in the Frame Specification,
// LIGO-T970130-B-E.
//
class History : public Base
{
public:

    /* Constructors */
    //-----------------------------------------------------------------------------
    //
    //: Default Constructor.
    //
    // This creates a History object with the following properties:
    // <ul>
    //     <li> name = "none" </li>
    //     <li> time = 0 </li>
    //     <li> comment = "" </li>
    // </ul>
    //
    //!exc: None.
    //
    History();


    //-----------------------------------------------------------------------------
    //
    //: Copy Constructor.
    //
    // This performs a deep copy of the object.
    //
    //!param: const History& history - Object to be copied.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    History( const History& history );

    //-----------------------------------------------------------------------------
    //
    //: Constructor.
    //
    //!param: const std::string& name - Name of history record.
    //!param: INT_4U time - Time of post-processing.
    //!param: const std::string& comment - Program name and relevant comment needed to define post-processing.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    History( const std::string& name, INT_4U time, const std::string& comment );
    
    /* Destructor */
    //!exc: None.   
    virtual ~History();

    /* Operator Overloads */

    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    //!param: const History& history - Object to be assigned.
    //
    //!return: const History& -- This object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const History& operator=( const History& history );

    //-----------------------------------------------------------------------------
    //
    //: += Operator.
    //
    //!param: const History& history - Object to be added.
    //
    //!return: const History& -- This object.
    //
    const History& operator+=( const History& history );

    //-----------------------------------------------------------------------------
    //
    //: Equal comparison.
    //
    // Determines whether two History objects are equal. For this comparison,
    // all of the data is compared.
    //
    //!param: const History& history - The object to compare with.
    //
    //!return: bool -- true if the objects are equal.
    //
    //!kw: const
    //
    //!exc: None.   
    bool operator==( const History& history ) const;
    bool operator!=( const History& history ) const;

    /* Required Methods */
    //!exc: None.   
    virtual INT_2U getClassId() const;

    //-----------------------------------------------------------------------------
    //
    //: Constructs the Structure Header for this class.
    //
    //!return: SH -- History Structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH3();

    /* Accessors */

    /* gcc 2.95.1 can't handle a typedef with this if the throw spec is
       present (e.g.
       typedef SearchContainer< History, &History::getName > dc; ).
       
       const std::string& getName() const throw(); */
    const std::string& getName() const;
   
    //!exc: None.   
    INT_4U getTime() const;
    //!exc: None.   
    const std::string& getComment() const;

private:
  friend class Dictionary;
  friend class FrameCPP::Version_6::FrHistory;


    //-----------------------------------------------------------------------------
    //
    //: Reads a History object from an FrameReaderBase object.
    //
    // This method creates a dynamically allocated object.
    //
    // <p>*NOTE* This method allocates memory and returns a pointer to it.  The
    // user is responsible for freeing this memory.
    //
    //!param: FrameReaderBase& in - Input data stream object.
    //
    //!return: History* -- Newly created.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //
    static History* read3( FrameReaderBase& in );

    //-----------------------------------------------------------------------------
    //
    //: Writes a History object to an Output object.
    //
    //!param: Output& out out - Output data stream object.
    //
    //!exc: write_failure - write failure occured.
    //
    virtual void write( Output& out ) const;

private:

    //: Name of history record.
    std::string mName;
    //: Time of post-processing.
    INT_4U mTime;
    //: oProgram name and relevant comment needed to define post-processing.
    std::string mComment;
};




/******************/
/* Inline Methods */
/******************/


//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Not-equal comparison.
//
//!param: const History& history - The object to compare with.
//
//!return: bool -- True if the objects are not equal.
//
//!kw: inline const
//
//!exc: None.   
//      
inline bool History::operator!=( const History& history ) const
{
    return !( *this == history );
}



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Gets the time of post-processing.
//
// This is the GPS time in integer seconds since the GPS standard epoch.
//
//!return: INT_4U -- Time
//
//!exc: None.   
//      
inline INT_4U History::getTime() const 
{
    return mTime;
}


//-----------------------------------------------------------------------------
//
//: Get the program name and relevant comments needed to define
//: post-processing.
//
//!return: const std::string& -- Comments.
//
//!exc: None.   
//      
inline const std::string& History::getComment() const
{
    return mComment;
}

 
} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // FrameCppHistoryHH
