#include "framecpp/config.h"

#include <sstream>
#include <cassert>

#include "input.hh"
#include "frame.hh"
#include "endofframe.hh"
#include "dictionary.hh"


using namespace FrameCPP::Version_3_4_5;
using namespace std;   


//-----------------------------------------------------------------------------
//
//: Constructor
//
// Creates an input object which reads the passed stream.
//
//!param: std::istream& in - The stream to read.
//!param: FileHeader* fh - Optional file header.
//
//!exc: None.   
Input::Input( std::istream* in, FileHeader* fh )
        : mStream( in ), mFileHeader( fh ),
	  dictionary( library.getCurrentVersionDictionary() ),
	  mErrorWatch( 0 ),
          mNBytes( 0 ), mErrors(), mCurrentSH( 0 ), mTransMap(),
          mIsHeaderLoaded( 0 ), mNextObjectHeader()
{
}


//-----------------------------------------------------------------------------
//
//: Destructor
//
// We only need to delete the FileHeader.
//
Input::~Input() 
{
    delete mFileHeader;
    mFileHeader = 0;
}


//-----------------------------------------------------------------------------
//
//: Get next object header.
//
// This returns the object header for the next item in the frame file.  This
// method skips over structure headers and structure elements, storing
// their information.
//
//!return ObjectHeader - Next object header.
//
//!exc: read_failure - Read failed.   
FrameCPP::Version_3_4_5::ObjectHeader Input::readNextObjectHeader()
{
    // If the header is already loaded, just return it
    if ( mIsHeaderLoaded )
    {
        return mNextObjectHeader;
    }
    
    INT_4U length;   // length of structure in bytes
    INT_2U classId;  // File Structure class
    INT_2U instance; // Instance count for this class type
    INT_2U localClassId = 0;

    mCurrentSH = 0;   // No structure header
    bool end = false; // Not at the end of the loop

    // Loop until we find the next header
    while ( !end )
    {
        // Read a header
        (*this) >> length >> classId >> instance;
        
        // Check the length
        if ( length < 8 )
        {
            ostringstream ss;
            ss << "Invalid object length: " << length
               << ".  Length must be >= 8.";
            addError( ss.str() );
            throw( read_failure( "Invalid object length", BADSTREAM ) );
        }
        
        // Check for a Structure Header
        if ( classId == ID_SH )
        {
            readStructureHeader();
        }
        // Check for a Structure Element
        else if ( classId == ID_SE )
        {
	    readStructureElement( length );
	    // skipCurrentObject();
        }
        else
        {
            // If we were reading a SH, close it.
            if ( mCurrentSH != 0 )
            {
                endDictionaryElement();
            }
            
            // Check to see if a SH for this class ID has been read.
            TransMap::iterator tIter = mTransMap.find( classId );
            if ( tIter == mTransMap.end() )
            {
                // If no SH exists, skip this object.
                addError( "Unknown class ID.  A class ID was found which "
                          "has no corresponding structure header." );
                throw( read_failure( "Unknown structure class ID", BADSTREAM ) );
            }
            else
            {
                // Ok, we found the header
                end = true;
                localClassId = tIter->second;
            }
        }
    }
    
    mNextObjectHeader = ObjectHeader(
        length, classId, instance, localClassId );
    
    mIsHeaderLoaded = true;
    return mNextObjectHeader;
}


//-----------------------------------------------------------------------------
//
//: Read Structure Header (Dictionary Element)
//
// This method reads a structure header.  the object header must have already
// been read.
//
//!exc: read_failure - Read failed.      
void Input::readStructureHeader() 
{
    // If we were alread reading a structure header then close it
    if ( mCurrentSH != 0 )
    {
        endDictionaryElement();
    }
    
    mCurrentSH = SH::read( *this );
}


//-----------------------------------------------------------------------------
//
//: Finish a Dictionary Element
//
// This closes a dictionary element and adds a translation entry to the
// translation map (this map stores a mapping between file class ID's and
// local class ID's).
//
void Input::endDictionaryElement()
{
    assert( mCurrentSH != 0 );

    if ( mCurrentSH->getClass() == 1 )
    {
        addError( "Invalid structure header.  A structure header appeared "
                  "for class ID = 1.  This class is reserved.  Ignoring "
                  "structure header." );
    }
    else if ( mCurrentSH->getClass() == 2 )
    {
        addError( "Invalid structure header.  A structure header appeared "
                  "for class ID = 2.  This class is reserved.  Ignoring "
                  "structure header." );
    }
    else
    {
        TransMap::iterator iter = mTransMap.find( mCurrentSH->getClass() );

	// Use existing structure header, it is the valid one
	//
        if ( iter == mTransMap.end() )
	  {
	    if ( dictionary->find( mCurrentSH->getName() ) == dictionary->end() )
	      {
		addError( "Unknown class type.  The structure header "
			  "for an unknown class appeared." );
	      }
	    else
	      {
		const SH& localSH(
		  dictionary->find( mCurrentSH->getName() )->second.getSH() );

		mTransMap.insert( std::pair< INT_2U, INT_2U >(
			 mCurrentSH->getClass(), localSH.getClass() ) );

		if ( !mCurrentSH->equal( localSH ) )
		  {
		    ostringstream ss;
		    ss << "Mismatched dictionary element.  The dictionary element "
		      "for " << localSH.getName() << " did not match the known "
		      "definition.";
		    addError( ss.str() );
		  }
	      }
	    
	  }


	/*
        if ( iter != mTransMap.end() )
        {
            addError( "Duplicate structure header.  The structure header "
                      "for a class appeared twice.  The original header is "
                      "discarded and the second one will be used." );

            mTransMap.erase( iter );
        }

        if ( dictionary->find( mCurrentSH->getName() ) == dictionary->end() )
        {
            addError( "Unknown class type.  The structure header "
                      "for an unknown class appeared." );
        }
        else
        {
            const SH& localSH(
                dictionary->find( mCurrentSH->getName() )->second.getSH() );
            
            mTransMap.insert( std::pair< INT_2U, INT_2U >(
                mCurrentSH->getClass(), localSH.getClass() ) );

            if ( *mCurrentSH != localSH )
            {
                ostringstream ss;
                ss << "Mismatched dictionary element.  The dictionary element "
                    "for " << localSH.getName() << " did not match the known "
                    "definition.";
                addError( ss.str() );
            }
        }
	*/
    }

    delete mCurrentSH;
    mCurrentSH = 0;
}
        
        
//-----------------------------------------------------------------------------
//
//: Read a Structure Element
//
//!param: INT_4U length - the length of the structure element.  This is needed
//+     in case this is a stray structure element, in which case the
//+     structure element will be skipped.
//
//!exc: read_failure - Read failed.
//
void Input::readStructureElement( INT_4U length )
{
    if ( mCurrentSH == 0 )
    {
        addError( "Misplaced SE.  A structure element was found without "
                  "a corresponding structure header." );
        if ( length < 8 )
        {
            addError( "Invalid length.  The specified length of an "
                      "object was less than 8.  Unable to skip." );
            throw( read_failure( "Unable to skip object", SKIPFAILED ) );
        }
        skip( length - 8 );
    }
    else
    {
        SE* se = SE::read( *this );
        mCurrentSH->appendSE( *se );
        delete se;
        se = 0;
    }
}


//-----------------------------------------------------------------------------
//
//: Skip bytes
//
//!param: unsigned int length - The number of bytes to skip.
//
//!exc: read_failure - Read failed.
//
void Input::skip( unsigned int length )
{
    if ( mIsHeaderLoaded )
    {
        mIsHeaderLoaded = false;
    }
    
    mStream->seekg( length, std::ios::cur );
}

    
//-----------------------------------------------------------------------------
//
//: Add an Error
//
//!param: const strin& msg - Error message.
//
//!exc: None.
//   
void Input::addError( const std::string& msg )
{    
    mErrors.push_back( msg );

    if ( mErrorWatch != 0 )
    {
        (*mErrorWatch)( msg );
    }
}


//-----------------------------------------------------------------------------
//
//: Read a STRING.
//
//!param: std::string& v - Target.
//
//!return: Input& - This object.
//
//!exc: read_failure - String read failed.
//
Input& Input::operator>>( std::string& v )
{
    // Initialize the file header object if necessary
    if ( mFileHeader == 0 )
    {
        mFileHeader = new FileHeader( *this );
	dictionary = mFileHeader->getDictionary();
    }

    // Read the length
    INT_2U length;
    (*this) >> length;

    // If an empty std::string, return nothing
    if ( length == 0 )
    {
        v.erase();
    } else {
      char*	buffer( (char*)NULL );
      try
      {
	buffer = new char[ length + 1 ];
	
	read( buffer, length );
	if ( buffer[ length - 1 ] != '\0' )
	{
	  buffer[ length ] = '\0';
	  std::ostringstream ss;
	  ss << "String missing null-terminator: ("
	     << mNextObjectHeader.getFileClassId() << ","
	     << mNextObjectHeader.getInstance() << ") \"" 
	     << buffer << "\".";
	  addError( ss.str() );
	}
	v = buffer;
	delete[] buffer;
        buffer = 0;
      }
      catch( ... )
      {
	delete[] buffer;	// release the resource
        buffer = 0;
	throw;			// rethrow the exception
      }
    }

    return *this;
}
    
    
//-----------------------------------------------------------------------------
//
//: Read Bytes
//
//!param: void* buffer - the location to store the bytes
//!param: unsigned int n - The number of bytes to read
//
//!exc: read_failure - Read failed.
//
void Input::read( void* buffer, unsigned int n ) 
{
    if ( mIsHeaderLoaded )
    {
        mIsHeaderLoaded =  false;
    }

    char* b( (char*)buffer );
    mStream->read( b, n );
    if ( !mStream->good() )
    {
        if ( mStream->eof() )
        {
            throw read_failure( "End of file reached", ENDOFFILE );
        }
        else
        {
            throw read_failure(
                "Stream error occurred.", BADSTREAM );
        }
    }
    mNBytes += n;
}


//-----------------------------------------------------------------------------
//
//: Set an Error Watch
//
// An error watch is a method which is called whenever an error is added to the
// Input class.
//
//!param: void (*f)(const std::string&) - The function to call when an error occurs.
//+     If this is null, then no function will be called.
//
//!exc: None.
//      
void Input::setErrorWatch( void(*f)(const std::string&) ) 
{
    mErrorWatch = f;
}


//-----------------------------------------------------------------------------
//
//: Reset the Input object
//
// This causes the input object to re-read the fileheader from the stream when
// the next frame is read.
//
//!exc: None.
//      
void Input::reset()
{
    mTransMap.erase( mTransMap.begin(), mTransMap.end() );
    mNBytes = 0;
    delete mCurrentSH;
    mCurrentSH = 0;
    clearErrors();
    delete mFileHeader;
    mIsHeaderLoaded = false;
    mFileHeader = 0;
}
