#ifndef FrameCppInputHH
#define FrameCppInputHH


#include "fileheader.hh"
#include "container.hh"
#include "refheadbase.hh"
#include "reference.hh"
#include "refhead.hh"
#include "sh.hh"
#include <vector>
#include <map>
#include <iostream>
#include "objectheader.hh"
#include "se.hh"
#include "location.hh"
#include "dictionary.hh"

namespace FrameCPP
{

namespace Version_6
{
  class IFrameStream;
}

namespace Version_3_4_5
{


class Base;
class Frame;


//-----------------------------------------------------------------------------
//
//: Input
//
// This class' responsibilities are:
// <p><ol type=1>
//     <li> Read object headers: <ol type=a>
//         <li> If the header corresponds to a Structure Header or Structure
//              Element then read it in and remember it. </li>
//         <li> Convert the file class ID to the local class ID for the object.
//              </li>
//         <li> If the object is unknown, skip it. </li></ol>
//     <li> Provide bulk reads (through the use of the FileHeader class). </li>
//     <li> Store information about errors and warnings which occur during the
//          read process. </li>
// </ol>
//
// <p>An instance of this class is constructed with an istream object.  When
// the user performs the first read request, the Input class attempts to read
// the File Header, once this is loaded then the read is processed.
//
class Input
{
public:

    enum ErrorCode
    {
        BADSTREAM,
        SKIPFAILED,
        ENDOFFILE
    };
    
    /* Typedef's */
    typedef std::vector< std::string >::const_iterator error_iterator;

    /* Constructors */
    //-----------------------------------------------------------------------------
    //
    //: Constructor
    //
    // Creates an input object which reads the passed stream.
    //
    //!param: istream& in - The stream to read.
    //!param: FileHeader* fh - Optional file header.
    //
    //!exc: None.
    //
    Input( std::istream* in=0, FileHeader* fh=0 );

    /* Destructor */
    //-----------------------------------------------------------------------------
    //
    //: Destructor
    //
    // We only need to delete the FileHeader.
    //
    virtual ~Input();

    /* I/O methods */

    //-----------------------------------------------------------------------------
    //
    //: Get next object header.
    //
    // This returns the object header for the next item in the frame file.  This
    // method skips over structure headers and structure elements, storing
    // their information.
    //
    //!return ObjectHeader - Next object header.
    //
    //!exc: read_failure - Read failed.
    //   
    virtual ObjectHeader readNextObjectHeader();

    //-----------------------------------------------------------------------------
    //
    //: Read Bytes
    //
    //!param: void* buffer - the location to store the bytes
    //!param: unsigned int n - The number of bytes to read
    //
    //!exc: read_failure - Read failed.
    //
    void read( void*, unsigned int n );

    //-----------------------------------------------------------------------------
    //
    //: Skip bytes
    //
    //!param: unsigned int length - The number of bytes to skip.
    //
    //!exc: read_failure - Read failed.
    //
    void skip( unsigned int n );
    //!exc: read_failure - Read failed.
    void skipCurrentObject();
    //: Get frame file data format vesion
    //!exc: None.
    INT_2U getDataFormatVersion() const;

    /* Operator Overloads */
    //!exc: read_failure - Read failed.   
    Input& operator>>( CHAR& v );
    //!exc: read_failure - Read failed.   
    Input& operator>>( CHAR_U& v );
    //!exc: read_failure - Read failed.   
    Input& operator>>( INT_2S& v );
    //!exc: read_failure - Read failed.   
    Input& operator>>( INT_2U& v );
    //!exc: read_failure - Read failed.   
    Input& operator>>( INT_4S& v );
    //!exc: read_failure - Read failed.   
    Input& operator>>( INT_4U& v );
    //!exc: read_failure - Read failed.   
    Input& operator>>( INT_8S& v );
    //!exc: read_failure - Read failed.   
    Input& operator>>( INT_8U& v );
    //!exc: read_failure - Read failed.   
    Input& operator>>( REAL_4& v );
    //!exc: read_failure - Read failed.   
    Input& operator>>( REAL_8& v );

    //-----------------------------------------------------------------------------
    //
    //: Read a STRING.
    //
    //!param: std::string& v - Target.
    //
    //!return: Input& -- This object.
    //
    //!exc: read_failure - String read failed.
    //
    Input& operator>>( std::string& v );

    /* Error methods */
    //!exc: std::range_error 
    const std::string& getError( unsigned int index ) const;
    //!exc: None.
    Input::error_iterator errorBegin() const;
    //!exc: None.   
    Input::error_iterator errorEnd() const;
    //!exc: None.   
    unsigned int errorQuantity() const;
    //!exc: None.   
    const std::string& errorFirst() const;
    //!exc: None.   
    const std::string& errorLast() const;
    //!exc: None.   
    void clearErrors();


    //-----------------------------------------------------------------------------
    //
    //: Set an Error Watch
    //
    // An error watch is a method which is called whenever an error is added to the
    // Input class.
    //
    //!param: void (*f)(const std::string&) - The function to call when an error occurs.
    //+     If this is null, then no function will be called.
    //
    //!exc: None.   
    void setErrorWatch( void(*f)(const std::string&) );

    /* Accessors */
    //!exc: None.   
    INT_4U getNBytes() const;

    /* Mutator */
    //-----------------------------------------------------------------------------
    //
    //: Reset the Input object
    //
    // This causes the input object to re-read the fileheader from the stream when
    // the next frame is read.
    //
    //!exc: None.   
    void reset();
    
protected:
    
    
    //-----------------------------------------------------------------------------
    //
    //: Add an Error
    //
    //!param: const strin& msg - Error message.
    //
    //!exc: None.   
    void addError( const std::string& msg );
    typedef std::map< INT_2U, INT_2U > TransMap;

private:

    /* Helper Methods */
    //-----------------------------------------------------------------------------
    //
    //: Read Structure Header (Dictionary Element)
    //
    // This method reads a structure header.  the object header must have already
    // been read.
    //
    //!exc: read_failure - Read failed.
    void readStructureHeader();

    //-----------------------------------------------------------------------------
    //
    //: Finish a Dictionary Element
    //
    // This closes a dictionary element and adds a translation entry to the
    // translation map (this map stores a mapping between file class ID's and
    // local class ID's).
    //
    void endDictionaryElement();
    //-----------------------------------------------------------------------------
    //
    //: Read a Structure Element
    //
    //!param: INT_4U length - the length of the structure element.  This is needed
    //+     in case this is a stray structure element, in which case the
    //+     structure element will be skipped.
    //
    //!exc: read_failure - Read failed.
    //
    void readStructureElement( INT_4U length );

    /* Primary data */
protected:
    //: Input data stream.
    std::istream* mStream;
    //: Frame file header.
    FileHeader* mFileHeader;
    //: Input frame file dictionary.
    Dictionary* dictionary;
private:
    void (*mErrorWatch)(const std::string&);
    
    /* Read Stats */
    //: Number of bytes read.
protected:
    unsigned int mNBytes;
private:
    //: Error messages.
    std::vector< std::string > mErrors;

    //: Read Internal State.
    SH* mCurrentSH;

protected:
    TransMap mTransMap;
    bool mIsHeaderLoaded;

private:
  friend class FrameCPP::Version_6::IFrameStream;

  ObjectHeader mNextObjectHeader;
};


//-----------------------------------------------------------------------------
//
//: Skip the length of the current object header
//
//!exc: read_failure - Read failed.
//
inline void Input::skipCurrentObject() {
  if ( mIsHeaderLoaded )
    {
      mIsHeaderLoaded = false;
      mStream->seekg( mNextObjectHeader.getLength() - 8, std::ios::cur );
    }
}


//-----------------------------------------------------------------------------
//
//: Get Number of Bytes Read
//
//!return: INT_4U -- The number of bytes read.
//
//!exc: None.
//   
inline INT_4U Input::getNBytes() const
{
    return mNBytes;
}


//-----------------------------------------------------------------------------
//
//: Read an INT_2S
//
//!param: INT_2S& v - Target.
//
//!return: Input& -- This object.
//
//!exc: read_failure - Read failed.
//
inline Input& Input::operator>>( INT_2S& v )
{
    if ( mFileHeader == 0 )
    {
        mFileHeader = new FileHeader( *this );
	dictionary = mFileHeader->getDictionary();
    }

    mFileHeader->readINT2S( *this, &v );
    return *this;
}

    
//-----------------------------------------------------------------------------
//
//: Read a CHAR
//
//!param: CHAR& v - Target.
//
//!return: Input& -- This object.
//
//!exc: read_failure - Read failed.
//
inline Input& Input::operator>>( CHAR& v )
{
    if ( mFileHeader == 0 )
    {
        mFileHeader = new FileHeader( *this );
	dictionary = mFileHeader->getDictionary();
    }

    read( &v, 1 );
    return *this;
}

    
//-----------------------------------------------------------------------------
//
//: Read a CHAR_U
//
//!param: CHAR_U& v - Target.
//
//!return: Input& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline Input& Input::operator>>( CHAR_U& v )
{
    if ( mFileHeader == 0 )
    {
        mFileHeader = new FileHeader( *this );
	dictionary = mFileHeader->getDictionary();
    }

    read( &v, 1 );
    return *this;
}

    
//-----------------------------------------------------------------------------
//
//: Read an INT_2U
//
//!param: INT_2U& v - Target.
//
//!return: Input& -- This object.
//
//!exc: read_failure - Read fialure occured.
//
inline Input& Input::operator>>( INT_2U& v )
{
    if ( mFileHeader == 0 )
    {
        mFileHeader = new FileHeader( *this );
	dictionary = mFileHeader->getDictionary();
    }
    
    mFileHeader->readINT2U( *this, &v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read an INT_4S
//
//!param: INT_4S& v - Target.
//
//!return: Input& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline Input& Input::operator>>( INT_4S& v )
{
    if ( mFileHeader == 0 )
    {
        mFileHeader = new FileHeader( *this );
	dictionary = mFileHeader->getDictionary();
    }
    
    mFileHeader->readINT4S( *this, &v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read an INT_4U
//
//!param: INT_4U& v - Target.
//
//!return: Input& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline Input& Input::operator>>( INT_4U& v )
{
    if ( mFileHeader == 0 )
    {
        mFileHeader = new FileHeader( *this );
	dictionary = mFileHeader->getDictionary();
    }
    
    mFileHeader->readINT4U( *this, &v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read an INT_8S
//
//!param: INT_8S& v - Target.
//
//!return: Input& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline Input& Input::operator>>( INT_8S& v )
{
    if ( mFileHeader == 0 )
    {
        mFileHeader = new FileHeader( *this );
	dictionary = mFileHeader->getDictionary();
    }
    
    mFileHeader->readINT8S( *this, &v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read an INT_8U
//
//!param: INT_8U& v - Target.
//
//!return: Input& -- This object.
// 
//!exc: read_failure - Read failure occured.
//
inline Input& Input::operator>>( INT_8U& v )
{
    if ( mFileHeader == 0 )
    {
        mFileHeader = new FileHeader( *this );
	dictionary = mFileHeader->getDictionary();
    }
    
    mFileHeader->readINT8U( *this, &v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read an REAL_4
//
//!param: REAL_4& v - Target.
//
//!return: Input& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline Input& Input::operator>>( REAL_4& v )
{
    if ( mFileHeader == 0 )
    {
        mFileHeader = new FileHeader( *this );
	dictionary = mFileHeader->getDictionary();
    }
    
    mFileHeader->readREAL4( *this, &v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read an REAL_8
//
//!param: REAL_8& v - Target.
//
//!return: Input& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline Input& Input::operator>>( REAL_8& v )
{
    if ( mFileHeader == 0 )
    {
        mFileHeader = new FileHeader( *this );
	dictionary = mFileHeader->getDictionary();
    }
    
    mFileHeader->readREAL8( *this, &v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Get an error
//
//!param: unsigned int index - Error index.
//
//!return: const std::string& -- Error message.
//
//!exc: std::range_error - Thrown if the requested error index does not exist.
//
inline const std::string& Input::getError( unsigned int index ) const
{
    if ( index >= mErrors.size() )
    {
        throw std::range_error( "Index out of range" );
    }

    return mErrors[ index ];
}


//-----------------------------------------------------------------------------
//
//: Error Begin
//
// Get an iterator pointing to the first iterator.
//
//!return: error_iterator
//
//!exc: None.
inline Input::error_iterator Input::errorBegin() const
{
    return mErrors.begin();
}


//-----------------------------------------------------------------------------
//
//: Error End
//
// Get an iterator pointing immediately past the last iterator.
//
//!return: error_iterator
//
//!exc: None.
inline Input::error_iterator Input::errorEnd() const
{
    return mErrors.end();
}


//-----------------------------------------------------------------------------
//
//: Error Quantity
//
// Get the number of errors.
//
//!return: unsigned int
//
//!exc: None.   
inline unsigned int Input::errorQuantity() const
{
    return mErrors.size();
}


//-----------------------------------------------------------------------------
//
//: Get first error
//
//!return: const std::string&
//
//!exc: None.   
inline const std::string& Input::errorFirst() const 
{
    return mErrors.front();
}


//-----------------------------------------------------------------------------
//
//: Get last error
//
//!return: const std::string&
//
//!exc: None.   
inline const std::string& Input::errorLast() const
{
    return mErrors.back();
}


//-----------------------------------------------------------------------------
//
//: Clear the errors
//
//!exc: None.   
inline void Input::clearErrors() 
{
    mErrors.erase( mErrors.begin(), mErrors.end() );
}

//-----------------------------------------------------------------------------
//
//: Get frame file data format vesion
//
//!return: INT_2U -- input frame data format version or 0 if not read yet
//
//!exc: None.   
inline INT_2U Input::getDataFormatVersion() const
{
    if ( mFileHeader == 0 ) 
      return 0;
    return mFileHeader->getDataFormatVersion();
}

 
} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppInputHH
