#include "framecpp/config.h"

#include <stdexcept>

#include "msg.hh"
#include "framereaderbase.hh"
#include "output.hh"


using FrameCPP::Version_3_4_5::Msg;
using std::runtime_error;


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates a Msg object with the following properties:
// <ul type=disc>
//     <li> alarm = "none" </li>
//     <li> message = "" </li>
//     <li> severity = 0 </li>
//     <li> gtime = 0 seconds and 0 nanoseconds </li>
// </ul>
//
//!exc: None.
//   
Msg::Msg() 
  : mAlarm( "none" ), mMessage( "" ), mSeverity( 0 ), mGTime( 0, 0 )
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
//!param: const Msg& msg - Object to be copied.
//
//!exc: std::bad_alloc - Memory allocation failed
//
Msg::Msg( const Msg& msg )
        : mAlarm( msg.mAlarm ), mMessage( msg.mMessage ),
          mSeverity( msg.mSeverity ),
	  mGTime( msg.mGTime )
{
}


//-----------------------------------------------------------------------------
//
//: Constructor
//
// This creates a log message with the specified values.
//
//!param: const std::string& alarm - Name of message, error flag or alarm state
//!param: const std::string& message - Message body
//!param: INT_4U severity - Message severity level
//!param: const Time& gtimes - GPS time
//
//!exc: std::bad_alloc - Memory allocation failed
//
Msg::Msg( const std::string& alarm, const std::string& message, INT_4U severity, const Time& gtime )
  : mAlarm( alarm ), mMessage( message ), mSeverity( severity ), mGTime( gtime )
{
}


//-----------------------------------------------------------------------------
//
//: Get Alarm
//
// Get the alarm.
//
//!return: const std::string& -- The alarm
//
const std::string& Msg::getAlarm() const
{
    return mAlarm;
}



//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
//!param: const Msg& msg - Object to be assigned.
//
//!return: const Msg& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const Msg& Msg::operator=( const Msg& msg )
{
    if ( this != & msg )
    {
        mAlarm = msg.mAlarm;
        mMessage = msg.mMessage;
        mSeverity = msg.mSeverity;
        mGTime = msg.mGTime;
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: += Operator.
//
//!param: const Msg& msg - Object to be added.
//
//!return: const Msg& - This object.
//
const Msg& Msg::operator+=( const Msg& msg )
{
    throw runtime_error( "Unimplemented Msg::operator+=() is called." );

    return *this;
}


//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two Msg objects are equal. For this comparison, all of
// the data is compared.
//
//!param: const Msg& msg - The object to compare with.
//
//!return: bool - true if the objects are equal.
//
//!kw: const
//
//!exc: None.
//      
bool Msg::operator==( const Msg& msg ) const
{
    if ( this == &msg )
    {
        return true;
    }

    return ( mAlarm == msg.mAlarm &&
             mMessage == msg.mMessage &&
             mSeverity == msg.mSeverity &&
             mGTime == msg.mGTime);
}



//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Get Structure Header
//
// Constructs the Structure Header for this class.
//
//!return: SH - Msg structure header.
//
//!exc: std::bad_alloc - Memory allocation failed
//
FrameCPP::Version_3_4_5::SH Msg::getSH()
{ 
    SH sh( "FrMsg", ID_MSG, "Message Log Data" );
    sh.appendSE( SE( "alarm", "STRING", "" ) );
    sh.appendSE( SE( "message", "STRING", "" ) );
    sh.appendSE( SE( "severity", "INT_4U", "" ) );
    sh.appendSE( SE( "GTimeS", "INT_4U", "" ) );
    sh.appendSE( SE( "GTimeN", "INT_4U", "" ) );
    sh.appendSE( SE( "next", "PTR_STRUCT(FrMsg *)", "" ) );

    return sh;
}


//-----------------------------------------------------------------------------
//
//: Get Structure Header for frames version 3 and 4
//
// Constructs the Structure Header for this class.
//
//!return: SH - Msg structure header.
//
//!exc: std::bad_alloc - Memory allocation failed
//
FrameCPP::Version_3_4_5::SH Msg::getSH3()
{ 
    SH sh( "FrMsg", ID_MSG, "Message Log Data" );
    sh.appendSE( SE( "alarm", "STRING", "" ) );
    sh.appendSE( SE( "message", "STRING", "" ) );
    sh.appendSE( SE( "severity", "INT_4U", "" ) );
    sh.appendSE( SE( "next", "PTR_STRUCT(FrMsg *)", "" ) );

    return sh;
}



//-----------------------------------------------------------------------------
//
//: Reads a Msg object from an Input object.
//
// This method creates a dynamically allocated object.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: Input& in - Input data stream object.
//
//!return: Msg* - Newly created.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//
Msg* Msg::read( FrameReaderBase& in )
{
  Msg& msg = *new Msg();
  in >> msg.mAlarm >> msg.mMessage >> msg.mSeverity >> msg.mGTime;
  in.readNextNodeReference();
  return &msg;
}


//: Frame format 3 and 4read call
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//   
Msg* Msg::read3( FrameReaderBase& in )
{
  Msg& msg = *new Msg();
  in >> msg.mAlarm >> msg.mMessage >> msg.mSeverity;
  in.readNextNodeReference();
  return &msg;
}


//-----------------------------------------------------------------------------
//
//: Writes a Msg object to an Output object.
//
//!param: Output& out - Output data object.
//
//!exc: write_failure - Write failure occured.
//
//!kw: const
//
void Msg::write( Output& out ) const
{
    out.startRecord( ID_MSG, 30
                     + mAlarm.size()
                     + mMessage.size(), this );
    out << mAlarm << mMessage << mSeverity << mGTime;
    out.endRecord( this );
}

   
//-----------------------------------------------------------------------------
//
//: Get Class ID
//
// Get the class ID.
//
//!return: INT_2U -- The class ID.
//
//!kw: inline const
//
//!exc: None.   
INT_2U Msg::getClassId() const
{
    return ID_MSG;
}

   
