#ifndef FrameCppObjectHeaderHH
#define FrameCppObjectHeaderHH


#include "types.hh"
#include "reference.hh"


namespace FrameCPP {

namespace Version_3_4_5
{


//-----------------------------------------------------------------------------
//
//: The header for an object in a file.
//
// This class represents the header which appears before every object stored in
// a frame file.  The header consists of three integers:
// <p><ol>
//     <li> The byte count of the object, including the header. </li>
//     <li> An id for the type of class represented by the object,
//          this is defined at run-time. </li>
//     <li> The instance counter for this class in this frame. </li>
// </ol>
//
// <p>This class stores this information and provides a mapping between the
// class ID defined in the file and the class ID defined in memory by FrameCPP.
//
class ObjectHeader
{
public:

    /* Constructors */
    //-----------------------------------------------------------------------------
    //
    //: Default Constructor.
    //
    //!todo: Does this need to exist?
    //
    ObjectHeader();

    //-----------------------------------------------------------------------------
    //
    //: Constructor.
    //
    //!param: INT_4U length - Structure length.
    //!param: INT_2U fileClass - Structure class ID.
    //!param: INT_2U instance - Structure instance number.
    //!param: INT_2U localClass - FrameCPP structure class ID.
    //
    ObjectHeader( INT_4U length, INT_2U fileClass, INT_2U instance,
                  INT_2U localClass );

    //-----------------------------------------------------------------------------
    //
    //: Copy Constructor.
    //
    //!param: const ObjectHeader& oh - An object to be copied.
    //
    ObjectHeader( const ObjectHeader& oh );

    /* Operator Overloads */

    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    //!param: const ObjectHeader& oh - An object to be assigned.
    //
    //!return: const ObjectHeader& -- This object.
    //
    const ObjectHeader& operator=( const ObjectHeader& oh );

    /* Accessors */
    //!exc: None.
    INT_4U getLength() const;
    //!exc: None.   
    INT_2U getFileClassId() const;
    //!exc: None.   
    INT_2U getInstance() const;
    //!exc: None.   
    INT_2U getLocalClassId() const;
    //!exc: None.   
    Reference getReference() const;

private:

    //: Structure length.
    INT_4U mLength;
    //: Structure class ID.
    INT_2U mFileClassId;
    //: Structure instance number.
    INT_2U mInstance;
    //: FrameCPP structure class ID.
    INT_2U mLocalClassId;
};


//-----------------------------------------------------------------------------
//
//: Get the structure length.
//
//!return: INT_4U - Structure length.
//
//!exc: None.
//   
inline INT_4U ObjectHeader::getLength() const
{
    return mLength;
}


//-----------------------------------------------------------------------------
//
//: Get the Class ID as defined in the file.
//
//!return: INT_2U - Structure class ID.
//
//!exc: None.
//      
inline INT_2U ObjectHeader::getFileClassId() const
{
    return mFileClassId;
}


//-----------------------------------------------------------------------------
//
//: Get the instance in the file.
//
//!return: INT_2U - Structure instance number.
//
//!exc: None.
//      
inline INT_2U ObjectHeader::getInstance() const
{
    return mInstance;
}


//-----------------------------------------------------------------------------
//
//: Get the Class ID as defined locally.
//
//!return: INT_2U - Local structure class ID.
//
//!exc: None.
//      
inline INT_2U ObjectHeader::getLocalClassId() const
{
    return mLocalClassId;
}


//-----------------------------------------------------------------------------
//
//: Get a reference to this object.
//
// This consists of the File Class ID and the instance.
//
//!return: Reference - Newly constructed reference.
//
//!exc: None.
//      
inline Reference ObjectHeader::getReference() const
{
    return Reference( mFileClassId, mInstance );
}
 
} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppObjectHeaderHH
