#include "framecpp/config.h"

#include "output.hh"
#include "fileheader.hh"
#include "endoffile.hh"
#include "endofframe.hh"
#include "frame.hh"
#include "dictionary.hh"


using namespace FrameCPP::Version_3_4_5;

   
ObjectInfo::~ObjectInfo()
{}
   

//: Constructs an output object and writes the fileheader.
//   
//!exc: write_failure - Write failed.   
//   
Output::Output( std::ostream& out )
        : dictionary(*library.getCurrentVersionDictionary()),
	  mOut( out ), mClosed( true ), mNFrames( 0 ), mNBytes( 0 ),
          mIsNode( false ),
          mNextNode( 0, 0 ), mObjectQueue(), mCurrentInfo( 0 ),
          mInstance(), mMaxInstance()
{}


//: Destructs the output object.
//   
//!exc: write_failure - Write failed.   
//      
Output::~Output()
{
    close();

    // Now deallocate any memory which has been allocated
    for( ObjectQueue::iterator iter = mObjectQueue.begin();
         iter != mObjectQueue.end(); ++iter )
    {
        delete iter->second;
        iter->second = 0;
    }
    // Erase all elements
    mObjectQueue.erase( mObjectQueue.begin(), mObjectQueue.end() );
}


//: Write bytes to the stream.
//   
//!exc: write_failure - Write failed.   
//      
void Output::write( const void* buffer, unsigned int size )
{
    if ( mClosed )
    {
        open();
    }
    
    mOut.write( (char *) buffer, size );
    mNBytes += size;
}

//--------------------------------------------------------------------------------
//
//: Serialize a frame
//
// Outputs a frame object in the format defined by the frame spec.
//
//!exc: write_failure - Write failed.   
//      
void Output::writeFrame( const Frame& frame )
{
    mIsNode = false;

    ((const Base&)frame).write( *this );

    EndOfFrame eofr( frame.getRun(), frame.getFrame(), 0 , 0 );
    Base& b( eofr );
    b.write( *this );
    
    ++mNFrames;

    // This should already be empty.  Maybe we could check it.
    for( ObjectQueue::iterator iter = mObjectQueue.begin();
         iter != mObjectQueue.end(); ++iter )
    {
        delete iter->second;
        iter->second = 0;
    }
    mObjectQueue.erase( mObjectQueue.begin(), mObjectQueue.end() );

    mCurrentInfo = 0;
    mIsNode = false;
    
    // These are not empty yet
    mInstance.erase( mInstance.begin(), mInstance.end() );
    mMaxInstance.erase( mMaxInstance.begin(), mMaxInstance.end() );
    mWritten.erase( mWritten.begin(), mWritten.end() );
}


//--------------------------------------------------------------------------------
//
//: Begins a record.
//
// This writes out the length, class and instance id of the
// object and also initializes the ObjectInfo structure.
//   
//!exc: write_failure - Write failed.   
//      
void Output::startRecord( INT_2U classId, INT_4U length, const Base* b )
{

#ifdef not_def
  // DBG: print out instance std::map
  std::map< const Base*, INT_2U >::const_iterator iter = mInstance.begin();
  cerr << endl << "classId=" << classId << endl;
  for (; iter != mInstance.end(); iter++)
    {
      cerr << iter->first << '\t' << iter->second << endl;
    }
#endif

    INT_2U instance = getInstance( b );
    
    {
        // Put in something for unknown here.
        if (( classId != ID_SH ) && ( classId != ID_SE ))
        {
	    if ( shWrittenSet.find( classId ) == shWrittenSet.end() )
	      {
		bool node = mIsNode;
		mIsNode = false;
            
		const SH& sh( dictionary.find( classId )->second.getSH() );
		((Base&)sh).write( *this );
		
		for ( SH::const_se_iterator iter = sh.getSE().begin();
		      iter != sh.getSE().end(); ++iter )
		  {
		    reinterpret_cast< Base* >( *iter )->write( *this );
		  }
		
		mIsNode = node;
		shWrittenSet.insert( classId );
	      }
	}
    }   

    mWritten.insert( b );
    (*this) << length << classId << instance;

    mCurrentInfo = new ObjectInfo();
    mObjectQueue.insert( std::pair< const Base*, ObjectInfo* >( b, mCurrentInfo ) );
}

//--------------------------------------------------------------------------------
//
//: Ends a record.
//
// This method is executed when an object is finished writing
// itself.  This method calls the write function for any objects it referred
// to and appends a pointer to the next object if it was a linked-list node.
//   
//!exc: write_failure - Write failed.   
//      
void Output::endRecord( const Base* b )
{
    mCurrentInfo = 0;
    
    // Check to see if this object was a node.  If so, we must write the
    // reference of the next node.
    if ( mIsNode )
    {
        (*this) << mNextNode;
        mIsNode = false;
    }

    // Find the information for this object
    ObjectQueue::iterator iter = mObjectQueue.find( b );
    if ( iter == mObjectQueue.end() )
    {
        throw write_failure( "Logic Error.  No information was found for "
                             "the object passed to endRecord." );
 }
    ObjectInfo* o( iter->second );


    // Write all of the referenced objects, including containers ALL IN PROPER ORDER
    //
    ObjectInfo::const_so_iterator siter = o->subobjects.begin();
    ObjectInfo::const_c_iterator citer = o->containers.begin();
    for(ObjectInfo::const_i_iterator iiter = o->isContainer.begin();
        iiter != o->isContainer.end(); ++iiter )
    {
      if (*iiter) // container
	{
	  const ContainerBase* base = *citer++;
	  unsigned int size = base->size();
	  for( unsigned int i = 0; i < size; ++i )
	    {
	      if ( i == size - 1  )
		{
		  mNextNode = Reference( 0, 0 );
		}
	      else
		{
		  mNextNode = Reference( (*base)[ i+1 ]->getClassId(),
					 getInstance( ((*base)[ i+1 ]) ) );
		}
	      
	      mIsNode = true;
	      (*base)[ i ]->write( *this );
	    }
	}
      else // subobject
	{
	  const Base* base = *siter++;
	  if ( mWritten.find( base ) == mWritten.end() )
	    {
	      base->write( *this );
	    }
	}
    }   

#ifdef not_def
    // Now write all of the objects which this object referred to.
    for(
        ObjectInfo::const_so_iterator siter = o->subobjects.begin();
        siter != o->subobjects.end(); ++siter )
    {
        if ( mWritten.find( *siter ) == mWritten.end() )
        {
            (*siter)->write( *this );
        }
    }
    
    // Ok, now write all of the containers
    for(
        ObjectInfo::const_c_iterator citer = o->containers.begin();
        citer != o->containers.end(); ++citer )
    {
        unsigned int size = (*citer)->size();
        for( unsigned int i = 0; i < size; ++i )
        {
            if ( i == size - 1  )
            {
                mNextNode = Reference( 0, 0 );
            }
            else
            {
                mNextNode = Reference( (**citer)[ i+1 ]->getClassId(),
                                       getInstance( ((**citer)[ i+1 ]) ) );
            }
            
            mIsNode = true;
            (**citer)[ i ]->write( *this );
        }
    }
#endif

    // Warning! The mObjectQueue container may have been modified, so we
    // shouldn't use the old iterator value, it might not be valid.
    delete o;
    o = 0;
    mObjectQueue.erase( b );
    }

//: Output a character
//   
//!exc: write_failure - Write failed.   
//      
Output& Output::operator<<( CHAR v )
{
    write( &v, sizeof( INT_2S ) );
    return *this;
}

//: Output an unsigned character
//   
//!exc: write_failure - Write failed.   
//      
Output& Output::operator<<( CHAR_U v )
{
    write( &v, sizeof( INT_2S ) );
    return *this;
}

//: Output short integer
//   
//!exc: write_failure - Write failed.   
//      
Output& Output::operator<<( INT_2S v )
{
    write( &v, sizeof( INT_2S ) );
    return *this;
}

//: Output short unsigned integer
//   
//!exc: write_failure - Write failed.   
//      
Output& Output::operator<<( INT_2U v )
{
    write( &v, sizeof( INT_2U ) );
    return *this;
}

//: Output signed integer
//   
//!exc: write_failure - Write failed.   
//      
Output& Output::operator<<( INT_4S v )
{
    write( &v, sizeof( INT_4S ) );
    return *this;
}

//: Output unsigned integer
//   
//!exc: write_failure - Write failed.   
//      
Output& Output::operator<<( INT_4U v )
{
    write( &v, sizeof( INT_4U ) );
    return *this;
}

//: Output long integer
//   
//!exc: write_failure - Write failed.   
//      
Output& Output::operator<<( const INT_8S& v )
{
    write( &v, sizeof( INT_8S ) );
    return *this;
}

//: Output long unsigned integer
//   
//!exc: write_failure - Write failed.   
//      
Output& Output::operator<<( const INT_8U& v )
{
    write( &v, sizeof( INT_8U ) );
    return *this;
}

//: Output real number
//   
//!exc: write_failure - Write failed.   
//      
Output& Output::operator<<( const REAL_4& v )
{
    write( &v, sizeof( REAL_4 ) );
    return *this;
}

//: Output double precision real number
//   
//!exc: write_failure - Write failed.   
//      
Output& Output::operator<<( const REAL_8& v )
{
    write( &v, sizeof( REAL_8 ) );
    return *this;
}

//: Output a std::string
//   
//!exc: write_failure - Write failed.   
//      
Output& Output::operator<<( const std::string& v )
{
    (*this) << INT_2U( v.size()+1 );
    write( v.c_str(), v.size() + 1 );
    return *this;
}

//: Output an object
//   
//!exc: write_failure - Write failed.   
//      
Output& Output::operator<<( const FrameCPP::Version_3_4_5::Base* b ) 
{
    if ( mCurrentInfo == 0 )
    {
        throw write_failure( "Logic error.  Attempt to write before "
                             "startRecord called." );
    }
    
    if ( b == 0 )
    {
        (*this) << Reference( 0, 0 );
        return *this;
    }
    
    (*this) << Reference( b->getClassId(), getInstance( b ) );
    mCurrentInfo->subobjects.push_back( b );
    mCurrentInfo->isContainer.push_back( false );

    return *this;
}

//: Output a frame
//   
//!exc: write_failure - Write failed.   
//      
Output& Output::operator<<( const Frame& frame )
{
    writeFrame( frame );
    return *this;
}

//: Output a reference
//   
//!exc: write_failure - Write failed.   
//      
Output& Output::operator<<( const Reference& ref )
{
    (*this) << ref.getClass() << ref.getInstance();
    return (*this);
}

//: Get object instance
INT_2U Output::getInstance( const Base* b )
{
    // Find the instance for this pointer
    std::map< const Base*, INT_2U >::const_iterator iter = mInstance.find( b );

    // If it is found, return the instance
    if ( iter != mInstance.end() )
    {
        return iter->second;
    }
    // This defaults to the first instance
//    INT_2U instance = 1;
    INT_2U instance = 0;

    // find the last instance value for this class type
    std::map< INT_2U, INT_2U >::iterator iter2 =
        mMaxInstance.find( b->getClassId() );
    if ( iter2 == mMaxInstance.end() )
    {
        // This class type wasn't present, so add it
//        mMaxInstance.insert( std::pair< INT_2U, INT_2U >( b->getClassId(), 1 ) );
        mMaxInstance.insert( std::pair< INT_2U, INT_2U >( b->getClassId(), 0 ) );
    }
    else
    {
        // Ok, this isn't the 1st instance.  Increment the value in
        // MaxInstance.  This is the instance for this object.
        (*iter2).second += 1;
        instance = iter2->second;
    }

    // Add this object to the instance list
    mInstance.insert( std::pair< const Base*, INT_2U >( b, instance ) );
        
    return instance;
}

//: Write file header out
void Output::open()
{
    mNFrames = 0;
    mNBytes = 0;
    mClosed = false;
    mIsNode = false;

    FileHeader::writeLocal( *this, dictionary.getVersion() );
}

//: Finish frame file and close
void Output::close( unsigned long tocSize )
{
    if ( mClosed )
    {
        return;
    }

    int deltaBytes = 0;
    
    if ( mMaxInstance.find( ID_ENDOFFILE ) == mMaxInstance.end() )
    {
        const SH& sh( dictionary.find( ID_ENDOFFILE )->second.getSH() );
        deltaBytes += 16 + sh.getName().size() + sh.getComment().size();

        for ( SH::const_se_iterator iter = sh.getSE().begin();
              iter != sh.getSE().end(); ++iter )
        {
            deltaBytes += 17 + (*iter)->getName().size()
                + (*iter)->getClass().size() + (*iter)->getComment().size();
        }
    }
    
    deltaBytes += 24 + 4;
    
    // First write the EndOfFile object
    EndOfFile eof( mNFrames, mNBytes + deltaBytes, false, 0,
		   tocSize? deltaBytes + tocSize: 0 );
    ((Base*)&eof)->write( *this );

    mClosed = true;
}

//: Place an object into the subobjects queue
//   
//!exc: write_failure - Write failed.   
//      
void Output::queueObject( const Base& o )
{
    if ( mCurrentInfo == 0 )
    {
        throw write_failure( "Logic error.  Attempt to write before "
                             "startRecord called." );
    }
    
    mCurrentInfo->subobjects.push_back( &o );
    mCurrentInfo->isContainer.push_back( false );
}
