#include "framecpp/config.h"

#include <stdexcept>

#include "procdata.hh"
#include "framereader.hh"
#include "output.hh"


using FrameCPP::Version_3_4_5::ProcData;
using FrameCPP::Version_3_4_5::Vect;
using FrameCPP::Version_3_4_5::Container;
using FrameCPP::Version_3_4_5::Base;

using std::runtime_error;


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates a ProcData object with the following properties:
// <ul>
//     <li> name = "none" </li>
//     <li> comment = "" </li>
//     <li> sampleRate = 0.0 </li>
//     <li> timeOffset = 0 seconds, 0 nanoseconds </li>
//     <li> fShift = 0.0 </li>
//     <li> phase = 0.0 </li>
//     <li> No other data. </li>
// </ul>
//
//!exc: None.
//      
ProcData::ProcData()
        : Base(), 
          mName( "none" ), mComment( "" ), mSampleRate( 0.0 ),
          mTimeOffset( 0, 0 ), mFShift( 0.0 ), mPhase ( 0. ),
          mData(), mTable(), mAux()
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
// This performs a deep copy of the object.
//
//!param: const ProcData& procData - Object to be copied.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
ProcData::ProcData( const ProcData& procData )
        : Base( procData ), 
          mName( procData.mName ),
          mComment( procData.mComment ), mSampleRate( procData.mSampleRate ),
          mTimeOffset( procData.mTimeOffset ), 
          mFShift( procData.mFShift ), mPhase( procData.mPhase ),
          mData( procData.mData ), mTable( procData.mTable ), mAux( procData.mAux )
{
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: const std::string& name - Data or channel name.
//!param: const REAL_8& sampleRate - Data acquisition rate, samples / s.
//!param: const Time& timeOffset - Integer seconds time relative to frame start.
//!param: const REAL_8& fShift - Frequency shift.
//!param: const REAL_4& phase - Phase.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
ProcData::ProcData( const std::string& name,
                    const REAL_8& sampleRate, const Time& timeOffset,
                    const REAL_8& fShift, const REAL_4& phase )
        : Base(), mName( name ), mComment(),
          mSampleRate( sampleRate ), mTimeOffset( timeOffset ),
          mFShift( fShift ), mPhase( phase ),
	  mData(), mTable(), mAux()
{
}

   
//-----------------------------------------------------------------------------
//
//: Destructor.
//
// This method currently does nothing.
//
//!exc: None.
//   
ProcData::~ProcData()
{
}


//-----------------------------------------------------------------------------
//
//: Get the data or channel name.
//
//!return: const std::string& -- Name.
//
const std::string& ProcData::getName() const
{
    return mName;
}


//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
//!param: const ProcData& procData - Object to be assigned.
//
//!return: const ProcData& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const ProcData& ProcData::operator=( const ProcData& procData )
{
    if ( this != &procData )
    {
        mName = procData.mName;
        mComment = procData.mComment;
        mSampleRate = procData.mSampleRate;
        mTimeOffset = procData.mTimeOffset;
        mFShift = procData.mFShift;
        mPhase = procData.mPhase;
        mData = procData.mData;
        mTable = procData.mTable;
        mAux = procData.mAux;
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: += Operator.
//
//!param: const ProcData& procData - Object to be added.
//
//!return: const ProcData& - This object.
//
const ProcData& ProcData::operator+=( const ProcData& procData )
{
    throw runtime_error( "Unimplemented ProcData::operator+=() is called." );

    return *this;
}



//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
//: Constructs the Structure Header for this class.  
//
//!return: SH - Proc Data structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH ProcData::getSH()
{
  SH sh( "FrProcData", ID_PROCDATA, "Post-processed Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "comment", "STRING", "" ) );
  sh.appendSE( SE( "sampleRate", "REAL_8", "" ) );
  sh.appendSE( SE( "timeOffsetS", "INT_4U", "" ) );
  sh.appendSE( SE( "timeOffsetN", "INT_4U", "" ) );
  sh.appendSE( SE( "fShift", "REAL_8", "" ) );
  sh.appendSE( SE( "phase", "REAL_4", "" ) );
  sh.appendSE( SE( "data", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "aux", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "table", "PTR_STRUCT(FrTable *)", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrProcData *)", "" ) );
  return sh;
}


//-----------------------------------------------------------------------------
//
//: Constructs frame data version 4 Structure Header for this class.  
//
//!return: SH - Proc Data structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH ProcData::getSH4()
{
  SH sh( "FrProcData", ID_PROCDATA, "Post-processed Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "comment", "STRING", "" ) );
  sh.appendSE( SE( "sampleRate", "REAL_8", "" ) );
  sh.appendSE( SE( "timeOffsetS", "INT_4U", "" ) );
  sh.appendSE( SE( "timeOffsetN", "INT_4U", "" ) );
  sh.appendSE( SE( "fShift", "REAL_8", "" ) );
  sh.appendSE( SE( "data", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "aux", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "table", "PTR_STRUCT(FrTable *)", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrProcData *)", "" ) );
  return sh;
}

//-----------------------------------------------------------------------------
//
//: Constructs frame data vesion 3 Structure Header for this class.  
//
//!return: SH - Proc data structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH ProcData::getSH3()
{
    SH sh( "FrProcData", ID_PROCDATA, "Post-processed Data" );
    sh.appendSE( SE( "name", "STRING", "" ) );
    sh.appendSE( SE( "comment", "STRING", "" ) );
    sh.appendSE( SE( "sampleRate", "REAL_8", "" ) );
    sh.appendSE( SE( "timeOffsetS", "INT_4U", "" ) );
    sh.appendSE( SE( "timeOffsetN", "INT_4U", "" ) );
    sh.appendSE( SE( "fShift", "REAL_8", "" ) );
    sh.appendSE( SE( "data", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "aux", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "next", "PTR_STRUCT(FrProcData *)", "" ) );

    return sh;
}

  
//-----------------------------------------------------------------------------
//
//: Sets a timeOffset.
//   
// This method is a temporary solution to frameAPI::adcData--->procData 
// conversion. It sets the timeoffset metadata after ProcData object has been
// constructed. 
//   
//!param: const INT_4S sec - Value of seconds. Default is 0.  
//!param: const INT_4U nan - Value of nanoseconds. Default is 0.     
//
//!return: Nothing.
//   
void ProcData::setTimeOffset( const INT_4S sec, const INT_4U nan )
{
   mTimeOffset.setSec( sec );
   mTimeOffset.setNSec( nan );
   return;
}

//-----------------------------------------------------------------------------
//
//: Reads an ProcData object from an FrameReaderBase object.
//
// This method creates a dynamically allocated object.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: FrameReaderBase& in - Input data stream object.
//
//!return: ProcData* - Newly allocated.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//
ProcData* ProcData::read( FrameReaderBase& in )
{
  ProcData& proc = *new ProcData();
  in >> proc.mName >> proc.mComment >> proc.mSampleRate >> proc.mTimeOffset
     >> proc.mFShift >> proc.mPhase
     >> proc.mData >>  proc.mAux >> proc.mTable;
  
  in.readNextNodeReference();
  return &proc;  
}

//: Version 4 frame input method.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//   
ProcData* ProcData::read4( FrameReaderBase& in )
{
  ProcData& proc = *new ProcData();
  in >> proc.mName >> proc.mComment >> proc.mSampleRate >> proc.mTimeOffset >> proc.mFShift
     >> proc.mData >>  proc.mAux >> proc.mTable;
  
  in.readNextNodeReference();
  return &proc;  
}

//: Version 3 frame input method.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//   
ProcData* ProcData::read3( FrameReaderBase& in )
{
    ProcData& proc = *new ProcData();
    in >> proc.mName >> proc.mComment >> proc.mSampleRate >> proc.mTimeOffset >> proc.mFShift
       >> proc.mData >> proc.mAux;

    in.readNextNodeReference();
    return &proc;
}


//-----------------------------------------------------------------------------
//
//: Writes an ProcData object to an Output object.
//
//!param: Output& out out - Output data stream object.
//
//!exc: write_failure - Write failure occured.
//
void ProcData::write( Output& out ) const
{
    out.startRecord( ID_PROCDATA,
		     58
		     + mName.size() + mComment.size(),
                     this );
    
    out << mName << mComment << mSampleRate << mTimeOffset
	<< mFShift << mPhase
        << mData  << mAux << mTable;
    
    out.endRecord( this );
}

    
//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two ProcData objects are equal. For this comparison,
// all of the data is compared.
//
//!param: const ProcData& procData - The object to compare with.
//
//!return: bool - true if the objects are equal.
//
//!kw: const
//
//!exc: None.
//   
bool ProcData::operator==( const ProcData& procData ) const
{
    return ( mName == procData.mName &&
             mComment == procData.mComment &&
             mSampleRate == procData.mSampleRate &&
             mTimeOffset == procData.mTimeOffset &&
             mFShift == procData.mFShift &&
             mPhase == procData.mPhase &&
             mData == procData.mData &&
             mTable == procData.mTable &&
             mAux == procData.mAux );
}

   
//-----------------------------------------------------------------------------
//
//: Gets the class ID.
//
//!return: INT_2U -- Proc data class ID.
//
//!exc: None.
//      
INT_2U ProcData::getClassId() const
{
    return ID_PROCDATA;
}

   
