#ifndef FrameCppProcDataHH
#define FrameCppProcDataHH


#include "framecpp/config.h"
#include "framecpp/Version3_4_5/base.hh"
#include "framecpp/Version3_4_5/sh.hh"
#include "framecpp/Version3_4_5/vect.hh"
#include "framecpp/Version3_4_5/table.hh"
#include "framecpp/Version3_4_5/time.hh"

#include "framecpp/Version3_4_5/searchcontainer.hh"


namespace FrameCPP {

namespace Version_6
{
  class FrProcData;
}

namespace Version_3_4_5 {

class FrameReaderBase;


//-----------------------------------------------------------------------------
//
//: This class represents post-processed data.
//
// It is modeled after the FrProcData structure in the Frame Specification,
// LIGO-T970130-B-E.
//


class ProcData : public Base
{
public:

    //!ignore_begin:
    DECLARE_DEFAULT_CONTAINER( Data, data, FrameCPP::Version_3_4_5::Vect, getName );
    DECLARE_CONTAINER( Table, table, Table, getName );
    DECLARE_CONTAINER( Aux, aux, FrameCPP::Version_3_4_5::Vect, getName );
    //!ignore_end:

    /* Constructors */
    //-----------------------------------------------------------------------------
    //
    //: Default Constructor.
    //
    // This creates a ProcData object with the following properties:
    // <ul>
    //     <li> name = "none" </li>
    //     <li> comment = "" </li>
    //     <li> sampleRate = 0.0 </li>
    //     <li> timeOffset = 0 seconds, 0 nanoseconds </li>
    //     <li> fShift = 0.0 </li>
    //     <li> phase = 0.0 </li>
    //     <li> No other data. </li>
    // </ul>
    //
    //!exc: None.
    //   
    ProcData();


    //-----------------------------------------------------------------------------
    //
    //: Copy Constructor.
    //
    // This performs a deep copy of the object.
    //
    //!param: const ProcData& procData - Object to be copied.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    ProcData( const ProcData& procData );

    //-----------------------------------------------------------------------------
    //
    //: Constructor.
    //
    //!param: const std::string& name - Data or channel name.
    //!param: const REAL_8& sampleRate - Data acquisition rate, samples / s.
    //!param: const Time& timeOffset - Integer seconds time relative to frame start.
    //!param: const REAL_8& fShift - Frequency shift.
    //!param: const REAL_4& phase - Phase.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    ProcData( const std::string& name,
              const REAL_8& sampleRate, const Time& timeOffset,
              const REAL_8& fShift, const REAL_4& phase );
    
    /* Destructor */
    //!exc: None.
    virtual ~ProcData();

    /* Operator Overloads */

    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    //!param: const ProcData& procData - Object to be assigned.
    //
    //!return: const ProcData& -- This object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const ProcData& operator=( const ProcData& procData );

    //-----------------------------------------------------------------------------
    //
    //: += Operator.
    //
    //!param: const ProcData& procData - Object to be added.
    //
    //!return: const ProcData& -- This object.
    //
    const ProcData& operator+=( const ProcData& procData );
    
    //-----------------------------------------------------------------------------
    //
    //: Equal comparison.
    //
    // Determines whether two ProcData objects are equal. For this comparison,
    // all of the data is compared.
    //
    //!param: const ProcData& procData - The object to compare with.
    //
    //!return: bool -- true if the objects are equal.
    //
    //!kw: const
    //
    //!exc: None.
    //      
    bool operator==( const ProcData& procData ) const;
    bool operator!=( const ProcData& procData ) const;
 
    /* Required Methods */
    //!exc: None.
    virtual INT_2U getClassId() const;


    //-----------------------------------------------------------------------------
    //
    //: Constructs the Structure Header for this class.  
    //
    //!return: SH -- Proc Data structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH();

    //: Constructs frame data vesion 4 Structure Header for this class.  
    //!exc: std::bad_alloc - Memory allocation failed.   
    static SH getSH4();

    //: Constructs fram data vesion 3 Structure Header for this class.  
    //!exc: std::bad_alloc - Memory allocation failed.   
    static SH getSH3();

    /* Accessors */

    /* gcc 2.95.1 can't handle a typedef with this if the throw spec is
       present (e.g.
       typedef SearchContainer< ProcData, &ProcData::getName > dc; ).
       
       const std::string& getName() const throw(); */
    const std::string& getName() const;
    //!exc: None.
    const std::string& getComment() const;
    //!exc: None.   
    REAL_8 getSampleRate() const;
    //!exc: None.   
    const Time& getTimeOffset() const;
    //!exc: None.   
    REAL_8 getFShift() const;
    //!exc: None.   
    REAL_4 getPhase() const;
   
    /* Data Accessors */
    //!exc: None.   
    DataContainer& refData();
    //!exc: None.   
    const DataContainer& refData() const;

    /* Table Accessors */
    //!exc: None.   
    TableContainer& refTable();
    //!exc: None.   
    const TableContainer& refTable() const;

    /* Aux Accessors */
    //!exc: None.   
    AuxContainer& refAux();
    //!exc: None.   
    const AuxContainer& refAux() const;

    /* Mutators */
    //!exc: std::bad_alloc - Memory allocation failed.
    void appendComment( const std::string& comment );
    void setTimeOffset( const INT_4S sec = 0, const INT_4U nan = 0 );   

private:
  friend class Dictionary;
  friend class FrameCPP::Version_6::FrProcData;

    //-----------------------------------------------------------------------------
    //
    //: Reads an ProcData object from an FrameReaderBase object.
    //
    // This method creates a dynamically allocated object.
    //
    // <p>*NOTE* This method allocates memory and returns a pointer to it.  The
    // user is responsible for freeing this memory.
    //
    //!param: FrameReaderBase& in - Input data stream object.
    //
    //!return: ProcData* -- Newly allocated.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //
    static ProcData* read( FrameReaderBase& in );

    //: Version 4 frame input method.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //   
    static ProcData* read4( FrameReaderBase& in );

    //: Version 3 frame input method.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //   
    static ProcData* read3( FrameReaderBase& in );

    //-----------------------------------------------------------------------------
    //
    //: Writes an ProcData object to an Output object.
    //
    //!param: Output& out out - Output data stream object.
    //
    //!exc: write_failure - Write failure occured.
    //
    virtual void write( Output& out ) const;
    
private:

    //: Data or channel name.
    std::string mName;
    //: Comment.
    std::string mComment;
    //: Sampling rate.
    REAL_8 mSampleRate;
    //: Integer seconds start time relative to frame start.
    Time mTimeOffset;
    //: Frequency shift.
    REAL_8 mFShift;
    //: Phase of heterodyning signal at start of dataset
    REAL_4 mPhase;
    //: Sampled data
    DataContainer mData;
    //: Tabled data.
    TableContainer mTable;
    //: Generic data.
    AuxContainer mAux;
};
    




/******************/
/* Inline Methods */
/******************/



//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------



//-----------------------------------------------------------------------------
//
//: Not-equal comparison.
//
//!param: const ProcData& procData - The object to compare with.
//
//!return: bool -- True if the objects are not equal.
//
//!kw: inline const
//
//!exc: None.
//      
inline bool ProcData::operator!=( const ProcData& procData ) const
{
    return !( *this == procData );
}



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------



//-----------------------------------------------------------------------------
//
//: Get the comment.
//
//!return: const std::string&  -- Comment.
//
//!exc: None.
//      
inline const std::string& ProcData::getComment() const
{
    return mComment;
}


//-----------------------------------------------------------------------------
//
//: Get the data acquisition rate, samples / sec.
//
//!return: REAL_8 -- Sampling rate.
//
//!exc: None.
//      
inline REAL_8 ProcData::getSampleRate() const
{
    return mSampleRate;
}


//-----------------------------------------------------------------------------
//
//: Get the time relative to the frame start.
//
// This is for triggered data lasting less than one frame, it is the integer seconds and nanoseconds.
//
//!return: const Time& -- Relative time.
//
//!exc: None.
//      
inline const Time& ProcData::getTimeOffset() const
{
    return mTimeOffset;
}


//-----------------------------------------------------------------------------
//
//: Get the frequency shift if the signal has been heterodyned before the ADC.
//
// fShift = (f_heterodyne - fNyquist@sampleRate).
//
//!return: REAL_8 -- Frequency shift.
//
//!exc: None.
//      
inline REAL_8 ProcData::getFShift() const
{
    return mFShift;
}


//-----------------------------------------------------------------------------
//
//: Get phase of heterodyning signal at start of dataset.
//
//!return: REAL_4 -- phase.
//
//!exc: None.
//      
inline REAL_4 ProcData::getPhase() const
{
    return mPhase;
}


//-----------------------------------------------------------------------------
//
//: Get the container which contains the data.
//
//!return: DataContainer& -- Data.
//
//!kw: inline
//
//!exc: None.
//      
inline ProcData::DataContainer&
ProcData::refData()
{
    return mData;
}


//-----------------------------------------------------------------------------
//
//: Get the container which contains the data.
//
//!return: const DataContainer& -- Data.
//
//!kw: inline const
//
//!exc: None.
//      
inline
const ProcData::DataContainer&
ProcData::refData() const
{
    return mData;
}


//-----------------------------------------------------------------------------
//
//: Get the container for user provided information.
//
//!return: AuxContainer& -- Generic data.
//
//!exc: None.
//      
inline ProcData::AuxContainer&
ProcData::refAux()
{
    return mAux;
}


//-----------------------------------------------------------------------------
//
//: Get the container for user provided information.
//
//!return: const AuxContainer& -- Generic data.
//
//!exc: None.
//      
inline
const ProcData::AuxContainer&
ProcData::refAux() const
{
    return mAux;
}

//-----------------------------------------------------------------------------
//
//: Get the container for user provided table data.
//
//!return: SearchContainer< Table, &Table::getName >& -- Table.
//
//!exc: None.
//      
inline ProcData::TableContainer&
ProcData::refTable()
{
  return mTable;
}


//-----------------------------------------------------------------------------
//
//: Get the container for user provided table data.
//
//!return: const SearchContainer< Table, &Table::getName >& -- Table.
//
//!exc: None.
//      
inline const ProcData::TableContainer&
ProcData::refTable() const
{
  return mTable;
}


//-----------------------------------------------------------------------------
// Mutators
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Appends a comment.
//
// This method adds a comment followed by a newline to the
// end of the current list of comments.
//
//!param: const std::string& comment - Comment.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
inline void ProcData::appendComment( const std::string& comment )
{
    if ( mComment.size() > 0 ) {
	mComment += "\n" + comment;
    } else
	mComment = comment;
}

 
} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // FrameCppProcDataHH
