#ifndef FrameCppReferenceHH
#define FrameCppReferenceHH


#include "types.hh"


namespace FrameCPP {

namespace Version_3_4_5 {


class Input;



//-----------------------------------------------------------------------------
//
//: A reference to a Frame structure object in a file.
//
class Reference
{
public:

    /* Constructors */
    //!exc: None.
    Reference();
    //!exc: None.   
    Reference( INT_2U classId, INT_2U instance );
    //!exc: None.   
    Reference( const Reference& ref );

    /* Operator Overloads */
    //-----------------------------------------------------------------------------
    //
    //: Assignment operator.
    //
    //!param: const Reference& ref - Object to be assigned.
    //
    //!return: const Reference& -- This object.
    //
    //!exc: None.   
    const Reference& operator=( const Reference& ref );


    //-----------------------------------------------------------------------------
    //
    //: Equal comparison operator.
    //
    //!param: const Reference& ref - Object with which to compare.
    //
    //!return: bool -- True if objects have the same data.
    //
    //!exc: None.   
    bool operator==( const Reference& ref ) const;

    //-----------------------------------------------------------------------------
    //
    //: Less than or equal to comparison operator.
    //
    //!param: const Reference& ref - Object with which to compare.
    //
    //!return: bool -- True if this object is less than or equal so `ref'
    //
    //!exc: None.   
    bool operator<=( const Reference& ref ) const;

    //-----------------------------------------------------------------------------
    //
    //: Greater than or equal to comparison operator.
    //
    //!param: const Reference& ref - Object with which to compare.
    //
    //!return: bool -- True if this object is greater that or equal to `ref'.
    //
    //!exc: None.   
    bool operator>=( const Reference& ref ) const;

    //-----------------------------------------------------------------------------
    //
    //: Not equal comparison operator.
    //
    //!param: const Reference& ref - Object with which to compare.
    //
    //!return: bool -- True if this object is not equal to `ref'.
    //
    //!exc: None.   
    bool operator!=( const Reference& ref ) const;

    //-----------------------------------------------------------------------------
    //
    //: Less than comparison operator.
    //
    //!param: const Reference& ref - Object with which to compare.
    //
    //!return: bool -- True if this object is less that `ref'.
    //
    //!exc: None.   
    bool operator<( const Reference& ref ) const;

    //-----------------------------------------------------------------------------
    //
    //: Greater than comparison operator.
    //
    //!param: const Reference& ref - Object with which to compare.
    //
    //!return: bool -- True is this object is greater that `ref'.
    //
    //!exc: None.   
    bool operator>( const Reference& ref ) const;

    /* Accessors */
    INT_2U getClass() const;
    INT_2U getInstance() const;

    /* Friends */
    //-----------------------------------------------------------------------------
    //
    //: Input extraction operator.
    //
    //!param: Input& in - Input data stream object.
    //!param: Reference& ref - Data target.
    //
    //!return: Input& -- Input data stream object.
    //
    friend Input& operator>>( Input&, Reference& ref );
    
private:
    //: Structure ID << 16 & Structure Instance.
    INT_4U mReference;
};


//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: INT_2U classId - Structure class ID.
//!param: INT_2U instance - Structure instance.
//
//!exc: None.   
//   
inline Reference::Reference( INT_2U classId, INT_2U instance )
{
    mReference = (INT_4U(classId)<<16) + INT_4U(instance);
}


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates a reference to the zeroth instance of class ID 0.
//
//!todo: Should this constructor exist?
//
//!exc: None.   
//      
inline Reference::Reference()
        : mReference( 0 )
{}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
//!param: const Reference& ref - Object to be copied.
//
//!exc: None.   
//      
inline Reference::Reference( const Reference& ref )
        : mReference( ref.mReference )
{}

//-----------------------------------------------------------------------------
//
//: Get the class ID for this reference.
//
//!return: INT_2U -- Class ID.
//
inline INT_2U Reference::getClass() const
{
    return (mReference >> 16);
}


//-----------------------------------------------------------------------------
//
//: Get the instance number for this reference.
//
//!return: INT_2U -- Class instance.
//
inline INT_2U Reference::getInstance() const
{
    return (mReference & 0xffff);
}


//-----------------------------------------------------------------------------
//
//: Less than comparison operator.
//
//!param: const Reference& ref - Object with which to compare.
//
//!return: bool -- True if this object is less that `ref'.
//
//!exc: None.   
//      
inline bool Reference::operator<( const Reference& ref ) const
{
    return ( mReference < ref.mReference );
}


//-----------------------------------------------------------------------------
//
//: Greater than comparison operator.
//
//!param: const Reference& ref -
//
//!return: bool -
//
//!exc: None.   
//      
inline bool Reference::operator>( const Reference& ref ) const
{
    return ( mReference > ref.mReference );
}

//-----------------------------------------------------------------------------
//
//: Equal comparison operator.
//
//!param: const Reference& ref - Object with which to compare.
//
//!return: bool -- True if objects have the same data.
//
//!exc: None.   
//      
inline bool Reference::operator==( const Reference& ref ) const
{
    return ( mReference == ref.mReference);
}

    
//-----------------------------------------------------------------------------
//
//: Less than or equal to comparison operator.
//
//!param: const Reference& ref - Object with which to compare.
//
//!return: bool -- True if this object is less than or equal so `ref'
//
//!exc: None.   
//      
inline bool Reference::operator<=( const Reference& ref ) const
{
    return !( *this > ref );
}


//-----------------------------------------------------------------------------
//
//: Greater than or equal to comparison operator.
//
//!param: const Reference& ref - Object with which to compare.
//
//!return: bool -- True if this object is greater that or equal to `ref'.
//
//!exc: None.   
//      
inline bool Reference::operator>=( const Reference& ref ) const
{
    return !( *this < ref );
}


//-----------------------------------------------------------------------------
//
//: Not equal comparison operator.
//
//!param: const Reference& ref - Object with which to compare.
//
//!return: bool -- True if this object is not equal to `ref'.
//
//!exc: None.   
//      
inline bool Reference::operator!=( const Reference& ref ) const
{
    return !( *this == ref );
}


Input& operator>>( Input& in, Reference& ref );


} // namespace - Version_3_4_5

} // namespace - FrameCPP




#endif // FrameCppReferenceHH
