#include "framecpp/config.h"

#include "se.hh"
#include "input.hh"
#include "output.hh"


using FrameCPP::Version_3_4_5::SE;


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Copy constructor.
//
//!param: const SE& se - The object to copy from.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
SE::SE( const SE& se )
        : mName( se.mName ), mClass( se.mClass ), mComment( se.mComment )
{
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
// This creates a Structure Element with the given attributes.
//
//!param: const std::string& name - The field name.
//!param: const std::string& type - The Class type
//!param: const std::string& comment - A comment.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
SE::SE( const std::string& name, const std::string& type, const std::string& comment )
        : mName( name ), mClass( type ), mComment( comment )
{
}


//-----------------------------------------------------------------------------
//
//: Input Constructor.
//
// This creates a Structure Element by reading it from an Input object.
//
//!param: Input& in - The Input object to read from.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failed.
//
//!kw: private
//
//!todo: Check about changing this to reading from a FrameReaderBase.
//
SE::SE( Input& in )
        : mName(), mClass(), mComment()
{
    in >> mName >> mClass >> mComment;
}


//-----------------------------------------------------------------------------
//
//: Get the field name.
//
//!return: const std::string& -- The name.
//
//!kw: inline const
//
//!exc: None.   
//      
const std::string& SE::getName() const
{
    return mName;
}



//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator..
//
//!param: const SE& se - The object to assign from.
//
//!return: const SE& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const SE& SE::operator=( const SE& se )
{
    if ( this != &se )
    {
        mName = se.mName;
        mClass = se.mClass;
        mComment = se.mComment;
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two SE objects are equal. For this comparison, the
// comments are not compared.
//
//!param: const SE& se - The object to compare with.
//
//!return: bool - true if the objects are equal.
//
//!kw: const
//
//!exc: None.   
//      
bool SE::operator==( const SE& se ) const
{
    return ( ( mName == se.mName ) && ( mClass == se.mClass ) );
}    


//-----------------------------------------------------------------------------
//
//: Instantiates a new SE object from a stream.
//
// This method is used only by the Input object.
//
//!param: Input& in - The object to read from.
//
//!return: SE* - The new object.  The user must destruct this object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failed.
//
//!kw: static
//
SE* SE::read( Input& in )
{
    return new SE( in );
}


//-----------------------------------------------------------------------------
//
//: Write this object.
//
// Writes this SE to an output object.
//
//!param: Output& out - The object to write to.
//
//!exc: write_failure - Write failed.
//
//!kw: const
//
void SE::write( Output& out ) const 
{
    out.startRecord( ID_SE, 17 + mName.size() + mClass.size()
                     + mComment.size(), this );
    out << mName << mClass << mComment;
    out.endRecord( this );
}


//-----------------------------------------------------------------------------
//
//: Get the Class ID associated with this class type.
//
//!return: INT_2U -- The SE class ID.
//
//!kw: inline const
//
//!exc: None.   
//   
INT_2U SE::getClassId() const
{
    return ID_SE;
}
   
   
