#include "framecpp/config.h"

#include <stdexcept>

#include "serdata.hh"
#include "framereader.hh"
#include "output.hh"


using FrameCPP::Version_3_4_5::SerData;
using FrameCPP::Version_3_4_5::Vect;
using FrameCPP::Version_3_4_5::Container;
using FrameCPP::Version_3_4_5::Base;

using std::runtime_error;


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates a SerData object with the following properties:
// <ul type=disc>
//     <li> name = "none" </li>
//     <li> time = 0 seconds, 0 nanoseconds </li>
//     <li> sampleRate = 0.0 </li>
//     <li> ASCII data = "" </li>
//     <li> No other data. </li>
// </ul>
//
//!exc: None.
//      
SerData::SerData()
        : Base(), 
          mName( "none" ), mTime( 0, 0 ), mSampleRate( 0.0 ),
          mData( "" ), mSerial(), mTable()
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
// This performs a deep copy for owned data.
//
//!param: const SerData& serData - The object to copy from
//
//!exc: std::bad_alloc - Memory allocation failed
//
SerData::SerData( const SerData& serData )
        : Base( serData ), 
          mName( serData.mName ), mTime( serData.mTime ),
          mSampleRate( serData.mSampleRate ), mData( serData.mData ),
          mSerial( serData.mSerial ), mTable( serData.mTable )
{
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: const std::string& name - Name of station producing serial data stream.
//!param: const Time& time - Time of data acquisition.
//!param: REAL_4& sampleRate - Sampling rate, sample / s.
//
//!exc: std::bad_alloc - Memory allocation failed.
//   
SerData::SerData( const std::string& name, const Time& time, REAL_4 sampleRate )
        : Base(), 
          mName( name ), mTime( time ), mSampleRate( sampleRate ),
          mData( "" ), mSerial(), mTable()
{
}


//-----------------------------------------------------------------------------
//
//: Get Name
//
// Gets the name.
//
// This returns the name of the station producing the serial data stream.
//
//!return: const std::string& -- Name.
//
const std::string& SerData::getName() const
{
    return mName;
}



//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
// This performs a deep copy for owned data.
//
//!param: const SerData& serData - The object to assign from.
//
//!return: const SerData& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const SerData& SerData::operator=( const SerData& serData )
{
    if ( this != &serData )
    {
        mName = serData.mName;
        mTime = serData.mTime;
        mSampleRate = serData.mSampleRate;
        mData = serData.mData;
        mSerial = serData.mSerial;
        mTable = serData.mTable;
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: += Operator.
//
//!param: const SerData& serData - The object to be added.
//
//!return: const SerData& - This object.
//
const SerData& SerData::operator+=( const SerData& serData )
{
    throw runtime_error( "Unimplemented SerData::operator+=() is called." );

    return *this;
}


//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Get Structure Header
//
// Constructs the Structure Header for this class.  
//
//!return: SH - Structure header.
//
//!exc: std::bad_alloc - Memory allocation failed
//
FrameCPP::Version_3_4_5::SH SerData::getSH()
{
  SH sh( "FrSerData", ID_SERDATA, "Serial Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "timeSec", "INT_4U", "" ) );
  sh.appendSE( SE( "timeNsec", "INT_4U", "" ) );
  sh.appendSE( SE( "sampleRate", "REAL_4", "" ) );
  sh.appendSE( SE( "data", "STRING", "" ) );
  sh.appendSE( SE( "serial", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "table", "PTR_STRUCT(FrTable *)", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrSerData *)", "" ) );
  return sh;
}

//-----------------------------------------------------------------------------
//
//: Get frame data format version 3 Structure Header
//
// Constructs the Structure Header for this class.  
//
//!return: SH - Structure header.
//
//!exc: std::bad_alloc - Memory allocation failed
//
FrameCPP::Version_3_4_5::SH SerData::getSH3() 
{
    SH sh( "FrSerData", ID_SERDATA, "Serial Data" );
    sh.appendSE( SE( "name", "STRING", "" ) );
    sh.appendSE( SE( "timeSec", "INT_4U", "" ) );
    sh.appendSE( SE( "timeNsec", "INT_4U", "" ) );
    sh.appendSE( SE( "sampleRate", "REAL_4", "" ) );
    sh.appendSE( SE( "data", "STRING", "" ) );
    sh.appendSE( SE( "serial", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "more", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "next", "PTR_STRUCT(FrSerData *)", "" ) );

    return sh;
}


//-----------------------------------------------------------------------------
//
//: Reads a SerData object from an FrameReaderBase object.
//
// This method creates a dynamically allocated object.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: FrameReaderBase& in - Input data stream object.
//
//!return: SerData* - Newly created.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//
SerData* SerData::read( FrameReaderBase& in ) 
{
  SerData& ser = *new SerData();
  in >> ser.mName >> ser.mTime >> ser.mSampleRate >> ser.mData >> ser.mSerial >> ser.mTable;
  in.readNextNodeReference();
  return &ser;
}

//: Version 3 frame input function.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//   
SerData* SerData::read3( FrameReaderBase& in )
{
    SerData& ser = *new SerData();
    INT_4U dummyReference;

    // This will read in Vect reference (class_id, intance_id), 4 bytes
    // It is possible to put this vector into a table and assign to mTable
    //

    in >> ser.mName >> ser.mTime >> ser.mSampleRate >> ser.mData >> ser.mSerial >> dummyReference;

    in.readNextNodeReference();

    return &ser;
}


//-----------------------------------------------------------------------------
//
//: Writes a SerData object to an Output object.
//
//!param: Output& out out - Output data stream object.
//
//!exc: write_failure - Write failure occured.
//
void SerData::write( Output& out ) const 
{
  out.startRecord( ID_SERDATA,
		   38
		   + mName.size() + mData.size(),
		   this );

  out << mName << mTime << mSampleRate << mData << mSerial << mTable;
  out.endRecord( this );
}    


//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two SerData objects are equal. For this comparison,
// all of the data is compared.
//
//!param: const SerData& serData - The object to compare with.
//
//!return: bool - true if the objects are equal.
//
//!exc: None.
//   
bool SerData::operator==( const SerData& serData ) const
{
    return ( mName == serData.mName &&
             mTime == serData.mTime &&
             mSampleRate == serData.mSampleRate &&
             mData == serData.mData &&
             mSerial == serData.mSerial &&
             mTable == serData.mTable );
}

   
//-----------------------------------------------------------------------------
//
//: Get Class ID
//
// Gets the class ID.
//
//!return: INT_2U -- Class ID.
//
//!exc: None.
//   
INT_2U SerData::getClassId() const
{
    return ID_SERDATA;
}

   
