#ifndef FrameCppSimEventHH
#define FrameCppSimEventHH


#include "framecpp/config.h"
#include "framecpp/Version3_4_5/base.hh"
#include "framecpp/Version3_4_5/sh.hh"
#include "framecpp/Version3_4_5/vect.hh"
#include "framecpp/Version3_4_5/table.hh"
#include "framecpp/SearchContainer.hh"


namespace FrameCPP {

namespace Version_6
{
  class FrSimEvent;
}

namespace Version_3_4_5 {

  class FrameReaderBase;


//-----------------------------------------------------------------------------
//
//: Simulated Event data.
//
// <P>SimEvent possesses the following attributes:
//
// <table border=2 cellpadding=4>
//   <tr><td> <B>Attribute</B></td><td><B>Description</B> </td></tr>
//   <tr><td> name          </td><td>Name of trigger                                 </td></tr>
//   <tr><td> comment       </td><td>Descriptor of trigger                           </td></tr>
//   <tr><td> inputs        </td><td>Input channel and filter parameters to trigger
//                           process                                         </td></tr>
//   <tr><td> GTime         </td><td>GPS time (sec,nsec residual) corresponding to
//                           maximum of trigger                              </td></tr>
//   <tr><td> timeBefore    </td><td>Signal duration before GTime.getSec()  </td></tr>
//   <tr><td> timeAfter     </td><td>Signal duration after GTime.getSec()   </td></tr>
//   <tr><td> amplitude     </td><td>Continuous output amplitude returned by trigger </td></tr>
// </table>
//                
// <p>The SimEvent object is a container for tow types of data:
//
// <table border=2 cellpadding=4>
//   <tr><td> <B>Data</B></td><td><B>Description</B>                             </td></tr>
//   <tr><td> data </td><td>A container of Vect objects which contain additional trigger
//                  results. </td></tr>
//   <tr><td> table</td><td>Table with additional event information              </td></tr>
// </table>
//
// <p>The data is available through the container objects inside SimEvent.
// These are accessed via the <code>refData</code> and <code>refTable</code>
// methods.
//
  class SimEvent : public Base
  {
  public:

    //!ignore_begin:
    DECLARE_DEFAULT_CONTAINER( Data, data, Vect, getName );
    DECLARE_CONTAINER( Table, table, Table, getName );
    //!ignore_end:

    /* Constructors/Destructor */
    //-----------------------------------------------------------------------------
    //
    //: Default Constructor.
    //
    // This creates a SimEvent object with the following properties:
    // <ul type=disc>
    //     <li> name = "none" </li>
    //     <li> comment = "" </li>
    //     <li> inputs = "" </li>
    //     <li> GTime = (0,0) </li>
    //     <li> timeBefore = 0.0 </li>
    //     <li> timeAfter = 0.0 </li>
    //     <li> amplitude = 0.0 </li>
    //     <li> No vector or table data. </li>
    // </ul>
    //
    //!exc: None.
    //
    SimEvent();

    //-----------------------------------------------------------------------------
    //
    //: Copy Constructor.
    //
    // This performs a deep copy of all owned data.
    //
    //!param: const SimEvent& SimEvent - the object to copy from.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    SimEvent( const SimEvent& SimEvent );

    //: Constructor

    //: Constructor.
    //!exc: None.
    SimEvent( const std::string& name, const std::string& inputs = "", 
              const Time& GTime = Time( 0, 0 ),
              REAL_4 timeBefore = 0.0, REAL_4 timeAfter = 0.0,
              REAL_4 amplitude = 0.0 );
 
    /* Operator Overloads */
    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    // This performs a deep copy of all owned data.
    //
    //!param: const SimEvent& SimEvent - The object to assign from.
    //
    //!return: const SimEvent& -- This object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const SimEvent& operator=( const SimEvent& SimEvent );

    //-----------------------------------------------------------------------------
    //
    //: += Operator.
    //
    //!param: const SimEvent& SimEvent - The object to be added.
    //
    //!return: const SimEvent& -- This object.
    //
    const SimEvent& operator+=( const SimEvent& SimEvent );

    //-----------------------------------------------------------------------------
    //
    //: Equal comparison.
    //
    //!param: const SimEvent& sim - The object to compare with.
    //
    //!return: bool -- true if the objects are equal.
    //
    //!kw: const
    //
    //!exc: None.
    //   
    bool operator==( const SimEvent& SimEvent ) const;
    //!exc: None.   
    bool operator!=( const SimEvent& SimEvent ) const;

    /* Required Methods */
    //!exc: None.   
    virtual INT_2U getClassId() const;

    //-----------------------------------------------------------------------------
    //
    //: Structure Header
    //
    // Constructs the Structure Header for this class.
    //
    //!return: SH  -- SimEvent class ID.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH();

    /* Accessors */
    const std::string& getName() const;
    //!exc: None.   
    const std::string& getComment() const;
    //!exc: None.   
    const std::string& getInputs() const;
    //!exc: None.   
    const Time& getGTime() const;
    //!exc: None.   
    REAL_4 getTimeBefore() const;
    //!exc: None.   
    REAL_4 getTimeAfter() const;
    //!exc: None.   
    REAL_4 getAmplitude() const;

    /* Data Accessors */
    //!exc: None.   
    SimEvent::DataContainer& refData();
    //!exc: None.   
    const SimEvent::DataContainer& refData() const;
    //!exc: None.   
    SimEvent::TableContainer& refTable();
    //!exc: None.   
    const SimEvent::TableContainer& refTable() const;

    /* Mutators */
    //!exc: std::bad_alloc - Memory allocation failed.
    void appendComment( const std::string& comment );
    
private:
    friend class Dictionary;
    friend class FrameCPP::Version_6::FrSimEvent;


    //-----------------------------------------------------------------------------
    //
    //: Reads a SimEvent object from an Input object.
    //
    // This method creates a dynamically allocated object.
    //
    // <p>*NOTE* This method allocates memory and returns a pointer to it.  The
    // user is responsible for freeing this memory.
    //
    //!param: FrameReaderBase& in - Input data stream object.
    //
    //!return: SimEvent* -- Newly created.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //
    static SimEvent* read( FrameReaderBase& in );


    //-----------------------------------------------------------------------------
    //
    // Writes a SimEvent object to an Output object.
    //
    //!param: Output& out out - Output data stream object.
    //
    //!exc: write_failure - Write failure occured.
    //
    virtual void write( Output& out ) const;

private:

    //: Name of event.
    std::string mName;
    //: Descriptor of event.
    std::string mComment;
    //: Input channels and filter parameters to event process.
    std::string mInputs;
    //: GPS time.
    Time   mGTime;
    //: Signal duration before mGtime.
    REAL_4 mTimeBefore;
    //: Signal duration after mGtime.
    REAL_4 mTimeAfter;
    //: Continuous output amplitude returned by event.
    REAL_4 mAmplitude;
    //: Additional event results.
    DataContainer mData;
    //: Additional event information.
    TableContainer mTable;
};    
    


/******************/
/* Inline Methods */
/******************/


//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Not-equal comparison.
//
//!param: const SimEvent& sed - The object to compare with.
//
//!return: bool -- True if the objects are not equal.
//
//!kw: inline const
//
//!exc: None.   
//   
inline bool SimEvent::operator!=( const SimEvent& sed ) const
{
    return !( (*this) == sed );
}



//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Comment.
//
// Get the comment.
//
//!return: const std::string& -- comment.
//
//!exc: None.   
//      
inline const std::string& SimEvent::getComment() const
{
    return mComment;
}


//-----------------------------------------------------------------------------
//
//: Inputs.
//
// Get inputs.
//
//!return: const std::string& -- Inputs.
//
//!exc: None.   
//      
inline const std::string& SimEvent::getInputs() const
{
    return mInputs;
}


//-----------------------------------------------------------------------------
//
//: Trigger maximum time.
//
// Get trigger maximum time.
//
//!return: const Time& -- Trigger max time.
//
//!kw: inline const
//
//!exc: None.   
//      
inline const Time& SimEvent::getGTime() const
{
    return mGTime;
}


//-----------------------------------------------------------------------------
//
//: Signal duration before GTime.getSec().
//
// Get signal duration before trigger maximum.
//
//!return: REAL_4 -- Signal duration before maximum.
//
//!exc: None.   
//      
inline REAL_4 SimEvent::getTimeBefore() const
{
    return mTimeBefore;
}

//-----------------------------------------------------------------------------
//
//: Signal duration after GTime.getSec().
//
// Get signal duration after trigger maximum.
//
//!return: REAL_4 -- Signal duration after maximum.
//
//!exc: None.   
//      
inline REAL_4 SimEvent::getTimeAfter() const
{
    return mTimeAfter;
}


//-----------------------------------------------------------------------------
//
//: Continuous output amplitude returned by trigger.
//
// Get continuous output amplitude returned by trigger.
//
//!return: REAL_4 -- Continuous output amplitude returned.
//
//!exc: None.   
//      
inline REAL_4 SimEvent::getAmplitude() const
{
    return mAmplitude;
}


//-----------------------------------------------------------------------------
//
//: Data Container.
//
// Get the container which contains the data.
//
//!return: SearchContainer< Vect, &Vect::getName >& -- Data.
//
//!kw: inline
//
//!exc: None.   
//      
inline SimEvent::DataContainer& SimEvent::refData()
{
  return mData;
}


//-----------------------------------------------------------------------------
//
//: const Data Container.
//
// Get the container which contains the data.
//
//!return: const SearchContainer< Vect, &Vect::getName >& -- Data.
//
//!kw: inline const
//
//!exc: None.   
//      
inline const
SimEvent::DataContainer& SimEvent::refData() const
{
  return mData;
}

//-----------------------------------------------------------------------------
//
//: Table Container.
//
// Get table container.
//
//!return: Container< Table >& -- Table.
//
//!exc: None.   
//      
inline
SimEvent::TableContainer& SimEvent::refTable()
{
  return mTable;
}


//-----------------------------------------------------------------------------
//
//: const Table Container.
//
// Get table container.
//
//!return: const Container< Table >& -- Table.
//
//!exc: None.   
//      
inline const
SimEvent::TableContainer& SimEvent::refTable() const 
{
  return mTable;
}


//-----------------------------------------------------------------------------
//
//: Appends a comment.
//
// This method adds a comment followed by a newline to the
// end of the current list of comments.
//
//!param: const std::string& comment - The comment to append.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
//!kw: inline
//
inline void SimEvent::appendComment( const std::string& comment )
{
    if ( mComment.size() > 0 ) {
	 mComment += ( comment + "\n" );
    } else
	 mComment = comment;
}



} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // FrameCppSimEventHH
