#ifndef FrameCppStatDataHH
#define FrameCppStatDataHH


#include "framecpp/config.h"
#include "framecpp/Version3_4_5/base.hh"
#include "framecpp/Version3_4_5/sh.hh"
#include "framecpp/Version3_4_5/vect.hh"
#include "framecpp/Version3_4_5/table.hh"

#include "framecpp/SearchContainer.hh"


namespace FrameCPP {

namespace Version_6
{
  class FrStatData;
}

namespace Version_3_4_5 {

class Detector;
class FrameReaderBase;


//-----------------------------------------------------------------------------
//
//: This class represents static data.
//
// It is modeled after the FrStatData structure in the Frame Specification,
// LIGO-T970130-B-E.
//
class StatData : public Base
{
public:

  //!ignore_begin:
    DECLARE_DEFAULT_CONTAINER( Data, data, Vect, getName );
    DECLARE_CONTAINER( Table, table, Table, getName );
  //!ignore_end:

    /* Constructors */
    //-----------------------------------------------------------------------------
    //
    //: Default Constructor.
    //
    // This creates an StatData object with the following properties:
    // <ul>
    //     <li> name = "none" </li>
    //     <li> comment = "" </li>
    //     <li> representation = "" </li>
    //     <li> timeStart = 0 </li>
    //     <li> timeEnd = 0 </li>
    //     <li> version = 0 </li>
    //     <li> detector = none (null pointer) </li>
    //     <li> No other data. </li>
    // </ul>
    //
    //!exc: None.   
    //      
    StatData();

    //-----------------------------------------------------------------------------
    //
    //: Copy Constructor.
    //
    // This performs a deep copy of the object.  However,
    // since the StatData object does not own the Detector, a deep copy is NOT
    // performed on this object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //         
    StatData( const StatData& statData );

    //: Constructor
    //!exc: std::bad_alloc - Memory allocation failed.   
    StatData(const std::string& name, const std::string& comment,
	     const std::string& representation,
	     INT_4U timeStart,INT_4U timeEnd,
	     INT_4U version );
    
    /* Destructor */
    //-----------------------------------------------------------------------------
    //
    //: Destructor.
    //
    // This method currently does nothing.
    //
    //
    //!exc: None.   
    //         
    ~StatData();

    /* Operator Overloads */
    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    //!param: const StatData& statData - Object to be assigned.
    //
    //!return: const StatData& -- This object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const StatData& operator=( const StatData& statData );

    //-----------------------------------------------------------------------------
    //
    //: += Operator.
    //
    //!param: const StatData& statData - Object to be added.
    //
    //!return: const StatData& -- This object.
    //
    const StatData& operator+=( const StatData& statData );


    //-----------------------------------------------------------------------------
    //
    //: Equal comparison.
    //
    // Determines whether two StatData objects are equal. For this comparison,
    // all of the data is compared.
    //
    //!param: const StatData& statData - The object to compare with.
    //
    //!return: bool -- true if the objects are equal.
    //
    //!kw: const
    //
    //!exc: None.   
    //         
    bool operator==( const StatData& statData ) const;
    //!exc: None.      
    bool operator!=( const StatData& statData ) const;
    
    /* Required Methods */
    //!exc: None.      
    virtual INT_2U getClassId() const;


    //-----------------------------------------------------------------------------
    //
    //: Constructs the Structure Header for this class.  
    //
    //!return: SH -- Static Data structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH();


    //-----------------------------------------------------------------------------
    //
    //: Get frame data format version 3 Structure Header
    //
    // Constructs the Structure Header for this class.  
    //
    //!return: SH -- Static Data structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed
    //
    static SH getSH3();

    /* Accessors */
    /* gcc 2.95.1 can't handle a typedef with this if the throw spec is
       present (e.g.
       typedef SearchContainer< StatData, &StatData::getName > dc; ).
       
       const std::string& getName() const throw(); */
    const std::string& getName() const;
    //!exc: None.      
    const std::string& getComment() const;
    //!exc: None.      
    const std::string& getRepresentation() const;
    //!exc: None.      
    INT_4U getTimeStart() const;
    //!exc: None.      
    INT_4U getTimeEnd() const;
    //!exc: None.      
    INT_4U getVersion() const;
    //!exc: None.      
    Detector* getDetector();
    //!exc: None.      
    const Detector* getDetector() const;
    //!exc: None.      
    DataContainer& refData();
    //!exc: None.      
    const DataContainer& refData() const;
    //!exc: None.      
    TableContainer& refTable();
    //!exc: None.      
    const TableContainer& refTable() const;

    /* Mutators */
    //!exc: None.      
    void setDetector( Detector* detector );
    
private:
  friend class Dictionary;
  friend class FrameCPP::Version_6::FrStatData;


    //-----------------------------------------------------------------------------
    //
    //: Reads a StatData object from an FrameReaderBase object.
    //
    // This method creates a dynamically allocated object.
    //
    // <p>*NOTE* This method allocates memory and returns a pointer to it.  The
    // user is responsible for freeing this memory.
    //
    //!param: FrameReaderBase& in - Input data stream object.
    //
    //!return: StatData* -- Newly created.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //
    static StatData* read( FrameReaderBase& in );

    //: Read framecpp version 3 static data object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //   
    static StatData* read3( FrameReaderBase& in );

    //-----------------------------------------------------------------------------
    //
    //: Writes an StatData object to an Output object.
    //
    //!param: Output& out out - Output data stream object.
    //
    //!exc: write_failure - Write failure occured.
    //
    virtual void write( Output& out ) const;

private:

    //: Static data name.
    std::string mName;
    //: Comment.	
    std::string mComment;
    //: Type of static data beeing represented.
    std::string mRepresentation;
    //: Start time of static data validity.
    INT_4U mTimeStart;
    //: End time of static data validity.
    INT_4U mTimeEnd;
    //: Version numner for this static structure.
    INT_4U mVersion;
    //: Detector with which this static data is associated.
    Detector* mDetector;
    //: Data.
    DataContainer mData;
    //: Table.
    TableContainer mTable;
};



/******************/
/* Inline Methods */
/******************/



//-----------------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------------



//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Not-equal comparison.
//
//!param: const StatData& statData - The object to compare with.
//
//!return: bool -- True if the objects are not equal.
//
//!kw: inline const
//
//!exc: None.
//      
inline bool StatData::operator!=( const StatData& statData ) const
{
    return !( *this == statData );
}



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Gets the comment.
//
//!return: const std::string&
//
//!exc: None.
//      
inline const std::string& StatData::getComment() const
{
    return mComment;
}


//-----------------------------------------------------------------------------
//
//: Gets the representation.
//
//!return: const std::string&
//
//!exc: None.
//      
inline const std::string& StatData::getRepresentation() const
{
    return mRepresentation;
}


//-----------------------------------------------------------------------------
//
//: Gets the start time of static data validity.
//
// This is GPS time in integer seconds since GPS standard epoch.
//
//!return: INT_4U
//
//!exc: None.
//      
inline INT_4U StatData::getTimeStart() const
{
    return mTimeStart;
}


//-----------------------------------------------------------------------------
//
//: Gets the end time of static data validity.
//
// This is the GPS time in integer seconds since the GPS standard epoch.  If
// this is unknown it is set to zero.
//
//!return: INT_4U
//
//!exc: None.
//      
inline INT_4U StatData::getTimeEnd() const
{
    return mTimeEnd;
}


//-----------------------------------------------------------------------------
//
//: Gets the version number for this static data.
//
//!return: INT_4U
//
//!exc: None.
//      
inline INT_4U StatData::getVersion() const
{
    return mVersion;
}


//-----------------------------------------------------------------------------
//
//: Gets the detector to shich this static data applies.
//
//!return: Detector*
//
//!exc: None.
//      
inline Detector* StatData::getDetector()
{
    return mDetector;
}


//-----------------------------------------------------------------------------
//
//: Gets the detector to shich this static data applies.
//
//!return: const Detector*
//
//!exc: None.
//      
inline const Detector* StatData::getDetector() const
{
    return mDetector;
}


//-----------------------------------------------------------------------------
//
//: Sets the Detector object to which this static data applies.
//
//!param: Detector* detector
//
//!exc: None.
//      
inline void StatData::setDetector( Detector* detector )
{
    mDetector = detector;
}


//-----------------------------------------------------------------------------
//
//: Get the container which contains the data.
//
//!return: DataContainer&
//
//!kw: inline
//
//!exc: None.
//      
inline StatData::DataContainer&
StatData::refData()
{
    return mData;
}


//-----------------------------------------------------------------------------
//
//: Get the container which contains the data.
//
//!return: const DataContainer&
//
//!kw: inline const
//
//!exc: None.
//      
inline const
StatData::DataContainer&
StatData::refData() const
{
    return mData;
}


//-----------------------------------------------------------------------------
//
//: Get the container holding the user-define table data.
//
//!return: TableContainer&
//
//!exc: None.
//      
inline
StatData::TableContainer&
StatData::refTable()
{
  return mTable;
}


//-----------------------------------------------------------------------------
//
//: Get the container holding the user-define table data.
//
//!return: const TableContainer&
//
//!exc: None.
//      
inline const
StatData::TableContainer&
StatData::refTable() const
{
  return mTable;
}

} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // StatDataHH
