#include "framecpp/config.h"

#include "table.hh"
#include "framereaderbase.hh"
#include <general/regexmatch.hh>
#include "reference.hh"
#include "output.hh"


using namespace FrameCPP::Version_3_4_5;


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates Table object with the following properties:
// <ul type=disc>
//     <li> name = "none" </li>
//     <li> comment = "" </li>
//     <li> nRows = 1 </li>
//     <li> no data </li>
// </ul>
//
//!exc: None.      
Table::Table()
  : Base(), 
    mName( "none" ), mComment( "" ),
    mNRows( 1 ), mColumns()
{
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
// This creates Table object with the specified properties.
// Additionally, the comment is set to "" (no comment).
//
//!param: const std::string& name - Chanel name.
//!param: INT_4U nRows - Number rows allowed in a column.
//
//!exc: None.      
Table::Table( const std::string& name, INT_4U nrows )
  : Base(), mName( name ),
    mComment( "" ), mNRows( nrows ),
    mColumns()
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
// A deep copy is performed for owned objects.
//
//!param: const Table& table - The object to be copied.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
Table::Table( const Table& table ) 
  :Base( table ), 
   mName( table.mName ),
   mComment( table.mComment ),
   mNRows( table.mNRows ),
   mColumns( table.mColumns )
{
}


//-----------------------------------------------------------------------------
//
//: Table name.
//
// Gets the table name.
//
//!return: const std::string& -- Table name.
//
const std::string& Table::getName() const
{
  return mName;
}


//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
// A deep copy is performed for owned objects.
//
//!param: const Table& table - The object to be assigned.
//
//!return: const Table& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const Table& Table::operator=( const Table& table )
{
  if ( this != &table )
  {
    mName = table.mName;
    mComment = table.mComment;
    mNRows = table.mNRows;
    mColumns = table.mColumns;
  } 
  return *this;
}

//-----------------------------------------------------------------------------
//
//: += Assignment Operator.
//
//!param: const Table& table -
//
//!return: const Table&
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: frame_mismatch - Table mismatch
//
const Table& Table::operator+=( const Table& table )
{
  // Check the attributes that should match
  //
  // TO BE DONE
  //
  return *this;
}


//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Structure Header
//
// Constructs the Structure Header for this class. 
//
//!return: SH - Table structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH Table::getSH()
{
  SH sh( "FrTable", ID_TABLE, "Table" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "comment", "STRING", "" ) );
  sh.appendSE( SE( "nColumn", "INT_2U", "" ) );
  sh.appendSE( SE( "nRow", "INT_4U", "" ) );
  sh.appendSE( SE( "columnName", "STRING[nColumn]", "" ) );
  sh.appendSE( SE( "column", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrTable *)", "" ) );
  
  return sh;
}


//: Frame format version 4 Structure Header
//
//!exc: std::bad_alloc - Memory allocation failed.
//   
FrameCPP::Version_3_4_5::SH Table::getSH4()
{
  SH sh( "FrTable", ID_TABLE, "Table" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "comment", "STRING", "" ) );
  sh.appendSE( SE( "nColumn", "INT_2U", "" ) );
  sh.appendSE( SE( "nRow", "INT_4U", "" ) );
  sh.appendSE( SE( "columnName", "STRING *", "" ) );
  sh.appendSE( SE( "column", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrTable *)", "" ) );
  
  return sh;
}


//-----------------------------------------------------------------------------
//
//: Reads Table object from an FrameReaderBase object.
//
//!param: FrameReaderBase& in - Input data stream object.
//
//!return: Table* - Newly allocated Table.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Red failure occured.
//
Table* Table::read4( FrameReaderBase& in )
{
  Table& table = *new Table();

  INT_2U nColumn;

  in >> table.mName >> table.mComment >> nColumn >> table.mNRows;

  for ( unsigned int i = 0; i < nColumn; i++ )
    {
      std::string junk;
      in >> junk;
    }

  in >> table.mColumns;

  in.readNextNodeReference();
  return &table;
}


//-----------------------------------------------------------------------------
//
//: Write Table object
//
// Writes Table object to an Output object.
//
//!param: Output& out out - Output data object.
//
//!exc: write_failure - Write failure occured.
//
void Table::write( Output& out ) const
{
  INT_4U len =
      8 * 1                           // One header
    + 3 * ( 2 + mColumns.size() )  // Strings (2 and an array of length mColumns)
    + 6                               // 
    + 4 * 2                           // Two pointers
    + mName.size() + mComment.size(); // String lengths

  // Add the length of all std::strings from the array
  for (unsigned int i = 0; i < mColumns.size(); i++)
    {
      len += mColumns[i]->getName().size();
    }  

  out.startRecord( ID_TABLE, len, this );

  out << mName << mComment << (INT_2U)mColumns.size() << mNRows;
  for (unsigned int i = 0; i < mColumns.size(); i++)
    {
      out << mColumns[i]->getName();
    }
  out << mColumns;
  
  out.endRecord( this );
}



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether Table objects are equal. For this comparison,
// all of the data is compared.
//
//!param: const Table& table - The object to compare with.
//
//!return: bool - true if the objects are equal.
//
//!kw: const
//
//!exc: None.
//   
bool Table::operator==( const Table& table ) const
{
    return ( mName       == table.mName       &&
             mComment    == table.mComment    &&
             mNRows      == table.mNRows      &&
             mColumns    == table.mColumns );
}

   
//-----------------------------------------------------------------------------
//
//: Class ID.
//
// Gets the class ID.
//
// This is the internal ID used to identify this class type.  It is specified
// in "types.hh".
//
//!return: INT_2U -- Table class ID.
//
//!exc: None.      
INT_2U Table::getClassId() const
{
  return ID_TABLE;
}


   
