#ifndef FrameCppTableHH
#define FrameCppTableHH

#include "framecpp/config.h"
#include "framecpp/Version3_4_5/base.hh"
#include "framecpp/Version3_4_5/sh.hh"
#include "framecpp/Version3_4_5/vect.hh"
#include "framecpp/Version3_4_5/time.hh"
#include "framecpp/SearchContainer.hh"


/*
  :TODO: all vectors are forced to be of mNRows length
  Maybe extend search container to allow inclusion of vectors of certain length only?
*/

namespace FrameCPP {

  namespace Version_6
  {
    class FrTable;
  }

namespace Version_3_4_5 {

class FrameReaderBase;

//-----------------------------------------------------------------------------
//
//: Table
//
// <P>Table class encapsulates the following attributes:
//
// <table border=2 cellpadding=4>
//   <tr><td> <B>Attribute</B></td><td><B>Description</B>                        </td></tr>
//   <tr><td> name      </td><td>Table name                                      </td></tr>
//   <tr><td> comment   </td><td>Comment                                         </td></tr>
//   <tr><td> mNRows    </td><td>Number of rows allowed in a column              </td></tr>
// </table>
//
// <p>Table object is also a container for two types of data:
//
// <table border=2 cellpadding=4>
//   <tr><td> <B>Data</B></td><td><B>Description</B>                             </td></tr>
//   <tr><td> columns</td><td>A container of Frame Vect table columns            </td></tr>
// </table>
//
class Table : public Base
{
public:

  //!ignore_begin:
  /* Container Macros */
  DECLARE_DEFAULT_CONTAINER( Columns, columns, FrameCPP::Version_3_4_5::Vect, getName );
  //!ignore_end:

  /* Constructors/Destructor */
  //-----------------------------------------------------------------------------
  //
  //: Default Constructor.
  //
  // This creates Table object with the following properties:
  // <ul type=disc>
  //     <li> name = "none" </li>
  //     <li> comment = "" </li>
  //     <li> nRows = 1 </li>
  //     <li> no data </li>
  // </ul>
  //
  //!exc: None.   
  Table();


  //-----------------------------------------------------------------------------
  //
  //: Copy Constructor.
  //
  // A deep copy is performed for owned objects.
  //
  //!param: const Table& table - The object to be copied.
  //
  //!exc: std::bad_alloc - Memory allocation failed.
  //
  Table( const Table& table );

  //-----------------------------------------------------------------------------
  //
  //: Constructor.
  //
  // This creates Table object with the specified properties.
  // Additionally, the comment is set to "" (no comment).
  //
  //!param: const std::string& name - Chanel name.
  //!param: INT_4U nRows - Number rows allowed in a column.
  //
  //!exc: None.      
  Table( const std::string& name, INT_4U nrows );

  /* Operator Overloads */
  //-----------------------------------------------------------------------------
  //
  //: Assignment Operator.
  //
  // A deep copy is performed for owned objects.
  //
  //!param: const Table& table - The object to be assigned.
  //
  //!return: const Table& -- This object.
  //
  //!exc: std::bad_alloc - Memory allocation failed.
  //
  const Table& operator=( const Table& table );

  //-----------------------------------------------------------------------------
  //
  //: Equal comparison.
  //
  // Determines whether Table objects are equal. For this comparison,
  // all of the data is compared.
  //
  //!param: const Table& table - The object to compare with.
  //
  //!return: bool -- true if the objects are equal.
  //
  //!kw: const
  //
  //!exc: None.      
  bool operator==( const Table& table ) const;
  //!exc: None.      
  bool operator!=( const Table& table ) const;

  //
  //: += Assignment Operator.
  //
  //!param: const Table& table -
  //
  //!return: const Table&
  //
  //!exc: std::bad_alloc - Memory allocation failed.
  //!exc: frame_mismatch - Table mismatch
  //
  const Table& operator+=( const Table& table );
    
  /* Required Methods */
  //!exc: None.      
  virtual INT_2U getClassId() const;


  //-----------------------------------------------------------------------------
  //
  //: Structure Header
  //
  // Constructs the Structure Header for this class. 
  //
  //!return: SH -- Table structure header.
  //
  //!exc: std::bad_alloc - Memory allocation failed.
  //
  static SH getSH();

  //: Frame format version 4 Structure Header
  //
  //!exc: std::bad_alloc - Memory allocation failed.
  //   
  static SH getSH4();

  /* General Accessors */
  // Return by value: to be consistent with Version6
  const std::string& getName() const;
  //!exc: None.         
  const std::string& getComment() const;
  //!exc: None.         
  INT_2U getNColumns() const;
  //!exc: None.         
  INT_4U getNRows() const;

  /* Data Accessors */
  //!exc: None.         
  Table::ColumnsContainer& refColumns();
  //!exc: None.         
  const Table::ColumnsContainer& refColumns() const;

  /* Mutators */
  //!exc: std::bad_alloc - Memory allocation failed.
  void appendComment( const std::string& comment );

private:
  friend class Dictionary;
  friend class FrameCPP::Version_6::FrTable;

  //-----------------------------------------------------------------------------
  //
  //: Reads Table object from an FrameReaderBase object.
  //
  //!param: FrameReaderBase& in - Input data stream object.
  //
  //!return: Table* -- Newly allocated Table.
  //
  //!exc: std::bad_alloc - Memory allocation failed.
  //!exc: read_failure - Red failure occured.
  //
  static Table* read( FrameReaderBase& in ){ return read4( in ); }

  //
  //!exc: std::bad_alloc - Memory allocation failed.
  //!exc: read_failure - Red failure occured.
  //   
  static Table* read4( FrameReaderBase& in );

  //-----------------------------------------------------------------------------
  //
  //: Write Table object
  //
  // Writes Table object to an Output object.
  //
  //!param: Output& out out - Output data object.
  //
  //!exc: write_failure - Write failure occured.
  //
  virtual void write( Output& out ) const;
private:
  
  //: Name of this table.
  std::string mName;
  //: Comment.
  std::string mComment;
  //: Number of rows.
  // All vectors forced to be of this length :TODO:
  INT_4U mNRows;
  //: Table columns
  ColumnsContainer mColumns;
};    



/******************/
/* Inline Methods */
/******************/


//-----------------------------------------------------------------------------
//
//: Not-equal comparison.
//
// Determines whether two table objects are not equal.  All contained data is
// compared.
//
//!param: const Table& table - The object to compare with.
//
//!return: bool -- True if the objects are not equal.
//
//!kw: inline const
//
//!exc: None.   
inline bool Table::operator!=( const Table& table ) const
{
  return !( (*this) == table );
}


//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Comment
//
// Gets the comment.
//
//!return: const std::string& -- Table comment.
//
//!exc: None.      
inline const std::string& Table::getComment() const
{
  return mComment;
}


//-----------------------------------------------------------------------------
//
//: Number of rows allowed per vector (column).
//
//!return: INT_4U
//
//!kw: inline const - The number of rows.
//
//!exc: None.      
inline INT_4U Table::getNRows() const
{
  return mNRows;
}


//-----------------------------------------------------------------------------
//
//: Number of columns table has.
//
//!return: INT_2U -- The number of columns.
//
//!kw: inline const
//
//!exc: None.      
inline INT_2U Table::getNColumns() const
{
  return mColumns.size();
}


//-----------------------------------------------------------------------------
//
//: Columns Container
//
// Get table columns vector container
//
//!return: Container< Vect >& -- Columns vector.
//
//!exc: None.      
inline Table::ColumnsContainer&
Table::refColumns()
{
  return mColumns;
}


//-----------------------------------------------------------------------------
//
//: Columns Container
//
// Get table columns vector container
//
//!return: Container< Vect >& -- Columns vector.
//
//!exc: None.      
inline const Table::ColumnsContainer&
Table::refColumns() const
{
  return mColumns;
}


//-----------------------------------------------------------------------------
// Mutators
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Appends a comment.
//
// This method adds a comment followed by a newline to the
// end of the current list of comments.
//
//!param: const std::string& comment - The comment to append.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
//!kw: inline
//
inline void Table::appendComment( const std::string& comment )
{
  if ( mComment.size() > 0 ) {
	 mComment += ( comment + "\n" );
  } else
	 mComment = comment;
}
 
} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // FrameCppTableHH
