#include "framecpp/config.h"

#include "time.hh"
#include <math.h>
#include "input.hh"
#include "output.hh"

using namespace FrameCPP::Version_3_4_5;

/* :TODO: this Time class is used in many place where seconds are INT_4U.
   Perhaps I should rename this class OffsetTime and write another called Time,
   where seconds are unsigned */

//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
//!param: const Time& time - Time to be assigned to.
//
//!return: const Time& - This object.
//
//!exc: bad_alloc - Memory allocation failed.
//
//!exc: None.
const Time& Time::operator=( const Time& time )
{
    if ( this != &time )
    {
        mSec = time.getSec();
        mNSec = time.getNSec();
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: Addition & assignment.
//
//!param: const Time& time - Time to be added.
//
//!return: const Time& time - This object.
//
//!exc: None.   
const Time& Time::operator+=( const Time& time )
{
    mSec += time.getSec();
    mNSec += time.getNSec();

    if ( mNSec >= 1000000000 )
    {
        mSec += INT_4S( mNSec / 1000000000 );
        mNSec = mNSec % 1000000000;
    }
    
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Subtraction and assignment.
//
//
//!param: const Time& time - Time to be subtracted.
//
//!return: const Time& time - This object.
//
//!exc: None.   
const Time& Time::operator-=( const Time& time )
{
    mSec -= time.getSec();

    if ( mNSec >= time.getNSec() )
    {
        mNSec -= time.getNSec();
    }
    else
    {
        --mSec;
        mNSec += 1000000000 - time.getNSec();
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: Multiplication and assignment.
//
//!param: const double& d - Multiplier.
//
//!return: const Time& time - This object.
//
//!todo: What happens if d is negative?
//
//!exc: None.   
const Time& Time::operator*=( const double& d )
{
    double ns = mNSec * d;
    mSec = INT_4S( mSec * d + ns / 1000000000 );
    mNSec = INT_4U( fmod( ns, 1000000000 ) );

    return *this;
}


//-----------------------------------------------------------------------------
//
//: Division and assignment.
//
//!param: const double& d - Divider.
//
//!return: const Time& time - This object.
//
//!todo: What happens if d is negative?
//
//!exc: None.   
const Time& Time::operator/=( const double& d ) 
{
    return ( (*this) *= ( 1.0 / d ) );
}


//-----------------------------------------------------------------------------
//
//: Division.
//
//!param: const Time& time - Divider.
//
//!return: double - This object time devided by `time'.
//
//!exc: None.   
double Time::operator/( const Time& time )
{
    return ( ( 1000000000 * mSec + mNSec ) 
           / ( 1000000000 * time.getSec() + time.getNSec() ) );
}


//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two Time objects are equal. 
//
//!param: const Time& time - The object to compare with.
//
//!return: bool - True if the objects are equal.
//
//!kw: const
//
//!exc: None.   
bool Time::operator==( const Time& time ) const
{
    return ( mSec == time.getSec() ) && ( mNSec == time.getNSec() );
}


//-----------------------------------------------------------------------------
//
//: Less than or equal to comparison.
//
//!param: const Time& time - The object to compare with.
//
//!return: bool - True if this object's time is less than ot equal to `time'.
//
//!kw: const
//
//!exc: None.   
bool Time::operator<=( const Time& time ) const
{
    return !( *this > time );
}


//-----------------------------------------------------------------------------
//
//: Greater than or equal to comparison.
//
//!param: const Time& time - The object to compare with.
//
//!return: bool - True if this object's time is greater or equal to `time'.
//
//!kw: const
//
//!exc: None.   
bool Time::operator>=( const Time& time ) const
{
    return !( *this < time );
}


//-----------------------------------------------------------------------------
//
//: Not equal comparison.
//
//!param: const Time& time - The object to compare with.
//
//!return: bool - True if this object's time is not equal to `time'.
//
//!kw: const
//
//!exc: None.
bool Time::operator!=( const Time& time ) const 
{
    return !( *this == time );
}


//-----------------------------------------------------------------------------
//
//: Less than comparison.
//
//!param: const Time& time - The object to compare with.
//
//!return: bool - True if this object's time is less than `time'.
//
//!kw: const
//
//!exc: None.   
bool Time::operator<( const Time& time ) const
{
    if ( mSec != time.getSec() )
    {
        return ( mSec < time.getSec() );
    }
    else
    {
        return ( mNSec < time.getNSec() );
    }
}

         
//-----------------------------------------------------------------------------
//
//: Greater than comparison.
//
//!param: const Time& time - The object to compare with.
//
//!return: bool - True if this object's time is greater than `time'.
//
//!kw: const
//
//!exc: None.   
bool Time::operator>( const Time& time ) const
{
    if ( mSec != time.getSec() )
    {
        return ( mSec > time.getSec() );
    }
    else
    {
        return ( mNSec > time.getNSec() );
    }
}
    

//-----------------------------------------------------------------------------
//
//: Addition.
//
//!param: const Time& t1
//!param: const Time& t2
//
//!return: Time - 
//
//!exc: None.   
Time FrameCPP::Version_3_4_5::operator+( const Time& t1, const Time& t2 )
{
    Time t3( t1 );
    return ( t3 += t2 );
}


//-----------------------------------------------------------------------------
//
//: Subtraction.
//
//!param: const Time& t1
//!param: const Time& t2
//
//!return: Time - 
//
//!exc: None.   
Time FrameCPP::Version_3_4_5::operator-( const Time& t1, const Time& t2 )
{
    Time t3( t1 );
    return ( t3 -= t2 );
}


//-----------------------------------------------------------------------------
//
//: Multiplication.
//
//!param: const Time& t
//!param: const double& d
//
//!return: Time - 
//
//!exc: None.   
Time FrameCPP::Version_3_4_5::operator*( const Time& t, const double& d ) 
{
    Time t3( t );
    return ( t3 *= d );
}
    

//-----------------------------------------------------------------------------
//
//: Division.
//
//!param: const Time& t
//!param: const double& d
//
//!return: Time - 
//
//!exc: None.   
Time FrameCPP::Version_3_4_5::operator/( const Time& t, const double& d ) 
{
    Time t3( t );
    return ( t3 /= d );
}


//-----------------------------------------------------------------------------
//
//: Multiplication.
//
//!param: const double& d
//!param: const Time& t
//
//!return: Time - 
//
//!exc: None.   
Time FrameCPP::Version_3_4_5::operator*( const double& d, const Time& t ) 
{
    Time t3( t );
    return ( t3 *= d );
}


//-----------------------------------------------------------------------------
//
//: Input extraction operator.
//
//!param: Input& in - Input data stream object.
//!param: Time& time - Time object to set to.
//
//!return: Input& - Inputs data stream object.
//
//!exc: read_failure - Read failed.
//
Input& FrameCPP::Version_3_4_5::operator>>( Input& in, Time& time )
{
    in >> time.mSec >> time.mNSec;
    return in;
}


//-----------------------------------------------------------------------------
//
//: Output insertion operator.
//
//!param: Output& out
//!param: const Time& time
//
//!return: Output&
//
//!exc: write_failure - Write failed.
//
Output& FrameCPP::Version_3_4_5::operator<<( Output& out, const Time& time )
{
    out << time.getSec() << time.getNSec();
    return out;
}
