#ifndef FrameCppTimeHH
#define FrameCppTimeHH


#include "framecpp/config.h"
#include "types.hh"
#include "errors.hh"

namespace FrameCPP {

namespace Version_3_4_5 {

class Input;
class Output;




//-----------------------------------------------------------------------------
//
//: A class storing seconds & nanoseconds.
//
// This is a very simple class storing time as integer seconds and nanoseconds.
//
class Time
{
public:

    /* Constructors */
    //!exc: None.   
    explicit Time( INT_4S sec = 0, INT_4U nsec = 0 );
    //!exc: None.      
    Time( const Time& time );

    /* Operator Overloads */
    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    //!param: const Time& time - Time to be assigned to.
    //
    //!return: const Time& - This object.
    //
    //!exc: None.
    //
    const Time& operator=( const Time& time );

    //-----------------------------------------------------------------------------
    //
    //: Addition & assignment.
    //
    //!param: const Time& time - Time to be added.
    //
    //!return: const Time& time - This object.
    //
    //!exc: None.      
    const Time& operator+=( const Time& time );


    //-----------------------------------------------------------------------------
    //
    //: Subtraction and assignment.
    //
    //
    //!param: const Time& time - Time to be subtracted.
    //
    //!return: const Time& time - This object.
    //
    //!exc: None.      
    const Time& operator-=( const Time& time );

    //-----------------------------------------------------------------------------
    //
    //: Multiplication and assignment.
    //
    //!param: const double& d - Multiplier.
    //
    //!return: const Time& time - This object.
    //
    //!todo: What happens if d is negative?
    //
    //!exc: None.      
    const Time& operator*=( const double& d );


    //-----------------------------------------------------------------------------
    //
    //: Division and assignment.
    //
    //!param: const double& d - Divider.
    //
    //!return: const Time& time - This object.
    //
    //!todo: What happens if d is negative?
    //
    //!exc: None.      
    const Time& operator/=( const double& d );


    //-----------------------------------------------------------------------------
    //
    //: Division.
    //
    //!param: const Time& time - Divider.
    //
    //!return: double - This object time devided by `time'.
    //
    //!exc: None.      
    double operator/( const Time& time );


    //-----------------------------------------------------------------------------
    //
    //: Equal comparison.
    //
    // Determines whether two Time objects are equal. 
    //
    //!param: const Time& time - The object to compare with.
    //
    //!return: bool - True if the objects are equal.
    //
    //!kw: const
    //
    //!exc: None.      
    bool operator==( const Time& time ) const;


    //-----------------------------------------------------------------------------
    //
    //: Less than or equal to comparison.
    //
    //!param: const Time& time - The object to compare with.
    //
    //!return: bool - True if this object's time is less than ot equal to `time'.
    //
    //!kw: const
    //
    //!exc: None.      
    bool operator<=( const Time& time ) const;


    //-----------------------------------------------------------------------------
    //
    //: Greater than or equal to comparison.
    //
    //!param: const Time& time - The object to compare with.
    //
    //!return: bool - True if this object's time is greater or equal to `time'.
    //
    //!kw: const
    //
    //!exc: None.      
    bool operator>=( const Time& time ) const;

    //-----------------------------------------------------------------------------
    //
    //: Not equal comparison.
    //
    //!param: const Time& time - The object to compare with.
    //
    //!return: bool - True if this object's time is not equal to `time'.
    //
    //!kw: const
    //
    //!exc: None.      
    bool operator!=( const Time& time ) const;

    //-----------------------------------------------------------------------------
    //
    //: Less than comparison.
    //
    //!param: const Time& time - The object to compare with.
    //
    //!return: bool - True if this object's time is less than `time'.
    //
    //!kw: const
    //
    //!exc: None.      
    bool operator<( const Time& time ) const;


         
    //-----------------------------------------------------------------------------
    //
    //: Greater than comparison.
    //
    //!param: const Time& time - The object to compare with.
    //
    //!return: bool - True if this object's time is greater than `time'.
    //
    //!kw: const
    //
    //!exc: None.      
    bool operator>( const Time& time ) const;
    
    /* Accessors */
    //!exc: None.      
    INT_4S getSec() const;
    //!exc: None.      
    INT_4U getNSec() const;

    /* Mutators */
    //: Set seconds
    //!exc: None.      
    INT_4S setSec(INT_4S s) { return mSec = s; }
    //: Set nanosecond residual
    //!exc: None.      
    INT_4U setNSec(INT_4U nsec) { return mNSec = nsec; }
  

    /* Friends */

    //-----------------------------------------------------------------------------
    //
    //: Input extraction operator.
    //
    //!param: Input& in - Input data stream object.
    //!param: Time& time - Time object to set to.
    //
    //!return: Input& - Inputs data stream object.
    //
    //!exc: read_failure - Read failed.
    //
   friend Input& operator>>( Input&, Time& );

private:
    //: Seconds.
    INT_4S mSec;
    //: Nanoseconds.
    INT_4U mNSec;
};



//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: INT_4S sec - number of seconds. Default: 0
//!param: INT_4U nsec - number of nanoseconds. Default: 0
//
//!exc: None.      
inline Time::Time( INT_4S sec, INT_4U nsec )
        : mSec( sec ), mNSec( nsec )
{
    if ( mNSec >= 1000000000 )
    {
        mSec += INT_4S( mNSec / 1000000000 );
        mNSec = mNSec % 1000000000;
    }
}


//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Get the number of seconds.
//
//!return: INT_4S - Number of seconds.
//
//!exc: None.         
inline INT_4S Time::getSec() const 
{
    return mSec;
}


//-----------------------------------------------------------------------------
//
//: Get the number of nanoseconds.
//
//!return: INT_4U - The number of nanoseconds.
//
//!exc: None.         
inline INT_4U Time::getNSec() const
{
    return mNSec;
}

   
//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
//!param: const Time& time - Object to be copied.
//
//!exc: None.         
inline Time::Time( const Time& time )
        : mSec( time.getSec() ), mNSec( time.getNSec() )
{
}
    


//!exc: None.         
Time operator+( const Time& t1, const Time& t2 );
//!exc: None.         
Time operator-( const Time& t1, const Time& t2 );
//!exc: None.         
Time operator*( const Time& t, const double& d );
//!exc: None.         
Time operator/( const Time& t, const double& d );
//!exc: None.         
Time operator*( const double& d, const Time& t );

//!exc: read_failure - Read failed.
Input& operator>>( Input& in, Time&time );
//!exc: write_failure - Write failed.   
Output& operator<<( Output& out, const Time& time );


} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // FrameCppTimeHH
