#ifndef FrameCppVectHH
#define FrameCppVectHH

#include <zlib.h>

#include <vector>

#include "framecpp/config.h"
#include "framecpp/Version3_4_5/base.hh"
#include "framecpp/Version3_4_5/dimension.hh"

#include "framecpp/util.hh"


namespace FrameCPP {

  namespace Version_6
  {
    class FrVect;
  }

namespace Version_3_4_5 {


class SH;
class FrameReaderBase;
class Output;


//-----------------------------------------------------------------------------
//
//: This class represents vector data.
//
// It is modeled after the FrVect structure in the Frame Specification,
// LIGO-T970130-B-E.
//
//!todo: Break out compression/decompression routines into external modules or
//!todo: files/classes so that adding new compression methods is easier.  That
//!todo: is, separate the comp/decomp methods from the Vect class.
//
class Vect : public Base
{
public:
    
    //: Compression Types.
    enum Compression
    {
        NONE        = 0,
        GZIP        = 1,
        DIFF        = 2,
        GZIP_DIFF   = 3,
	DIFF_ZERO_SUPPRESS_SHORT = 5,
	DIFF_ZERO_SUPPRESS_SHORT_DIFF_GZIP_OTHER = 6,
	ZERO_SUPPRESS_INT_FLOAT = 8,
	BZIP2       = 9
    };

    //: Data Types
    enum Type
    {
        FR_VECT_C   =  0,
        FR_VECT_2S  =  1,
        FR_VECT_8R  =  2,
        FR_VECT_4R  =  3,
        FR_VECT_4S  =  4,
        FR_VECT_8S  =  5,
	FR_VECT_8C  =  6,
	FR_VECT_16C =  7,
	FR_VECT_STR =  8,
        FR_VECT_2U  =  9,
        FR_VECT_4U  = 10,
        FR_VECT_8U  = 11,
        FR_VECT_1U  = 12
    };

    //: Data Byte-Order
    enum ByteOrder
    {
        BIG,
        LITTLE
    };

    //: Error Status
    enum Error
    {
        OK = 0,
        MEM_ERROR,
        DATA_ERROR
    };
    
    /* Constructors */
    //-----------------------------------------------------------------------------
    //
    //: Default Constructor.
    //
    // This creates a Vect object with the following properties:
    // <ul>
    //     <li> name = "none" </li>
    //     <li> compression = 0 (none) </li>
    //     <li> type = FR_VECT_2S (INT_2S) </li>
    //     <li> nData = 0 </li>
    //     <li> nBytes = 0 </li>
    //     <li> data = 0 (null) </li>
    //     <li> nDim = 0 </li>
    //     <li> No dimension information) </li>
    //     <li> unitY = "" </li>
    // </ul>
    //
    //!exc: None.   
    //      
    Vect();

    //-----------------------------------------------------------------------------
    //
    //: Copy Constructor.  
    //
    //!param: const Vect& vect - Copied vector.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    Vect( const Vect& vect );


    //-----------------------------------------------------------------------------
    //
    //: Constructor.
    //
    //!param: const std::string& name - The name of the data.
    //!param: INT_2U type - The data type.
    //!param: INT_4U nDim - The number of dimensions.
    //!param: const Dimension* dims - A pointer to 'nDim' Dimension objects
    //+       containing information about the dimensions for this data.
    //!param: const ByteOrder byte_order - Byte order of the data. Default is 
    //+       BIGENDIAN.   
    //!param: const void* data - A pointer to the data.  Default: 0
    //!param: const std::string& unitY - Units for the data.  Default: ""
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //    
    Vect( const std::string& name, INT_2U type, INT_4U nDim, const Dimension* dims,
          const ByteOrder byte_order = BIG, const void* data = 0, 
          const std::string& unitY = "" );

    //-----------------------------------------------------------------------------
    //
    //: Constructor.
    //
    //!param: const std::string& name - The name of the data.
    //!param: INT_2U type - The data type.
    //!param: INT_4U nDim - The number of dimensions.
    //!param: const Dimension* dims - A pointer to 'nDim' Dimension objects
    //+     containing information about the dimensions for this data.
    //!param: const ByteOrder byte_order - Byte order of the data. Default is 
    //+       BIGENDIAN.      
    //!param: const void* data - A pointer to the data.  Default: 0
    //!param: const std::string& unitY - Units for the data.  Default: ""
    //!param: bool allocate - Whether to allocate data. Default: true.
    //!param: bool owns - Whether to destruct data in the destructor. Default: true.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    Vect( const std::string& name, INT_2U type, INT_4U nDim, const Dimension* dims,
          const ByteOrder byte_order = BIG, void* data = 0, const std::string& unitY = "", 
          bool allocate = true, bool owns = true );

    //-----------------------------------------------------------------------------
    //
    //: Constructor.
    //
    //!param: const std::string& name - The name of the data.
    //!param: INT_2U type - The data type.
    //!param: INT_4U nDim - The number of dimensions.
    //!param: const INT_4U* dims - A pointer to 'nDim' Dimension objects
    //+     containing information about the dimensions for this data.
    //!param: const ByteOrder byte_order - Byte order of the data. Default is 
    //+       BIGENDIAN.      
    //!param: const void* data - A pointer to the data.  Default: 0
    //!param: const std::string& unitY - Units for the data.  Default: ""
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    Vect( const std::string& name, INT_2U type, INT_4U nDim, const INT_4U* dims,
          const ByteOrder byte_order = BIG, const void* data = 0, 
          const std::string& unitY = "" );

    //-----------------------------------------------------------------------------
    //
    //: Constructor.
    //
    //!param: const std::string& name - The name of the data.
    //!param: INT_2U type - The data type.
    //!param: INT_4U nDim - The number of dimensions.
    //!param: const INT_4U* dims - A pointer to 'nDim' Dimension objects
    //+     containing information about the dimensions for this data.
    //!param: const ByteOrder byte_order - Byte order of the data. Default is 
    //+       BIGENDIAN.      
    //!param: const void* data - A pointer to the data.  Default: 0
    //!param: const std::string& unitY - Units for the data.  Default: ""
    //!param: bool allocate - Whether to allocate data. Default: true.
    //!param: bool owns - Whether to destruct data in the destructor. Default: true.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    Vect( const std::string& name, INT_2U type, INT_4U nDim, const INT_4U* dims,
          const ByteOrder byte_order = BIG, void* data = 0, 
          const std::string& unitY = "", bool allocate = true, bool owns = true );

    //!exc: std::bad_alloc - Memory allocation failed.   
    template< class T >
    Vect( const std::string& name, INT_4U nDim = 1, const Dimension* dims = 0,
          const T* data = 0, const std::string& unitY = "" );

    //!exc: std::bad_alloc - Memory allocation failed.   
    template< class T >
    Vect( const std::string& name, INT_4U nDim = 1, const Dimension* dims = 0,
          T* data = 0, const std::string& unitY = "", bool allocate = true,
          bool owns = true );

    //!exc: std::bad_alloc - Memory allocation failed.   
    template< class T >
    Vect( const std::string& name, INT_4U nDim = 1, const INT_4U* dims = 0,
          const T* data = 0, const std::string& unitY = "" );

    //!exc: std::bad_alloc - Memory allocation failed.   
    template< class T >
    Vect( const std::string& name, INT_4U nDim = 1, const INT_4U* dims = 0,
          T* data = 0, const std::string& unitY = "", bool allocate = true,
          bool owns = true );
    
    //-----------------------------------------------------------------------------
    //
    //: Destructor.
    //
    // Deletes data, if it is owned.
    // 
    //!exc: None
    virtual ~Vect();

    /* Operator Overloads */
    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    //!param: const Vect& vect - Vector to be assigned.
    //
    //!return: const Vect& -- This vector.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const Vect& operator=( const Vect& vect );

    //-----------------------------------------------------------------------------
    //
    //: += Assignment Operator.
    //
    //!param: const Vect& vect - Vector with which to concatenate.
    //
    //!return: const Vect& -- This vector.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: frame_mismatch - Vectors are incompatible.
    //
    const Vect& operator+=( const Vect& vect );


    //-----------------------------------------------------------------------------
    //
    //: Equal comparison operator.
    //
    // Determines whether two Vect objects are equal. For this comparison,
    // all of the data is compared.
    //
    //!param: const Vect& v - The object with which to compare.
    //
    //!return: bool -- True if the objects are equal.
    //
    //!exc: None.
    bool operator==( const Vect& vect ) const;
    //!exc: None.   
    bool operator!=( const Vect& vect ) const;
    
    /* Required Methods */
    //!exc: None.   
    virtual INT_2U getClassId() const;

    //-----------------------------------------------------------------------------
    //
    //: Constructs the Structure Header for this class.
    //
    //!return: SH -- Vector class structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH();

    //: Constructs frame data format version 4 Structure Header for this class.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //   
    static SH getSH4();

    //: Constructs frame data format version 3 Structure Header for this class.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //   
    static SH getSH3();

    /* Accessors */

    /* gcc 2.95.1 can't handle a typedef with this if the throw spec is
       present (e.g. typedef SearchContainer< Vect, &Vect::getName > dc; ).
       
       const std::string& getName() const throw(); */
    const std::string& getName() const;
    //!exc: None.   
    Compression getCompress() const;
    //!exc: None.      
    ByteOrder getByteOrder() const;
    //!exc: None.      
    INT_2U getType() const;
    //!exc: None.      
    INT_4U getNData() const;
    //!exc: None.      
    INT_4U getNBytes() const;
    //!exc: None.      
    void* getData( bool decompress = true );
    //!exc: None.      
    const void* getData( bool decompress = true ) const;
    //!exc: None.      
    INT_4U getNDim() const;
    //!exc: std::range_error      
    Dimension& getDimension( size_t index );
    //!exc: std::range_error      
    const Dimension& getDimension( size_t index ) const;
    //!exc: None.         
    const std::string& getUnitY() const;
    //!exc: None.         
    bool isDataOwned() const;

    //-----------------------------------------------------------------------------
    //
    //: Expand compressed data.
    //
    // In general, uncompressing follows three steps:
    //
    // 1) Gunzip the data.
    // 2) Fix byte-ordering.
    // 3) Integrate.
    //
    // Depending upon the compression type and byte-ordering differences, not all
    // of these steps are needed.
    //    
    void expand();

    //-----------------------------------------------------------------------------
    //
    //: Compress the data for this Vect.
    //
    //!param: Compression c - Compressions type.
    //!param: size_t level - Compressions level.
    //
    void compress( Compression c, size_t level = 6 );

    //-----------------------------------------------------------------------------
    //: Perform byte swapping operation on vector data
    //
    void byteSwapData();

    /* Mutators */
    //----------------------------------------------------------------------------
    //
    //: Set the dimensions.
    //
    //!param: std::vector<Dimensions>& dims - New dimensions.
    //
    //!exc: None.         
    void setDimension( std::vector<Dimension>& dims ) { mDimension = dims; }

    //----------------------------------------------------------------------------
    //
    //: Set the number of data points.
    //
    //!param: INT_4U nd - New number of data points.
    //
    //!exc: None.         
    void setNData(INT_4U nd) { mNData = nd; }


    //----------------------------------------------------------------------------
    //
    //: Set the number of data bytes.
    //
    //!param: INT_4U nb - New number of data bytes.
    //
    //!exc: None.         
    void setNBytes(INT_4U nb) { mNBytes = nb; }

    //----------------------------------------------------------------------------
    //
    //: Set the data.
    //
    //!param: const void* data - Data to be copied to the vector.
    //
    //!exc: bad alloc - Memory allocation failed.
    //
    void setData( const void* data );

    //----------------------------------------------------------------------------
    //
    //: Set or adopt the data.
    //
    //!param: const void* data - Data to be copied to the vect.
    //!param: bool allocate - Whether to allocate data. Default: true.
    //!param: bool owns - Whether to destruct data in the destructor. Default: true.
    //
    //!exc: bad alloc - Memory allocation failed.
    //
    void setData( void* data, bool allocate = true, bool owns = true );
    //!exc: None.         
    void setDataOwnership( bool owns );
    //!exc: None.         
    void setDataNeedsByteSwap( bool swap );

    //-----------------------------------------------------------------------------
    //
    //: Get size of single data point of particular size.
    //
    //!param: INT_2U type - Data type ID.
    //
    //!return: INT_2U -- Data size.
    //
    //!exc: None.         
    static size_t getTypeSize( INT_2U type );
    
private:
  friend class Dictionary;
  friend class FrameReadPlan;
  friend class FrameCPP::Version_6::FrVect;

    /* This method will be accessed by the Dictionary class but is
       not available for public use. */
    //-----------------------------------------------------------------------------
    //
    //: Reads an Vect object from an FrameReaderBase object.
    //
    // This method creates a dynamically allocated object.
    //
    // <p>*NOTE* This method allocates memory and returns a pointer to it.  The
    // user is responsible for freeing this memory.
    //
    //!param: FrameReaderBase& in - Data source.
    //
    //!return: Vect* -- Newly created vector.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Data read failure occured.
    //
    static Vect* read( FrameReaderBase& in ){ return read4(in); }

    //: Read legacy data vector.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Data read failure occured.
    //   
    static Vect* read4( FrameReaderBase& in );

    //: Read legacy data vector.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Data read failure occured.
    //   
    static Vect* read3( FrameReaderBase& in );

    //-----------------------------------------------------------------------------
    //
    //: Writes a Vect object to an Output object.
    //
    //!param: Output& out out - Output data stream.
    //
    //!exc: write_failure - Data write failure occured.
    //
    virtual void write( Output& out ) const;

    //: Delete owned data.
    //!exc: None.   
    void deleteData();

    //-----------------------------------------------------------------------------
    //
    //: Template function to get the type ID corresponding to certain data type.
    //
    //!return: Type
    //
    template< class T > static Type getDataType();

    //-----------------------------------------------------------------------------
    //
    //: Differentially compress the data.
    //
    // The data must be in an uncompressed state before this is called.
    //
    void diffCompress();

    //-----------------------------------------------------------------------------
    //
    //: Differentially compress the data.
    //
    // The data must be in an uncompressed state before this is called.
    //
    template< class T > void diffCompressType();


    //-----------------------------------------------------------------------------
    //
    //: Compress the data with GZip.
    //
    // The data must be in an uncompressed state before this is called.
    //
    //!param: size_t level - Compression level.
    //
    void gzipCompress( size_t level );

    //-----------------------------------------------------------------------------
    //
    //: Suppress zeroes.
    //
    void zCompress();

    //-----------------------------------------------------------------------------
    //
    //: Suppress zeroes.
    //
    // Borrowed from FrameL.c, v3r85
    //
    static int FrVectZComp( unsigned short *out, 
			    unsigned long *compL,
			    short *data,
			    int nData,
			    int bSize );

    //-----------------------------------------------------------------------------
    //
    //: Expand zero suppressed data.
    //
    // Borrowed from FrameL.c, v3r85
    //
    static void FrVectZExpand( short *out, 
			       unsigned short *data, 
			       int nData );
  
    //: Vector name.
    std::string mName;
    //: Compression algorithm number.
    INT_2U mCompress;
    //: Vector class.
    INT_2U mType;
    //: Number of sample elements in data series.
    INT_4U mNData;
    //: Number of bytes in the compressed vector.
    INT_4U mNBytes;
    //: Data.
    unsigned char* mData;
    //: Dimensionality of data vector.
    INT_4U mNDim;
    //: Dimension info.
    std::vector< Dimension > mDimension;
    //: How to interpret value of each element.
    std::string mUnitY;

    //: Whether this vector owns its data.
    bool mOwnsData;
    //: Whether data needs to be byte swapped before use.
    bool dataNeedsByteSwap;

    //: BZIP2 compression parameter blockSize100k specifies the block size to be used for compression.
    // It should be a value between 1 and 9 inclusive, and the actual block size used is 100000 x this figure.
    // 9 gives the best compression but takes most memory.
    static const int bzip2_blockSize100k;

    //: BZIP2 compression parameter workFactor controls how the compression phase behaves when presented with worst case, highly repetitive, input data.
    // If compression runs into difficulties caused by repetitive data,
    // the library switches from the standard sorting algorithm to a fallback algorithm.
    // The fallback is slower than the standard algorithm by perhaps a factor of three,
    // but always behaves reasonably, no matter how bad the input. 
    // The default value of 30 gives reasonable behaviour over a wide range of circumstances. 
    static const int bzip2_workFactor;
};


// Template specifications   
template<> Vect::Type Vect::getDataType< CHAR >();
template<> Vect::Type Vect::getDataType< INT_2S >();
template<> Vect::Type Vect::getDataType< REAL_8 >();
template<> Vect::Type Vect::getDataType< REAL_4 >();
template<> Vect::Type Vect::getDataType< INT_4S >();
template<> Vect::Type Vect::getDataType< INT_8S >();
template<> Vect::Type Vect::getDataType< COMPLEX_8 >();
template<> Vect::Type Vect::getDataType< COMPLEX_16 >();
template<> Vect::Type Vect::getDataType< INT_2U >();
template<> Vect::Type Vect::getDataType< INT_4U >();
template<> Vect::Type Vect::getDataType< INT_8U >();
template<> Vect::Type Vect::getDataType< CHAR_U >();

//!exc: std::bad_alloc   
template<> Vect::Vect( const std::string& name, INT_4U nDim, const Dimension* dims,
   const COMPLEX_8* data, const std::string& unitY );
//!exc: std::bad_alloc      
template<> Vect::Vect( const std::string& name, INT_4U nDim, const Dimension* dims,
   const COMPLEX_16*, const std::string& unitY );

//!exc: std::bad_alloc      
template<> Vect::Vect( const std::string& name, INT_4U nDim, const Dimension* dims,
   COMPLEX_8* data, const std::string& unitY, bool allocate,
   bool owns );
//!exc: std::bad_alloc      
template<> Vect::Vect( const std::string& name, INT_4U nDim, const Dimension* dims,
   COMPLEX_16* data, const std::string& unitY, bool allocate,
   bool owns );

//!exc: std::bad_alloc      
template<> Vect::Vect( const std::string& name, INT_4U nDim, const INT_4U* dims,
   const COMPLEX_8* data, const std::string& unitY );
//!exc: std::bad_alloc      
template<> Vect::Vect( const std::string& name, INT_4U nDim, const INT_4U* dims,
   const COMPLEX_16* data, const std::string& unitY );

//!exc: std::bad_alloc      
template<> Vect::Vect( const std::string& name, INT_4U nDim, const INT_4U* dims,
   COMPLEX_8* data, const std::string& unitY, bool allocate,
   bool owns );
//!exc: std::bad_alloc      
template<> Vect::Vect( const std::string& name, INT_4U nDim, const INT_4U* dims,
   COMPLEX_16* data, const std::string& unitY, bool allocate,
   bool owns );

template<> void Vect::diffCompressType< COMPLEX_8 >();   
template<> void Vect::diffCompressType< COMPLEX_16 >();      

//-----------------------------------------------------------------------------
//
//: Constructor.
//
// This creates a Vect object from a pointer to data.
//
//!param: const std::string& name - The name for the data.
//!param: INT_4U nDim - The number of dimensions.
//!param: const Dimension* dims - A pointer to 'nDim' Dimension objects
//+     containing information about the dimensions for this data.
//!param: const T* data - A pointer to the data.  This will be copied into
//+     the Vector.
//!param: const std::string& unitY - Units for the data.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
template< class T >
Vect::Vect( const std::string& name, INT_4U nDim, const Dimension* dims,
                      const T* data, const std::string& unitY ) 
        : mName( name ), 
	  mCompress( 
#ifdef WORDS_BIGENDIAN
		0
#else
		256
#endif
	  ),
          mType( /* Vect::template */ getDataType< T >() ), mNData( 1 ),
          mNBytes( 0 ), mData( 0 ), mNDim( nDim ), mDimension(),
          mUnitY( unitY ), mOwnsData( true ), dataNeedsByteSwap( false )
{
    // if dims is specified, copy them
    if ( dims != 0 )
    {
        // calculate mNData
        for ( size_t i = 0; i < nDim; i++ )
        {
            mNData *= dims[ i ].getNx();
        }

        // add the dimension information
        for ( size_t i = 0; i < mNDim; ++i )
        {
            mDimension.push_back( dims[ i ] );
        }
    }
    // otherwise assume that there is just one data point
    else
    {
        mNData = 1;

        for( size_t i = 0; i < nDim; ++i )
        {
            // Dimension size = 1, scale factor = 1.0, no units
            mDimension.push_back( Dimension( 1, 1.0, "" ) );
        }
    }

    // calculate number of bytes
    mNBytes = mNData * sizeof( T );

    // Allocate memory for the data
    mData = reinterpret_cast< unsigned char* >( new T[ mNData ] );
    
    // if data was provided then copy it
    if ( data != 0 )
    {
        memcpy( mData, data, mNBytes );
    }
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
// This creates a Vect object from a pointer to data.
//
//!param: const std::string& name - The name for the data.
//!param: INT_4U nDim - The number of dimensions.
//!param: const Dimension* dims - A pointer to 'nDim' Dimension objects
//+     containing information about the dimensions for this data.
//!param: const T* data - A pointer to the data.  This will be copied into
//+     the Vector.
//!param: const std::string& unitY - Units for the data.
//!param: bool allocate - Whether to allocate data. Default: true.
//!param: bool owns - Whether to destruct data in the destructor. Default: true.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
template< class T >
Vect::Vect( const std::string& name, INT_4U nDim, const Dimension* dims,
   T* data, const std::string& unitY, bool allocate, bool owns )
        : mName( name ), 
	  mCompress( 
#ifdef WORDS_BIGENDIAN
		0
#else
		256
#endif
	  ),
          mType( /* Vect::template */ getDataType< T >() ), mNData( 1 ),
          mNBytes( 0 ), mData( 0 ), mNDim( nDim ), mDimension(),
          mUnitY( unitY ), mOwnsData( owns ), dataNeedsByteSwap( false )
{
    // if dims is specified, copy them
    if ( dims != 0 )
    {
        // calculate mNData &
        // add the dimension information
        for ( size_t i = 0; i < nDim; i++ )
        {
            mDimension.push_back( dims[ i ] );
            mNData *= dims[ i ].getNx();
        }
    }
    // otherwise assume that there is just one data point
    else
    {
        mNData = 1;

        for( size_t i = 0; i < nDim; ++i )
        {
            // Dimension size = 1, scale factor = 1.0, no units
            mDimension.push_back( Dimension( 1, 1.0, "" ) );
        }
    }

    // calculate number of bytes
    mNBytes = mNData * sizeof( T );

    if ( allocate )
    {
        // Allocate memory for the data
        mData = reinterpret_cast< unsigned char* >( new T[ mNData ] );

        // if data was provided then copy it
        if ( data != 0 )
        {
            memcpy( mData, data, mNBytes );
        }
    }
    else
    {
        mData = (unsigned char *) data;
    }
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
// This creates a Vect object from a pointer to data.
//
//!param: const std::string& name - The name for the data.
//!param: INT_4U nDim - The number of dimensions.
//!param: const INT_4U* dims - A pointer to 'nDim' integers
//+     containing the sizes for each dimension.
//!param: const T* data - A pointer to the data.  This will be copied into
//+     the Vector.
//!param: const std::string& unitY - Units for the data.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
template< class T >
Vect::Vect(
    const std::string& name, INT_4U nDim, const INT_4U* dims,
    const T* data, const std::string& unitY )
        : mName( name ),
	  mCompress( 
#ifdef WORDS_BIGENDIAN
		0
#else
		256
#endif
	  ),
          mType( /* Vect::template */ getDataType< T >() ), mNData( 1 ),
          mNBytes( 0 ), mData( 0 ), mNDim( nDim ), mDimension(),
          mUnitY( unitY ), mOwnsData( true ), dataNeedsByteSwap( false )
{
    // if dims is specified, copy them
    if ( dims != 0 )
    {
        // calculate mNData
        for ( size_t i = 0; i < nDim; i++ )
        {
            mNData *= dims[ i ];
        }

        // add the dimension information
        for ( size_t i = 0; i < mNDim; ++i )
        {
            mDimension.push_back( Dimension( dims[ i ], 1.0, "" ) );
        }
    }
    // otherwise assume that there is just one data point
    else
    {
        mNData = 1;

        for( size_t i = 0; i < nDim; ++i )
        {
            // Dimension size = 1, scale factor = 1.0, no units
            mDimension.push_back( Dimension( 1, 1.0, "" ) );
        }
    }

    // calculate number of bytes
    mNBytes = mNData * sizeof( T );

    // Allocate memory for the data
    mData = reinterpret_cast< unsigned char* >( new T[ mNData ] );
    
    // if data was provided then copy it
    if ( data != 0 )
    {
        memcpy( mData, data, mNBytes );
    }
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
// This creates a Vect object from a pointer to data.
//
//!param: const std::string& name - The name for the data.
//!param: INT_4U nDim - The number of dimensions.
//!param: const INT_4U* dims - A pointer to 'nDim' size_tegers
//+     containing the sizes for each dimension.
//!param: const T* data - A pointer to the data.  This will be copied into
//+     the Vector.
//!param: const std::string& unitY - Units for the data.
//!param: bool allocate - Whether to allocate data. Default: true.
//!param: bool owns - Whether to destruct data in the destructor. Default: true.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
template< class T >
Vect::Vect(
    const std::string& name, INT_4U nDim, const INT_4U* dims,
    T* data, const std::string& unitY, bool allocate, bool owns )
        : mName( name ), 
	  mCompress(
#ifdef WORDS_BIGENDIAN
		0
#else
		256
#endif
	  ),
          mType( /* Vect::template */ getDataType< T >() ), mNData( 1 ),
          mNBytes( 0 ), mData( 0 ), mNDim( nDim ), mDimension(),
          mUnitY( unitY ), mOwnsData( owns ), dataNeedsByteSwap( false )
{
    // if dims is specified, copy them
    if ( dims != 0 )
    {
        // calculate mNData &
        // add the dimension information
        for ( size_t i = 0; i < nDim; i++ )
        {
            mDimension.push_back( Dimension( dims[ i ], 1.0, "" ) );
            mNData *= dims[ i ];
        }
    }
    // otherwise assume that there is just one data point
    else
    {
        mNData = 1;

        for( size_t i = 0; i < nDim; ++i )
        {
            // Dimension size = 1, scale factor = 1.0, no units
            mDimension.push_back( Dimension( 1, 1.0, "" ) );
        }
    }

    // calculate number of bytes
    mNBytes = mNData * sizeof( T );

    if ( allocate )
    {
        // Allocate memory for the data
        mData = reinterpret_cast< unsigned char* >( new T[ mNData ] );
        
        // if data was provided then copy it
        if ( data != 0 )
        {
            memcpy( mData, data, mNBytes );
        }
    }
    else
    {
        mData = (unsigned char *) data;
    }
}



/******************/
/* Inline Methods */
/******************/


//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------



//-----------------------------------------------------------------------------
//
//: Get the compression algorithm.
//
//!return: INT_2U -- Compression algorithm ID.
//
//!exc: None.      
inline Vect::Compression Vect::getCompress() const 
{
    return Compression( mCompress >= 256 ? mCompress - 256 : mCompress );
}


//-----------------------------------------------------------------------------
//
//: Get the byte ordering of the data.
//
//!return: ByteOrder -- Endianness of vector's data.
//
//!exc: None.      
inline Vect::ByteOrder Vect::getByteOrder() const 
{
    // to check if data is compressed need to call getCompress()
    // method explicitely, otherwise mCompress != NONE is 
    // the same as mCompress != 0( BIGENDIAN = 0, NONE = 0 )
    if ( getCompress() != NONE )
    {
        if ( mCompress >= 256 )
        {
            return LITTLE;
        }
        else
        {
            return BIG;
        }
    }
    else
    {
       // data is not compressed: should be 0 or 256
       if( mCompress == 0 )
       {
          return BIG;
       }
       else
       {
          return LITTLE;
       }
    }
}

    
//-----------------------------------------------------------------------------
//
//: Get the vector data type.
//
//!return: INT_2U -- Vector data type ID.
//
//!exc: None.      
inline INT_2U Vect::getType() const
{
    return mType;
}


//-----------------------------------------------------------------------------
//
//: Get the number of sampled elements in the data series.
//
//!return: INT_4U -- Number of data points.
//
//!exc: None.      
inline INT_4U Vect::getNData() const
{
    return mNData;
}


//-----------------------------------------------------------------------------
//
//: Get the number of bytes in the compressed vector.
//
//!return: INT_4U -- (Compressed) Data size.
//
//!exc: None.      
inline INT_4U Vect::getNBytes() const
{
    return mNBytes;
}


//-----------------------------------------------------------------------------
//
//: Get a pointer to the data.
//
//!param: bool decompress - Whether to expand compressed data.
//
//!return: void* -- Vector data.
//
//!exc: None.      
inline void* Vect::getData( bool decompress )
{
    if ( decompress )
    {
        expand();
    }
    
    return mData;
}


//-----------------------------------------------------------------------------
//
//: Get a pointer to the data.
//
//!param: bool decompress - Whether to expand compressed data.
//
//!return: const void* -- Vector data.
//
//!exc: None.      
inline const void* Vect::getData( bool decompress ) const
{
    if ( decompress )
    {
        const_cast< Vect* >( this )->expand();
    }
    
    return mData;
}


//-----------------------------------------------------------------------------
//
//: Get the dimensionality of the data vector.
//
//!return: mNDim -- Number of data dimensions.
//
//!exc: None.      
inline INT_4U Vect::getNDim() const
{
    return mNDim;
}


//-----------------------------------------------------------------------------
//
//: Get dimension information.
//
//!param: size_t index - Which dimension.
//
//!return: Dimension& -- Reuqested dimension's info.
//
//!exc: std::range_error - Bad dimensions index given.
//
inline Dimension& Vect::getDimension( size_t index )
{
    if ( index >= mNDim )
    {
        throw std::range_error( "Index out of range." );
    }

    return mDimension[ index ];
}


//-----------------------------------------------------------------------------
//
//: Get dimension information.
//
//!param: size_t index - Which dimension.
//
//!return: const Dimension& -- Reuqested dimension's info.
//
//!exc: std::range_error - Bad dimensions index given.
//
inline const Dimension& Vect::getDimension(
    size_t index ) const 
{
    if ( index >= mNDim )
    {
        throw std::range_error( "Index out of range." );
    }

    return mDimension[ index ];
}


//-----------------------------------------------------------------------------
//
//: Get the std::string describing how to interpret the value of each element.
//
//!return: const std::string& -- Data units.
//
//!exc: None.      
inline const std::string& Vect::getUnitY() const 
{
    return mUnitY;
}


//-----------------------------------------------------------------------------
//
//: Not-equal comparison.
//
//!param: const Vect& vect - The object to compare with.
//
//!return: bool -- True if the objects are not equal.
//
//!exc: None.      
inline bool Vect::operator!=( const Vect& vect ) const 
{
    return !( (*this) == vect );
}

//-----------------------------------------------------------------------------
//: Set data ownership flag.
//
//!param bool owns - True if data needs to be destructed in the destructor.
//
//!exc: None.      
inline void Vect::setDataOwnership( bool owns ) 
{
    mOwnsData = owns;
}

//-----------------------------------------------------------------------------
//: Set byteswap required flag.
//
//!param bool swap - True if data needs to byte swapped.
//
//!exc: None.      
inline void Vect::setDataNeedsByteSwap( bool swap ) 
{
  dataNeedsByteSwap = swap;
}

//-----------------------------------------------------------------------------
//: See if data is owned by this vector
//
//!return: bool -- Data is owned or not.
//
//!exc: None.      
inline bool Vect::isDataOwned() const 
{
    return mOwnsData;
}

} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppVectHH
    
