#ifndef FrameCPP_VERSION_6_FrAdcData_HH
#define FrameCPP_VERSION_6_FrAdcData_HH

#include <memory>
#include <vector>

#include "general/gpstime.hh"

#include "framecpp/Version6/Container.hh"
#include "framecpp/Version6/Dimension.hh"
#include "framecpp/Version6/FrBase.hh"
#include "framecpp/Version6/FrVect.hh"
#include "framecpp/Version6/STRING.hh"

namespace FrameCPP
{
  namespace Version_6
  {
    const REAL_4 FR_ADC_DATA_DEFAULT_BIAS = 0.0;
    const REAL_4 FR_ADC_DATA_DEFAULT_SLOPE = 1.0;
    const REAL_8 FR_ADC_DATA_DEFAULT_FSHIFT = 0.0;
    const REAL_8 FR_ADC_DATA_DEFAULT_TIME_OFFSET = 0.0;
    const INT_2U FR_ADC_DATA_DEFAULT_DATA_VALID = 0;
    const REAL_4 FR_ADC_DATA_DEFAULT_PHASE = 0.0;
    inline const char* FR_ADC_DATA_DEFAULT_UNITS( )
    {
      return "counts";
    }

    class FrAdcData
      : public FrBase
    {
    public:

      typedef Container< FrVect >	data_type;
      typedef Container< FrVect >	aux_type;

      typedef data_type::const_iterator const_iterator;
      typedef data_type::iterator iterator;
      typedef aux_type::const_iterator const_aux_iterator;
      typedef aux_type::iterator aux_iterator;

      static const char* getStaticName();

      //: Default constructor
      FrAdcData( );

      FrAdcData( const FrAdcData& Source );

      FrAdcData( const std::string& name, INT_4U group,
		 INT_4U channel, INT_4U nBits,
		 REAL_8 sampleRate,
		 REAL_4 bias = FR_ADC_DATA_DEFAULT_BIAS,
		 REAL_4 slope = FR_ADC_DATA_DEFAULT_SLOPE,
		 const std::string& units = FR_ADC_DATA_DEFAULT_UNITS( ),
		 REAL_8 fShift = FR_ADC_DATA_DEFAULT_FSHIFT,
		 REAL_8 timeOffset = FR_ADC_DATA_DEFAULT_TIME_OFFSET,
		 INT_2U dataValid = FR_ADC_DATA_DEFAULT_DATA_VALID,
		 REAL_4 phase = FR_ADC_DATA_DEFAULT_PHASE );

      ~FrAdcData( );

      //: Comparison operator
      bool operator==( const FrAdcData& RHS ) const;
      bool operator!=( const FrAdcData& RHS ) const;

      //: Number of bytes needed to write this structure
      INT_8U Bytes( ) const;

      // Accessors
      const std::string& GetName() const; 
      const std::string& GetComment( ) const { return m_data.comment; }
      INT_4U GetChannelNumber( ) const { return m_data.channelNumber; }
      INT_4U GetChannelGroup( ) const { return m_data.channelGroup; }
      INT_4U GetNBits( ) const { return m_data.nBits; }
      REAL_4 GetBias( ) const { return m_data.bias; }
      REAL_4 GetSlope( ) const { return m_data.slope; }
      STRING GetUnits( ) const { return m_data.units; }
      REAL_8 GetSampleRate( ) const { return m_data.sampleRate; }
      REAL_8 GetTimeOffset( ) const { return m_data.timeOffset; }
      REAL_8 GetFShift( ) const { return m_data.fShift; }
      REAL_4 GetPhase( ) const { return m_data.phase; }
      INT_2U GetDataValid( ) const { return m_data.dataValid; }
      const aux_type& RefAux( ) const { return m_data.aux; }
      aux_type& RefAux( ) { return m_data.aux; }
      const data_type& RefData( ) const { return m_data.data; }
      data_type& RefData( ) { return m_data.data; }
   
      // Mutators
      void AppendComment( const std::string& comment );

      //: Merge with another FrAdcData
      FrAdcData& Merge( const FrAdcData& RHS );

      void SetDataValid( INT_2U flag )
      {
         m_data.dataValid = flag;
      }

      void SetBias( REAL_4 bias )
      {
         m_data.bias = bias;
      }

      void SetSlope( REAL_4 slope )
      {
         m_data.slope = slope;
      }

      void SetTimeOffset( REAL_8 time )
      {
         m_data.timeOffset = time;
      }

      std::auto_ptr< FrAdcData >
      SubFrAdcData( REAL_8 Offset, REAL_8 Dt ) const;
   
      //: Read the structure from stream
      static FrAdcData* ReadDynamic( IFrameStream& Stream,
				  const FrBase& Base );

      //: Write the structure to the stream
      virtual void Write( OFrameStream& Stream ) const;

      //: Write the structure description to the stream
      virtual void WriteDescription( OFrameStream& Stream ) const;

    private:
      struct data_definition_type {
	STRING	name;
	STRING	comment;
	INT_4U	channelGroup;
	INT_4U	channelNumber;
	INT_4U	nBits;
	REAL_4	bias;
	REAL_4	slope;
	STRING	units;
	REAL_8	sampleRate;
	REAL_8	timeOffset;
	REAL_8	fShift;
	REAL_4	phase;
	INT_2U	dataValid;
	data_type	data;
	aux_type	aux;

	data_definition_type( );
	data_definition_type( const data_definition_type& Source );

	//: Comparison operator
	bool operator==( const data_definition_type& RHS ) const;
	bool operator!=( const data_definition_type& RHS ) const;

	void copy_core( const data_definition_type& Source );
      } m_data;

      void read( IFrameStream& Stream );

#if ENABLE_DEPRICATED_FRAMECPP
    public:
      //-----------------------------------------------------------------
      // Version 5 compatability
      //-----------------------------------------------------------------
      typedef data_type DataContainer;
      typedef aux_type AuxContainer;

      FrAdcData( const std::string& name, INT_4U group,
		 INT_4U channel, INT_4U nBits,
		 const REAL_8& sampleRate,
		 REAL_4 bias = 0.0,
		 REAL_4 slope = 1.0,
		 const std::string& units = "counts",
		 const REAL_8& fShift = 0.0,
		 const GPSTime& timeOffset = GPSTime( 0, 0 ),
		 INT_2U dataValid = 0,
		 const REAL_4& phase = 0.0 );

      const std::string& getName( ) const;

      const std::string& getComment( ) const
      {
	return GetComment( );
      }

      INT_4U
      getChannelNumber( ) const
      {
	return GetChannelNumber( );
      }

      INT_4U getChannelGroup( ) const
      {
	return GetChannelGroup( );
      }

      INT_4U getNBits( ) const
      {
	return GetNBits( );
      }
      REAL_4 getBias( ) const
      {
	return GetBias( );
      }
      REAL_4 getSlope( ) const
      {
	return GetSlope( );
      }
      STRING getUnits( ) const
      {
	return GetUnits( );
      }
      REAL_8 getSampleRate( ) const
      {
	return GetSampleRate( );
      }
      const GPSTime& getTimeOffset( ) const
      {
	offsetVersion5 = GPSTime( ) + m_data.timeOffset;
	
	return offsetVersion5;
      }
      REAL_8 getFShift( ) const
      {
	return GetFShift( );
      }
      REAL_4 getPhase( ) const
      {
	return GetPhase( );
      }
      INT_2U getDataValid( ) const
      {
	return GetDataValid( );
      }

      void appendComment( const std::string& comment )
      {
	AppendComment( comment );
      }

      void setDataValid( INT_2U flag )
      {
	SetDataValid( flag );
      }
      void setBias( REAL_4 bias )
      {
	SetBias( bias );
      }
      void setSlope( REAL_4 slope )
      {
	SetSlope( slope );
      }
      void setTimeOffset( const GPSTime& time )
      {
	SetTimeOffset( time - GPSTime( 0, 0 ) );
      }

      const FrAdcData::AuxContainer& refAux( ) const
      {
	return RefAux( );
      }

      FrAdcData::AuxContainer& refAux( )
      {
	return RefAux( );
      }
      
      const FrAdcData::DataContainer& refData( ) const
      {
	return RefData( );
      }

      FrAdcData::DataContainer& refData( )
      {
	return RefData( );
      }
    private:
      mutable GPSTime	offsetVersion5;

#endif /* ENABLE_DEPRICATED_FRAMECPP */
    };

    inline bool FrAdcData::data_definition_type::
    operator!=( const data_definition_type& RHS ) const
    {
      return ! ( *this == RHS );
    }

    inline bool FrAdcData::
    operator==( const FrAdcData& RHS ) const
    {
      return ( this->m_data == RHS.m_data );
    }

    inline bool FrAdcData::
    operator!=( const FrAdcData& RHS ) const
    {
      return ! ( *this == RHS );
    }
  } // namespace - Version_6
} // namespace FrameCPP
#endif /* FrameCPP_VERSION_6_FrAdcData_HH */
