#include "framecpp/config.h"

#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/FrBase.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/OFrameStream.hh"

using namespace FrameCPP::Version_6;

#define	LM_DEBUG 0

FrBase::TooSmallException::
TooSmallException( const FrBase& Base, INT_4U Version )
  : runtime_error( format( Base, Version ) )
{
}

std::string FrBase::TooSmallException::
format( const FrBase& Base, INT_4U Version )
{
  std::ostringstream	msg;
  
  msg << "FrBase " << Base << " with length " << Base.GetLength( )
      << " byte" << ( ( Base.GetLength( ) == 1 ) ? "" : "s")
      << " is impossibly small."
      << " The minimal size is " << Base.Bytes( Version ) << " bytes.";
  return msg.str( );
}

FrBase::
FrBase( INT_2U Class )
  : m_instance_set( false )
{
  m_data.length = 0;
  m_data.class_id = Class;
  m_data.instance = 0;
}

FrBase::
FrBase( const FrBase& Source )
  : m_instance_set( Source.m_instance_set )
{
  m_data = Source.m_data;
}

FrBase::
~FrBase( )
{
}

INT_8U FrBase::
Bytes( ) const
{
  return Bytes( DATA_FORMAT_VERSION );
}

INT_8U FrBase::
Bytes( INT_2U Version )
{
  INT_8U	base_size( 0 );
  switch( Version )
  {
  case 6:
    base_size =
      sizeof( INT_8U ) + // length;
      sizeof( INT_2U ) + // class_id;
      sizeof( INT_4U );  // instance;
    break;
  case 5:
  case 4:
  case 3:
    base_size =
      sizeof( INT_4U ) + // length
      sizeof( INT_2U ) + // class_id
      sizeof( INT_2U ); // instance
    break;
  default:
    {
      std::ostringstream	msg;

      msg << "FrBase::Bytes( Version ) does not support version: "
	  << Version;
      throw std::runtime_error( msg.str( ) );
    }
    break;
  }
  return base_size;
}

std::string FrBase::
ErrorInfo( ) const
{
  return "";
}

INT_8U FrBase::
Skip( const IFrameStream& Stream ) const
{
  return m_data.length - Bytes( Stream.GetVersion( ) );
}

void FrBase::
read( IFrameStream& Stream )
{
  switch( Stream.GetVersion( ) )
  {
  case DATA_FORMAT_VERSION:
    Stream >> m_data.length
	   >> m_data.class_id
	   >> m_data.instance;
    break;
  case 5:
  case 4:
  case 3:
    {
      INT_4U	length;
      INT_2U	instance;

      Stream >> length
	     >> m_data.class_id
	     >> instance;
      m_data.length = length;
      m_data.instance = instance;
    }
    break;
  }
  if ( ( Stream.good( ) ) &&
       ( GetLength( ) < Bytes( Stream.GetVersion( ) ) ) )
  {
    throw TooSmallException( *this, Stream.GetVersion( ) );
  }
#if LM_DEBUG
  std::cerr << "DEBUG: FrBase: Read " << m_data.length
	    << " class_id: " << m_data.class_id
	    << " instance: " << m_data.instance
	    << std::endl;
#endif
}

void FrBase::
Write( OFrameStream& Stream ) const
{
  SetLength( );
  Stream << m_data.length
	 << m_data.class_id
	 << m_data.instance;
}

void FrBase::
WriteDescription( OFrameStream& Stream ) const
{
}

const FrBase& FrBase::
operator=( const FrBase& Base )
{
  m_data.length = Base.m_data.length;
  m_data.class_id = Base.m_data.class_id;
  m_data.instance = Base.m_data.instance;
  return *this;
}

std::ostream&
operator<<( std::ostream& Stream,
	    const FrBase& Base )
{
  Stream << "(" << Base.GetClassId( ) << "-" << Base.GetInstance( )
	 << ":" << Base.GetLength( )
	 << "[" << GetClassName(Base.GetClassId( ) )<< "])";
  return Stream;
}
