#include "framecpp/config.h"

#include "framecpp/Version3_4_5/endoffile.hh"

#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/Frame5ReaderBase.hh"
#include "framecpp/Version6/FrEndOfFile.hh"
#include "framecpp/Version6/Frame5ReaderBase.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/OFrameStream.hh"

using namespace FrameCPP::Version_6;


const CHAR* FrEndOfFile::getStaticName()
{
   static const CHAR* class_name( "FrEndOfFile" );
   return class_name;
}


FrEndOfFile::
FrEndOfFile( )
  : FrBase( CLASS_FR_END_OF_FILE )
{
  m_data.nFrames = 0;
  m_data.nBytes = 0;
  m_data.chkType = 0;
  m_data.chkSum = 0;
  m_data.seekTOC = 0;
}

INT_8U FrEndOfFile::
Bytes( ) const
{
  return Bytes( DATA_FORMAT_VERSION );
}

INT_8U FrEndOfFile::
Bytes( INT_2U Version )
{
  INT_8U	ret( FrBase::Bytes( Version ) );
  switch( Version )
  {
  case 6:
    ret +=
      sizeof( INT_4U ) + // nFrame
      sizeof( INT_8U ) + // nBytes
      sizeof( INT_4U ) + // chkType
      sizeof( INT_4U ) + // chkSum
      sizeof( INT_8U );	// seekTOC
    break;
  case 5:
  case 4:
    //:TRICKY: purposely drop through
    ret +=
      sizeof( INT_4U );	// seekTOC
  case 3:
    ret +=
      sizeof( INT_4U ) + // nFrame
      sizeof( INT_4U ) + // nBytes
      sizeof( INT_4U ) + // chkType
      sizeof( INT_4U );  // chkSum
    break;
  }
  return ret;
}

FrEndOfFile* FrEndOfFile::
ReadDynamic( IFrameStream& Stream, const FrBase& Base )
{
  std::auto_ptr< FrEndOfFile > ret( new FrEndOfFile );
  ret->read( Stream );
  ret->SetInstance( Base.GetInstance( ) );
  return ret.release( );
}

void FrEndOfFile::
Write( OFrameStream& Stream ) const
{
  if ( Stream.IsCalculatingChecksumFile() )
  {
    CheckSum&	cksum( Stream.GetCheckSumFile( ) );
    m_data.chkType = cksum.GetType( );
  }
  else
  {
    m_data.chkType = FrameCPP::CheckSum::NONE;
  }

  Stream << m_data.nFrames
	 << m_data.nBytes
	 << m_data.chkType;

  m_data.chkSum = 0;
  if ( Stream.IsCalculatingChecksumFile( ) )
  {
    Stream.calcChecksumFile( &m_data.chkSum, sizeof( m_data.chkSum ) ); 
    Stream.calcChecksumFile( &m_data.seekTOC, sizeof( m_data.seekTOC ) ); 
    m_data.chkSum = Stream.GetCheckSumFile( ).value( );
  }

  Stream << m_data.chkSum
	 << m_data.seekTOC;
}

void FrEndOfFile::
WriteDescription( OFrameStream& Stream ) const
{
  Stream.WriteDictionary( FrSH( getStaticName(), CLASS_FR_END_OF_FILE,
				"End of File Data Structure" ) );
  Stream.WriteDictionary( FrSE( "nFrames", "INT_4U",
				"Number of frames in this file") );
  Stream.WriteDictionary( FrSE( "nBytes", "INT_8U",
				"Total number of bytes in this file ( 0 if NOT computed )") );
  Stream.WriteDictionary( FrSE( "chkType", "INT_4U",
				"Checksum schemes: 0=none, 1=CRC, >2 presently unsued" ) );
  Stream.WriteDictionary( FrSE( "chkSum", "INT_4U", "File checksum." ) );
  Stream.WriteDictionary( FrSE( "seekTOC", "INT_8U",
				"Byes to back up to the beginning of the table of contents structure."
				" If seekTOC == 0, then there is no TOC for this file." ) );
}

void FrEndOfFile::
read( IFrameStream& Stream )
{
  using namespace FrameCPP::Version_3_4_5;

  std::auto_ptr< EndOfFile > eof5;
  switch( Stream.GetVersion( ) )
  {
  case DATA_FORMAT_VERSION:
    Stream >> m_data.nFrames
	   >> m_data.nBytes
	   >> m_data.chkType;
    //-------------------------------------------------------------------
    // Calculate the checksum for 0
    //-------------------------------------------------------------------
    m_data.chkSum = 0;
    //Stream.calcChecksums( &(m_data.chkSum), sizeof( m_data.chkSum ) );
    if ( Stream.IsCalculatingChecksumFile() )
    {
      Stream.calcChecksumFile( &(m_data.chkSum), sizeof( m_data.chkSum ) );
    }
    Stream.read( &m_data.chkSum, sizeof( m_data.chkSum ), false );
    if ( Stream.IsCalculatingMD5Sum( ) )
    {
      Stream.calcMD5Sum( &(m_data.chkSum), sizeof( m_data.chkSum ) );
    }
    Stream.m_header.Reorder( m_data.chkSum );
    //-------------------------------------------------------------------
    // Continue with normal reading
    //-------------------------------------------------------------------
    Stream >> m_data.seekTOC;
    break;
  case 5:
    {
      eof5.reset( EndOfFile::read( Stream.GetExposedFrame5Reader( ) ) );
    }
    break;
  case 4:
    {
      eof5.reset( EndOfFile::read4( Stream.GetExposedFrame5Reader( ) ) );
    }
    break;
  case 3:
    {
      eof5.reset( EndOfFile::read3( Stream.GetExposedFrame5Reader( ) ) );
    }
    break;
  }
  if ( eof5.get( ) )
  {
    m_data.nFrames = eof5->getNFrames( );
    m_data.nBytes = eof5->getNBytes( );
    m_data.chkType = eof5->getChkType( );
    m_data.chkSum = eof5->getChkSum( );
    m_data.seekTOC = eof5->getSeekTOC( );
  }
}

