#include "framecpp/config.h"

#include "framecpp/Version3_4_5/endofframe.hh"

#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/FrEndOfFrame.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/OFrameStream.hh"

using namespace FrameCPP::Version_6;


const CHAR* FrEndOfFrame::getStaticName()
{
   static const CHAR* class_name( "FrEndOfFrame" );
   return class_name;
}


FrEndOfFrame::
FrEndOfFrame( )
  : FrBase( CLASS_FR_END_OF_FRAME )
{
  m_data.chkType = 0;
  m_data.chkSum = 0;
}

FrEndOfFrame::
FrEndOfFrame( INT_4S Run, INT_4U Frame )
  : FrBase( CLASS_FR_END_OF_FRAME )
{
  m_data.run = Run;
  m_data.frame = Frame;
}

INT_8U FrEndOfFrame::
Bytes( ) const
{
  return FrBase::Bytes( ) +
    sizeof( m_data.run ) +
    sizeof( m_data.frame ) +
    sizeof( m_data.chkType ) +
    sizeof( m_data.chkSum );
}

FrEndOfFrame* FrEndOfFrame::
ReadDynamic( IFrameStream& Stream, const FrBase& Base )
{
  std::auto_ptr< FrEndOfFrame > ret( new FrEndOfFrame );
  ret->read( Stream );
  ret->SetInstance( Base.GetInstance( ) );
  return ret.release( );
}

void FrEndOfFrame::
Write( OFrameStream& Stream ) const
{
  if ( Stream.IsCalculatingChecksumFrame() )
  {
    CheckSum&	cksum( Stream.GetCheckSumFrame( ) );
    m_data.chkType = cksum.GetType( );
  }
  else
  {
    m_data.chkType = FrameCPP::CheckSum::NONE;
  }

  Stream << m_data.run
	 << m_data.frame
	 << m_data.chkType;

  m_data.chkSum = 0;
  if ( Stream.IsCalculatingChecksumFrame() )
  {
    Stream.calcChecksumFrame( &m_data.chkSum, sizeof( m_data.chkSum ) ); 
    m_data.chkSum = Stream.GetCheckSumFrame( ).value( );
  }
  Stream << m_data.chkSum;
}

void FrEndOfFrame::
WriteDescription( OFrameStream& Stream ) const
{
  Stream.WriteDictionary
    ( FrSH( getStaticName(), CLASS_FR_END_OF_FRAME,
	"End of File Data Structure" ) );
  Stream.WriteDictionary
    ( FrSE( "run", "INT_4S",
	"Run number; same as in FrameHeader run number datum." ) );
  Stream.WriteDictionary
    ( FrSE( "frame", "INT_4U",
	    "Frame number, monotonically increasing until end of run;"
	    " same as in Frame Header run number datum" ) );
  Stream.WriteDictionary
    ( FrSE( "chkType", "INT_4U",
	    "Checksum schemes: 0=none, 1=CRC, >2 presently unsued" ) );
  Stream.WriteDictionary
    ( FrSE( "chkSum", "INT_4U", "File checksum." ) );
}

void FrEndOfFrame::
read( IFrameStream& Stream )
{
  using namespace FrameCPP::Version_3_4_5;
  std::auto_ptr< EndOfFrame > eof5;
  switch( Stream.GetVersion( ) )
  {
  case DATA_FORMAT_VERSION:
    Stream >> m_data.run
	   >> m_data.frame
	   >> m_data.chkType;
    //-------------------------------------------------------------------
    // Calculate the checksum for 0
    //-------------------------------------------------------------------
    m_data.chkSum = 0;
    if ( Stream.IsCalculatingChecksumFrame( ) )
    {
      Stream.calcChecksumFrame( &(m_data.chkSum), sizeof( m_data.chkSum ) );
    }
    Stream.read( &m_data.chkSum, sizeof( m_data.chkSum ), false );
    if ( Stream.IsCalculatingChecksumFile( ) )
    {
      Stream.calcChecksumFile( &(m_data.chkSum), sizeof( m_data.chkSum ) );
    }
    if ( Stream.IsCalculatingMD5Sum( ) )
    {
      Stream.calcMD5Sum( &(m_data.chkSum), sizeof( m_data.chkSum ) );
    }
    Stream.m_header.Reorder( m_data.chkSum );
    break;
  case 5:
    eof5.reset( EndOfFrame::read( Stream.GetFrame5Reader( ) ) );
    break;
  case 4:
    eof5.reset( EndOfFrame::read4( Stream.GetFrame5Reader( ) ) );
    break;
  case 3:
    eof5.reset( EndOfFrame::read3( Stream.GetFrame5Reader( ) ) );
    break;
  }
  if ( eof5.get( ) )
  {
    m_data.run = eof5->getRun( );
    m_data.frame = eof5->getFrame( );
    m_data.chkType = eof5->getChkType( );
    m_data.chkSum = eof5->getChkSum( );
  }
}
