#include "framecpp/config.h"

#include <iostream>
#include <algorithm>
#include <stdexcept>

#include "general/autoarray.hh"

#include "framecpp/Common/Compression.hh"

#include "framecpp/Version3_4_5/vect.hh"

#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/Frame5ReaderBase.hh"
#include "framecpp/Version6/FrVect.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/OFrameStream.hh"
#include "framecpp/Version6/PTR_STRUCT.hh"

using namespace FrameCPP::Version_6;

using FrameCPP::ZeroSuppressCompressShort;
using FrameCPP::ZeroSuppressCompressInt;
using FrameCPP::ZeroSuppressUncompressInt;
using namespace FrameCPP::Compression;

#define LM_DEBUG 0

#if LM_DEBUG
#define AT( ) std::cerr << "INFO: " << __FILE__ << " " << __LINE__ << std::endl;
#else
#define AT( )
#endif

namespace
{
  //=====================================================================
  // Local enumerations
  //=====================================================================
  enum {
    MODE_RAW = FrVect::BIGENDIAN_RAW,
    MODE_GZIP = FrVect::BIGENDIAN_GZIP,
    MODE_DIFF = FrVect::BIGENDIAN_DIFF,
    MODE_DIFF_GZIP = FrVect::BIGENDIAN_DIFF_GZIP,
    MODE_ZERO_SUPPRESS_SHORT = FrVect::BIGENDIAN_ZERO_SUPPRESS_SHORT,
    MODE_ZERO_SUPPRESS_SHORT_GZIP_OTHER =
    FrVect::BIGENDIAN_ZERO_SUPPRESS_SHORT_GZIP_OTHER,
    MODE_ZERO_SUPPRESS_INT_FLOAT = FrVect::BIGENDIAN_ZERO_SUPPRESS_INT_FLOAT,
    MODE_ZERO_SUPPRESS_OTHERWISE_GZIP
      = FrVect::BIGENDIAN_ZERO_SUPPRESS_OTHERWISE_GZIP
  };

  enum {
    MODE_BIGENDIAN = FrVect::BIGENDIAN_RAW,
    MODE_LITTLEENDIAN = FrVect::LITTLEENDIAN_RAW,
    MODE_HOSTENDIAN = FrVect::NONE
  };

  //=====================================================================
  // Local classes
  //=====================================================================
  class DimBuffer
  {
  public:
    DimBuffer( )
    {
    }

    DimBuffer( const std::vector< Dimension >& Dims );

    void copy( std::vector< Dimension >& Dims );

    static INT_8U GetMinimumBytes( );

  private:
    friend class ::FrameCPP::Version_6::IFrameStream;
    friend class ::FrameCPP::Version_6::OFrameStream;

    std::vector< INT_8U >	m_nx;
    std::vector< REAL_8 >	m_dx;
    std::vector< REAL_8 >	m_startX;
    std::vector< STRING >	m_unitX;
  };

  inline INT_8U DimBuffer::
  GetMinimumBytes( )
  {
    return
      sizeof( INT_8U ) +
      sizeof( REAL_8 ) +
      sizeof( REAL_8 ) +
      STRING::Bytes( "" );
  }

  //=====================================================================
  // Local function prototypes
  //=====================================================================
  FrameCPP::Compression::data_type_mapping& data_type_map_init( );
  FrameCPP::Compression::compress_type_mapping& compress_type_map_init( );

  //=====================================================================
  // Local variables
  //=====================================================================
  static const FrameCPP::Compression::data_type_mapping&
  data_type_map = data_type_map_init( );

  static const FrameCPP::Compression::compress_type_mapping&
  compress_type_map = compress_type_map_init( );

  //=====================================================================
  // Local functions
  //=====================================================================
  INT_8U
  calc_nData( const std::vector< Dimension >& Dimensions )
  {
    INT_8U	ret( 0 );
    if ( Dimensions.size( ) > 0 )
    {
      ret = 1;
      for ( std::vector< Dimension >::const_iterator
	      i( Dimensions.begin( ) );
	    i != Dimensions.end( );
	    i++ )
      {
	ret *= (*i).GetNx( );
      }
    }
    return ret;
  }

  FrameCPP::Compression::compress_type_mapping&
  compress_type_map_init( )
  {
    static FrameCPP::Compression::compress_type_mapping	m;

#define INIT( x ) m[ x ] = FrameCPP::Compression::x

    INIT( MODE_RAW );
    INIT( MODE_GZIP );
    INIT( MODE_DIFF_GZIP );
    INIT( MODE_ZERO_SUPPRESS_SHORT );
    INIT( MODE_ZERO_SUPPRESS_INT_FLOAT );
    INIT( MODE_ZERO_SUPPRESS_OTHERWISE_GZIP );

#undef INIT
    return m;
  }
  
  FrameCPP::Compression::data_type_mapping&
  data_type_map_init( )
  {
    static FrameCPP::Compression::data_type_mapping	m;

#define INIT( x ) m[ FrameCPP::Version_6::FrVect::x ] \
      = FrameCPP::Compression::x

    INIT( FR_VECT_C );
    INIT( FR_VECT_2S );
    INIT( FR_VECT_8R );
    INIT( FR_VECT_4R );
    INIT( FR_VECT_4S );
    INIT( FR_VECT_8S );
    INIT( FR_VECT_8C );
    INIT( FR_VECT_16C );
    INIT( FR_VECT_STRING );
    INIT( FR_VECT_2U );
    INIT( FR_VECT_4U );
    INIT( FR_VECT_8U );
    INIT( FR_VECT_1U );

#undef INIT
    return m;
  }
  
  
}// namespace - anonymous

namespace
{
  INT_8U
  Bytes( const std::vector< Dimension >& Dims )
  {
    INT_8U	ret( sizeof( INT_4U ) );	// nDim
    for( std::vector< Dimension >::const_iterator d( Dims.begin( )  );
	 d != Dims.end( );
	 d++ )
    {
      ret += sizeof( INT_8U ) +	// nx
	sizeof( REAL_8 ) +	// dx
	sizeof( REAL_8 ) +	// startX
	(*d).GetUnitX( ).Bytes( );
    }
    return ret;
  } // Bytes

  DimBuffer::
  DimBuffer( const std::vector< Dimension >& Dims )
  {
    INT_4U	nDim( Dims.size( ) );
      
    m_nx.resize( nDim );
    m_dx.resize( nDim );
    m_startX.resize( nDim );
    m_unitX.resize( nDim );

    for ( INT_4U x = 0; x < nDim; x++ )
    {
      m_nx[ x ] = Dims[ x ].GetNx( );
      m_dx[ x ] = Dims[ x ].GetDx( );
      m_startX[ x ] = Dims[ x ].GetStartX( );
      m_unitX[ x ] = Dims[ x ].GetUnitX( );
    }
  }

  void DimBuffer::
  copy( std::vector< Dimension >& Dims )
  {
    INT_4U	nDim( m_nx.size( ) );
    
    Dims.resize( nDim );
    for (  INT_4U x = 0; x < nDim; x++ )
    {
      Dims[ x ] = Dimension( m_nx[ x ],
			     m_dx[ x ],
			     m_unitX[ x ],
			     m_startX[ x ] );
    }
  }

} // namespace - anonymous

namespace FrameCPP
{
  namespace Version_6
  {
    template< >
    OFrameStream& OFrameStream::
    operator<<( const DimBuffer& Dims )
    {
      INT_4U	nDim( Dims.m_nx.size( ) );
      *this << nDim
	    << Dims.m_nx
	    << Dims.m_dx
	    << Dims.m_startX
	    << Dims.m_unitX
	;
      return *this;
    }
  
    template< >
    IFrameStream& IFrameStream::
    operator>>( DimBuffer& Dims )
    {
      INT_4U	nDim( 0 );
      *this >> nDim;
      // Sanity check
      CheckReadBoundry( nDim * DimBuffer::GetMinimumBytes( ) );

      this->ReadVector( Dims.m_nx, nDim );
      this->ReadVector( Dims.m_dx, nDim );
      this->ReadVector( Dims.m_startX, nDim );
      this->ReadVector( Dims.m_unitX, nDim );
      return *this;
    }
  } // namespace - Version_6
} // namespace - FrameCPP


const CHAR* FrVect::getStaticName()
{
  static const CHAR* class_name( "FrVect" );
  return class_name;
}


#define TRACE_MEMORY 0
#if TRACE_MEMORY
#define MEM_ALLOCATE() std::cerr << "MEMORY: Allocate: " << FrVect::getStaticName() << " " << (void*)this << " Line: " << __LINE__ << std::endl;
#define MEM_DELETE() std::cerr << "MEMORY: Delete: " << FrVect::getStaticName() << " " << (void*)this << " Line: " << __LINE__ << std::endl;
#else
#define MEM_ALLOCATE()
#define MEM_DELETE()
#endif


const int FrVect::DEFAULT_GZIP_LEVEL( 6 );


FrVect::data_definition_type::
data_definition_type()
  : name( "" ), type( FR_VECT_C ), data( 0 ), dimension( ), unitY( "" )
{
}


FrVect::data_definition_type::
data_definition_type( const std::string& n, INT_2U t,
		      const byte_order_type byte_order, INT_4U ndim,
		      const Dimension* dims, const std::string& unit )
  : name( n ), type( t ), data( 0 ), dimension(), unitY( unit )
{
  compress = ( ( byte_order == BYTE_ORDER_BIG_ENDIAN ) ? BIGENDIAN_RAW :
	       LITTLEENDIAN_RAW );

  for( INT_4U x = 0; x < ndim; ++x, ++dims )
  {
    dimension.push_back( *dims );
  }

  nData = calc_nData( dimension );
  nBytes = GetTypeSize( type ) * nData;
}


#if ENABLE_DEPRICATED_FRAMECPP
FrVect::data_definition_type::
data_definition_type( const std::string& n, INT_2U t, 
		      const byte_order_type byte_order, INT_4U ndim,
		      const INT_4U* dims, const std::string& unit )
  :  name( n ), type( t ), data( 0 ), dimension(), unitY( unit )
{
  compress = ( ( byte_order == BYTE_ORDER_BIG_ENDIAN ) ? BIGENDIAN_RAW :
	       LITTLEENDIAN_RAW );

  for( INT_4U x = 0; x < ndim; ++x, ++dims )
  {
    dimension.push_back( Dimension( *dims ) );
  }

  nData = calc_nData( dimension );
  nBytes = GetTypeSize( type ) * nData;
}
#endif


void FrVect::data_definition_type::
copy_core( const data_definition_type& Source )
{
  name = Source.name;
  compress = Source.compress;
  type = Source.type;
  nData = Source.nData;
  nBytes = Source.nBytes;

  for( std::vector< Dimension >::const_iterator
	 d( Source.dimension.begin( ) ),
	 d_end( Source.dimension.end() );
       d != d_end; ++d )
  {
    dimension.push_back( *d );
  }

  unitY = Source.unitY;
}

bool FrVect::data_definition_type::
operator==( const FrVect::data_definition_type& RHS ) const
{
#if LM_INFO
  if ( name != RHS.name )
  {
    std::cerr << "DEBUG: " << __LINE__ << std::endl;
  }
  if ( compress != RHS.compress )
  {
    std::cerr << "DEBUG: " << __LINE__ << std::endl;
  }
  if ( type != RHS.type )
  {
    std::cerr << "DEBUG: " << __LINE__ << std::endl;
  }
  if ( nData != RHS.nData )
  {
    std::cerr << "DEBUG: " << __LINE__ << std::endl;
  }
  if ( nBytes != RHS.nBytes )
  {
    std::cerr << "DEBUG: " << __LINE__ << " nBytes: " << nBytes << " RHS.nBytes: " << RHS.nBytes << std::endl;
  }
  if ( dimension != RHS.dimension )
  {
    std::cerr << "DEBUG: " << __LINE__ << std::endl;
  }
  if ( unitY != RHS.unitY )
  {
    std::cerr << "DEBUG: " << __LINE__ << std::endl;
  }
  if ( memcmp( data, RHS.data, nBytes ) != 0 )
  {
    std::cerr << "DEBUG: " << __LINE__ << std::endl;
  }
  std::cerr << "DEBUG: nData: " << nData << std::endl;
#endif
  return ( ( this == &RHS ) ||
	   (  ( name == RHS.name ) &&
	      ( compress == RHS.compress ) &&
	      ( type == RHS.type ) &&
	      ( nData == RHS.nData ) &&
	      ( nBytes == RHS.nBytes ) &&
	      ( dimension == RHS.dimension ) &&
	      ( unitY == RHS.unitY ) &&
	      ( std::equal( data, data + nBytes, RHS.data ) ) ) );
}


FrVect::
FrVect( )
  : FrBase( CLASS_FR_VECT ),
    m_data(), m_owns( true )
{
  MEM_ALLOCATE( );
}

FrVect::
FrVect( const FrVect& Source )
  : FrBase( CLASS_FR_VECT ),
    m_data(), m_owns( true )
{
  MEM_ALLOCATE( );

  m_data.copy_core( Source.m_data );
  copy_data( Source.m_data.data, m_data.nBytes, true, true );
}

FrVect::
FrVect( const std::string& name, INT_2U type, INT_4U nDim,
	const Dimension* dims,
	const byte_order_type byte_order, const void* data, 
	const std::string& unitY )
  : FrBase( CLASS_FR_VECT ),
    m_data( name, type, byte_order, nDim, dims, unitY ), m_owns( true )
{
  MEM_ALLOCATE( );
  copy_data( const_cast< CHAR_U* >( static_cast< const CHAR_U* >( data ) ),
	     m_data.nBytes, true, true );
}

FrVect::
FrVect( const std::string& name, INT_2U type, INT_4U nDim,
	const Dimension* dims,
	const byte_order_type byte_order,
	void* data,
	const std::string& unitY, 
	bool allocate, bool owns )
  : FrBase( CLASS_FR_VECT ),
    m_data( name, type, byte_order, nDim, dims, unitY ), m_owns( owns )
{
  MEM_ALLOCATE( );
  copy_data( static_cast< CHAR_U* >( data ), m_data.nBytes, allocate, owns );
}


#if ENABLE_DEPRICATED_FRAMECPP
FrVect::
FrVect( const std::string& name, INT_2U type, INT_4U nDim,
	const INT_4U* dims,
	const byte_order_type byte_order, const void* data, 
	const std::string& unitY )
  : FrBase( CLASS_FR_VECT ),
    m_data( name, type, byte_order, nDim, dims, unitY ), m_owns( true )
{
  MEM_ALLOCATE( );

  m_data.data = new CHAR_U[ m_data.nBytes ];
  const CHAR_U* char_data( reinterpret_cast< const CHAR_U* >( data ) );

  std::copy( char_data, char_data + m_data.nBytes, m_data.data );
}


FrVect::
FrVect( const std::string& name, INT_2U type, INT_4U nDim,
	const INT_4U* dims,
	const byte_order_type byte_order,
	void* data, 
	const std::string& unitY,
	bool allocate, bool owns )
  : FrBase( CLASS_FR_VECT ), 
    m_data( name, type, byte_order, nDim, dims, unitY ), m_owns( true )
{
  MEM_ALLOCATE( );

  m_data.data = new CHAR_U[ m_data.nBytes ];
  const CHAR_U* char_data( reinterpret_cast< const CHAR_U* >( data ) );

  std::copy( char_data, char_data + m_data.nBytes, m_data.data );
}
   
void FrVect::expand ( )
{
  Uncompress( );
}
   
   
   
#endif	/* ENABLE_DEPRICATED_FRAMECPP */

FrVect::
~FrVect( )
{
  MEM_DELETE( );
  if ( m_owns )
  {
    delete[] m_data.data;
    m_data.data = (CHAR_U*)NULL;
  }
}


const std::string& FrVect::GetName( ) const
{
  return m_data.name;
}


void FrVect::
Compress( compression_scheme_type Scheme, int Level )
{
  AT( );
  if ( Scheme == GetCompress( ) )
  {
    // Nothing to do. Just return
    AT( );
    return;
  }

  //---------------------------------------------------------------------
  // Ensure the buffer is fully expanded
  //---------------------------------------------------------------------

  AT( );
  Uncompress( );
  
  //---------------------------------------------------------------------
  // Now compress the data
  //---------------------------------------------------------------------

  INT_4U	compress = ( Scheme & 0xFF );

  FrameCPP::Compression::Compress( compress,
				   Level,
				   compress_type_map,
				   GetType( ),
				   data_type_map,
				   m_data.data,
				   m_data.nData,
				   m_data.nBytes
				   );
  //---------------------------------------------------------------------
  // Save the compression mode
  //---------------------------------------------------------------------

  m_data.compress = ( MODE_HOSTENDIAN | compress );
}

//-----------------------------------------------------------------------------
//
//: Expand compressed data.
//
// In general, uncompressing follows three steps:
//
// 1) Gunzip the data.
// 2) Fix byte-ordering.
// 3) Integrate.
//
// Depending upon the compression type and byte-ordering differences, not all
// of these steps are needed.
//
void FrVect::
Uncompress( )
{
  AT( );
  if ( GetCompress( ) == NONE )
  {
    // Nothing to do. The buffer is already uncompressed
    AT( );
    return;
  }

  FrameCPP::Compression::Expand( ( ( m_data.compress & 0x100 )
				   == MODE_HOSTENDIAN ),
				 m_data.compress & 0xFF,
				 compress_type_map,
				 GetType( ),
				 data_type_map,
				 m_data.data,
				 m_data.nData,
				 m_data.nBytes
				 );
  m_data.compress = NONE;
}

   
bool FrVect::operator!=( const FrVect& RHS ) const
{
  return ! ( *this == RHS );
}

bool FrVect::operator==( const FrVect& RHS ) const
{
  return ( this->m_data == RHS.m_data );
}

FrVect& FrVect::
operator+=( const FrVect& RHS )
{
  //---------------------------------------------------------------------
  // Perform sanity checks
  //---------------------------------------------------------------------
  if ( ( this->GetName( ) != RHS.GetName( ) ) ||
       ( this->GetType( ) != RHS.GetType( ) ) ||
       ( this->GetNDim( ) != RHS.GetNDim( ) ) ||
       ( this->GetUnitY( ) != RHS.GetUnitY( ) ) )
  {
    std::ostringstream	msg;
    bool		comma;

    msg << "Unable to concat the FrVect structures because: ";
    comma = false;
    if ( this->GetName( ) != RHS.GetName( ) )
    {
      msg << "name of objects differ ("
	  << this->GetName( ) << " vs. "
	  << RHS.GetName( ) << ")";
      comma = true;
    }
    if ( this->GetType( ) != RHS.GetType( ) )
    {
      if ( comma )
      {
	msg << ", ";
      }
      msg << "type of objects differ ("
	  << this->GetType( ) << " vs. "
	  << RHS.GetType( ) << ")";
      comma = true;
    }
    if ( this->GetNDim( ) != RHS.GetNDim( ) )
    {
      if ( comma )
      {
	msg << ", ";
      }
      msg << "number of dimensions differ ("
	  << this->GetNDim( ) << " vs. "
	  << RHS.GetNDim( ) << ")";
      comma = true;
    }
    if ( this->GetUnitY( ) != RHS.GetUnitY( ) )
    {
      if ( comma )
      {
	msg << ", ";
      }
      msg << "y units differ (" << this->GetUnitY( ) << " vs. "
	  << RHS.GetUnitY( ) << ")";
      comma = true;
    }
    throw std::domain_error( msg.str( ) );
  }
  for ( INT_4U cur = 0,
	  last = this->GetNDim( );
	cur != last;
	++cur )
  {
    const Dimension& ld( this->GetDim( cur ) );
    const Dimension& rd( RHS.GetDim( cur ) );

    if ( cur > 0 )
    {
      if ( ld != rd )
      {
	std::ostringstream	msg;

	msg << "Unable to concat the FrVect structures";
	//:TODO: Need to have a list of reasons
	throw std::domain_error( msg.str( ) );
      }
    }
    else
    {
      if ( ( ld.GetNx( ) != rd.GetNx( ) ) ||
	   ( ld.GetStartX( ) != rd.GetStartX( ) ) ||
	   ( ld.GetUnitX( ) != rd.GetUnitX( ) ) )
      {
	std::ostringstream	msg;

	msg << "Unable to concat the FrVect structures";
	//:TODO: Need to have a list of reasons
	throw std::domain_error( msg.str( ) );
      }
    }
  }
  //---------------------------------------------------------------------
  // Calculate how much space is required
  //---------------------------------------------------------------------
  const INT_2U	type_size( GetTypeSize( GetType( ) ) );
  this->GetDim( 0 ).SetNx( this->GetDim( 0 ).GetNx( ) +
			   RHS.GetDim( 0 ).GetNx( ) );
  const INT_8U	s = GetNData( ) + RHS.GetNData( );

  //---------------------------------------------------------------------
  // Allocate the space, fill it with the existing data and then store
  //   it.
  //---------------------------------------------------------------------
  General::AutoArray< CHAR_U >	data( new CHAR_U[ s ] );
  std::copy( GetData( ), GetData( ) + ( GetNData( ) * type_size ),
	     data.get( ) );
  std::copy( RHS.GetData( ), RHS.GetData( ) + ( RHS.GetNData( ) * type_size ),
	     data.get( ) + ( GetNData( ) * type_size ) );
  copy_data( data.release( ), s, false, true );
  //---------------------------------------------------------------------
  // Return
  //---------------------------------------------------------------------
  return *this;
}
   
INT_8U FrVect::
Bytes( ) const
{
  return FrBase::Bytes( ) +
    m_data.name.Bytes( ) +
    sizeof( m_data.compress ) +
    sizeof( m_data.type ) +
    sizeof( m_data.nData ) +
    sizeof( m_data.nBytes ) +
    m_data.nBytes +	// data
    ::Bytes( m_data.dimension ) +
    m_data.unitY.Bytes( ) +
    PTR_STRUCT::Bytes( )	// next
    ;
}


const CHAR_U* FrVect::
GetData( bool Decompress ) const
{
  if ( Decompress )
  {
    const_cast<FrVect*>( this )->Uncompress( );
  }
  return m_data.data;
}


CHAR_U* FrVect::
GetData( bool Decompress )
{
  if ( Decompress )
  {
    Uncompress( );
  }

  return m_data.data;
}


namespace FrameCPP
{
  namespace Version_6
  {
    template< >
    INT_2U FrVect::
    GetDataType< CHAR >( )
    {
      return FR_VECT_C;
    }

    template< >
    INT_2U FrVect::
    GetDataType< CHAR_U >( )
    {
      return FR_VECT_1U;
    }

    template< >
    INT_2U FrVect::
    GetDataType< INT_2S >( )
    {
      return FR_VECT_2S;
    }

    template< >
    INT_2U FrVect::
    GetDataType< INT_2U >( )
    {
      return FR_VECT_2U;
    }

    template< >
    INT_2U FrVect::
    GetDataType< INT_4S >( )
    {
      return FR_VECT_4S;
    }

    template< >
    INT_2U FrVect::
    GetDataType< INT_4U >( )
    {
      return FR_VECT_4U;
    }

    template< >
    INT_2U FrVect::
    GetDataType< INT_8S >( )
    {
      return FR_VECT_8S;
    }

    template< >
    INT_2U FrVect::
    GetDataType< INT_8U >( )
    {
      return FR_VECT_8U;
    }

    template< >
    INT_2U FrVect::
    GetDataType< REAL_4 >( )
    {
      return FR_VECT_4R;
    }

    template< >
    INT_2U FrVect::
    GetDataType< REAL_8 >( )
    {
      return FR_VECT_8R;
    }

    template< >
    INT_2U FrVect::
    GetDataType< COMPLEX_8 >( )
    {
      return FR_VECT_8C;
    }

    template< >
    INT_2U FrVect::
    GetDataType< COMPLEX_16 >( )
    {
      return FR_VECT_16C;
    }

    template< >
    INT_2U FrVect::
    GetDataType< std::string >( )
    {
      return FR_VECT_STR;
    }

  } // namespace - Version_6
} // namespace - FrameCPP
   
#if ENABLE_DEPRICATED_FRAMECPP   
template< class T >
INT_2U FrVect::getDataType( )
{
  return GetDataType< T >( );
}
#endif   
   
   
INT_8U FrVect::
GetNData( ) const
{
  return m_data.nData;
}
//-----------------------------------------------------------------------------
//
//: Get size of single data point of particular size.
//
//!param: INT_2U type - Data type ID.
//
//!return: INT_2U - Data size.
//
//!exc: None.   
size_t FrVect::
GetTypeSize( INT_2U type ) 
{
  switch( type )
  {
  case FR_VECT_C:
    return sizeof( CHAR );
  case FR_VECT_2S:
    return sizeof( INT_2S );
  case FR_VECT_8R:
    return sizeof( REAL_8 );
  case FR_VECT_4R:
    return sizeof( REAL_4 );
  case FR_VECT_4S:
    return sizeof( INT_4S );
  case FR_VECT_8S:
    return sizeof( INT_8S );
  case FR_VECT_8C:
    return ( sizeof( REAL_4 ) * 2 );
  case FR_VECT_16C:
    return ( sizeof( REAL_8 ) * 2 );
    /*
      ????
      case FR_VECT_STR:
      return sizeof( std::std::string );
      ????
    */
  case FR_VECT_2U:
    return sizeof( INT_2U );
  case FR_VECT_4U:
    return sizeof( INT_4U );
  case FR_VECT_8U:
    return sizeof( INT_8U );
  case FR_VECT_1U:
    return sizeof( CHAR_U );
  }
  // None of above: Perhaps add type for throw?
  return 0;
}

FrVect& FrVect::
Merge( const FrVect& RHS )
{
  //:TODO: Need to implement Merge routine
  std::string msg( "Merge currently not implemented for " );
  msg += getStaticName( );

  throw std::domain_error( msg );
  return *this;
}

FrVect* FrVect::
ReadDynamic( IFrameStream& Stream, const FrBase& Base )
{

  std::auto_ptr< FrVect > ret( new FrVect );
  ret->read( Stream );
  ret->SetInstance( Base.GetInstance( ) );
  return ret.release( );
}

std::auto_ptr< FrVect > FrVect::
SubFrVect( INT_4U Start, INT_4U Stop ) const
{
  //---------------------------------------------------------------------
  // Create new FrVect
  //---------------------------------------------------------------------
  std::auto_ptr< FrVect > retval( new FrVect );
  //---------------------------------------------------------------------
  // Calculate the block size
  //---------------------------------------------------------------------
  INT_4U	block_size = GetTypeSize( GetType( ) );
  const INT_4U	ndim = GetNDim( );

  for ( INT_4U x = 1; x < ndim; ++x )
  {
    block_size *= GetDim( x ).GetNx( );
  }
  //---------------------------------------------------------------------
  // Copy the core of the information
  //---------------------------------------------------------------------
  retval->m_data.copy_core( this->m_data );
  //---------------------------------------------------------------------
  // Set the Dimension information for the sink
  //---------------------------------------------------------------------
  retval->GetDim( 0 ).SetNx( Stop - Start );
  retval->SetNData( Stop - Start );
  //---------------------------------------------------------------------
  // Allocate space in the sink
  //---------------------------------------------------------------------
  retval->copy_data( new CHAR_U[ block_size * ( Stop - Start ) ],
		     block_size * ( Stop - Start ),
		     false, true );
  retval->m_data.compress = RAW;
  //---------------------------------------------------------------------
  // Copy the data
  //---------------------------------------------------------------------
  const CHAR_U*	data_source( GetData( ) ); // Get pointer to decompressed data
  CHAR_U* data_sink( retval->GetData( ) ); // Get pointer to decompressed data

  for ( INT_8U x = Start; x < Stop; ++x )
  {
    //-------------------------------------------------------------------
    // Copy 1 block's worth of data
    //-------------------------------------------------------------------
    std::copy( data_source + ( x * block_size ),
	       data_source + ( ( x + 1 ) * block_size ),
	       data_sink + ( ( x - Start ) * block_size ) );
  }
  //---------------------------------------------------------------------
  // Return the newly constructed FrVect
  //---------------------------------------------------------------------
  return retval;
}

void FrVect::
Write( OFrameStream& Stream ) const
{
  DimBuffer	dims( m_data.dimension );
  Stream << m_data.name
	 << m_data.compress
	 << m_data.type
	 << m_data.nData
	 << m_data.nBytes;
  Stream.write( m_data.data, m_data.nBytes );
  Stream << dims
	 << m_data.unitY
	 << GetNext( )
    ;
}

void FrVect::
WriteDescription( OFrameStream& Stream ) const
{
  Stream.WriteDictionary( FrSH( getStaticName(), CLASS_FR_VECT,
				"Vector Data Structure" ) );
  Stream.WriteDictionary( FrSE( "name", "STRING", "") );
  Stream.WriteDictionary( FrSE( "compress", "INT_2U", "") );
  Stream.WriteDictionary( FrSE( "type", "INT_2U", "") );
  Stream.WriteDictionary( FrSE( "nData", "INT_8U", "") );
  Stream.WriteDictionary( FrSE( "nBytes", "INT_8U", "") );
  Stream.WriteDictionary( FrSE( "data", "CHAR[nBytes]", "") );
  Stream.WriteDictionary( FrSE( "nDim", "INT_4U", "") );
  Stream.WriteDictionary( FrSE( "nx", "INT_8U[nDim]", "") );
  Stream.WriteDictionary( FrSE( "dx", "REAL_8[nDim]", "") );
  Stream.WriteDictionary( FrSE( "startX", "REAL_8[nDim]", "") );
  Stream.WriteDictionary( FrSE( "unitX", "STRING[nDim]", "") );
  Stream.WriteDictionary( FrSE( "unitY", "STRING", "") );

  Stream.WriteDictionary( FrSE( "next", PTR_STRUCT::Desc( getStaticName() ) ) );

  return;
}

void FrVect::
copy_data( CHAR_U* Data, INT_8U NBytes, bool Allocate, bool Owns )
{
  if ( m_owns )
  {
    delete[] m_data.data;
    m_data.data = (CHAR_U*)NULL;
  }
  m_data.nBytes = NBytes;
  if ( Allocate )
  {
    m_data.data = new CHAR_U[ NBytes ];
    std::copy( Data, Data + NBytes, m_data.data );
    m_owns = true;
  }
  else
  {
    m_data.data = Data;
    m_owns = Owns;
  }
}

void FrVect::
read( IFrameStream& Stream )
{
  using namespace FrameCPP::Version_3_4_5;

  std::auto_ptr< Vect >	ver5;
  PTR_STRUCT		next;
      
  switch( Stream.GetVersion( ) )
  {
  case DATA_FORMAT_VERSION:
    {
      DimBuffer	dims;
      Stream >> m_data.name
	     >> m_data.compress
	     >> m_data.type
	     >> m_data.nData
	     >> m_data.nBytes;
      if ( m_data.data )
      {
	delete[] m_data.data;
      }
      m_data.data = new CHAR_U[ m_data.nBytes ];
      Stream.read( (char*)(m_data.data), m_data.nBytes );
      Stream >> dims
	     >> m_data.unitY
	     >> next
	;
      dims.copy( m_data.dimension );
    }
    break;
  case 5:
    ver5.reset( Vect::read( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  case 4:
    ver5.reset( Vect::read4( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  case 3:
    ver5.reset( Vect::read3( Stream.GetFrame5Reader( ) ) );
    next = Stream.GetNextNode5( );
    break;
  }
  if ( ver5.get( ) )
  {
    m_data.name = ver5->mName;
    m_data.compress = ver5->mCompress;
    m_data.type = ver5->mType;
    m_data.nData = ver5->mNData;
    m_data.nBytes = ver5->mNBytes;
    m_data.data = new CHAR_U[ m_data.nBytes ];
    std::copy( ver5->mData, &(ver5->mData[ m_data.nBytes ]), m_data.data );
    m_data.dimension.resize( 0 );
    for ( std::vector< FrameCPP::Version_3_4_5::Dimension >::const_iterator
	    d( ver5->mDimension.begin( ) );
	  d != ver5->mDimension.end( );
	  d++ )
    {
      m_data.dimension.push_back( Dimension( (*d).getNx( ),
					     (*d).getDx( ),
					     (*d).getUnitX( ),
					     (*d).getStartX( ) ) );
    }
    m_data.unitY = ver5->mUnitY;
  }
       
#if LM_DEBUG
  std::cerr << "DEBUG: name: " << m_data.name << std::endl
	    << m_data.compress << std::endl
	    << m_data.type << std::endl
	    << m_data.nData << std::endl
	    << m_data.nBytes << std::endl
	    << "DEBUG: nDims: " << m_data.dimension.size( ) << std::endl
	    << "DEBUG: unitY: " << m_data.unitY << std::endl
    ;
#endif /* LM_DEBUG */
  //---------------------------------------------------------------------
  // Add the reference to the next
  //---------------------------------------------------------------------
  SetNext( next );
}

#if ! defined( __SUNPRO_CC ) || ( __SUNPRO_CC > 0x550 )
#include "framecpp/Version6/FrVect.icc"

#if ENABLE_DEPRICATED_FRAMECPP
#define	INSTANTIATE_DEPRECATED( LM_TYPE ) ;				\
  template FrVect::							\
  FrVect( const std::string& name, INT_4U nDim, const INT_4U* dims,	\
	  LM_TYPE* data, const std::string& unitY,			\
	  bool allocate, bool owns );					\
									\
  template FrVect::							\
  FrVect( const std::string& name, INT_4U nDim, const INT_4U* dims,	\
	  const LM_TYPE* data, const std::string& unitY );		\
  template INT_2U FrVect::getDataType< LM_TYPE >()   
#else
#define	INSTANTIATE_DEPRECATED( LM_TYPE )
#endif

#define	INSTANTIATE( LM_TYPE )					\
  template FrVect::						\
  FrVect( const std::string& name, INT_4U nDim,			\
	  const Dimension* dims,				\
	  const LM_TYPE* data, const std::string& unitY );	\
								\
  template FrVect::						\
  FrVect( const std::string& name, INT_4U nDim,			\
	  const Dimension* dims,				\
	  LM_TYPE* data, const std::string& unitY,		\
	  bool allocate, bool owns )				\
    INSTANTIATE_DEPRECATED( LM_TYPE )

#include "TypeInstantiation.icc"

#undef INSTANTIATE

#endif /* */
