#ifndef FrameCPP_VERSION_6_IFrameStream_HH
#define FrameCPP_VERSION_6_IFrameStream_HH

#include <memory>
#include <stdexcept>
#include <sstream>
#include <string>
#include <vector>

#include "general/types.hh"

#include "framecpp/CheckSum.hh"

#include "framecpp/Common/FrameStreamBase.hh"

#include "framecpp/Version6/Container.hh"
#include "framecpp/Version6/Dictionary.hh"
#include "framecpp/Version6/Frame5ReaderBase.hh"
#include "framecpp/Version6/FrameStreamBase.hh"
#include "framecpp/Version6/FrEndOfFile.hh"
#include "framecpp/Version6/FrEndOfFrame.hh"
#include "framecpp/Version6/FrStatData.hh"
#include "framecpp/Version6/FrTOC.hh"
#include "framecpp/Version6/PTR_STRUCT.hh"
#include "framecpp/Version6/SearchContainer.hh"
#include "framecpp/Version6/Util.hh"

namespace FrameCPP
{
  namespace Version_3_4_5
  {
    class FrameReaderBase;
    class RefHeadBase;
  }

  namespace Version_6
  {
    class FrBase;
    class FrameH;
    class FrAdcData;
    class FrEvent;
    class FrProcData;
    class FrRawData;
    class FrSerData;
    class FrSimData;
    class FrSimEvent;

    //: Class to handle frame I/O.
    class IFrameStream
      :	public FrameStreamBase,
	public Common::FrameStreamBase
    {
    public:
      static const int MODE_TOC;
      static const int MODE_DAQ;
      static const int MODE_RANDOM_ACCESS;
      static const int MODE_SEQUENTIAL;

      template< class struct_5_type, class struct_6_type >
      class MapRef
      {
      public:
	MapRef( IFrameStream& Stream,
		const struct_5_type* Struct5,
		struct_6_type& Struct6 )
	  : m_stream( Stream ),
	    m_struct_5( Struct5 ),
	    m_struct_6( Struct6 )
	{
	}

	virtual ~MapRef( )
	{
	}

        void Map();

      protected:
	IFrameStream&		m_stream;
	const struct_5_type*	m_struct_5;
	struct_6_type&		m_struct_6;

	virtual void operator()( const void* Ref5Ptr,
				 const PTR_STRUCT& PtrStruct ) = 0;
      };

      //: Constructor
      //
      //!param: std::iostream& Stream - Input or Output stream
      IFrameStream( std::istream& Stream,
		    int Mode = ( MODE_TOC | MODE_DAQ | MODE_RANDOM_ACCESS ) );

      IFrameStream( std::istream& Stream,
		    const std::vector< Dictionary::io_calls_type >& LibraryDefs,
		    INT_4U NumberOfClassTypes,
		    int Mode = ( MODE_TOC | MODE_DAQ | MODE_RANDOM_ACCESS ) );

      //: Destructor
      //
      ~IFrameStream( );

      template< class T >
      void AddContainer( FrameCPP::ContainerBasic< T >& C,
			 const PTR_STRUCT& Ref );
      
      template< class T >
      void RemoveContainer( FrameCPP::ContainerBasic< T >& C );
      
      void AddPointer( FrBase** Pointer, const PTR_STRUCT& Ref );

      void RemovePointer( FrBase** Pointer );

      //: Ensure that the read boundry is somewhere beyond the data to be read
      //!param: INT_8U Offset - Number of bytes to read
      //!return: nothing - This method will throw an exception if reading
      //+	Offset number of bytes would extend beyond the current
      //+	value of m_road_block
      void CheckReadBoundry( INT_8U Offset );

      const FrTOC* GetTOC( ) const;

      //: Retrieve the Instance of the next pointer for version 5 frames
      PTR_STRUCT GetNextNode5( );

      INT_4U GetNumberOfFrames( ) const;

      INT_8U GetPositionH( INT_4U Frame );

      //: Retrieve the originator of the frame
      const char* GetOriginator( ) const;

      ::FrameCPP::Version_3_4_5::FrameReaderBase& GetFrame5Reader( );

      Frame5ReaderBase& GetExposedFrame5Reader( );

      INT_8U GetLengthOfLastObject( ) const;

      virtual void SetChecksumMethodFile( const CheckSum::kind_type Method );

      FrameH* ReadFullFrame( INT_4U Index );

      FrameH* ReadNextFrame( );

      //: Read the base of most objects
      FrBase* ReadObjectBase( bool SkipUnknown = true );

      //: Read the next object
      FrBase* ReadObject( );

      //: Read the next object
      FrBase* ReadObject( FrBase& Base );

#if 0
      template< class T >
      IFrameStream& ReadVector( T& Data, INT_4U Size );
#endif

      template< class T >
      IFrameStream& ReadVector( std::vector< T >& Data, INT_4U Size );

      //: Establish the read boundry
      void SetReadBoundry( INT_8U Boundry = INT_8U( 0 ) );

      //: Skip over unknown objects.
      void SkipObject( const FrBase& Base ) const;

      template< class T >
      IFrameStream& operator>>( T* Data );

      //: Read container from the stream
      template< class T >
      IFrameStream& operator>>( Container< T >& Data );

      //: Read data from the stream
      template< class T >
      IFrameStream& operator>>( T& Data );

      // Read in a number of bytes
      //
      //!param: void *Dest - Pointer to buffer receiving the data.
      //!param: INT_4U Length - Number of bytes to read.
      //!param: bool CalcChecksum - true if checksum is to be calcuted,
      //+	false otherwise
      //!return: INT_4U - Number of bytes actually read.
      INT_4U read( void* Dest, INT_4U Length, bool CalcChecksum = true );

      FrameH* ReadFrameH( INT_4U Frame, INT_4U ContainerSet );

      FrAdcData* ReadFrAdcData( INT_4U Frame, const std::string& Channel );
      FrAdcData* ReadFrAdcData( INT_4U Frame, INT_4U Channel );
      FrAdcData* ReadFrAdcStruct( INT_4U Frame, const std::string& Channel );
      FrAdcData* ReadFrAdcStruct( INT_4U Frame, INT_4U Channel );

      FrEvent* ReadFrEvent( INT_4U Frame, const std::string& Channel );

      FrProcData* ReadFrProcData( INT_4U Frame, const std::string& Channel );
      FrProcData* ReadFrProcData( INT_4U Frame, INT_4U Channel );
      FrProcData* ReadFrProcStruct( INT_4U Frame, const std::string& Channel );
      FrProcData* ReadFrProcStruct( INT_4U Frame, INT_4U Channel );

      FrSerData* ReadFrSerData( INT_4U Frame, const std::string& Channel );
      FrSerData* ReadFrSerStruct( INT_4U Frame, const std::string& Channel );

      FrSimData* ReadFrSimStruct( INT_4U Frame, const std::string& Channel );
      FrSimData* ReadFrSimData( INT_4U Frame, const std::string& Channel );

      FrSimEvent* ReadFrSimEvent( INT_4U Frame, const std::string& Channel );

      //: Read FrStatData from the input file
      INT_4U ReadFrStatData( const std::string& NamePattern,
			     const std::string& DetectorPattern,
			     const General::GPSTime& StartTime,
			     const General::GPSTime& EndTime,
			     const INT_4U Version,
			     FrStatData::Query& QueryResults );
      //: Read FrStatData at a given position in the file
      std::auto_ptr< FrStatData > ReadFrStatData( INT_8U Position );

      FrRawData* ReadFrRawData( INT_4U Frame );

      void clear( );

      bool eof( ) const;

      bool good( ) const;

      INT_8U tellg( ) const;

    private:
      friend class FrEndOfFile;
      friend class FrEndOfFrame;
      friend void FrTOC::BuildDynamically( IFrameStream& );
      friend FrBase* Dictionary::Read( IFrameStream& Stream,
				       const FrBase& Base );


      // Input stream.
      std::istream&	m_stream;

      // Mode of stream
      int		m_mode;
      INT_8U		m_length_of_last_object;
      //: Do not read beyond this value
      //
      // This value is intended to be a soft marker of where the end of a
      //   structure should be
      INT_8U		m_road_block;

      // Reader for version 3, 4 and 5 frames
      Frame5ReaderBase*	m_frame_5_reader;

      std::auto_ptr< FrTOC >	m_toc;

      INT_4U		m_next_frame;

      void init( );

      FrBase* read_object( const PTR_STRUCT& Object );

      void read_recursive( );

      template< class FrData >
      std::auto_ptr< FrData > read_fr_data( INT_8U Position );

      template< int ClassId, class FrData, class Map >
      std::auto_ptr< FrData > read_fr_data( INT_4U Frame,
					    const typename Map::const_iterator Data );

      template< int ClassId, class FrData, class Map >
      FrData* read_fr_struct( INT_4U Frame,
			      const typename Map::const_iterator Data );

      FrBase* read_objects_through( INT_2U id );

      FrBase* skip_till_class( INT_2U id );
    private:
      IFrameStream( const IFrameStream& );
      const IFrameStream& operator=( const IFrameStream& );
    }; // class - IFrameStream

    template< class T >
    inline void IFrameStream::
    AddContainer( FrameCPP::ContainerBasic< T >& C, const PTR_STRUCT& Ref )
    {
      m_dictionary.m_reference_info.AddContainer( C, Ref );
    }

    inline void IFrameStream::
    CheckReadBoundry( INT_8U Offset )
    {
      if ( ( m_road_block > 0 ) &&
	   ( ( tellg( ) + Offset ) > m_road_block ) )
      {
	std::ostringstream	oss;
	oss << "Attempt to read beyond the end of the structure block";
	
	throw std::runtime_error( oss.str( ) );
      }
    }

    inline FrameH* IFrameStream::
    ReadFullFrame( INT_4U Index )
    {
      m_next_frame = Index;
      return ReadNextFrame( );
    }

    template< class T >
    inline void IFrameStream::
    RemoveContainer( FrameCPP::ContainerBasic< T >& C )
    {
      m_dictionary.m_reference_info.RemoveContainer( C );
    }

    inline void IFrameStream::
    AddPointer( FrBase** Pointer, const PTR_STRUCT& Ref )
    {
      m_dictionary.m_reference_info.AddPointer( Pointer, Ref );
    }

    inline void IFrameStream::
    RemovePointer( FrBase** Pointer )
    {
      m_dictionary.m_reference_info.RemovePointer( Pointer );
    }

    inline INT_4U IFrameStream::
    GetNumberOfFrames( ) const
    {
      if ( m_toc.get( ) )
      {
	return m_toc->GetNFrame( );
      }
      throw std::runtime_error( "GetNumberOfFrames called with no toc" );
    }

    inline const char* IFrameStream::
    GetOriginator( ) const
    {
      return m_header.GetOriginator( );
    }

    inline const FrTOC* IFrameStream::
    GetTOC( ) const
    {
      return m_toc.get( );
    }
#if 0
    //: Read search container from the stream
    template< class T, const std::string& (T::*F)() const, class C >
    inline IFrameStream& IFrameStream::
    operator >>( FrameCPP::SearchContainerBasic< T, F, C >& Data )
    {
      *this >> static_cast< C& >( Data );
    }
#endif

    inline INT_8U IFrameStream::
    GetLengthOfLastObject( ) const
    {
      return m_length_of_last_object;
    }

    inline void IFrameStream::
    SetReadBoundry( INT_8U Boundry )
    {
      m_road_block = Boundry;
    }

    inline void IFrameStream::
    clear( )
    {
      return m_stream.clear( );
    }

    inline bool IFrameStream::
    eof( ) const
    {
      return m_stream.eof( );
    }

    inline bool IFrameStream::
    good( ) const
    {
      return m_stream.good( );
    }

    inline INT_8U IFrameStream::
    tellg( ) const
    {
      return m_stream.tellg( );
    }

   
   template< class struct_5_type, class struct_6_type >   
   void IFrameStream::MapRef< struct_5_type, struct_6_type >::Map( )
   {
      Frame5ReaderBase::LinkedListStart&
        containers( m_stream.GetExposedFrame5Reader( ).mLinkedListStart );
   	  
      for ( Frame5ReaderBase::LinkedListStart::iterator i( containers.begin( ) );
    	 i != containers.end( ); ++i )
      {
         const void* ref5( (const void*)( (*i).first->getContainer( ) ) );
         PTR_STRUCT ps( (*i).second.getClass( ),
   	   	     (*i).second.getInstance( ) );
         this->operator()( ref5, ps );
      
         delete (*i).first;
         (*i).first = (Version_3_4_5::RefHeadBase*)NULL;
      }
      
      containers.resize( 0 );
      
      return;
   }   
   
    // ------------------------------------------------------------------
    // Forward declaration of template specialization
    // ------------------------------------------------------------------
    template < >IFrameStream& IFrameStream::
    ReadVector( std::vector< STRING >& Data, INT_4U Size );

  } // namespace - Version_6
} // namespace - FrameCPP

#endif /* FrameCPP_VERSION_6_IFrameStream_HH */



