#if HAVE_CONFIG_H
#include "framecpp/config.h"
#endif /* HAVE_CONFIG_H */

#include <unistd.h>

#include <list>
#include <memory>
#include <sstream>

#include "general/fstream.hh"
#include "general/gpstime.hh"
#include "general/unittest.h"

#include "framecpp/FrameH.hh"
#include "framecpp/FrDetector.hh"
#include "framecpp/FrStatData.hh"
#include "framecpp/IFrameStream.hh"
#include "framecpp/OFrameStream.hh"

General::UnitTest	Test;

using General::GPSTime;

using FrameCPP::FrameH;
using FrameCPP::FrDetector;
using FrameCPP::FrStatData;
using FrameCPP::IFrameStream;
using FrameCPP::OFrameStream;

//=======================================================================
// Testing framework
//=======================================================================
void
test( const std::string& Leader, INT_4U Results,
      IFrameStream& FrameStream,
      const std::string& NamePattern,
      const std::string& DetectorPattern,
      const GPSTime& Start, const GPSTime& Stop,
      INT_4U Version )

{
  FrStatData::Query	results;

  FrameStream.ReadFrStatData( NamePattern, DetectorPattern,
			      Start, Stop, Version,
			      results );
  Test.Check( Results == results.size( ) )
    << Leader << " Read " << results.size( ) << " of the expected " << Results
    << std::endl;

  for ( INT_4U
	  x = 0,
	  last = results.size( );
	x < last;
	++x )
  {
    delete results[ x ];
  }
}

//=======================================================================
// main -- Main testing loop
//=======================================================================
std::ostringstream	filename;

int
main( int argc, char* argv[] )
try
{
  //---------------------------------------------------------------------
  // Prepare for testing
  //---------------------------------------------------------------------
  Test.Init( argc, argv );

  GPSTime	gtime( 600000000, 0 );

  //---------------------------------------------------------------------
  // :TODO: Initialize by creating the test frame
  //---------------------------------------------------------------------
  {
    std::auto_ptr< FrameH >
      frame( new FrameH( "FrStatDataSampleFrame",
			 -1,
			 0,
			 gtime,
			 0,
			 16.0 ) );

    frame->RefDetectProc( ).
      append( new FrDetector( "d1",	// name
			      "d1",	// prefix
			      0.0,	// longitude
			      0.0,	// latitude
			      0.0,	// elevation
			      0.0,	// armXazimuth
			      0.0,	// armYazimuth
			      0.0,	// armXaltitude
			      0.0,	// armYaltitude
			      0.0,	// armXmidpoint
			      0.0,	// armYmidpoint
			      0 ),	// localTime
	      false,
	      true );
    frame->RefDetectProc( ).
      append( new FrDetector( "d2",	// name
			      "d2",	// prefix
			      2.0,	// longitude
			      2.0,	// latitude
			      2.0,	// elevation
			      2.0,	// armXazimuth
			      2.0,	// armYazimuth
			      2.0,	// armXaltitude
			      2.0,	// armYaltitude
			      2.0,	// armXmidpoint
			      2.0,	// armYmidpoint
			      2 ),	// localTime
	      false,
	      true );

    //-------------------------------------------------------------------
    // Create StatData information
    //-------------------------------------------------------------------
    std::list< FrStatData > stat_data;

    stat_data.push_back( FrStatData( "SimpleStat", 	// name
				     "This is simple",	// comment
				     "No repr.",	// representation
				     gtime.GetSeconds( ),// timeStart
				     gtime.GetSeconds( ) + 999, // timeEnd
				     1 ) );		// version
    stat_data.back( ).SetDetector( frame->RefDetectProc( )[ 0 ] );

    stat_data.push_back( FrStatData( "SimpleStat", 	// name
				     "This is simple",	// comment
				     "No repr.",	// representation
				     gtime.GetSeconds( ),// timeStart
				     gtime.GetSeconds( ) + 999, // timeEnd
				     1 ) );		// version
    stat_data.back( ).SetDetector( frame->RefDetectProc( )[ 1 ] );

    stat_data.push_back( FrStatData( "SimpleStat", 	// name
				     "This is simple",	// comment
				     "No repr.",	// representation
				     gtime.GetSeconds( ) + 1000,// timeStart
				     gtime.GetSeconds( ) + 1999, // timeEnd
				     1 ) );		// version
    stat_data.back( ).SetDetector( frame->RefDetectProc( )[ 0 ] );

    stat_data.push_back( FrStatData( "SimpleStat", 	// name
				     "This is simple",	// comment
				     "No repr.",	// representation
				     gtime.GetSeconds( ) + 1000,// timeStart
				     gtime.GetSeconds( ) + 1999, // timeEnd
				     1 ) );		// version
    stat_data.back( ).SetDetector( frame->RefDetectProc( )[ 1 ] );

    stat_data.push_back( FrStatData( "SimpleStat", 	// name
				     "This is simple",	// comment
				     "No repr.",	// representation
				     gtime.GetSeconds( ) + 1000,// timeStart
				     gtime.GetSeconds( ) + 1999, // timeEnd
				     2 ) );		// version
    stat_data.back( ).SetDetector( frame->RefDetectProc( )[ 1 ] );
    //-------------------------------------------------------------------
    // Create output file
    //-------------------------------------------------------------------
    filename << "Z-FrStatDataFrame-"
	     << frame->GetGTime( ).GetSeconds( ) 
	     << "-16.gwf";
    unlink( filename.str( ).c_str( ) );
    
    General::ofstream out( filename.str( ).c_str( ) );
    OFrameStream fw( out );

    for ( std::list< FrStatData >::iterator
	      cur = stat_data.begin( ),
	      last = stat_data.end( );
	  cur != last;
	  ++cur )
    {
	fw.WriteFrStatData( *cur );
    }

    fw.WriteFrame( *frame );

    fw.close( );
    out.close( );

  }

  //---------------------------------------------------------------------
  // Testing
  //---------------------------------------------------------------------
  General::ifstream in( filename.str( ).c_str( ) );
  IFrameStream fr( in );

  test( "Retrieval of all All",
	5, fr, "^.*$", "^.*$",
	GPSTime( 0, 0 ), GPSTime( 0, 0 ),
	FrStatData::ALL_VERSIONS );

  test( "Retrieval of all d1",
	2, fr, "^.*$", "^d1$",
	GPSTime( 0, 0 ), GPSTime( 0, 0 ),
	FrStatData::ALL_VERSIONS );

  test( "Retrieval of all d2",
	3, fr, "^.*$", "^d2$",
	GPSTime( 0, 0 ), GPSTime( 0, 0 ),
	FrStatData::ALL_VERSIONS );

  test( "Retrieval for time 1000 - 1999",
	3, fr, "^.*$", "^.*$",
	gtime + 1000, gtime + 1999,
	FrStatData::ALL_VERSIONS );

  test( "Retrieval of latest",
	4, fr, "^.*$", "^.*$",
	GPSTime( 0, 0 ), GPSTime( 0, 0 ),
	FrStatData::LATEST_VERSION );

  test( "Retrieval of latest d2",
	2, fr, "^.*$", "^d2$",
	GPSTime( 0, 0 ), GPSTime( 0, 0 ),
	FrStatData::LATEST_VERSION );

  test( "Retrieval of any interferometer with a single leter name",
	0, fr, "^.*$", "^.$",
	GPSTime( 0, 0 ), GPSTime( 0, 0 ),
	FrStatData::ALL_VERSIONS );

  in.close( );
  

  //---------------------------------------------------------------------
  // Exit with the proper exit status
  //---------------------------------------------------------------------
  Test.Exit( );
} catch (...) {
  Test.Check( false ) << "Caught an exception" << std::endl;
  Test.Exit( );
}
