/*
  frcat -- Concatenate frame files.
*/

#if HAVE_CONFIG_H
#include "framecpp/config.h"
#endif /* HAVE_CONFIG_H */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>

#include <fstream>
#include <iostream>
#include <sstream>
#include <vector>

#include "framecpp/Version3_4_5/frame.hh"
#include "framecpp/Version3_4_5/framewritertoc.hh"
#include "framecpp/Version3_4_5/framereader.hh"
#include "framecpp/Version3_4_5/tocreader.hh"
#include "framecpp/Version3_4_5/errors.hh"

#include "framedir.hh"

using namespace FrameCPP::Version_3_4_5;
using namespace std;

// The name of the program executable
char* programname = 0;

// Input file wilcard pattern
const char* ifdir = 0;

// input frame directory
FrameDir *fdir=0;

// output frame file name
const char *ofname=0;

// output frame file
std::ostream *ofs=0;

// required ADCs
std::vector<char*> adcs;

bool verbose = false;

// wipe the output file and exit
extern "C" {
  void cleanup(int foo) { unlink(ofname); exit(1);}
}

void
fatal( std::string msg ) {
  std::cerr << msg << std::endl;
  cleanup(0);
}

void
error_watch (const std::string& msg)
{
  std::cerr << "error: " << msg << std::endl;
}

using namespace std;

void
usage( char* f )
{
  if ( ! strncasecmp( f, "-h", 2 ) 
       || ! strncasecmp( f, "--h", 3 ) )
    {
      cout << "Concatenate frame files." << endl;
      cout << "Optionally select required ADCs." << endl;
      cout << "Usage: " << (string)programname << " [-d frame_wildcard] [ -o outframe ] [ADC1 ... ADCN]" << endl;
      cout << "Where `frame_wildcard' is a frame file shell wildcard pattern,\n"
	   << "for which the default is `*', ie. all files in current directory.\n"
	   << "`outframe' is an output frame file name (frame.out is the dafault).\n";
      cout << "Regular expressions in the ADC names are supported." << endl;
      cout << "Empty list of the ADC channels means concatenate all data." << endl;
      exit(0);
    }
}


int
main( int argc, char* argv[] )
{  
  /* Determine program name (strip filesystem path) */
  for (programname = argv [0] + strlen (argv [0]) - 1;
       programname != argv [0] && *programname != '/';
       programname--)
    ;

  if (*programname == '/') 
    programname++;

  //  usage help
  if (argc > 1)
    usage(argv[1]);

  bool d_seen = false;

  // clean up the output file on the termination
  signal(SIGTERM, cleanup);
  signal(SIGINT, cleanup);

  // parse command line arguments
  for (int i = 1; i < argc; i++) {
    if (!strcmp(argv[i],"-d")) {
	d_seen = true;
	i++;
	if (i < argc)
	  fdir = new FrameDir(ifdir = argv[i]);
    } else if (!strcmp(argv[i],"-o")) {
 	i++;
	if (i < argc) {
#if !old
	  if ( strcmp( argv[ i ], "-" ) == 0 )
	  {
	    ofs = &( std::cout );
	  }
	  else
	  {
	    ofs = new std::ofstream( argv[ i ] );
	  }
#else
    	  int ofd = strcmp(argv[i],"-")?
			open( argv[i], O_CREAT|O_WRONLY|O_TRUNC, 0644 ):1;
  	  if ( ofd < 0 )
      		fatal( "could not open `" + (string)argv[i] + "' for output" );
#if __GNUC__ >= 3
    FILE *cfile = fdopen(ofd,"w");
    std::ofstream::__filebuf_type* outbuf
        = new std::ofstream::__filebuf_type (cfile, std::ios_base::out);
    std::ofstream::__streambuf_type* oldoutbuf
        = ((std::ofstream::__ios_type* )ofs)->rdbuf (outbuf);
    delete oldoutbuf;
#else
	  ofs = new std::ofstream(ofd);
#endif
#endif
	  ofname = argv[i];
	}
    } else if (!strcmp(argv[i],"-v")) {
	verbose = true;
    } else {
	// ADCs spec
	for (; i < argc; i++)
	  adcs.push_back(argv[i]);
    }
  }
  
  // input file glob pattern default to the star
  if (!d_seen)
    fdir = new FrameDir(ifdir = "*");

  // bail out if nothing matched
  if ( !fdir || fdir->end() == fdir->begin())
    fatal("no input frame files matched `" + std::string(ifdir) + "'");
  
  // output frame file default
  if (!ofs)
    ofs = new ofstream(ofname="frame.out");
     
  FrameWriterTOC fw( *ofs );

  // cat all the frames in all the files
  for (FrameDir::file_iterator iter = fdir->begin();
       iter != fdir->end(); iter++) { // over all frame files
    if (verbose)
    	cout << iter->second.getFile() << endl;
    ifstream in(iter->second.getFile());
    FrameReader fr( in );
    fr.setErrorWatch( error_watch );
    try {
      for(;;) { // over all frame in the file
        Frame* frame = fr.readFrame();
        if (!frame)
	  break;
  	if (adcs.size()) {
	  // wipe out all of the ADC that are not requested
	  RawData *new_rawdata = new RawData;
	  RawData::AdcDataContainer &sc
	 		= frame->getRawData()->refAdc();
	  for (unsigned int i = 0; i < adcs.size(); i++) {
	    if (verbose)
		cerr << "Searching for " << adcs[i] << endl;
	    RawData::adcData_iterator iter = sc.begin();
	    for (;;) {
	       iter = sc.regexFind(adcs[i], iter);
	       if (iter == sc.end())
		 break;
	       new_rawdata->refAdc().append(*iter);
	       if (verbose)
	       	cerr << "\t" << (*iter)->getName() << endl;
	       iter++;
	    }
	  }
	  frame->setRawData(new_rawdata,false);
	}
        fw.writeFrame( *frame );
        delete frame;
      }
    } catch ( read_failure& e ) {
      if ( e.code() != fr.ENDOFFILE )
	fatal( e.what() );
    }
  }
  fw.close();   // build TOC
  if ( dynamic_cast< std::ofstream* >( ofs ) )
  {
    dynamic_cast< std::ofstream* >(ofs)->close();
  }
  return 0;
}
