/*
  Create and write out several frame files. These frames
  are to be used for testing various LDAS functions. Specifically,
  we would like to be able to test currently unused frame data
  structures before they appear in frame builder written frames.
*/

#include "framecpp/config.h"

#include <stdlib.h>
#include<unistd.h>

#include <fstream>
#include <iostream>
#include <memory>
#include <sstream>

#include "framecpp/Dimension.hh"
#include "framecpp/FrameH.hh"
#include "framecpp/FrameWriterImage.hh"
#include "framecpp/FrameWriterTOC.hh"
#include "framecpp/FrAdcData.hh"
#include "framecpp/FrDetector.hh"
#include "framecpp/FrEvent.hh"
#include "framecpp/FrHistory.hh"
#include "framecpp/FrMsg.hh"
#include "framecpp/FrProcData.hh"
#include "framecpp/FrRawData.hh"
#include "framecpp/FrSerData.hh"
#include "framecpp/FrSimData.hh"
#include "framecpp/FrSimEvent.hh"
#include "framecpp/FrStatData.hh"
#include "framecpp/FrSummary.hh"
#include "framecpp/FrTable.hh"
#include "framecpp/FrVect.hh"
#include "framecpp/Time.hh"

using namespace FrameCPP;

char *stdTestFrameName = "std_test_frame";
bool ramp = false;
double amplitude = 1.0;
bool use_image_writer = false;
// First frame time stamp
unsigned long secini = 666666667;
// Frames per file
unsigned long fpf = 1;
// Frame length
unsigned long dt = 1;
// Sequence length
unsigned long Dt = 10;

FrameH*
makeFrame(unsigned long gps, unsigned int frame_number, unsigned long dt)
{
  std::auto_ptr<FrameH> fr( new FrameH( "LIGO",
					7, /* run */
					frame_number, /* frame number */
					Time( gps, 0 ),
					13,
					(double)dt ) );

  FrameCPP::FrRawData	rawData( "" );

  fr->setRawData( rawData );
  if (ramp)
  {
    FrAdcData b( "ramp",
		 0, 1, 64,
		 16.0*1024.0, 0.0, 1.00,
		 "counts",
		 0.0,
		 Time( 0, 0 ),
		 1 );
    fr->getRawData( )->refAdc( ).append( b );
    {
      std::auto_ptr< REAL_8 >	d( new REAL_8[ 16 * 1024 * dt] );
      for (unsigned long j = 0; j < dt; j++)
      {
	for ( unsigned long i = 0 ; i < 16 * 1024; i++)
	{
	  (d.get())[j*16*1024 + i] =
	    amplitude * (((REAL_8)i)/(16.0*1024.0));
	}
      }
      FrameCPP::Dimension ddims[ 1 ] = {
        FrameCPP::Dimension( 16*1024*dt, 1.00/(16*1024*dt), "time", 1.00 )
      };
      FrameCPP::FrVect dvect( "ramp", 1, ddims, d.get( ), "counts" );

      // AdcData
      fr->getRawData()->refAdc()[0]->refData().append( dvect );


      // ProcData
      FrameCPP::FrProcData p( "ramp", 16.0*1024.0, Time( 0, 0 ), .00, .0 );
      fr->refProcData().append( p );
      fr->refProcData()[0]->refData().append( dvect );

      // SimData
      FrameCPP::FrSimData sd( "ramp", "", 16.0*1024.0, .0, .0 );
      fr->refSimData().append( sd );
      fr->refSimData()[0]->refData().append( dvect );
    }
  } else {

    // Detector (both detectors are the same)
    //	
    // :TODO: put real Hanford values in here
    FrameCPP::FrDetector d("TesT",
			 1.0, 2.0,
			 .0,
			 .0, .0,
			 .0, .0,
			 .0, .0,
			 3600, 0, "");
    fr->refDetectorSim().append( d );
    fr->refDetectorProc().append( d );

    // History 
    fr->refHistory().append( FrameCPP::FrHistory( "History teaches us lessons", 666666777, "absolutely" ));
    fr->refHistory().append( FrameCPP::FrHistory( "...", 666666888, "x" ));

    // AdcData
    FrameCPP::FrAdcData a( "TesT", 2, 2, 2, 2.00, 1.00, 1.00, "TesT", 2.00,
			 Time(-2, 2), 1 );
    a.appendComment( "TesT" );
    fr->getRawData()->refAdc().append( a );
    FrameCPP::FrAdcData b( "TesT1", 0, 1, 32, 16.00, 1.00, 1.00, "Volts", .0,
			 Time(0,0), 1 );
    fr->getRawData()->refAdc().append( b );
    
    // append long data vector to the seconds ADC
    {
      REAL_4 d[ 16* 32 ];
      for ( int i = 0 ; i < 16 * 32; i++)
	d[i]= (REAL_4)i;
      FrameCPP::Dimension ddims[ 1 ] = {
        FrameCPP::Dimension( 16*32, 1.00, "", 1.00 )
      };
      FrameCPP::FrVect dvect( "a", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[1]->refData().append( dvect );
    }

    // character vector
    {
      CHAR d[ 4 ] = { 'a', 'a', 'a', 'a' };
      FrameCPP::Dimension ddims[ 1 ] = {
	FrameCPP::Dimension( 2, 2.00, "TesT", 2.00 )
      };
      FrameCPP::FrVect dvect( "TesT", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );
    }

    // unsigned character vector
    {
      CHAR_U d[ 4 ] = { 'b', 'b', 'b', 'b' };
      FrameCPP::Dimension ddims[ 1 ] = {
	FrameCPP::Dimension( 2, 2.00, "TesT", 2.00 )
      };
      FrameCPP::FrVect dvect( "TesT", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );
    }

    // short vector
    {
      INT_2S d[ 4 ] = { -1, -1, -1, -1 };
      FrameCPP::Dimension ddims[ 1 ] = {
	FrameCPP::Dimension( 2, 2.00, "TesT", 2.00 )
      };
      FrameCPP::FrVect dvect( "TesT", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );
    }

    // unsigned short vector
    {
      INT_2U d[ 4 ] = { 1, 1, 1, 1 };
      FrameCPP::Dimension ddims[ 1 ] = {
	FrameCPP::Dimension( 2, 2.00, "TesT", 2.00 )
      };
      FrameCPP::FrVect dvect( "TesT", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );
    }

    // int vector
    {
      INT_4S d[ 4 ] = { -2, -2, -2, -2 };
      FrameCPP::Dimension ddims[ 1 ] = {
	FrameCPP::Dimension( 2, 2.00, "TesT", 2.00 )
      };
      FrameCPP::FrVect dvect( "TesT", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );
    }

    // unsigned int vector
    {
      INT_2U d[ 4 ] = { 2, 2, 2, 2 };
      FrameCPP::Dimension ddims[ 1 ] = {
	FrameCPP::Dimension( 2, 2.00, "TesT", 2.00 )
      };
      FrameCPP::FrVect dvect( "TesT", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );
    }

    // real vector
    {
      REAL_4 d[ 4 ] = { 1.00, 1.00, 1.00, 1.00 };
      FrameCPP::Dimension ddims[ 1 ] = {
	FrameCPP::Dimension( 2, 2.00, "TesT", 2.00 )
      };
      FrameCPP::FrVect dvect( "TesT", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );
    }

    // real vector
    {
      REAL_8 d[ 4 ] = { 2.00, 2.00, 2.00, 2.00 };
      FrameCPP::Dimension ddims[ 1 ] = {
	FrameCPP::Dimension( 2, 2.00, "TesT", 2.00 )
      };
      FrameCPP::FrVect dvect( "TesT", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );
    }

    {
      COMPLEX_8 cdata[ 4 ];
      cdata[0] = COMPLEX_8(1.0, 1.2);
      cdata[1] = COMPLEX_8(1.3, 1.4);
      cdata[2] = COMPLEX_8(1.5, 1.6);
      cdata[3] = COMPLEX_8(1.7, 1.8);
      FrameCPP::Dimension cdims[ 1 ] = {
	FrameCPP::Dimension( 2, 2.00, "TesT", 2.00 )
      };
      FrameCPP::FrVect cvect( "TesT", 1, cdims, cdata, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( cvect );
    }

    {
      COMPLEX_16 cdata[ 4 ];
      cdata[0] = COMPLEX_16(1.0, 1.2);
      cdata[1] = COMPLEX_16(1.3, 1.4);
      cdata[2] = COMPLEX_16(1.5, 1.6);
      cdata[3] = COMPLEX_16(1.7, 1.8);
      FrameCPP::Dimension cdims[ 1 ] = {
	FrameCPP::Dimension( 2, 2.00, "TesT", 2.00 )
      };
      FrameCPP::FrVect cvect( "TesT", 1, cdims, cdata, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( cvect );
    }

    // Log Message (a couple of them)
    FrameCPP::FrMsg m( "TesT", "TesT", 2, Time(0, 0) );
    fr->getRawData()->refMsg().append( m );
    fr->getRawData()->refMsg().append( m );

    // ProcData
    FrameCPP::FrProcData p( "TesT", 2.00, Time( 2, 2 ), 2.00, .0 );
    p.appendComment("TesT");
    fr->refProcData().append( p );
    fr->refProcData().append( p );

    // SimData
    FrameCPP::FrSimData sd( "TesT", "TesT", 1.00, .0, .0 );
    fr->refSimData().append( sd );
    fr->refSimData().append( sd );

    // SerData
    FrameCPP::FrSerData srd( "TesT", Time( 2, 2 ), 1.00 );
    fr->getRawData()->refSerial().append( srd );
    fr->getRawData()->refSerial().append( srd );

    // StatData
    FrameCPP::FrStatData statData("TesT", "TesT", "TesT", 2, 2, 2 );
    // real vector
    {
      REAL_8 d[ 4 ] = { 2.00, 2.00, 2.00, 2.00 };
      FrameCPP::Dimension ddims[ 1 ] = {
	FrameCPP::Dimension( 2, 2.00, "TesT", 2.00 )
      };
      FrameCPP::FrVect dvect( "TesT", 1, ddims, d, "TesT" );
      statData.refData().append( dvect );
    }

    statData.setDetector(fr->refDetectorSim()[0]);
    fr->refDetectorSim()[0]->refStatData().append( statData );
    fr->refDetectorSim()[0]->refStatData().append( statData );

    statData.setDetector(fr->refDetectorProc()[0]);
    fr->refDetectorProc()[0]->refStatData().append( statData );
    fr->refDetectorProc()[0]->refStatData().append( statData );

    // Event
    FrameCPP::FrEvent event(
        "TesT", "TesT", "TesT", Time( 2, 2 ),
	1.0, 1.0,
	2, 1.00,
        1.00, "TesT",
	std::vector<std::pair<std::string, REAL_4> >());
    fr->refEvent().append( event );
    fr->refEvent().append( event );

    // Summary
    FrameCPP::FrSummary summary( "TesT", "TesT", "TesT", Time(0, 0) );
    fr->refSummary().append( summary );
    fr->refSummary().append( summary );

    // SimEvent
    FrameCPP::FrSimEvent sed( "TesT", "TesT", Time(2, 2), 1.00, 1.00, 1.00 );
    sed.appendComment("TesT");
    fr->refSimEvent().append( sed );


    // Table
    FrameCPP::FrTable table( "TesT", 1 );
    table.appendComment("TesT");
    fr->refTable().append( table );
  }
  return fr.release( );
}

void
usage()
{
  cout << endl << "Function: Create a sequence of frame files with meaningfull data in them." << endl;
  cout << "The purpose of this is to allow testing of new frame structures in the LDAS" << endl;
  cout << "before the frame builder is modified to write them out. Thus we are" << endl;
  cout << "trying to achive a smooth LDAS transition to new data in the frames." << endl;
  cout << endl << "Usage: takes numeric argument which is start GPS time" << endl;
  cout << endl << "The following options are supported:" << endl;
  cout << '\t' << "-r\t-generate ramp data" << endl;
  cout << '\t' << "-a amplitude\t-set ramp signal amplitude" << endl;
  cout << '\t' << "-i gps\t-set initial frame timestamp (default values is " << secini << endl;
  cout << '\t' << "-f nframes\t-set the number of frames per file (default is 1)" << endl;
  cout << '\t' << "-l dt\t-set the number of seconds of data per frame (default is 1)" << endl;
  cout << '\t' << "-p Dt\t-set overall length in seconds" << endl;
  cout << '\t' << "-I\t-use image frame writer" << endl;
}


// :TODO: include same data as in ADC into Proc and Sim structures.

int
main(int argc, char* argv[]) try
{
  int c;
  extern char *optarg;
  extern int optind;
  bool errflg = false;
  while ((c = getopt(argc, argv, "i:a:f:l:p:?hrI")) != EOF) {
    switch (c) 
      {
      case 'I':
	use_image_writer = true;
	break;
      case 'a':
	{
	  char *endptr;
	  double amp = strtod(optarg, &endptr);
	  if (endptr == optarg) {
	    cerr << "Invalid numeric argument given to -" << c <<" option" << endl;
	    errflg = true;
	  } else
	    amplitude = amp;
	  break;
	}
      case 'r': {
	ramp = true;
	break;
      }
      case '?':
      case 'h':
	errflg = true;
	break;
      case 'f': 
      case 'l': 
      case 'p':
      case 'i':
	{
	  char *endptr;
	  unsigned long f = strtol(optarg, &endptr, 10);
	  if (endptr == optarg || f == 0) {
	    cerr << "Invalid numeric argument given to -" << c <<" option" << endl;
	    errflg = true;
	  } else {
	    switch (c)
	      {
	      case 'f':
		fpf = f;
		break;
	      case 'l': 
		dt = f;
		break;
	      case 'p':
		Dt = f;
		break;
	      case 'i':
		secini = f;
		break;
	      }
	  }
	    break;
	}
      }
  }
  if (errflg || optind < argc) {
    usage();
    exit (2);
  }

  unsigned long nframes = Dt/dt;
  if (nframes < 1)
    {
      cerr << "Invalid number of frames (Dt/dt)" << endl;
      exit(1);
    }

  std::strstream ost;
  int image_size = 0;
  FrameWriterImage fw(ost);
  for (unsigned int i = 0; i < nframes; i += fpf) {
    std::ostringstream	filename;
    unsigned long tst = secini+i*dt;
    filename << "H-" << tst << "-" << fpf << "-" << dt << ".F";
    std::ofstream out( filename.str( ).c_str( ) );
    if (use_image_writer) {
      if (ost.pcount( ) == 0) {
        for (unsigned int j = 0; j < fpf; j++) {
          FrameH* frame = makeFrame(tst + j*dt, j + i, dt);
          fw.writeFrame( *frame );
          delete frame;
        }
        fw.close();
        image_size = ost.pcount( );
      } 
      fw.setFrameFileAttributes(7, i, 0, tst, dt, 0, 13, 8);
      out.write( ost.str(), image_size );
    } else {
      FrameWriterTOC fw( out );

      for (unsigned int j = 0; j < fpf; j++) {
        FrameH* frame = makeFrame(tst + j*dt, j + i, dt);
        fw.writeFrame( *frame );
        delete frame;
      }
      fw.close();
    }
    out.close();
  }
  return 0;
} catch (...) {
  return 1;
}
