/* -*- mode: c++; c-basic-offset: 2; -*- */
#ifndef GENERAL__SINGLETON_HH
#define GENERAL__SINGLETON_HH

#include "general/mutexlock.hh"

//!ignore_begin:
// This provides the definition for the singleton portion of the class
//
// The only interface exposed to the developer is the Instance() function.
//
// The default constructor being private ensures that this class cannot
//   be used as the base of any other class.
//
// The copy constructor being private prevents asignment.
//
// An example of the use of this class would be:
//
// class a
// {
//  SINGLETON_DECL( a ); // Important declarations for a singleton
//
//  public:
//    void x( );  // Method to do some work
//
//  private:
//    // Data members and other methods
// };
//
// SINGLETON_INST( a ); // Create code for Instance( ) method.
//
// a::
// a( )
// {
//   // Do initialization code here
// }
//
// void a::
// x( )
// {
//   // Do something here
// }
#define SINGLETON_DECL( Class_ )	\
public:					\
  static Class_& Instance( );		\
private:				\
  Class_( );				\
  Class_( const Class_& );		\
					\
  static void singleton_suicide( );     \
  static Class_* xxyyzz_singleton_instance

// Providing a seperate mutex lock for each singleton prevents most cases
//    of deadlock. The only case that still exists (in its simplest form)
//    is if class A requires class B and class B requires class A and
//    both class A is a singleton. This case, however, is bad coding and
//    needs to be avoided.
//
#define	SINGLETON_TS_DECL( Class_ ) \
SINGLETON_DECL( Class_ ); \
static pthread_mutex_t	xxyyzz_singleton_lock

// This provides the code for the singleton portion of the class
#define	SINGLETON_INST( Class_ )		\
						\
Class_& Class_::Instance( )			\
{						\
  if ( _instance == 0 )				\
  {						\
    xxyyzz_singleton_instance = new Class_;	\
  }						\
  return *xxyyzz_singleton_instance;		\
}						\
						\
void Class_::singleton_suicide( )               \
{						\
  delete xxyyzz_singleton_instance;		\
  xxyyzz_singlton_instance = (Class_*)NULL;	\
}						\
						\
Class_* Class_::_single_instance = 0

// This provides the code for the singleton portion of the class
#define	SINGLETON_TS_INST( Class_ )			\
							\
Class_& Class_::Instance( )				\
{							\
  if ( xxyyzz_singleton_instance == 0 )			\
  {							\
    MutexLock lock( xxyyzz_singleton_lock );		\
							\
    if ( xxyyzz_singleton_instance == 0 )		\
    {							\
      xxyyzz_singleton_instance = new Class_;		\
    }							\
  }							\
  return *xxyyzz_singleton_instance;			\
}							\
							\
void Class_::singleton_suicide( )                      	\
{							\
  MutexLock lock( xxyyzz_singleton_lock );		\
							\
  delete xxyyzz_singleton_instance;			\
  xxyyzz_singleton_instance = (Class_*)NULL;		\
}							\
							\
Class_* Class_::xxyyzz_singleton_instance = 0;		\
pthread_mutex_t Class_::xxyyzz_singleton_lock = PTHREAD_MUTEX_INITIALIZER
  
//!ignore_end:
#endif /* GENERAL__SINGLETON_HH */
