/* -*- mode: c++; c-basic-offset: 2; -*- */

#include "general/config.h"

#include <signal.h>

#include <iostream>
#include <typeinfo>

#include "general/Task.hh"
#include "general/TaskThread.hh"

namespace General
{
  TaskThread::
  TaskThread( )
    : m_done( false ),
      m_task( (Task*)NULL ),
      m_task_type_name( (const char*)NULL ),
      m_cancel_signal( CANCEL_SIGNAL_UNINITIALIZED ),
      m_state( TASK_THREAD_NOT_STARTED )
  {
  }

  TaskThread::
  ~TaskThread( )
  {
    m_done = true;
    {
      m_task_available.Broadcast( );
      if ( m_task )
      {
	//---------------------------------------------------------------
	// Ensure exclusive rights to critical data structures
	//---------------------------------------------------------------
	MutexLock lock( Mutex( ) );
	//---------------------------------------------------------------
	// Release any work that has been scheduled.
	//---------------------------------------------------------------
	m_task = (Task*)NULL;
	//---------------------------------------------------------------
	// Let the thread know that it should not accept any more work.
	//---------------------------------------------------------------
      }
      //-----------------------------------------------------------------
      // Wake up the thread if it is waiting for for to do.
      //-----------------------------------------------------------------
      m_task_available.Broadcast( );
    }
  }

  void TaskThread::
  AddTask( Task* TaskToDo )
  {
    {
      //-----------------------------------------------------------------
      // Ensure exclusive rights to critical data structures
      //-----------------------------------------------------------------
      MutexLock lock( MutexLock( ) );
      //-----------------------------------------------------------------
      // Assign next task.
      //-----------------------------------------------------------------
      m_task = TaskToDo;
    }
    //-------------------------------------------------------------------
    // Inform thread of the need to do work
    //-------------------------------------------------------------------
    m_task_available.Signal( );
  }

  void TaskThread::
  CriticalSection( bool IsCriticalSection )
  {
    if ( IsCriticalSection )
    {
      MutexLock::Lock( Mutex( ) );
    }
    else
    {
      MutexLock::UnLock( Mutex( ) );
    }
  }

  void TaskThread::
  TimedOut( )
  {
    CriticalSection( true );
    if ( m_state == TASK_THREAD_RUNNING )
    {
      m_state = TASK_THREAD_TIMEDOUT; 
   }
    CriticalSection( false );
  }

  bool TaskThread::
  Halt( ) const
  {
    bool retval = true;
    switch( m_cancel_signal )
    {
    case CANCEL_SIGNAL_UNINITIALIZED:
      // Task never got started in loop
      break;
    case CANCEL_SIGNAL_ABANDON:
      if ( m_state == TASK_THREAD_TIMEDOUT )
      {
	// The task took too long. It is time to orphan the thread so
	//   it will complete in the backgroiund
	m_done = true;
	retval = false;
	Detach( );
      }
      break;
    case CANCEL_SIGNAL_CANCEL:
      //-----------------------------------------------------------------
      // Cancellation model
      //-----------------------------------------------------------------
      Cancel( );
    default:
      //-----------------------------------------------------------------
      // Signal model
      //-----------------------------------------------------------------
      Kill( m_cancel_signal );
      break;
    }
    return retval;
  }

  void TaskThread::
  Wait( )
  {
    MutexLock	lock( Mutex( ) );

    if ( m_task == (Task*)NULL )
    {
      return;
    }
    m_task_available.Wait( );
  }

  // This method does the work of the thread and runs in the controlled thread.
  // It does a task and then waits for another task.
  // This allows for the thread to be reused.
  void TaskThread::
  run( )
  {
    bool delete_on_completion = false;
    for( ;; )
    {
      CriticalSection( true );
      m_state = TASK_THREAD_SLEEPING;
      while( (m_task == (Task*)NULL  )
	     && ( !m_done ) )
      {
	//---------------------------------------------------------------
	// Waiting for something to do
	//---------------------------------------------------------------
	m_task_available.Wait( );
      }
      if ( m_done )
      {
	//---------------------------------------------------------------
	// There is no more work for this thread. It is time to
	//   shut down the thread. Before doing so, release critical
	//   resources.
	//---------------------------------------------------------------
	m_state = TASK_THREAD_EXITING;
	CriticalSection( false );
	pthread_exit( NULL );
      }
      //-----------------------------------------------------------------
      // Some task is being requested.
      //-----------------------------------------------------------------
      m_state = TASK_THREAD_RUNNING;
      Task* t = m_task;

      m_cancel_signal = t->CancelSignal( );
      sigset_t	old_signal_set;

      if ( m_cancel_signal > 0 )
      {
	sigset_t	signal_set;

	sigemptyset( &signal_set );
	sigaddset( &signal_set, m_cancel_signal );

	//---------------------------------------------------------------
	// Preserve the old signal mask
	//---------------------------------------------------------------
	pthread_sigmask( SIG_SETMASK, NULL, &old_signal_set );
	//---------------------------------------------------------------
	// Set the new signal mask
	//---------------------------------------------------------------
	pthread_sigmask( SIG_UNBLOCK, &signal_set, NULL );
      }
      //-----------------------------------------------------------------
      // Perform the request
      //-----------------------------------------------------------------
      CriticalSection( false );
      try
      {
        if ( t )
        {
	  m_task_type_name = typeid( *t ).name( );
	  delete_on_completion = t->DeleteOnCompletion( );
          (*t)( );
        }
      }
      catch( ... )
      {
	//--------------------------------------------------------------
	// Ignore any exception trown by the thread since there
	//   is nothing that can be done about it.
	// :TODO: capture and rethrow exceptions.
	//--------------------------------------------------------------
      }
      {
	MutexLock lock( Mutex( ) );

	if ( m_state == TASK_THREAD_RUNNING )
	{
	  //-------------------------------------------------------------
	  // Normal completion
	  //-------------------------------------------------------------
	  m_state = TASK_THREAD_FINISHED;
	}
	  
	if ( m_cancel_signal > 0 )
	{
	  //-------------------------------------------------------------
	  // Restore the signal mask
	  //-------------------------------------------------------------
	  pthread_sigmask( SIG_SETMASK, &old_signal_set, NULL );
	}
	//---------------------------------------------------------------
	// Call Task Completion code.
	//---------------------------------------------------------------
	if ( t )
	{
	  if ( delete_on_completion )
	  {
	    //-----------------------------------------------------------
	    // Caller has entrusted us to clean up the memory.
	    //-----------------------------------------------------------
	    t->OnCompletion( m_state );
	    delete t;
	  }
	}
	m_cancel_signal = CANCEL_SIGNAL_UNINITIALIZED;
	delete_on_completion = false;
	if ( m_task == t )
	{
	  m_task = (Task*)NULL;
	}
      }
      m_task_available.Signal( );
    }
  }
}
