/* -*- mode: c++; c-basic-offset: 4; -*- */

#ifndef GeneralErrorInfoHH
#define GeneralErrorInfoHH


// System Includes
#include <string>

//-----------------------------------------------------------------------------
//
//: Error Information
//
// This class stores information about an error:
//
// <p>
//   <li>mLibrary - The library in which the error occurred (int).</li>
//   <li>mCode    - The error code (int)</li>
//   <li>mMsg     - A string representation of the error code (string).</li>
//   <li>mInfo    - Additional information specific to the error (string)</li>
//   <li>mFile    - The file in which the error was thrown (const char*)</li>
//   <li>mLine    - The line on which the error was thrown (size_t)</li>
//
//
// <p>This class is used by the LdasException class to store information about
// a thrown exception.
//
class ErrorInfo
{
public:

    /* Constructors */
    ErrorInfo(
        int library, int code, const std::string& msg, const std::string& info = "",
        const char* file = 0, size_t line = 0 );
    ErrorInfo( const ErrorInfo& e );

    /* Destructor */
    ~ErrorInfo();
   
    /* Operator Overloads */
    const ErrorInfo& operator=( const ErrorInfo& e );
    bool operator==( const ErrorInfo& e ) const;
    bool operator!=( const ErrorInfo& e ) const;    
    
    /* Accessors */
    int getLibrary() const;
    int getCode() const;
    const std::string& getMessage() const;
    const std::string& getInfo() const;
    const char* getFile() const;
    size_t getLine() const;

private:

    int mLibrary;
    int mCode;
    std::string mMsg;
    std::string mInfo;
    char* mFile;
    size_t mLine;
};



//-----------------------------------------------------------------------------
//
//: Not-Equal Comparison
//
//!param: const ErrorInfo& e
//
//!return: bool
//
inline bool ErrorInfo::operator!=( const ErrorInfo& e ) const
{
    return !operator==( e );
}


//-----------------------------------------------------------------------------
//
//: Get Library
//
//!return: int - The number of the library which threw the exception.
//
inline int ErrorInfo::getLibrary() const
{
    return mLibrary;
}


//-----------------------------------------------------------------------------
//
//: Get Error Code
//
//!return: int
//
inline int ErrorInfo::getCode() const 
{
    return mCode;
}


//-----------------------------------------------------------------------------
//
//: Get Error Message
//
// Returns a text representation of the error code.
//
//!return const std::string&
//
inline const std::string& ErrorInfo::getMessage() const
{
    return mMsg;
}


//-----------------------------------------------------------------------------
//
//: Get Error Information
//
// Returns additional information about the error.
//
//!return: const std::string&
//
inline const std::string& ErrorInfo::getInfo() const 
{
    return mInfo;
}


//-----------------------------------------------------------------------------
//
//: Get File Name
//
// Get the name of the file in which the error was thrown.
//
//!return: const char*
//
inline const char* ErrorInfo::getFile() const
{
    return mFile;
}


//-----------------------------------------------------------------------------
//
//: Get Line Number
//
// Get the line number on which the error was thrown.
//
//!return: size_t
//
inline size_t ErrorInfo::getLine() const
{
    return mLine;
}



#endif // GeneralErrorInfoHH
