/* -*- mode: c++; c-basic-offset: 4; -*- */

#ifndef MUTEXLOCK_HH
#define MUTEXLOCK_HH

#include <assert.h>
#include <errno.h>
#include <string.h>

#include <pthread.h>

#include <iostream>

//
// class MutexLock
//
// This class ensures that the given mutex is unlocked whenever an
// exception is thrown but not caught in a locked section.
//
// It should be used like this:
//
// {
//     MutexLock lock(pthread_mutex_t& mutex); // mutex is now locked
//     do whatever...
//     ...
// } // MutexLock has been destructed so mutex is now unlocked
//

// Creating two MutexLock objects with the same mutex in the same scope
// will cause a deadlock.
//
class MutexLock {
public:
    typedef pthread_mutex_t	lock_type;

    //: Constructor
    MutexLock(lock_type& mutex,
	      const char* const File = "", const unsigned int Line = 0,
	      bool Logging = true );

    //: Constructor
    MutexLock(lock_type* const mutex_ptr,
	      const char* const File = "", const unsigned int Line = 0,
	      bool Logging = true );

    //: Destructor
    ~MutexLock( );

    //: Release the lock before object is destructed
    void Release( const char* const File = "", const unsigned int Line = 0 );

    inline static void Destroy( lock_type& Lock )
    {
	pthread_mutex_destroy( &Lock );
    }

    inline static lock_type Initialize( )
    {
	static const lock_type retval = PTHREAD_MUTEX_INITIALIZER;

	return retval;
    }

    inline static void Initialize( lock_type& Lock )
    {
	pthread_mutex_init( &Lock,
			    static_cast< pthread_mutexattr_t* >( NULL ) );
    }

    inline static void Lock( lock_type& Lock )
    {
	pthread_mutex_lock( &Lock );
    }

    inline static void UnLock( lock_type& Lock )
    {
	pthread_mutex_unlock( &Lock );
    }

private:
    //: Default Constructor
    //
    // Default constructor is private to prevent copying of classes
    // internal data.
    MutexLock();
    //: Copy Constructor
    //
    // Copy constructor is private to prevent copying of classes internal data.
    MutexLock(const MutexLock&);

    //: Assigment operator
    //
    // Assignment operator is privat to prevent copying of classes internal
    // data
    const MutexLock& operator=(const MutexLock&);

    //: Mutex lock to prevent multiple access to critical sections of code.
    lock_type* m_mutex_ptr;

    bool	m_logging;
};

#endif // MUTEXLOCK_HH
