/* -*- mode: c++; c-basic-offset: 4; -*- */

#ifndef GeneralObjectRegistryHH
#define GeneralObjectRegistryHH

#include "general/config.h"
#include "general/ReadWriteLock.hh"
#include <set>

//-----------------------------------------------------------------------------
//
//: Object Registry
//
// This class implements a registry for objects.  The registry provides a
// mechanism for keeping track of pointers.  The class contains methods to
// add/remove pointers to the registry, as well as to check for the existence
// of a pointer in the registry.
//
// <p>The registry owns its objects.  That is, when the registry is destructed
// it destructs any objects which are registered with it.
//
// <p>A STL set is used to store pointers to the registered objects.
//
// <p>This class is used by the LDAS API's to keep track of pointer objects
// returned to the TCL layer.  To avoid segmentation faults, when the C++ layer
// receives a pointer from TCL, the appropriate ObjectRegistry object is
// checked to ensure that the pointer exists.  Furthermore, the resetApi
// command causes the C++ layer to empty the ObjectRegistry objects, thereby
// ensuring that memory isn't leaked by the TCL layer failing to destruct C++
// objects.
//
template< class T >
class ObjectRegistry
{
public:

/* typedefs */
typedef typename std::set< T* >::iterator iterator;
typedef typename std::set< T* >::const_iterator const_iterator;

/* Constructor/Destructor */
ObjectRegistry();
~ObjectRegistry();

/* Accessors */
bool isRegistered( const T* o );
size_t size() const;

/* Manipulators */
void reset();
void registerObject( T* o );
bool destructObject( T* o );
bool removeObject( T* o );

/* Algorithms */
template<class Op>
Op for_each( Op Function ) const;

protected:
    //: Ensure only one thread accesses the registry at a time
    mutable General::ReadWriteLock::lock_type m_lock;

    //: Obtain the start of the list
    iterator begin();
    //: Obtain the end of the list
    iterator end();
    //: Obtain the start of the list (read-only)
    const_iterator begin() const;
    //: Obtain the end of the list (read-only)
    const_iterator end() const;

private:
    //: Collection of data members
    std::set< T* > mData;
};


#include "objectregistry.icc"

#endif // GeneralObjectRegistryHH
