#include "general/config.h"

#if HAVE_BYTESWAP_H
#include "byteswap.h"
#elif HAVE_SYS_BYTEORDER_H
#include "sys/types.h"
#include "sys/byteorder.h"

#ifndef bswap_16
#define bswap_16(x) BSWAP_16(x)
#endif /* bswap_16 */
#ifndef bswap_32
#define bswap_32(x) BSWAP_32(x)
#endif /* bswap_32 */
#ifndef bswap_64
#define bswap_64(x) BSWAP_64(x)
#endif /* bswap_64 */
#else
#error No optimized byte swapping macros provided
#endif /* HAVE_SYS_BYTEORDER_H */

#include <cctype>
#include <iostream>   
#include <sstream>
#include <stdexcept>
#include <typeinfo>

#include "util.hh"

#include "general/types.hh"

using namespace std;

//-----------------------------------------------------------------------------
//: Case-Insensitve String comparison
//
// Compares the strings S1 and S2 and returns an integer less than, equal to,
// greater than zero.
//
//!param: const std::string& S1 - Left hand side of comparison
//!param: const std::string& S2 - Right hand side of comparison
//
int General::
cmp_nocase( const std::string& S1, const std::string& S2 )
{
  std::string::const_iterator	p1 = S1.begin();
  std::string::const_iterator	p2 = S2.begin();

  while( p1 != S1.end() && p2 != S2.end() )
  {
    if ( toupper( *p1 ) != toupper( *p2 ) )
    {
      return ( ( toupper( *p1 ) < toupper( *p2 ) ) ? -1 : 1 );
    }
    p1++;
    p2++;
  }
  return ( S2.size() == S1.size() )
    ? 0
    : ( S1.size() < S2.size() )
    ? -1
    : 1;
}

//-----------------------------------------------------------------------------
//
//: Reverse 1 byte
//
// This does nothing.
//
//!param: void* dest
//!param: size_t n
//
template<>
void reverse< 1 >( void* dest, size_t n )
{
    return;
}


//-----------------------------------------------------------------------------
//
//: Reverse 2 bytes
//
// Reverses the order of two bytes.  This is a template specialization.
//
//!param: void* dest
//!param: size_t n
//
template<>
void reverse< 2 >( void* dest, size_t n )
{
    register INT_2U* d = reinterpret_cast< INT_2U* >( dest );
    register INT_2U* end = d + n;

    while ( d != end )
    {
        *d = bswap_16( *d );
        ++d;
    }
}


//-----------------------------------------------------------------------------
//
//: Reverse 4 bytes
//
// Reverses the order of four bytes.  This is a template specialization.
//
//!param: void* dest
//!param: size_t n
//
template<>
void reverse< 4 >( void* dest, size_t n )
{
    register INT_4U* d = reinterpret_cast< INT_4U* >( dest );
    register INT_4U* end = d + n;

    while ( d != end )
    {
        *d = bswap_32( *d );
        ++d;
    }
}


//-----------------------------------------------------------------------------
//
//: Reverse 8 bytes
//
// Reverses the order of eight bytes.  This is a template specialization.
//
//!param: void* dest
//!param: size_t n
//
template<>
void reverse< 8 >( void* dest, size_t n )
{
    register INT_8U* cur = reinterpret_cast< INT_8U* >( dest );
    register INT_8U* end = cur + n;

    while ( cur != end )
    {
        *cur = bswap_64( *cur );
	++cur;
    }
}

std::string slower( const std::string &s )
{
   std::string	str;

   for( std::string::const_iterator i = s.begin();
        i != s.end();
        ++i )
   {
      str += tolower( *i );
   }

   return str;
}
   
   
//-----------------------------------------------------------------------------
//
//: Convert to Lower-case
//
//!param std::string& s
//
void string2lower( std::string& s )
{
    std::string::iterator iter = s.begin();
    for ( std::string::size_type size = s.size(); size != 0; --size )
    {
        *iter = tolower( *iter );
        ++iter;
    }
}


//-------------------------------------------------------------------------------
//
//: Is character a space?
//
//!param: const char c - Character to test.
//
//!return: bool - True if argument is whitespace, false otherwise.
//   
const bool is_space( const char c )
{
   return isspace( c );
}

   
//-------------------------------------------------------------------------------
//
//: Is character a not space?
//
//!param: const char c - Character to test.
//
//!return: bool - True if argument is not a whitespace, false otherwise.
//      
const bool not_space( const char c )
{
   return !isspace( c );
}   

//-----------------------------------------------------------------------
//: Rethrow a std::exception adding information to the front
//
//!param: const std::string& Prefix - Text to be prepended to the
//+	exception text.
//!param: const std::exception& Exception - Origional exception
//+	that was thrown and used as type of exception to throw.
//-----------------------------------------------------------------------
void General::
Rethrow( const std::string& Prefix, const std::exception& Exception )
{
  std::ostringstream	err;
  err << Prefix << ": " << Exception.what( );
  try
  {
    throw( Exception );
  }
  // logic_error: begin
  catch( const std::length_error& e )
  {
    throw std::length_error( err.str( ) );
  }
  catch( const std::domain_error& e )
  {
    throw std::domain_error( err.str( ) );
  }
  catch( const std::out_of_range& e )
  {
    throw std::out_of_range( err.str( ) );
  }
  catch( const std::invalid_argument& e )
  {
    throw std::invalid_argument( err.str( ) );
  }
  catch( const std::logic_error& e )
  {
    throw std::logic_error( err.str( ) );
  }
  // logic_error: end
  // runtime_error: begin
  catch( const std::range_error& e )
  {
    throw std::range_error( err.str( ) );
  }
  catch( const std::overflow_error& e )
  {
    throw std::overflow_error( err.str( ) );
  }
  catch( const std::underflow_error& e )
  {
    throw std::underflow_error( err.str( ) );
  }
  catch( const std::runtime_error& e )
  {
    throw std::runtime_error( err.str( ) );
  }
  // runtime_error: end
  // others: begin
  catch( const std::bad_alloc& e )
  {
    throw std::bad_alloc( e );
  }
  catch( const std::bad_exception& e )
  {
    throw std::bad_exception( e );
  }
  catch( const std::bad_cast& e )
  {
    throw std::bad_cast( e );
  }
  catch( const std::bad_typeid& e )
  {
    throw std::bad_typeid( e );
  }
  catch( const std::ios_base::failure& e )
  {
    throw std::ios_base::failure( err.str( ) );
  }
  // others: end
  catch( const std::exception& e )
  {
    throw std::runtime_error( err.str( ) );
  }
}
